//
// Alert.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Alert.h"

#include "MsgDialog.h"
#include "Application.h"
#include "RedrawGC.h"
#include "CheckBox.h"

#include "vos\Errors.h"
#include "vos\Preferences.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	using namespace vos;

	/**************************************************************************
	
	>	class AlertDialog : public MsgDialog
	
		Author:		Tim Browse
		Date:		5/5/2002
	
		Purpose:	Dialog class to implement Alert dialogs, which are simple
					dialogs that exist to give the user a message with an OK
					button.
	
	**************************************************************************/
	
	class AlertDialog : public MsgDialog
	{
		public:
			AlertDialog(const char *szAlertTitle,
						const char *szAlertText,
						Window *pParent,
						bool bShowCheckBox);

			virtual void InitTheDialog();

			virtual void OnOK();

			bool TellUserNextTime();

		public:
			void RepositionControls(int* pDeltaX, int* pDeltaY);

		protected:
			String m_AlertText;
			String m_AlertTitle;
			Window m_AlertMessage;

			// The "don't tell me again" checkbox.
			CheckBox m_CheckBox;
			
			// Flag to indicate if we should display a check box
			// under the alert text, saying "Don't tell about this again."
			bool m_bShowDontTellMeAgainCheckBox;

			// Flag to indicate what the user decided on the "Don't tell
			// me again" option
			bool m_bTellUserNextTime;
	};

	/**************************************************************************

	>	AlertDialog::AlertDialog(const char *szAlertTitle,
								 const char *szAlertText,
								 Window* pParent,
								 const char *szPreference)

		Inputs:		szAlertTitle - string resource to use as the alert dialog title.
					szAlertText - string resource to use as the alert message.
					pParent - parent Window of this dialog (may be NULL)
					szPreference - optional: name of a preference to use to
								   implement an automatic "Don't tell me about
								   this again." checkbox.  Can be NULL.

		Purpose:	Constructor

	**************************************************************************/

	AlertDialog::AlertDialog(const char *szAlertTitle,
							 const char *szAlertText,
							 Window* pParent,
							 bool bShowCheckBox)

									 : MsgDialog(Alert::s_AlertDialogID, 
												 pParent,
												 Alert::s_AlertTextControl),
									   m_AlertTitle(szAlertTitle),
									   m_AlertText(szAlertText),
									   m_bShowDontTellMeAgainCheckBox(bShowCheckBox),
									   m_bTellUserNextTime(true)
	{
		// Set up the dialog
		SetTitle(m_AlertTitle);
		SetMessage(m_AlertText);
	}

	/**************************************************************************
	
	>	void AlertDialog::InitTheDialog()
	
		Author:		Tim Browse
		Date:		5/5/2002
	
		Purpose:	Initialise the dialog controls.
	
	**************************************************************************/
	
	void AlertDialog::InitTheDialog()
	{
		// Attach checkbox if necessary
		if ((Alert::s_AlertDontTellMeAgainCheckBox != NOCTL) &&
			m_bShowDontTellMeAgainCheckBox)
		{
			m_CheckBox.Attach(this, Alert::s_AlertDontTellMeAgainCheckBox);
		}
	}

	/**************************************************************************

	>	void AlertDialog::RepositionControls(int* pDeltaX, int* pDeltaY)


		Author:		Tim Browse
		Date:		5/5/2002
		Inputs:		pDeltaX, pDeltaY - the amount that the dialog will
									   be rsized to accomodate the alert message.
		Outputs:	pDeltaX, pDeltaY - the possibly adjusted deltas, in case this
									   function needs to add/remove any controls.

		Purpose:	Over-rides the MsgDialog function to do specific control
					moving for the alert dialog.

					At the moment, all we do is move the OK button, and 
					optionally hide the "Don't show me again" checkbox.

	**************************************************************************/

	void AlertDialog::RepositionControls(int* pDeltaX, int* pDeltaY)
	{
		// Does the alert dialog have a "Don't tell me again" checkbox?
		if (Alert::s_AlertDontTellMeAgainCheckBox != NOCTL)
		{
			// Yes - do we actually want a "Don't tell me again" checkbox?
			if (m_bShowDontTellMeAgainCheckBox)
			{
				// Yes, so centre it
				MoveControlRelative(Alert::s_AlertDontTellMeAgainCheckBox, 
									(*pDeltaX)/2, *pDeltaY);
			}
			else
			{
				// No, so hide the control, and adjust DeltaY accordingly
				HideControl(Alert::s_AlertDontTellMeAgainCheckBox);

				Rect CheckboxPosition;
				GetControlPosition(Alert::s_AlertDontTellMeAgainCheckBox, 
								   &CheckboxPosition);

				int Diff = CheckboxPosition.bottom - m_MessageArea.bottom;

				*pDeltaY -= Diff;

				// Grow message area to cover checkbox area.
				AdjustMessageArea(0, Diff);
			}
		}

		// We only use DeltaX/2 so we keep the button centred in the dialog.
		MoveControlRelative(Alert::s_AlertOKButton, (*pDeltaX)/2, *pDeltaY);
	}

	/**************************************************************************
	
	>	bool AlertDialog::TellUserNextTime()
	
		Author:		Tim Browse
		Date:		5/5/2002
		Returns:	true => tell the user about this event the next time
					false => user has requested not to be told again.
	
		Purpose:	Find out if we should tell the user about this event
					next time it happens.  This is controlled by an (optional)
					"Don't tell me again" checkbox in the alert dialog.
	
	**************************************************************************/
	
	bool AlertDialog::TellUserNextTime()
	{
		return m_bTellUserNextTime;
	}

	/**************************************************************************
	
	>	void AlertDialog::OnOK()
	
		Author:		Tim Browse
		Date:		5/5/2002
	
		Purpose:	Intercept the OK button, so we can find out if the
					user chose the "Don't tell me again" option.
	
	**************************************************************************/
	
	void AlertDialog::OnOK()
	{
		if ((Alert::s_AlertDontTellMeAgainCheckBox != NOCTL) &&
			m_bShowDontTellMeAgainCheckBox)
		{
			// Logic is inverted
			if (m_CheckBox)
			{
				// User does not want to be told again
				m_bTellUserNextTime = false;
			}
			else
			{
				// They didn't check the box
				m_bTellUserNextTime = true;
			}
		}

		MsgDialog::OnOK();
	}

	//
	// Alert static variables...
	//

	// The dialog template ID to use for the alert dialogs.
	DialogResourceID Alert::s_AlertDialogID = 0;

	// IDs of the controls in the above template.
	ControlID Alert::s_AlertTextControl = NOCTL;
	ControlID Alert::s_AlertOKButton = NOCTL;
	ControlID Alert::s_AlertDontTellMeAgainCheckBox = NOCTL;	// Optional

	/**************************************************************************

	>	void Alert::Initialise(const DialogResourceID& AlertDialogID,
							   ControlID AlertTextControl,
							   ControlID AlertOKButton,
							   ControlID AlertDontTellMeAgainCheckBox)


		Author:		Tim Browse
		Date:		5/5/2002
		Inputs:		AlertDialogID - the ID of the dialog template to use.
					AlertTextControl - the control ID of the text message control.
					AlertOKButton - the control ID of the OK button
					AlertDontTellMeAgainCheckBox - control ID of the "don't tell
												   me this again" checkbox.

		Purpose:	Initialise the Alert system to use a particular dialog
					template when showing Alert dialogs.

					You MUST call this function before calling Alert::Show()
					in your app - do it at the start with any other initialisation
					code.

	**************************************************************************/

	void Alert::Initialise(const DialogResourceID& AlertDialogID,
						   ControlID AlertTextControl,
						   ControlID AlertOKButton,
						   ControlID AlertDontTellMeAgainCheckBox)
	{
		s_AlertDialogID = AlertDialogID;
		s_AlertTextControl = AlertTextControl;
		s_AlertOKButton = AlertOKButton;
		s_AlertDontTellMeAgainCheckBox = AlertDontTellMeAgainCheckBox;
	}

	/**************************************************************************

	>	void Alert::Show(const StringResourceID& AlertTitle,
						 const StringResourceID& AlertText,
						 Window *pParent,
						 const char* szPreferenceName)


		Author:		Tim Browse
		Date:		6/5/2002
		Inputs:		AlertTitle - title text to use for the alert
					AlertText - message text to use for the alert
					pParent - parent window of the Alert dialog (may be NULL)
					szPreference - name of a preference value to use to implement
								   "don't tell me this again" functionality.
								   (use NULL if you don't want this feature)

		Purpose:	Show an alert dialog to let the user know that something
					has happened.  Alert dialogs have an OK button but no
					way of cancelling, as cancelling them doesn't mean anything.

					The dialog can be linked to a preference value (bool) that
					indicates whether the alert should be shown or not.  This
					is enabled when the user clicks "Don't tell me this again"
					in the Alert dialog, if provided.

	**************************************************************************/

	void Alert::Show(const StringResourceID& AlertTitle,
					 const StringResourceID& AlertText,
					 Window *pParent,
					 const char* szPreferenceName)
	{
		// Load the strings
		vos::String strAlertText, strAlertTitle;
		AlertTitle.Load(&strAlertTitle);
		AlertText.Load(&strAlertText);

		// And call main function.
		Show(strAlertTitle.c_str(), strAlertText.c_str(), pParent, szPreferenceName);
	}

	/**************************************************************************

	>	void Alert::Show(const char *szAlertTitle,
						 const char *szAlertText,
						 Window *pParent,
						 const char* szPreferenceName)



		Author:		Tim Browse
		Date:		6/5/2002
		Inputs:		szAlertTitle - title text to use for the alert
					szAlertText - message text to use for the alert
					pParent - parent window of the Alert dialog (may be NULL)
					szPreference - name of a preference value to use to implement
								   "don't tell me this again" functionality.
								   (use NULL if you don't want this feature)

		Purpose:	Show an alert dialog to let the user know that something
					has happened.  Alert dialogs have an OK button but no
					way of cancelling, as cancelling them doesn't mean anything.

					The dialog can be linked to a preference value (bool) that
					indicates whether the alert should be shown or not.  This
					is enabled when the user clicks "Don't tell me this again"
					in the Alert dialog, if provided.

	**************************************************************************/

	void Alert::Show(const char *szAlertTitle,
					 const char *szAlertText,
					 Window *pParent,
					 const char* szPreferenceName)
	{
		// By default, we don't show the user the "Don't tell me again"
		// checkbox in the alert dialog
		bool bShowCheckBox = false;

		// Ok, check to see if we should show this alert
		if (szPreferenceName != NULL)
		{
			// Read the preference value (true => show the user the alert)
			bool bShowUser = true;
			vos::GetPreferences().Read(szPreferenceName, &bShowUser);

			if (bShowUser)
			{
				// Yes, show the user, and give them the chance to opt-out
				// for next time
				bShowCheckBox = true;
			}
			else
			{
				// User has elected not to see these alerts
				return;
			}
		}

		// Sanity checks
		COMPLAIN_IF(s_AlertDialogID.IsNull(), "No dialog ID for Alert class - have you called Alert::Initialise()?");
		COMPLAIN_IF(s_AlertTextControl == NOCTL, "No control ID for Alert class - have you called Alert::Initialise()?");

		// If no parent passed in, use main app window as the owner/parent
		if (pParent == NULL)
		{
			pParent = GetApp()->GetMainWindow();
		}

		// Show the alert
		AlertDialog TheAlert(szAlertTitle, szAlertText, pParent, bShowCheckBox);
		TheAlert.DoModal();

		// Update the preference if necessary
		if (szPreferenceName != NULL)
		{
			// true => show them the alert next time
			vos::GetPreferences().Write(szPreferenceName, TheAlert.TellUserNextTime());
		}
	}
};

