//
// Application.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Application.h"

#include "ApplicationImp.h"
#include "Dialog.h"

#include "vOS\FileSpec.h"
#include "vOS\Preferences.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{

	using namespace vos;

	IdleHandler::IdleHandler() : m_bWantsIdleEvents(false)
	{
		// Install the idle handler
		GetApp()->RegisterIdleEventHandler(this);
	}

	IdleHandler::~IdleHandler()
	{
		// Uninstall the idle handler
		GetApp()->UnregisterIdleEventHandler(this);
	}

	void IdleHandler::ClaimIdleEvents()
	{
		m_bWantsIdleEvents = true;
	}

	void IdleHandler::ReleaseIdleEvents()
	{
		m_bWantsIdleEvents = false;
	}

	
	/**************************************************************************
	
	>	bool IdleHandler::WantsIdleEvents()
	
		Author:		Tim Browse
		Date:		2/5/2002
		Returns:	true => idle handler wants idle events
					false => otherwise
	
		Purpose:	Find out if this handler currently wants to receive idles
	
	**************************************************************************/
	
	bool IdleHandler::WantsIdleEvents()
	{
		return m_bWantsIdleEvents;
	}

	//
	// The application singleton
	//
	Application* Application::s_pTheApp = NULL;

	Application* GetApp()
	{
		return Application::GetApp();
	}

	//
	// Application message map
	//

	BEGIN_VGUI_MESSAGE_MAP(Application, MessageTarget)
		//
		// It's empty!
		//
	END_VGUI_MESSAGE_MAP()


	//
	// Member functions
	//

	Application::Application() : m_pImp(NULL),
								 m_pMainWindow(NULL),
								 m_pMRUFileList(NULL),
								 m_iIdleEvent(0),
								 m_DefaultCommand(0)
	{
		// TCILB
	}

	void Application::Init()
	{
		// Create the platform dependent application implementation.
		// For base class, we use standard application type.
		m_pImp = TheFactory().MakeApplication();

		// Link back to this object
		m_pImp->SetAppLink(this);
	}

	Application::~Application()
	{
		delete m_pMainWindow;
		delete m_pImp;

		// Shut down the MRU file list object, if we have one
		delete m_pMRUFileList;
	}

	void Application::CreateApplication()
	{
		// Call user defined function to create correct type of app.
		s_pTheApp = vgui::CreateApplication();

		// Initialise the application object.
		s_pTheApp->Init();
	}

	bool Application::CreateMainWindowObject()
	{
		try
		{
			m_pMainWindow = new Window;
			return true;
		}
		catch (std::bad_alloc)
		{
			// Problem creating main window
			return false;
		}
	}

	bool Application::CreateMainWindow()
	{
		if (!CreateMainWindowObject())
		{
			// problem creating window - abort
			return false;
		}

		if (!m_pMainWindow->Create(NULL,
								   ws_MainAppWindow, 
								   m_resAppTitle))
		{
			// Window creation failed
			return false;
		}

		if (IsAccessory())
		{
			// Install the accessory
			if (!GetImp()->InstallAccessory(m_resIcon, 
											m_resAccessoryMenu, 
											m_resAppTitle,
											m_DefaultCommand))
			{
				// Error occured
				return false;
			}

			// Ok, do we show the main window?
			if (m_Options.m_bShowWindowOnAppStart)
			{
				m_pMainWindow->ShowWindow(sw_SHOW);
				m_pMainWindow->UpdateWindow();
			}
		}
		else
		{
			// Normal app - show and paint the window
			m_pMainWindow->ShowWindow(sw_SHOW);
			m_pMainWindow->UpdateWindow();
		}

		// All ok
		return true;
	}

	Window* Application::GetMainWindow()
	{
		Application* pApp = GetApp();

		// Check for NULL - this means we can safely call this function 
		// during startup - e.g. to report errors with the normal dialog
		// functions.
		if (pApp != NULL)
			return pApp->m_pMainWindow;
		else
			return NULL;
	}

	void Application::OnStartUp()
	{
		// Check for filename on command line
		String CommandLine = GetCommandLine();

		// If the string is non-empty, try to load the file
		if (!CommandLine.IsEmpty())
		{
			// Remove quotes if necessary
			if ((CommandLine[0] == '"') && 
				(CommandLine[CommandLine.GetLength()-1] == '"'))
			{
				CommandLine = CommandLine.Mid(1, CommandLine.GetLength() - 2);
			}

			FileSpec FS;
			FS.SetFromString(CommandLine);
			OnOpenFile(FS);
		}
	}

	OSAppHandle Application::GetOSAppHandle()
	{
		return m_pImp->GetOSAppHandle();
	}

	Menu* Application::GetMainMenu()
	{
		return m_pImp->GetMainMenu();
	}

	int Application::Run()
	{
		// Call implementation to run the app
		int iResult = m_pImp->Run();

		// App is exiting now...

		// Implementation dependent exit.
		m_pImp->Exit();

		// Allow app to do clean up on app closedown
		OnCloseDown();

		// Tell all registered handlers that the app is about to exit
		ServiceShutdownHandlers(true);

		// Clean up any systems used by the app
		CleanUp();

		// Return result code of app
		return iResult;
	}

	void Application::CleanUp()
	{
		// Clean up data allocated by the preferences system.
		vos::Preferences::GetPreferences().CleanUp();
	}

	bool Application::Exit()
	{
		// Exit application - start it off by destroying the main window
		if(m_pMainWindow != NULL)
		{
			m_pMainWindow->Destroy();
		}
		
		// We return true so this can be installed as a command handler.
		return true;
	}

	void Application::OnCloseDown()
	{
		// No default handling of close down...derived classes may over-ride.
	}

	void Application::ProcessMessages()
	{
		m_pImp->ServiceMessageQueue();
	}

	void Application::GetAppTitle(String& strAppTitle)
	{
		m_resAppTitle.Load(&strAppTitle);
	}



	bool Application::OnCommand(int nCommand)
	{
		//
		// Route this command in this order:
		//
		// 1. Main application window
		// 2. MRU File list object (if installed)
		// 3. This application object
		//

		if ((m_pMainWindow != NULL) && (m_pMainWindow->OnCommand(nCommand)))
		{
			// Main window handled command
			return true;
		}

		if ((m_pMRUFileList != NULL) && (m_pMRUFileList->OnCommand(nCommand)))
		{
			// MRU file list handled command
			return true;
		}

		// Route command through our own message map
		return MessageTarget::OnCommand(nCommand);
	}

	void Application::OnOpenFile(const vos::FileSpec& FS)
	{
		// Vanilla applications have no predefined response to opening a named
		// file - it is ignored.
	}

	void Application::SetNotifyIcon(const Icon& NotifyIcon, 
									const String& ToolTip,
									int NotifyCommand)
	{
		GetImp()->SetNotifyIcon(NotifyIcon, ToolTip, NotifyCommand);
	}

	void Application::SetFlashingNotifyIcon(const Icon& NotifyIcon, 
											const String& ToolTip,
											int NotifyCommand)
	{
		GetImp()->SetFlashingNotifyIcon(NotifyIcon, ToolTip, NotifyCommand);
	}

	void Application::ClearNotifyIcon()
	{
		GetImp()->ClearNotifyIcon();
	}


	String Application::GetCommandLine() const
	{
		String CmdLine;
		GetImp()->GetCommandLine(&CmdLine);
		return CmdLine;
	}

	void Application::RegisterIdleEventHandler(IdleHandler* pHandler)
	{
		// Register this idle handler
		m_IdleHandlers.insert(pHandler);
	}

	void Application::UnregisterIdleEventHandler(IdleHandler* pHandler)
	{
		m_IdleHandlers.erase(pHandler);
	}

	/**************************************************************************
	
	>	bool Application::ServiceIdleHandlers()
	
		Author:		Tim Browse
		Date:		2/5/2002
		Returns:	true => one or more idle handlers still have work to do
					false => all idle handlers have finished their work; the
							 application can go to sleep until the next
							 GUI message arrives.
	
		Purpose:	Service all idle handlers that have currently claimed
					idle events.
	
	**************************************************************************/
	
	bool Application::ServiceIdleHandlers()
	{
		IdleHandlers::iterator it = m_IdleHandlers.begin();

		int cActiveIdleHandlers = 0;

		while (it != m_IdleHandlers.end())
		{
			// Service this idle event handler (if enabled)
			if ((*it)->WantsIdleEvents())
			{
				if ((*it)->OnIdle(m_iIdleEvent))
				{
					// Idle handler wants more time, so it is still active
					cActiveIdleHandlers++;
				}
			}

			// Try next handler
			++it;
		}

		if (cActiveIdleHandlers > 0)
		{
			// Still some work to do
			m_iIdleEvent++;
			return true;
		}
		else
		{
			// End of idle processing - initialise event ID ready for
			// next round of idle processing
			m_iIdleEvent = 0;
			return false;
		}
	}

	void Application::RegisterShutdownHandler(ShutdownHandler *pHandler)
	{
		m_ShutdownHandlers.insert(pHandler);
	}

	void Application::ServiceShutdownHandlers(bool bNormalShutdown)
	{
		ShutdownHandlers::iterator it = m_ShutdownHandlers.begin();

		while (it != m_ShutdownHandlers.end())
		{
			// Service this shutdown handler
			(*it)->OnShutdown(bNormalShutdown);

			// Try next handler
			++it;
		}
	}

	void Application::InstallMRU(MessageTarget *pMRUList)
	{
		m_pMRUFileList = pMRUList;
	}

	ApplicationSingleWindow::ApplicationSingleWindow()
	{
	}

	ApplicationDialog::ApplicationDialog()
	{
	}

	bool ApplicationDialog::CreateMainWindowObject()
	{
		try
		{
			m_pMainWindow = CreateTheDialog();
			return (m_pMainWindow != NULL);
		}
		catch (std::bad_alloc)
		{
			// Problem creating main window
			return false;
		}
	}

	bool ApplicationDialog::CreateMainWindow()
	{
		if (!CreateMainWindowObject())
		{
			// problem creating window - abort
			return false;
		}

		// Hello!  if you're staring at this next line of code because an 
		// exception is being thrown when it is executed, then you probably 
		// need to enable RTTI on your C++ project :-)
		Dialog* pDialog = dynamic_cast<Dialog*>(m_pMainWindow);

		if (pDialog != NULL)
		{
			if (IsAccessory())
			{
				// We need to create the window before installing the
				// accessory, as the window handle may be needed to install
				// the accessory
				pDialog->DoModeless();
				
				// Install the accessory
				if (!GetImp()->InstallAccessory(m_resIcon, 
												m_resAccessoryMenu, 
												m_resAppTitle,
												m_DefaultCommand))
				{
					// Error occured
					return false;
				}
			}
			else
			{
				// Dialog for a normal main app window is treated as a modeless dialog,
				// so that we get to control the message loop.
				pDialog->DoModeless();
			}
		}

		// All ok
		return true;
	}
}; // namespace vgui


