//
// Application.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_Application_h
#define vgui_Application_h

#include "vgui.h"


#include "Window.h"
#include "String.h"
#include "ResourceID.h"
#include "MessageMap.h"
#include "MessageTarget.h"
#include "AppOptions.h"

#include <vector>
#include <set>

#include "vos\Application.h"

namespace vgui
{
	class Application;
	class ApplicationImp;
	class Icon;
	class Document;
	class View;
	class Menu;

	typedef enum
	{
		IdleHandlerStatus_Finished,
		IdleHandlerStatus_StillProcessing,
		IdleHandlerStatus_ReleaseIdleEvents
	}
	TIdleHandlerStatus;

	// Interface for claiming idles - derive your handler from this class.
	class IdleHandler
	{
		public:
			IdleHandler();
			virtual ~IdleHandler();

			// Called on idle events.
			// 
			// Return true to deregister this idle handler;
			// false to continue receiving idle events.
			virtual bool OnIdle(uint iIdleEvent) = 0;

			// Used for claiming idle events
			void ClaimIdleEvents();

			// Use to release idle events - do NOT call from within
			// OnIdle() call (return IdleHandlerStatus_ReleaseIdleEvents instead).
			void ReleaseIdleEvents();

			// Find out if this handler currently wants to receive idles
			bool WantsIdleEvents();

		private:
			// true => we should poll this handler on idle events
			bool m_bWantsIdleEvents;
	};

	typedef std::set<IdleHandler*> IdleHandlers;

	// Interface for registering a shutdown handler
	class ShutdownHandler
	{
		public:
			virtual void OnShutdown(bool bNormalShutdown) = 0;
	};

	typedef std::set<ShutdownHandler*> ShutdownHandlers;

	typedef void* OSAppHandle;

	// This function is not defined - you must define it yourself to 
	// create the correct type of application for your program.
	extern Application *CreateApplication();

	class Application : public MessageTarget,
						public vos::Application
	{
		public:

			//
			// Constructors
			//

			Application();
			~Application();

			//
			// Operations
			//
			static void CreateApplication();
			virtual void Init();
			int Run();
			bool Exit();
			virtual void OnStartUp();
			virtual void OnCloseDown();

			virtual bool CreateMainWindow();
			virtual bool CreateMainWindowObject();

			// Access to the application singleton.
			static Application* GetApp() { return s_pTheApp; }

			static Window* GetMainWindow();

			ApplicationImp *GetImp() const { return m_pImp; }

			const AppOptions& GetOptions() { return m_Options; }

			OSAppHandle GetOSAppHandle();

			Menu* GetMainMenu();

			bool IsAccessory() const { return m_Options.m_bAccessory; }

			//
			// Message handling
			//

			virtual bool OnCommand(int nCommand);

			virtual void OnOpenFile(const vos::FileSpec& FS);

			// process outstanding messages for this app.
			void ProcessMessages();

			void GetAppTitle(vos::String& strAppTitle);

			vos::String GetCommandLine() const;

			// Resource ID access
			StringResourceID		GetAppTitleID()			{ return m_resAppTitle; }
			MenuResourceID			GetMenuID()				{ return m_resMenu; }
			MenuResourceID			GetAccessoryMenuID()	{ return m_resAccessoryMenu; }
			IconResourceID			GetIconID()				{ return m_resIcon; }
			IconResourceID			GetSmallIconID()		{ return m_resSmallIcon; }
			AcceleratorResourceID	GetAcceleratorsID()		{ return m_resAccelerators; }

			// Notification icons
			void SetNotifyIcon(const Icon& NotifyIcon, 
							   const vos::String& ToolTip,
							   int NotifyCommand = 0);

			void SetFlashingNotifyIcon(const Icon& NotifyIcon, 
									   const vos::String& ToolTip,
									   int NotifyCommand = 0);

			void ClearNotifyIcon();

			// Service idle event handlers
			bool ServiceIdleHandlers();

			// Register a handler to be called when the app shuts down
			// (for whatever reason)
			void RegisterShutdownHandler(ShutdownHandler *pHandler);

			// Service all shutdown handlers (called when app terminates)
			void ServiceShutdownHandlers(bool bNormalShutdown);

			// MRU file object
			void InstallMRU(MessageTarget *pMRUList);

		protected:
			// Data...

			// The main application window
			Window *m_pMainWindow;

			// Resource ID of the application title
			StringResourceID m_resAppTitle;

			// Resource ID of the main menu
			MenuResourceID m_resMenu;

			// Resource ID of the main icons
			IconResourceID m_resIcon;
			IconResourceID m_resSmallIcon;		// Not used on all platforms.

			// Resource ID of the accelerator table
			AcceleratorResourceID m_resAccelerators;
			
		
			//
			// Data for apps that use accessory functionality (e.g. a system
			// tray app on Win32)...
			//

			// The menu to use to open the accessory.
			// (m_resMenu is still used on the main window itself,
			//  although can be left out for accessories, e.g. if your
			//  main window is a dialog and has no menu.)
			MenuResourceID m_resAccessoryMenu;

			// Dialog resource for the main window (accessory windows
			// are always dialogs - at least for now).
			DialogResourceID m_resAccessoryDialog;

			// The command to issue when the user performs the primary
			// action on the accessory - e.g. when they click on the
			// system tray icon under Windows
			int m_DefaultCommand;

		protected:
			// The implementation object for the application
			// (Pattern: Bridge)
			ApplicationImp *m_pImp;

			// Application singleton
			static Application* s_pTheApp;

			// Application has a message map
			DECLARE_VGUI_MESSAGE_MAP()

			// Application options
			AppOptions m_Options;

			// Idle event counter
			uint m_iIdleEvent;

		private:
			// Tell any sub-systems we use to clean their data up
			void CleanUp();

			//
			// Idle events
			//
			// used by IdleHandler class to register for idle events.
			void RegisterIdleEventHandler(IdleHandler* pHandler);
			void UnregisterIdleEventHandler(IdleHandler* pHandler);

			// List of idle handlers
			IdleHandlers m_IdleHandlers;

			// Grant access to idle handler class.
			friend class IdleHandler;

			// List of shutdown handlers
			ShutdownHandlers m_ShutdownHandlers;

			// MRU file list handler, if one exists
			MessageTarget* m_pMRUFileList;
	};

	extern Application* GetApp();

	//
	// Concrete application types.
	//

	class ApplicationSingleWindow : public Application
	{
		public:
			ApplicationSingleWindow();
	};

	class ApplicationDialog : public Application
	{
		public:
			ApplicationDialog();

			// Implement this function to create the correct dialog
			// object for your application.
			virtual Dialog* CreateTheDialog() = 0;

		public:
			virtual bool CreateMainWindowObject();
			virtual bool CreateMainWindow();
	};

};

//
// Macro to simplify declaring which app class to use.
//

#define VGUI_DECLARE_APP(MyAppClass)	\
namespace vgui							\
{										\
	Application *CreateApplication()	\
	{									\
		return new MyAppClass;			\
	}									\
};


#endif // !vgui_Application_h


