//
// AutoPrefs.cpp - simple classes to link preferences to dialog controls.
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "AutoPrefCheckBox.h"

#include "vos\Preferences.h"
#include "vos\Errors.h"

namespace vgui
{

	/**************************************************************************

	>	AutoPrefCheckBox::AutoPrefCheckBox()

		Author:		Tim Browse
		Date:		10/5/2002

		Purpose:	Constructor

	**************************************************************************/

	AutoPrefCheckBox::AutoPrefCheckBox()
	{
	}

	/**************************************************************************

	>	void AutoPrefCheckBox::Attach(Dialog* pDialog,
									  ControlID CheckBoxID,
									  const vos::String& PrefName,
									  bool bDefaultValue = true)


		Author:		Tim Browse
		Date:		10/5/2002
		Inputs:		pDialog - the dialog containing the control
					ID - ID of the checkbox control to link to
					PrefName - name of the preference to link the checkbox
							   to - e.g. "UI.ShowHelp"
					bDefaultValue - the default value to use, if the preference
									does not yet exist.

		Purpose:	Attach the AutoPref object to a checkbox, and initialise
					the checkbox to reflect the current preference.

	**************************************************************************/

	void AutoPrefCheckBox::Attach(Dialog* pDialog,
								  ControlID ID,
								  const vos::String& PrefName,
								  bool bDefaultValue)
	{
		// Attach the checkbox to the control
		m_CheckBox.Attach(pDialog, ID);

		// Tell the dialog about this autopref object
		pDialog->AddAutoPref(this);

		// Read the preference
		m_PrefName = PrefName;
		vos::GetPreferences().Read(m_PrefName, &bDefaultValue);

		// Set checkbox to reflect current preference value
		m_CheckBox = bDefaultValue;
	}

	/**************************************************************************

	>	void AutoPrefCheckBox::CommitPreference()


		Author:		Tim Browse
		Date:		10/5/2002

		Purpose:	Read the current state of the checkbox, and write that
					value back into the named preference value.

	**************************************************************************/

	void AutoPrefCheckBox::CommitPreference()
	{
		// Read the checkbox state, and write it back to the named preference.
		vos::GetPreferences().Write(m_PrefName, m_CheckBox.GetState());
	}


	/**************************************************************************
	
	>	AutoPrefCheckBox& AutoPrefCheckBox::operator=(bool NewState)
	
		Author:		Tim Browse
		Date:		21/5/2002
		Inputs:		NewState - true => check the checkbox
							  false => uncheck the checkbox
		Returns:	Reference to the object
	
		Purpose:	Mimic CheckBox API.

					This allows you to assign a true/false value to a check
					box, and have the checkbox udpate accordingly.
	
	**************************************************************************/
	
	AutoPrefCheckBox& AutoPrefCheckBox::operator=(bool NewState)
	{
		m_CheckBox = NewState;
		return *this;
	}

	/**************************************************************************
	
	>	AutoPrefCheckBox& AutoPrefCheckBox::operator=(ControlState NewState)
	
		Author:		Tim Browse
		Date:		21/5/2002
		Inputs:		NewState - see ControlState typedef
		Returns:	Reference to the object
	
		Purpose:	Mimic CheckBox API.

					This allows you to assign a 'tri-state' value value to a check
					box, and have the checkbox udpate accordingly.
	
	**************************************************************************/
	
	AutoPrefCheckBox& AutoPrefCheckBox::operator=(ControlState NewState)
	{
		m_CheckBox.operator=(NewState);
		return *this;
	}

	/**************************************************************************
	
	>	AutoPrefCheckBox::operator bool()
	
		Author:		Tim Browse
		Date:		21/5/2002
		Returns:	true => check box is currently 'on'
					false => otherwise
	
		Purpose:	Convert a checkbox into a bool value that represents
					its current state.
	
	**************************************************************************/
	
	AutoPrefCheckBox::operator bool()
	{
		return m_CheckBox.GetState();
	}

	/**************************************************************************
	
	>	bool AutoPrefCheckBox::GetState() const
	
		Author:		Tim Browse
		Date:		21/5/2002
		Returns:	true => check box is currently 'on'
					false => otherwise
	
		Purpose:	Convert a checkbox into a bool value that represents
					its current state.
	
	**************************************************************************/
	
	bool AutoPrefCheckBox::GetState() const
	{
		return m_CheckBox.GetState();
	}

	/**************************************************************************
	
	>	ControlState AutoPrefCheckBox::GetTriState() const
	
		Author:		Tim Browse
		Date:		21/5/2002
		Returns:	Current state of the checkbox (see ControlState typedef)
	
		Purpose:	Get the state of a checkbox as a 'tri-state' value.
	
	**************************************************************************/
	
	ControlState AutoPrefCheckBox::GetTriState() const
	{
		return m_CheckBox.GetTriState();
	}
	
	void AutoPrefCheckBox::EnableLinkedControls(Dialog* pParentDialog)
	{
		m_CheckBox.EnableLinkedControls(pParentDialog);
	}

	void AutoPrefCheckBox::AddLinkedControl(ControlID ID)
	{
		m_CheckBox.AddLinkedControl(ID);
	}

	void AutoPrefCheckBox::AddLinkedControls(const ControlID* pIDs)
	{
		m_CheckBox.AddLinkedControls(pIDs);
	}

};
