//
// AutoPrefDropDownList.cpp - simple classes to link preferences to dialog controls.
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "AutoPrefDropDownList.h"

#include "vos\Preferences.h"
#include "vos\Errors.h"

namespace vgui
{
	const char kPref_Item[] = "Item%d";
	const char kPref_ItemCount[] = "ItemCount";
	const char kPref_ItemText[] = "ItemText";

	AutoPrefDropDownList& AutoPrefDropDownList::operator=(const String& ItemName)
	{
		if (!SelectByString(ItemName))
			SetText(ItemName);
		
		return *this;
	}

	AutoPrefDropDownList& AutoPrefDropDownList::operator=(const StringResourceID& ItemName)
	{
		if (!SelectByResID(ItemName))
			SetText(ItemName.Load());

		return *this;
	}

	void AutoPrefDropDownList::Attach(Dialog* pDialog,
									  ControlID ListControlID,
									  const vos::String& PrefName,
									  bool bEnableCommitEvents,
									  bool bEnableHistoryBehaviour)
	{
		// Attach control, and initialise from the preference name provided
		DropDownList::Attach(pDialog, ListControlID);

		// Tell the dialog about this autopref object
		pDialog->AddAutoPref(this);

		// Set control contents from preferences.
		m_PrefName = PrefName;
		RestorePreferences();

		// Set up control according to flags
		if (bEnableCommitEvents)
			EnableCommitNotification();

		if (bEnableHistoryBehaviour)
			EnableHistoryBehaviour();
	}

	void AutoPrefDropDownList::CommitPreference()
	{
		if (HasHistoryBehaviour())
			CommitCurrentTextToHistory();

		vos::Preferences* pPrefSection = vos::GetPreferences().FindSection(m_PrefName);

		if (pPrefSection == NULL)
		{
			COMPLAIN("AutoPrefDropDownList() - Unable to find preference!");
			return;
		}

		// Write out any preferences as children of this section
		int cItems = GetCount();
		vos::String PrefName, ItemText;

		for (int iItem = 0; iItem < cItems; iItem++)
		{
			// Get this item's text
			ItemText.Empty();
			GetItemText(iItem, &ItemText);
			
			// Write the text to the preferences system.
			PrefName.Printf(kPref_Item, iItem);
			pPrefSection->Write(PrefName, ItemText);
		}

		// Save out the number of items, and the text of the currently
		// selected item.
		pPrefSection->Write(kPref_ItemCount, cItems);

		ItemText.Empty();
		GetText(&ItemText);
		pPrefSection->Write(kPref_ItemText, ItemText);
	}

	/**************************************************************************
	
	>	void AutoPrefDropDownList::RestorePreferences()
	
		Author:		Tim Browse
		Date:		8/6/2002
	
		Purpose:	Load the preferences for this control from the registry, 
					and set up the list control to reflect these values.
	
	**************************************************************************/
	
	void AutoPrefDropDownList::RestorePreferences()
	{
		vos::Preferences* pPrefSection = vos::GetPreferences().FindSection(m_PrefName);

		if (pPrefSection == NULL)
		{
			COMPLAIN("AutoPrefDropDownList() - Unable to find preference!");
			return;
		}

		// First, find out how many items are in the list.
		int cItems;
		if (!pPrefSection->Read(kPref_ItemCount, &cItems))
		{
			// No pref available yet.
			return;
		}

		// Now read the items from the preferences, and add them to the list.
		vos::String PrefName, ItemText;

		for (int iItem = 0; iItem < cItems; iItem++)
		{
			// Get this item's text
			ItemText.Empty();
			PrefName.Printf(kPref_Item, iItem);
			if (pPrefSection->Read(PrefName, &ItemText))
			{
				// Add this text to the list
				int iNewItem;
				AddItem(ItemText, &iNewItem);
			}
		}

		// Ok, now set the main text
		ItemText.Empty();
		if (pPrefSection->Read(kPref_ItemText, &ItemText))
		{
			SetText(ItemText);
		}
	}
};
