//
// AutoPrefs.cpp - simple classes to link preferences to dialog controls.
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "AutoPrefRadioGroup.h"

#include "vos\Preferences.h"
#include "vos\Errors.h"

namespace vgui
{
	/**************************************************************************

	>	AutoPrefRadioGroup::AutoPrefRadioGroup()

		Author:		Tim Browse
		Date:		10/5/2002

		Purpose:	Constructor

	**************************************************************************/

	AutoPrefRadioGroup::AutoPrefRadioGroup() : m_pRadioGroup(NULL),
												 m_pRadioButtonIDs(NULL),
												 m_pszPrefValues(NULL),
												 m_cRadioButtons(0)
	{
		// TCILB
	}

	/**************************************************************************

	>	void AutoPrefRadioGroup::Attach(Dialog* pDialog,
										 ControlID* pRadioButtonIDs,
										 const char** szPrefValues,
										 const vos::String& PrefName,
										 const char* szDefaultValue = NULL)


		Author:		Tim Browse
		Date:		10/5/2002
		Inputs:		pDialog - the dialog holdig the radio buttons.
					pRadioButtonIDs - an array of ControlID values, terminated
									  by vgui::NOCTL.
					szPrefValues - an array of value names to use for each
								   radio button, e.g. if the 3rd radio button
								   is selected, the third item in this array
								   is written into the user preference.
					PrefName - name of the user preference to use.
					szDefaultValue - the value to use if the preference does
									 not yet exist.  (If NULL, the first
									 radio button is assumed to be the default.)

		Purpose:	Attach the AutoPref to a set of radio buttons, and link
					it to a named preference.

					The radio buttons are passed in as an array of ControlID
					values, and an array of textual values is passed in, each
					one used for the corresponding radio button.

	**************************************************************************/

	void AutoPrefRadioGroup::Attach(Dialog* pDialog,
									ControlID* pRadioButtonIDs,
									const char* pszPrefValues[],
									const vos::String& PrefName,
									const char* szDefaultValue)
	{
		// Sanity checks
		COMPLAIN_IF(pDialog == NULL, "Attach() - NULL dialog pointer!");
		COMPLAIN_IF(pRadioButtonIDs == NULL, "Attach() - NULL radio button IDs!");
		COMPLAIN_IF(pszPrefValues == NULL, "Attach() - NULL preference values!");

		// Tell the dialog about this autopref object
		pDialog->AddAutoPref(this);

		// Work out how many buttons are to be used
		uint iButton = 0;

		while (pRadioButtonIDs[iButton] != NOCTL)
		{
			iButton++;
		}

		// Ok, remember this info
		m_cRadioButtons = iButton;
		m_pRadioButtonIDs = pRadioButtonIDs;
		m_pszPrefValues = pszPrefValues;
		m_PrefName = PrefName;

		// Now Create radio group
		m_pRadioGroup = new RadioGroup(pDialog, pRadioButtonIDs);

		// Read initial preference state
		vos::String PrefValue;
		if (szDefaultValue != NULL)
			PrefValue = szDefaultValue;

		vos::GetPreferences().Read(m_PrefName, &PrefValue);

		// And set radio buttons accordingly (default to first one)
		ControlID ButtonToSet = m_pRadioButtonIDs[0];

		iButton = 0;
		while (iButton < m_cRadioButtons)
		{
			if (PrefValue.CompareText(m_pszPrefValues[iButton]))
			{
				// Match found
				ButtonToSet = m_pRadioButtonIDs[iButton];
				break;
			}

			++iButton;
		}

		// Set the radio button
		m_pRadioGroup->SetState(ButtonToSet);
	}

	/**************************************************************************

	>	void AutoPrefRadioGroup::CommitPreference()

		Author:		Tim Browse
		Date:		10/5/2002

		Purpose:	Read the current preference state from the radio button
					controls, and write this value to the user preferences.

	**************************************************************************/

	void AutoPrefRadioGroup::CommitPreference()
	{
		// Ok, search for current state in our list of IDs
		uint iButton = 0;

		ControlID CurrentState = m_pRadioGroup->GetState();

		while (iButton < m_cRadioButtons)
		{
			if (m_pRadioButtonIDs[iButton] == CurrentState)
			{
				// Found it!
				vos::String PrefValue(m_pszPrefValues[iButton]);
				vos::GetPreferences().Write(m_PrefName, PrefValue);

				// All done
				return;
			}

			++iButton;
		}

		// Preference value not found - is radio control ID missing from
		// the radio group IDs?
		COMPLAIN("CommitPreference() - unknown radio button ID!");
	}

};
