//
// Dialog.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Dialog.h"

#include "DialogImp.h"

#include "MessageMap.h"
#include "IAutoPref.h"
#include "DynamicControlHandler.h"
#include "Errors.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

//
// If IDOK/IDCANCEL have not been defined yet, then define them.
//
// NB. These are standard Win32 values - other platforms may
//     use different values.
//
#ifndef IDOK
#define IDOK                1
#define IDCANCEL            2
#endif

namespace vgui
{

	BEGIN_VGUI_MESSAGE_MAP(Dialog, Window)
		ON_VGUI_BN_CLICKED(IDOK, OnOK)
		ON_VGUI_BN_CLICKED(IDCANCEL, OnCancel)
	END_VGUI_MESSAGE_MAP()

	Dialog::Dialog(const DialogResourceID& ID, 
				   Window *pParent) : m_DialogID(ID),
									  Window(TheFactory().MakeDialog())
	{
		// Link back to this dialog
		m_pWindowImp->SetWindowLink(this);

		// Remember parent
		m_pParent = pParent;
	}

	Dialog::~Dialog()
	{
	}

	ControlID Dialog::DoModal(WindowPos Pos)
	{
		return GetImp()->DoModalDialog(m_DialogID, m_pParent, Pos);
	}

	void Dialog::DoModeless(WindowPos Pos, TShowWindow Show)
	{
		GetImp()->DoModelessDialog(m_DialogID, m_pParent, Pos, Show);
	}

	bool Dialog::Create()
	{
		return GetImp()->Create(m_DialogID, m_pParent);
	}

	void Dialog::EndDialog(ControlID ID)
	{
		GetImp()->EndDialog(ID);
	}

	void Dialog::OnOK()
	{
		// User clicked OK, so we update any AutoPrefs
		CommitAutoPreferences();

		EndDialog(IDOK);
	}

	void Dialog::OnCancel()
	{
		EndDialog(IDCANCEL);
	}

	void Dialog::OnPreInitDialog()
	{
		// Get the minimum size for this dialog - by default it is assumed
		// to be the size specified in the dialog template.
		//
		// We do this before OnInitDialog() has been called, in case the
		// client resizes the dialog during initialisation.
		vgui::Rect Pos;
		GetClientArea(&Pos);
		m_InitialSize.m_Width = Pos.GetWidth();
		m_InitialSize.m_Height = Pos.GetHeight();
	}

	void Dialog::OnInitDialog()
	{
		// ILB
	}

	void Dialog::OnPostInitDialog()
	{
		// ILB
	}

	void Dialog::AlterInitialSizeBy(int WidthDelta, int HeightDelta)
	{
		m_InitialSize.m_Width += WidthDelta;
		m_InitialSize.m_Height += HeightDelta;
	}
	
	void Dialog::MoveControl(ControlID ID, int x, int y)
	{
		GetImp()->MoveControl(ID, x, y);
	}

	void Dialog::MoveControlRelative(ControlID ID, int DeltaX, int DeltaY)
	{
		GetImp()->MoveControlRelative(ID, DeltaX, DeltaY);
	}

	void Dialog::ResizeControl(ControlID ID, int Width, int Height)
	{
		GetImp()->ResizeControl(ID, Width, Height);

		// In case this is a dynamically resizing/moving control, we
		// let the dynamic control data structures know about the change.
		ResizeDynamicControl(ID, Width, Height);
	}

	void Dialog::EnableControl(ControlID ID)
	{
		GetImp()->EnableControl(ID);
	}

	void Dialog::DisableControl(ControlID ID)
	{
		GetImp()->DisableControl(ID);
	}

	void Dialog::ShowControl(ControlID ID)
	{
		GetImp()->ShowControl(ID);
	}

	void Dialog::HideControl(ControlID ID)
	{
		GetImp()->HideControl(ID);
	}

	void Dialog::EnableControlGroup(const ControlID* pIDs)
	{
		while (*pIDs != NOCTL)
			EnableControl(*pIDs++);
	}

	void Dialog::DisableControlGroup(const ControlID* pIDs)
	{
		while (*pIDs != NOCTL)
			DisableControl(*pIDs++);
	}

	void Dialog::ShowControlGroup(const ControlID* pIDs)
	{
		while (*pIDs != NOCTL)
			ShowControl(*pIDs++);
	}

	void Dialog::HideControlGroup(const ControlID* pIDs)
	{
		while (*pIDs != NOCTL)
			HideControl(*pIDs++);
	}

	void Dialog::GetControlPosition(ControlID ID, Rect* pPosition)
	{
		GetImp()->GetControlPosition(ID, pPosition);
	}

	void Dialog::SetControlPosition(ControlID ID, const Rect& Position)
	{
		GetImp()->SetControlPosition(ID, Position);
	}

	void Dialog::SetControlState(ControlID ID, ControlState NewState)
	{
		GetImp()->SetControlState(ID, NewState);
	}

	ControlState Dialog::GetControlState(ControlID ID)
	{
		return GetImp()->GetControlState(ID);
	}

	bool Dialog::IsOn(ControlID ID)
	{
		return (GetControlState(ID) == cs_Checked);
	}

	bool Dialog::IsOff(ControlID ID)
	{
		return (GetControlState(ID) != cs_Checked);
	}

	void Dialog::TurnOn(ControlID ID)
	{
		SetControlState(ID, cs_Checked);
	}

	void Dialog::TurnOff(ControlID ID)
	{
		SetControlState(ID, cs_Unchecked);
	}

	bool Dialog::OnResizing(vgui::Size* pNewSize)
	{
		if (m_DynamicWindows.empty())
		{
			// No dynamic controls, so we let the dialog resize
			// go ahead
			return false;
		}

		// Veto sizes that are too small (i.e. that would make the dialog
		// smaller than its original size).
		pNewSize->m_Width = max(pNewSize->m_Width,  m_InitialSize.m_Width);
		pNewSize->m_Height= max(pNewSize->m_Height, m_InitialSize.m_Height);
		return true;
	}

	bool Dialog::FindControlWindow(ControlID ID, Window** ppWindow)
	{
		return GetImp()->FindControlWindow(ID, ppWindow);
	}

	bool Dialog::OnResize(int NewWidth, int NewHeight)
	{
		if (m_DynamicWindows.size() > 0)
		{
			DynamicWindows::iterator it = m_DynamicWindows.begin();

			while (it != m_DynamicWindows.end())
			{
				Window *pControlWindow;
				
				if (FindControlWindow(it->m_ID, &pControlWindow))
				{
					pControlWindow->OnControlResizing();
				}

				++it;
			}

			GetImp()->LockControlUpdate();

			// Work out the size deltas
			int DeltaX = NewWidth - m_InitialSize.m_Width;
			int DeltaY = NewHeight - m_InitialSize.m_Height;

			// If we have any dynamic controls, update them
			it = m_DynamicWindows.begin();

			while (it != m_DynamicWindows.end())
			{
				// How should we update the position?
				vgui::Rect Pos = it->m_InitialPosition;

				if (it->m_XBehaviour == ControlResize_Move)
				{
					Pos.Translate(DeltaX, 0);
				}
				else if (it->m_XBehaviour == ControlResize_MoveCentre)
				{
					// We move it, but by Delta/2, so that a centred control
					// remains centred
					Pos.Translate(DeltaX/2, 0);
				}
				else if (it->m_XBehaviour == ControlResize_Stretch)
				{
					Pos.right += DeltaX;
				}

				if (it->m_YBehaviour == ControlResize_Move)
				{
					Pos.Translate(0, DeltaY);
				}
				else if (it->m_YBehaviour == ControlResize_MoveCentre)
				{
					// We move it, but by Delta/2, so that a centred control
					// remains centred
					Pos.Translate(0, DeltaY/2);
				}
				else if (it->m_YBehaviour == ControlResize_Stretch)
				{
					Pos.bottom += DeltaY;
				}

				// Ok, update the position (if it has moved)
				vgui::Rect CurrentPos;
				if (it->IsControl())
				{
					GetControlPosition(it->m_ID, &CurrentPos);

					if (Pos != CurrentPos)
					{
						SetControlPosition(it->m_ID, Pos);
					}
				}
				else
				{
					it->m_pWindow->GetPosition(&CurrentPos);

					if (Pos != CurrentPos)
					{
						it->m_pWindow->Move(Pos);
					}
				}

				// Do the next dynamic window
				++it;
			}

			GetImp()->UnlockControlUpdate();

			it = m_DynamicWindows.begin();

			while (it != m_DynamicWindows.end())
			{
				Window *pControlWindow;
				
				if (FindControlWindow(it->m_ID, &pControlWindow))
				{
					pControlWindow->OnControlResize();
				}

				++it;
			}
		}

		// Allow normal processing; don't claim message
		return false;
	}

	void Dialog::AddDynamicControl(ControlID ID, 
								   TControlResize XBehaviour,
								   TControlResize YBehaviour)
	{
		// Create record for this control
		TDynamicWindow Info;
		Info.m_ID = ID;
		Info.m_pWindow = NULL;
		Info.m_XBehaviour = XBehaviour;
		Info.m_YBehaviour = YBehaviour;
		GetControlPosition(ID, &Info.m_InitialPosition);

		// Allow dynamic control handlers to modify/veto this request
		for (DynamicControlHandlers::iterator it = m_DynamicControlHandlers.begin();
			 it != m_DynamicControlHandlers.end();
			 ++it)
		{
			if (!(*it)->AddDynamicControl(this, &Info))
			{
				// Request was vetoed
				return;
			}
		}

		// Ok, request was allowed (although it may have been modified)
		// so we go ahead and add the control to our list
		m_DynamicWindows.push_back(Info);
	}

	/**************************************************************************
	
	>	void Dialog::AddDynamicControls(DynamicControlInfo* pControls)
	
		Author:		Tim Browse
		Date:		22/8/2002
		Inputs:		pControls - the table of information about which controls
								to set up as dynamic controls in this dialog.
								(NB. last entry must have ID of NOCTL, or this
								function will crash).
	
		Purpose:	Add a number of dynamic controls to the dialog.

					This function just makes it simpler to add dynamic
					controls to a dialog.  See DynamicControls.h for some
					macros that make it easy to build a table of control
					information to pass to this function.

					NB. You must place a sentinel entry at the end of the
						table with a control ID of NOCTL, or this function
						will crash.

						Just use the macros from DynamicControls.h to be safe.

	**************************************************************************/
	
	void Dialog::AddDynamicControls(DynamicControlInfo* pControls)
	{
		while (pControls->m_ID != NOCTL)
		{
			AddDynamicControl(pControls->m_ID,
							  pControls->m_XBehaviour,
							  pControls->m_YBehaviour);

			pControls++;
		}
	}

	void Dialog::AddDynamicWindow(Window* pWindow,
								  TControlResize XBehaviour,
								  TControlResize YBehaviour)
	{
		COMPLAIN("AddDynamicWindow() - this function has not been tested yet, "
				 "so do not use it.");
		return;
					
		// Create record for this window
		TDynamicWindow Info;
		Info.m_ID = NOCTL;
		Info.m_pWindow = pWindow;
		Info.m_XBehaviour = XBehaviour;
		Info.m_YBehaviour = YBehaviour;
		pWindow->GetPosition(&Info.m_InitialPosition);

		// Add to our list
		m_DynamicWindows.push_back(Info);
	}

	/**************************************************************************
	
	>	void Dialog::ResizeDynamicControl(ControlID ID, int Width, int Height)
	
		Author:		Tim Browse
		Date:		3/5/2002
		Inputs:		ID - ID of the control that is being resized.
					Width, Height - the new size of the control.
	
		Purpose:	When a control is moved with ResizeControl(), this function
					checks to see if it has been registered as a dynamic
					control.  If so, then the dynamic control record is udpated
					with the control's new size, so it continues to be
					resized correctly.

					The control being resized is not necessarily dynamic - we
					check the dialog data to see if it is - if it isn't, this
					function has no effect.
	
	**************************************************************************/
	
	void Dialog::ResizeDynamicControl(ControlID ID, int Width, int Height)
	{
		if (m_DynamicWindows.size() > 0)
		{
			// If we have any dynamic controls, update them
			DynamicWindows::iterator it = m_DynamicWindows.begin();

			while (it != m_DynamicWindows.end())
			{
				if (it->IsControl() && (it->m_ID == ID))
				{
					// Matching dynamic control exists - update
					// original position accordingly
					it->m_InitialPosition.SetWidth(Width);
					it->m_InitialPosition.SetHeight(Height);

					// All done
					return;
				}

				// Check next window for match
				++it;
			}
		}
	}

	/**************************************************************************
	
	>	void Dialog::RegisterDynamicControlHandler(DynamicControlHandler* pHandler)
	
		Author:		Tim Browse
		Date:		25/7/2002
		Inputs:		The handler to register
	
		Purpose:	Register a handler for controlling dynamic dialog controls.
					This handler will be called (along with any other 
					registered handlers) when dynamic controls are
					processed.
	
	**************************************************************************/
	
	void Dialog::RegisterDynamicControlHandler(DynamicControlHandler* pHandler)
	{
		m_DynamicControlHandlers.insert(pHandler);
	}

	/**************************************************************************
	
	>	void Dialog::UnregisterDynamicControlHandler(DynamicControlHandler* pHandler)
	
		Author:		Tim Browse
		Date:		25/7/2002
		Inputs:		The handler to unregister
	
		Purpose:	Unregister a handler for controlling dynamic dialog controls.
	
	**************************************************************************/
	
	void Dialog::UnregisterDynamicControlHandler(DynamicControlHandler* pHandler)
	{
		m_DynamicControlHandlers.erase(pHandler);
	}

	/**************************************************************************
	
	>	bool Dialog::GetDialog(Dialog** ppDialog)
	
		Author:		Tim Browse
		Date:		25/7/2002
		Outputs:	ppDialog - the dialog pointer
		Returns:	true => the window is a Dialog object, and ppDialog contains
							a pointer to the dialog.
					false => this window is not a dialog.
	
		Purpose:	Access this window object as a dialog, if it is actually
					a dialog.

					This saves you from doing fiddly checks and casts etc.
	
	**************************************************************************/
	
	bool Dialog::GetDialog(Dialog** ppDialog)
	{
		// This is a dialog
		*ppDialog = this;
		return true;
	}

	/**************************************************************************
	
	>	void Dialog::AddAutoPref(IAutoPref* pAutoPref)
	
		Author:		Tim Browse
		Date:		10/5/2002
		Inputs:		pAutoPref - the new AutoPref to link to this dialog.
	
		Purpose:	Tell the Dialog that an AutoPref object has been
					linked to it.  This means this AutoPref will be asked
					to commit its preference when CommitPreferences() is
					called.
	
	**************************************************************************/
	
	void Dialog::AddAutoPref(IAutoPref* pAutoPref)
	{
		m_AutoPrefs.push_back(pAutoPref);
	}

	/**************************************************************************
	
	>	void Dialog::CommitAutoPreferences()
	
		Author:		Tim Browse
		Date:		10/5/2002
	
		Purpose:	Get all AutoPref objects linked to this dialog (via
					AddAutoPref()) to commit their preference values to the
					user preferences system.

					Called by the default implementation of OnOK() - but
					add anywhere else you need the preferences to be
					committed.
	
	**************************************************************************/
	
	void Dialog::CommitAutoPreferences()
	{
		AutoPrefs::iterator itPref = m_AutoPrefs.begin();

		while (itPref != m_AutoPrefs.end())
		{
			(*itPref)->CommitPreference();
			++itPref;
		}
	}

	bool Dialog::IsDynamicDialog() const
	{
		return !m_DynamicWindows.empty();
	}

	void Dialog::DisableAutoResizeGripper()
	{
		GetImp()->DisableAutoResizeGripper();
	}

};

