//
// Dialog.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_Dialog_h
#define vgui_Dialog_h

#include "Window.h"
#include "Controls.h"
#include "DynamicControls.h"
#include "Rect.h"

#include <set>

namespace vgui
{
	class Font;
	class IAutoPref;
	class DynamicControlHandler;

	// List of dynamic control handlers
	typedef std::set<DynamicControlHandler*> DynamicControlHandlers;

	// List of auto prefs
	typedef std::vector<IAutoPref*> AutoPrefs;

	class Dialog : public Window
	{
		public:
			DECLARE_VGUI_MESSAGE_MAP()

		public:
			Dialog(const DialogResourceID& ID, Window *pParent = NULL);
			~Dialog();

		public:
			ControlID DoModal(WindowPos Pos = wp_Centred);

			// NB. Modeless dialogs currently delete themselves when closed.
			void DoModeless(WindowPos Pos = wp_Centred, TShowWindow Show = sw_SHOW);

			// Just create the dialog - mainly used for child dialogs.
			bool Create();

			void EndDialog(ControlID);

			//
			// Lightweight access to controls in the dialog - you don't need
			// to create a control object to perform these actions.
			//
			void MoveControl(ControlID ID, int x, int y);
			void MoveControlRelative(ControlID ID, int DeltaX, int DeltaY);
			void ResizeControl(ControlID ID, int Width, int Height);

			void EnableControl(ControlID ID);
			void DisableControl(ControlID ID);
			void ShowControl(ControlID ID);
			void HideControl(ControlID ID);

			// Terminate the array of control IDs with NOCTL
			void EnableControlGroup(const ControlID* pIDs);
			void DisableControlGroup(const ControlID* pIDs);
			void ShowControlGroup(const ControlID* pIDs);
			void HideControlGroup(const ControlID* pIDs);

			void GetControlPosition(ControlID ID, Rect* pPosition);
			void SetControlPosition(ControlID ID, const Rect& Position);
			void SetControlState(ControlID ID, ControlState NewState);
			ControlState GetControlState(ControlID ID);
			bool FindControlWindow(ControlID ID, Window** ppWindow);

			bool IsOn(ControlID ID);
			bool IsOff(ControlID ID);

			void TurnOn(ControlID ID);
			void TurnOff(ControlID ID);

			bool GetDialog(Dialog** ppDialog);
			
			//
			// Simple facility for moving/resizing controls when a dialog is resized.
			//
			void AddDynamicControl(ControlID ID, 
								   TControlResize XBehaviour,
								   TControlResize YBehaviour);

			// Add a table of controls - last entry in table should have
			// control ID of NOCTL.
			void AddDynamicControls(DynamicControlInfo* pControls);

			void AddDynamicWindow(Window* pWindow,
								  TControlResize XBehaviour,
								  TControlResize YBehaviour);

			bool IsDynamicDialog() const;

			// Call this if you don't want the Dialog manager to automatically
			// add a resize gripper control to resizable dialogs
			void DisableAutoResizeGripper();

			// Adjust the initial size of the dialog - used when adding controls
			// to a esizable dialog, and we don't want to permit it to be
			// smaller than this size.
			void AlterInitialSizeBy(int WidthDelta, int HeightDelta);

			// Facility to allow objects to modify behaviour of the dynamic controls
			void RegisterDynamicControlHandler(DynamicControlHandler* pHandler);
			void UnregisterDynamicControlHandler(DynamicControlHandler* pHandler);

			// 
			// Automatic preference management
			//
			void AddAutoPref(IAutoPref* pAutoPref);
			void CommitAutoPreferences();

			// Called by DialogImp object to init the dialog
			void OnPreInitDialog();
			void OnPostInitDialog();

		public:
			// Message handlers
			virtual void OnOK();
			virtual void OnCancel();
			virtual void OnInitDialog();

			virtual bool OnResizing(vgui::Size* pNewSize);
			virtual bool OnResize(int NewWidth, int NewHeight);

		public:
			DialogImp* GetImp() const { return (DialogImp*) m_pWindowImp; }

		protected:
			// Resource of the dialog
			DialogResourceID m_DialogID;

		private:
			// Used to keep dynamic control records up to date when a
			// control is resized with ResizeControl().
			void ResizeDynamicControl(ControlID ID, int Width, int Height);

		private:
			// The initial size for the dialog (based on initial size
			// specified by dialog template).
			vgui::Size m_InitialSize;

			// Array of dynamically positioned/resized controls/windows (if any)
			DynamicWindows m_DynamicWindows;

			// List of registered dynamic control handlers
			DynamicControlHandlers m_DynamicControlHandlers;

			// Array of auto prefs (if any)
			AutoPrefs m_AutoPrefs;
	};
};

#endif // !vgui_Dialog_h
