//
// Document.cpp - basic document class for vgui-based document-centric apps.
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Document.h"

#include "View.h"
#include "Errors.h"
#include "SaveDialog.h"
#include "Application.h"
#include "Msgs.h"
#include "Question.h"

#include "vos\DiskFile.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	// Number used to generate numbered titles for documents.
	int Document::s_nDocumentNumber = 1;

	Document::Document() : m_bModified(false)
	{
	}

	Document::~Document()
	{
	}

	bool Document::Init()
	{
		// Ensure document classes are set up
		SetupDocumentClass();

		// All done
		return true;
	}

	void Document::SetModified()
	{
		m_bModified = true;
	}

	void Document::ClearModified()
	{
		m_bModified = false;
	}

	// Add a new view to this object.  The document is responsible for
	// initialising the view object and providing linkage between the
	// objects.
	void Document::AddView(View* pNewView)
	{
		// Initialise the new view.
		pNewView->Initialise(this);

		// Add this view to our list
		m_Views.push_back(pNewView);
	}

	bool Document::CloseAllViews()
	{
		// Repeatedly get the first view and try to close it.
		// When we run out of views, or a view refuses to close, we stop.
		while (m_Views.size() > 0)
		{
			// Close first view
			View *pView = m_Views.front();

			if (!OnCloseView(pView))
			{
				// View did not close - stop trying to close views, and
				// tell caller that we didn't close all views.
				return false;
			}
		}

		// If we get here, we closed all the views.
		// NB. Caller should delete this object/remove it from the active documents list.
		return true;
	}

	bool Document::OnCloseView(View *pViewToDelete)
	{
		// Check this is one of ours :-)
		Views::iterator it = m_Views.begin();

		while (it != m_Views.end())
		{
			// is this a match?
			if (*it == pViewToDelete)
			{
				// Yes, so we're done searching.
				break;
			}

			// Try next view
			++it;
		}

		// Did we find this view in our list?
		if (it == m_Views.end())
		{
			// No - complain and return
			COMPLAIN("Document::OnCloseView() was asked to close a view that "
					 "it did not own.");
			return false;
		}

		// If this is the last view open onto the document, then we must close the
		// document.
		if (m_Views.size() == 1)
		{
			// Has the document been modified?
			m_bModified = true;
			if (m_bModified)
			{
				// Yes, so don't allow it to be closed without first asking
				// the user if they want to save changes.
				String ConfirmMsg;
				ConfirmMsg.Printf(msgDocument_SaveChanges, m_Title);
				const char *szButtons[] = 
				{
					msg_Yes,
					msg_No,
					msg_Cancel,
					0
				};

				Question SaveChangesQuestion(msgDocument_SaveChangesTitle,
											 ConfirmMsg.c_str(), 
											 pViewToDelete);

				int iReply = SaveChangesQuestion.Ask(szButtons);

				if (szButtons[iReply] == msg_Cancel)
				{
					// Cancel - don't close the view.
					return false;
				}

				if (szButtons[iReply] == msg_Yes)
				{
					// User wants to save the document, so save it.
					OnSave();

					// If that failed somehow (signified by the modified flag
					// still being set) then don't close the view.
					if (m_bModified)
					{
						return false;
					}
				}

				// If we get here, then it's ok to close the document...
			}
		}

		// Remove this from our list
		m_Views.erase(it);

		// Destroy the view.
		delete pViewToDelete;

		// The view was closed
		return true;
	}

	void Document::AutoAssignTitle(Document *pPrototypeDocument)
	{
		// If document has no title, give it one automatically.
		if (m_Title.IsEmpty())
		{
			// Use prototype's title as a template
			m_Title = pPrototypeDocument->GetTitle();
			
			// If no title, just fall-back to "Document"
			if (m_Title.IsEmpty())
			{
				m_Title = msgDocument_DefaultName;
			}

			// Add a distinguishing number...
			m_Title += String(s_nDocumentNumber++);
		}
	}

	bool Document::OnSave()
	{
		if (m_FileSpec.IsEmpty())
		{
			// No filename for this document yet - use SaveAs() instead.
			return OnSaveAs();
		}

		// Ok, try to save the document to its current filename
		if (!Save(m_FileSpec))
		{
			// Something went wrong with the save - tell the user.
			REPORT_ERROR();
		}

		// All done - message handled.
		return true;
	}

	bool Document::OnSaveAs()
	{
		// Create a "Save As" dialog...
		FileDialogParams Params;
		SaveDialog TheSaveDialog(&Params, Application::GetMainWindow());

		// Show files of this document type in the dialog.
		Params.m_FileClasses.push_back(m_Class);

		// Show it and ask user for the save filename...
		FileSpec DocumentFilename;
		if (TheSaveDialog.GetFilename(&DocumentFilename))
		{
			// Filename obtained successfully - so save this document to specified file.
			if (!Save(DocumentFilename))
			{
				// An error occured - report it to the user
				REPORT_ERROR();
			}
		}
		else
		{
			// user cancelled operation
		}

		// Event handled
		return true;
	}

	Error Document::Save(const vos::FileSpec& DocumentFilename)
	{
		// Try to open the specified file
		vos::DiskFile DocumentFile;

		bool bSuccess = false;

		if (DocumentFile.Open(DocumentFilename, vos::fam_Write))
		{
			// File opened ok, so make sure this filename sticks, in case
			// the file save fails, so that the user can try to fix the
			// problem (e.g. free up disk space) and hit save again without
			// having to specify the filename again.
			m_FileSpec = DocumentFilename;

			// Also, change document title to reflect new filename
			SetTitle(DocumentFilename.GetLeafname());

			// Ok, write data to the file
			Error Result = SaveData(DocumentFile);

			// If it worked, then clear the modify flag
			if (Result == SUCCESS)
			{
				ClearModified();
			}

			// Close file and return result to caller
			DocumentFile.Close();
			return Result;
		}
		else
		{
			// Failed to open the file for writing - tell the user.
			return LASTERROR;
		}
	}

	void Document::SetTitle(const String& NewTitle)
	{
		// Update our title
		m_Title = NewTitle;

		// Tell all our new views about the title change.
		Views::iterator it = m_Views.begin();

		while (it != m_Views.end())
		{
			// Tell this view to update its title from the document
			(*it)->NotifyDocumentTitleChanged();

			// Try next view
			++it;
		}
	}

};
