//
// Document.h - basic document class for vgui-based document-centric apps.
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_Document_h
#define vgui_Document_h

#include "vgui.h"

#include "String.h"
#include "MessageTarget.h"

#include "vos\FileSpec.h"

#include "DocumentClass.h"
#include "View.h"

#include <vector>

namespace vgui
{
	class Document : public MessageTarget
	{
		public:
			Document();
			virtual ~Document();

			// Initialise a document.
			virtual bool Init();

		public:
			String GetTitle() const { return m_Title; }
			void SetTitle(const String& NewTitle);

			DocumentClass GetClass() const { return m_Class; }

			// Used by the DocumentApplication class to determine when
			// to destroy a document.
			int GetNumberOfViews() const { return m_Views.size(); }

			//
			// Document save handling...
			//

			// GUI handlers
			virtual bool OnSave();
			virtual bool OnSaveAs();

			// Function to actually save the document to a file.
			virtual Error Save(const vos::FileSpec& DocumentFilename);

			// You must implement this function in derived Document classes
			// to save your document's data.
			virtual Error SaveData(vos::File& DocumentFile) = 0;

			// Modification handling
			void SetModified();
			void ClearModified();

		protected:
			// Basic attributes - document title and filename
			String m_Title;
			vos::FileSpec m_FileSpec;

			// Class of document - you should initialise this in your 
			// derived class's constructor.
			DocumentClass m_Class;

			// true if the document has been modified
			bool m_bModified;

		public:
			// Used for Prototype pattern - every time a new document is needed,
			// the Application's prototype document is cloned.
			virtual Document* ClonePrototype() = 0;

			// Called for each document - the document class type and class
			// description must be set up in this function.
			virtual void SetupDocumentClass() = 0;

			// Add a new view to this object.  The document is responsible for
			// initialising the view object and providing linkage between the
			// objects.
			virtual void AddView(View* pNewView);

			// Delete a view from this object - the view is disposed of by
			// this function.  If it is the last view, the document deletes itself.
			// Returns true if the view was actually closed; false otherwise (e.g. user
			// chose not to close an unsaved document).
			virtual bool OnCloseView(View *pViewToDelete);

			// Close all views on this document.
			// NB. It is caller's responsilbity to then delete this Document object.
			bool CloseAllViews();

			// Handy way of providing numbered document titles.
			void AutoAssignTitle(Document *pPrototypeDocument);

		protected:
			// The views currently displaying this document
			// (usually only one per document)
			Views m_Views;

			static int s_nDocumentNumber;

	};

	typedef std::vector<Document*> Documents;
};

#endif // !vgui_Document_h
