//
// DocumentApplication.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "DocumentApplication.h"

#include "Document.h"
#include "View.h"
#include "DocAppWindow.h"

#include "Errors.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	using namespace vos;

	// The document/view templates.  This is one or more pairs of
	// document and view prototype objects.  These objects are cloned
	// whenever we need to create a new document/view pair.
	//
	// When there is more than one doc/view pair, the user is given
	// a choice as to what type of document they want to create.
	//
	// These objects are never used other than as prototypes.
	//
	DocViewTemplates DocumentApplication::s_DocViewTemplates;

	DocumentApplication::DocumentApplication() : m_pActiveView(NULL)
	{
	}

	DocumentApplication::~DocumentApplication()
	{
	}

	bool DocumentApplication::OnClose()
	{
		// User is trying to close the application - close all views currently 
		// open on our active documents
		Documents::iterator it = m_Documents.begin();

		while (it != m_Documents.end())
		{
			// Close all views on this document (which will auto delete
			// the document).
			if (!(*it)->CloseAllViews())
			{
				// Failed to close all views on this document; the user has probably
				// chosen to cancel closing the app, so stop now, and tell caller that
				// we've handled this message.
				return true;
			}

			// Do next document
			++it;
		}

		// Closed all views on all documents without problems, so tell caller
		// to handle this message in the usual way - ie. let it close this window.
		return false;
	}

	bool DocumentApplication::CreateMainWindowObject()
	{
		try
		{
			m_pMainWindow = new DocAppWindow(this);
			return true;
		}
		catch (std::bad_alloc)
		{
			// Problem creating main window
			return false;
		}
	}

	void DocumentApplication::NotifyViewActivated(View* pActiveView)
	{
		m_pActiveView = pActiveView;
	}

	void DocumentApplication::NotifyViewDeactivated(View* pActiveView)
	{
		// If the view is the current active view, deregister it.
		if (pActiveView == m_pActiveView)
		{
			m_pActiveView = NULL;
		}
		else
		{
			// If we get here, then the active view is out of sync.
			COMPLAIN("A non-active view was deactivated.");
		}
	}

	bool DocumentApplication::OnCommand(int nCommand)
	{
		//
		// Route this command in this order:
		//
		// 1. The active view
		// 2. The active document
		// 3. Defer to standard application routine
		//

		if (m_pActiveView != NULL)
		{
			// Try the active view then the document
			if (m_pActiveView->OnCommand(nCommand) ||
				m_pActiveView->GetDocument()->OnCommand(nCommand))
			{
				// Active view or document handled command
				return true;
			}
		}

		// Send command through standard application message route.
		return Application::OnCommand(nCommand);
	}

	bool DocumentApplication::OnNewDocument()
	{
		// If we don't have a prototype document, things have gone wrong!
		if (s_DocViewTemplates.size() == 0)
		{
			COMPLAIN("The application does not have any Document/View templates. "
					 "Are you sure you have added any in your Application's "
					 "CreateDocViewTemplates() function?");
			return true;
		}

		// Get the template we want to use - default to the first for now.
		// TODO: Allow user to choose the template (document type) that they
		//       want.
		DocViewTemplate *pTemplate = *(s_DocViewTemplates.begin());

		// Get the prototype document
		Document* pPrototypeDocument = pTemplate->GetDocument();

		// Make a new document
		Document* pNewDocument = pPrototypeDocument->ClonePrototype();

		// Initialise it
		if (!pNewDocument->Init())
		{
			// Initialisation failed - clean up and tell caller
			delete pNewDocument;
			return false;
		}

		// Make sure it has a title
		pNewDocument->AutoAssignTitle(pPrototypeDocument);

		// Add it to our list of active documents
		m_Documents.push_back(pNewDocument);

		// Make a view for it
		View* pNewView = pTemplate->GetView()->ClonePrototype();

		// And attach it to the document
		pNewDocument->AddView(pNewView);

		// All done
		return true;
	}

	void DocumentApplication::AddDocViewTemplate(Document* pDocumentPrototype,
												 View* pViewPrototype)
	{
		// Initialise the Document object
		pDocumentPrototype->Init();

		// Create a new document/view template.
		DocViewTemplate* pNewTemplate = new DocViewTemplate(pDocumentPrototype,
															pViewPrototype);

		// Add it to our list.
		s_DocViewTemplates.push_back(pNewTemplate);
	}

	bool DocumentApplication::OnCloseView(View* pViewToClose)
	{
		// Ask document to close this view
		Document *pDoc = pViewToClose->GetDocument();

		if (pDoc->OnCloseView(pViewToClose))
		{
			// View was closed...

			// If this view was our currently selected view (which it virtually
			// always is), then clear our selected view pointer.
			if (m_pActiveView == pViewToClose)
				m_pActiveView = NULL;

			// Does this document have any views left?
			if (pDoc->GetNumberOfViews() == 0)
			{
				// No - so get rid of it.
				Documents::iterator it = m_Documents.begin();

				while (it != m_Documents.end())
				{
					// Is this the document?
					if (*it == pDoc)
					{
						// Yes, so lose it
						m_Documents.erase(it);
						break;
					}

					// Try next document in the list
					++it;
				}

				// Destroy the document
				delete pDoc;
			}

			// All done - view was closed.
			return true;
		}
		else
		{
			// View was not closed (possibly unsaved document, etc.)
			return false;
		}
	}

	ApplicationMDI::ApplicationMDI()
	{
	}

	void ApplicationMDI::Init()
	{
		// Create the platform dependent application implementation.
		m_pImp = TheFactory().MakeApplicationMDI();

		// Create the document/view templates for this MDI app
		CreateDocViewTemplates();
	}


}; // namespace vgui


