//
// DynamicControls.h - types to manage dynamic dialog controls
//

#ifndef vgui_DynamicControls_h
#define vgui_DynamicControls_h

#include "Rect.h"
#include "ControlID.h"

#include <vector>

namespace vgui
{
	class Window;

	//
	// TControlResize
	//
	// Enum to specify how dynamically positioned/sized controls
	// should rect when the dialog is resized.
	//
	// (One behaviour for each of the X and Y dimensions)
	//
	typedef enum
	{
		ControlResize_Static,
		ControlResize_Stretch,
		ControlResize_Move,
		ControlResize_MoveCentre
	}
	TControlResize;

	//
	// Structure used to batch up declarations of dynamic controls
	//
	class DynamicControlInfo
	{
		public:
			// Control ID of the window (if it is a control; NOCTL otherwise)
			ControlID m_ID;

			// How it behaves on resizing
			TControlResize m_XBehaviour;
			TControlResize m_YBehaviour;
	};

	//
	// Internal structure used to keep track of dynamic controls/windows
	//
	typedef struct
	{
		// Control ID of the window (if it is a control; NOCTL otherwise)
		ControlID m_ID;

		// If it's not a control, this is the Window object
		Window* m_pWindow;

		// How it behaves on resizing
		TControlResize m_XBehaviour;
		TControlResize m_YBehaviour;

		// Initial size/position
		vgui::Rect m_InitialPosition;

		// Find out if this is a control or normal window
		bool IsControl() { return (m_ID != NOCTL); }
	}
	TDynamicWindow;

	//
	// Compact name for vector of dynamic window records.
	//
	typedef std::vector<TDynamicWindow> DynamicWindows;
	
	// 
	// Macros for building tables of dynamic controls
	//
	//
	// Example usage:
	//
	//	To declare the table of dynamic controls, do something like this:
	//
	//		VGUI_BEGIN_DYNAMIC_CONTROLS(s_MyControlTableName)
	//			VGUI_DYNAMIC_CONTROL(IDC_MYEDITBOX,  Stretch, Static)
	//			VGUI_DYNAMIC_CONTROL(IDC_MYDROPLIST, Static,  Move)
	//			VGUI_DYNAMIC_CONTROL(IDC_BUTTON1,	 Move,    Static)
	//			VGUI_DYNAMIC_CONTROL(IDC_BUTTON2,	 Move,    Static)
	//		VGUI_END_DYNAMIC_CONTROLS()
	//
	//	(Valid styles are Static, Stretch, Move, and MoveCentre)
	//
	//  Then to tell the dialog about the controls, just do this:
	//
	//		pDialog->AddDynamicControls(s_MyControlTableName)
	//
	//  or more often, just:
	//
	//		AddDynamicControls(s_MyControlTableName)
	//
	//  as you'll usually call this from within the dialog class itself.
	//
#define VGUI_BEGIN_DYNAMIC_CONTROLS(TableName)	\
		static vgui::DynamicControlInfo TableName[] = \
		{

#define VGUI_DYNAMIC_CONTROL(ID, XStyle, YStyle) \
			{ID, vgui::ControlResize_##XStyle, vgui::ControlResize_##YStyle	},

#define VGUI_END_DYNAMIC_CONTROLS()		\
			{vgui::NOCTL, vgui::ControlResize_Static, vgui::ControlResize_Static}, \
		};
};

#endif // !vgui_DynamicControls_h
