//
// ErrorReporter.cpp - the standard vgui Error reporter object
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "ErrorReporter.h"

#include <stdlib.h>

#include "Dialog.h"
#include "Application.h"
#include "RedrawGC.h"
#include "Font.h"

#include "vOS\Debugger.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif


namespace vgui
{
	using namespace vos;

	//
	// The dialogs used to report errors.
	//

	class ErrorDialog : public Dialog
	{
		public:
			DECLARE_VGUI_MESSAGE_MAP()

		public:
			ErrorDialog(const DialogResourceID& rDialog,
						const ErrorInfo& Error,
						const ErrorReporterInfo& Info);

			virtual bool OnRedraw();

		public:
			// Message handlers
			bool OnButtonClicked(TMessageType MsgType, int nMsg, int nCtrlID);
			virtual void OnInitDialog();

		protected:
			virtual void ShowErrorType();
			void UpdateDisplayControls();

		protected:
			const ErrorInfo* m_pError;
			const ErrorReporterInfo& m_Info;

			Window m_ErrorMessage;
			Window m_ErrorDetails;

			// More info/prev error buttons for following error chain.
			Window m_MoreInfo;
			Window m_PrevError;
	};

	BEGIN_VGUI_MESSAGE_MAP(ErrorDialog, Dialog)
		// Use the generic button handler, as we don't know the control IDs until run-time.
		ON_VGUI_BN_CLICKED_ALL(OnButtonClicked)
	END_VGUI_MESSAGE_MAP()

	ErrorDialog::ErrorDialog(const DialogResourceID& rDialog,
							 const ErrorInfo& Error,
							 const ErrorReporterInfo& Info) : Dialog(rDialog, Application::GetMainWindow()),
															  m_pError(&Error),
															  m_Info(Info)
	{
	}

	void ErrorDialog::UpdateDisplayControls()
	{
		// PrevError button is disabled if we're at the head
		// of the list of errors.
		if (m_Info.m_PrevErrorButton != NOCTL)
		{
			if (m_pError->GetNextError())
			{
				// There is another error - make button visible.
				m_PrevError.ShowWindow(sw_SHOW);
			}
			else
			{
				// There are no more errors - hide button.
				m_PrevError.ShowWindow(sw_HIDE);
			}
		}

		// More info is available if there is another error
		// in the chain.
		if (m_Info.m_MoreInfoButton != NOCTL)
		{
			if (m_pError->GetPrevError())
			{
				// There is another error - make button visible.
				m_MoreInfo.ShowWindow(sw_SHOW);
			}
			else
			{
				// There are no more errors - hide button.
				m_MoreInfo.ShowWindow(sw_HIDE);
			}
		}

		// Force redraw of error text
		Rect TextArea;
		m_ErrorDetails.GetClientArea(&TextArea);
		m_ErrorDetails.ClientToWindow(this, &TextArea);
		ForceRedraw(TextArea);
	}

	bool ErrorDialog::OnButtonClicked(TMessageType MsgType, int nMsg, int ButtonPressed)
	{
		if (ButtonPressed == m_Info.m_MoreInfoButton)
		{
			// Move to previous error
			if (m_pError->GetPrevError(&m_pError))
			{
				// Update the error display
				UpdateDisplayControls();
			}
		}
		else if (ButtonPressed == m_Info.m_PrevErrorButton)
		{
			// Move to more recent error
			if (m_pError->GetNextError(&m_pError))
			{
				// Update the error display
				UpdateDisplayControls();
			}
		}
		else
		{
			// Otherwise, we just end the dialog on the button click.
			EndDialog(ButtonPressed);
		}

		return true;
	}

	bool ErrorDialog::OnRedraw()
	{
		// This is currently on hold while I work out how to draw over controls
		// in a dialog.
		RedrawGC gc(this);

		// Create a font
		Font* pFont = GetFont();
		gc.SetFont(pFont);

		// Get position of the text control
		Rect TextArea;
		m_ErrorDetails.GetClientArea(&TextArea);
		m_ErrorDetails.ClientToWindow(this, &TextArea);

		gc.Draw3DEdge(TextArea, edge3d_ThinSunken);

		TextArea.Deflate(3, 3);

		gc.DrawText(m_pError->GetText(), 
					TextArea, 
					tf_VCentre|tf_HCentre|tf_Multiline);

		gc.RestoreDefaultFont();

		return true;
	}

	void ErrorDialog::OnInitDialog()
	{
		// Set up the dialog to contain the error text.
		ShowErrorType();
		m_ErrorDetails.Attach(this, m_Info.m_ErrorDetailsText);
		m_ErrorDetails.SetText(m_pError->GetText());

		// Attach error chain navigation buttons if they exist.
		if (m_Info.m_PrevErrorButton != NOCTL)
		{
			m_PrevError.Attach(this, m_Info.m_PrevErrorButton);
		}

		if (m_Info.m_MoreInfoButton != NOCTL)
		{
			m_MoreInfo.Attach(this, m_Info.m_MoreInfoButton);
		}

		// Work out what buttons to show...
		UpdateDisplayControls();
	
	}

	void ErrorDialog::ShowErrorType()
	{
		String ErrMsg;

		switch (m_pError->GetType())
		{
#ifdef _DEBUG
			case Error_Complaint:
				ErrMsg = "A complaint has occurred.";
				break;
#endif

			case Error_Normal:
				ErrMsg = "An error has occurred.";
				break;

			case Error_Critical:
				ErrMsg = "A critical error has occurred.";
				break;

			case Error_Terminal:
				ErrMsg = "A terminal error has occured.  The program must now exit.";
				break;
		}

		m_ErrorMessage.Attach(this, m_Info.m_ErrorText);
		m_ErrorMessage.SetText(ErrMsg);
	}

	class ErrorDialog_Dbg : public ErrorDialog
	{
		public:
			ErrorDialog_Dbg(const DialogResourceID& rDialog,
							const ErrorInfo& Error,
							const ErrorReporterInfo& Info);

		protected:
			virtual void ShowErrorType();
	};

	ErrorDialog_Dbg::ErrorDialog_Dbg(const DialogResourceID& rDialog,
									 const ErrorInfo& Error,
									 const ErrorReporterInfo& Info) : ErrorDialog(rDialog, Error, Info)
	{
	}

	void ErrorDialog_Dbg::ShowErrorType()
	{
		String ErrMsg;

		switch (m_pError->GetType())
		{
#ifdef _DEBUG
			case Error_Complaint:
				ErrMsg = "A complaint has occurred.";
				break;
#endif

			case Error_Normal:
				ErrMsg = "An error has occurred.";
				break;

			case Error_Critical:
				ErrMsg = "A critical error has occurred.";
				break;

			case Error_Terminal:
				ErrMsg = "A terminal error has occurred.  The program must now exit.";
				break;
		}

		// Work out where this error happened in the source code.
		String SourceFile = m_pError->GetSourceFile();

		if (SourceFile.GetLength() > 0)
		{
			// Tell user where the error originated.
			ErrMsg += "  (";
			ErrMsg += SourceFile;
			ErrMsg += ", at line ";
			ErrMsg += String(m_pError->GetSourceLine());
			ErrMsg += ")";
		}
		else
		{
			// We don't have source file/line number information.
			ErrMsg += " (source location of error unknown)";
		}

		// Show the text in the dialog
		m_ErrorMessage.Attach(this, m_Info.m_ErrorText);
		m_ErrorMessage.SetText(ErrMsg);
	}

	/////////////////////////////////////////////////////////////////////////////////

	vGuiErrorReporter::vGuiErrorReporter(const ErrorReporterInfo& Info) : m_Info(Info)
	{
	}

	void vGuiErrorReporter::ReportError(vos::ErrorStack* pErrors)
	{
#ifdef _DEBUG

		// Create the dialog used to report errors in debug builds.
		DialogResourceID rDialog(m_Info.m_ErrorDialog_Debug);

		ErrorDialog_Dbg *pDialog = new ErrorDialog_Dbg(rDialog, 
													   *pErrors->Peek(), 
													   m_Info);
		
		ControlID ButtonPressed = pDialog->DoModal();

		if (ButtonPressed == m_Info.m_ContinueButton)
		{
			// Carry on - don't do anything.
		}
		else if (ButtonPressed == m_Info.m_DebugButton)
		{
			// Drop into debugger
			Debugger::Interrupt();
		}
		else if (ButtonPressed == m_Info.m_AbortButton)
		{
			// Abort the process - use standard abort() function so the
			// client can install a SIGABRT handler in the usual (portable)
			// way, if they want to.
			abort();
		}
		else
		{
			// Unknown error - just output to trace to avoid
			// infinite loops in error handler.
			// TODO: Implement trace facilities in vOS and use them!
		}
#else

		// Create the dialog used to report errors in release builds.
		DialogResourceID rDialog(m_Info.m_ErrorDialog_Release);

		ErrorDialog *pDialog = new ErrorDialog(rDialog, *pErrors->Peek(), m_Info);
		
		// Show the dialog to the user - they just get to click OK in confirmation.
		pDialog->DoModal();
#endif
	}

};


