//
// GC.h - Graphics Context
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_GC_h
#define vgui_GC_h

#include "String.h"

#include "vgui.h"

#include "Colour.h"
#include "Point.h"
#include "Rect.h"

namespace vgui
{
	class GCImp;
	class Font;
	class Icon;
	//
	// Enumeration types for GC control
	//

	typedef enum
	{
		//
		// Control of each part of the edge.
		//
		edge3d_RaisedInner			= BIT(1),
		edge3d_SunkenInner			= BIT(2),
		edge3d_RaisedOuter			= BIT(3),
		edge3d_SunkenOuter			= BIT(4),

		//
		// Macros for edge types.
		//
		edge3d_Bump			= edge3d_RaisedOuter|edge3d_SunkenInner,
		edge3d_Etched		= edge3d_SunkenOuter|edge3d_RaisedInner,
		edge3d_Raised		= edge3d_RaisedOuter|edge3d_RaisedInner,
		edge3d_Sunken		= edge3d_SunkenOuter|edge3d_SunkenInner,

		edge3d_ThinRaised	= edge3d_RaisedOuter,
		edge3d_ThinSunken	= edge3d_SunkenOuter,
	}
	Edge3DType;

#ifdef VOS_MSWINDOWS
	//
	// Enum values optimised for Win32 text rendering
	//
	enum
	{
		// Horizontal alignment
		tf_Left				= DT_LEFT,
		tf_HCentre			= DT_CENTER,
		tf_Right			= DT_RIGHT,

		// Vertical alignment
		tf_Top				= DT_TOP,
		tf_VCentre			= DT_VCENTER,
		tf_Bottom			= DT_BOTTOM,

		// Layout
		tf_Multiline		= DT_WORDBREAK
	};
#else
	// Platform independent values
	enum
	{
		// Horizontal alignment
		tf_Left				= BIT(1),
		tf_HCentre			= BIT(2),
		tf_Right			= BIT(3),

		// Vertical alignment
		tf_Top				= BIT(4),
		tf_VCentre			= BIT(5),
		tf_Bottom			= BIT(6),

		// Layout
		tf_Multiline		= BIT(7)
	};
#endif

	class GC
	{
		public:

			//
			// Constructor/Destructor
			//
			GC();
			~GC();

			//
			// Drawing Attributes
			//

			// Save/restore entire GC state
			void Save();
			void Restore();

			// Clip rectangle
			void GetClipRect(Rect* pClipRect);

			typedef enum
			{
				ss_SOLID,
				ss_DASHED,
				ss_DOTTED,
				ss_DASHDOT
			}
			StrokeStyle;

			void SetStrokeStyle(StrokeStyle Style);

			typedef enum
			{
				fs_SOLID,		// Fills with solid colour
				fs_NONE,		// No filling
			}
			FillStyle;

			void SetFillStyle(FillStyle Style);

			void SetFillColour(const Colour& FillColour);
			void SetStrokeColour(const Colour& StrokeColour);
			void SetBgColour(const Colour& BgColour);

			Colour SetTextColour(const Colour& TextColour);

			void SetStrokeWidth(int Width);

			// Drawing control
			typedef enum
			{
				bg_Opaque,
				bg_Transparent
			}
			BgMode;

			void SetBgMode(BgMode Mode);

			typedef enum
			{
				dt_Stroke,
				dt_Fill,
				dt_FillAndStroke,
				dt_StrokeAndFill
			}
			DrawType;

			// Drawing lines
			void MoveTo(int x, int y);
			void LineTo(int x, int y);

			// Simple shapes
			void DrawRect(const Rect& rc, 
						  DrawType Type = dt_FillAndStroke);

			void DrawCircle(int cx, int cy, int Radius, 
							DrawType Type = dt_FillAndStroke);

			// Polygons
			void DrawPolygon(const Point* pts,
							 int cPoints,
							 DrawType Type = dt_FillAndStroke);

			void DrawPolygon(const Points& pts,
							 DrawType Type = dt_FillAndStroke);

			// Text rendering

			void SetFont(Font* pFont);
			void RestoreDefaultFont();

			void DrawText(const vos::String& Text, int x, int y);

			void DrawText(const vos::String& Text, 
						  const Rect& TextArea,
						  int FormatFlags);

			void CalcTextArea(const vos::String& Text,
							  int FormatFlags,
							  Rect* pTextArea);

			// Specialised rendering

			void Draw3DEdge(const Rect& TheRect, Edge3DType EdgeType);
			void DrawFocusRect(const Rect& TheRect);

			// Image/bitmap rendering
			void DrawIcon(const Icon& Icon, int X, int Y);

			// GC Focus
			
			void BeginDrawing();
			void EndDrawing();

			// Co-ordinate functions

		protected:
			// Protected constructor for use by Window class.
			GC(Window *pWindow);
			friend class Window;

		public:
			GCImp* GetGCImp() { return m_pGCImp; }

		protected:
			GC(GCImp *pImp);

			GCImp* GetImp() { return m_pGCImp; }

		private:
			// The implementation of this GC.
			// (Pattern: Bridge)
			GCImp *m_pGCImp;

		private:
			// Prohibit copying/assignment of GCs - clients should use
			// pointers/references to GCs if they are to be shared.
			GC(const GC&);
			GC& operator=(const GC&);
	};
};

#endif // !vgui_GC_h


