//
// HistoryList.cpp - simple class to provide a historical list of strings
//				     (and optional 32-bit data for each string)
//

#include "StdHdr.h"

#include "HistoryList.h"

#include "Errors.h"

namespace vgui
{
	/**************************************************************************
	
	>	HistoryList::HistoryList() : m_cMaxItems(20)
	
		Author:		Tim Browse
		Date:		27/7/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	HistoryList::HistoryList() : m_cMaxItems(20)
	{
		// TCILB
	}
	
	/**************************************************************************
	
	>	void HistoryList::SetHistoryLimit(uint cMaxItems)
	
		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		cMaxItems - maximum number of items to hold in the history list.
	
		Purpose:	Set the maximum number of items that can be in a history list.
	
	**************************************************************************/
	
	void HistoryList::SetHistoryLimit(uint cMaxItems)
	{
		m_cMaxItems = cMaxItems;
		LimitNumberOfItems();
	}

	/**************************************************************************
	
	>	uint HistoryList::GetHistoryLimit() const
	
		Author:		Tim Browse
		Date:		27/7/2002
		Returns:	Maximum number of items
	
		Purpose:	Find out the maximum number of items that are allowed
					in this list.
	
	**************************************************************************/
	
	uint HistoryList::GetHistoryLimit() const
	{
		return m_cMaxItems;
	}

	/**************************************************************************
	
	>	void HistoryList::LimitNumberOfItems()
	
		Author:		Tim Browse
		Date:		27/7/2002
	
		Purpose:	Make sure that the history list is not holding more than
					the maximum number of items (items are removed from the
					tail of the list if there are too many).
	
	**************************************************************************/
	
	void HistoryList::LimitNumberOfItems()
	{
		while (size() >= m_cMaxItems)
		{
			pop_back();
		}
	}

	/**************************************************************************
	
	>	void HistoryList::AddToHistory(const vos::String& Text, UserData TheUserData)
	
		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		Text - text of the new item
					TheUserData - user-defined data for this item, if required.
	
		Purpose:	Add a new entry to the head of the history list, using the
					data provided.  Any other instances of this text are removed
					from the list.
	
	**************************************************************************/
	
	void HistoryList::AddToHistory(const vos::String& Text, UserData TheUserData)
	{
		// Remove any existing instances of this text
		HistoryList::iterator it = begin();

		while (it != end())
		{
			HistoryList::iterator itNextItem = it;
			++itNextItem;

			if (it->m_Text == Text)
			{
				erase(it);
			}

			it = itNextItem;
		}

		// Add to the start of the list
		HistoryItem NewItem;
		NewItem.m_Text = Text;
		NewItem.m_UserData= TheUserData;
		push_front(NewItem);

		// Make sure list does not exceed history limit
		LimitNumberOfItems();
	}

	/**************************************************************************

	>	bool HistoryList::AddItem(const vos::String& Text, 
								  UserData TheUserData, 
								  int* piItem)


		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		Text - text of the new item
					TheUserData - user-defined data for this item, if required.
		Outputs:	piItem - the (0-based) index of the new item (if not NULL)
		Returns:	true => item added ok
					false => item not added; history list already full

		Purpose:	Add an item to the end of the history list, using the
					data provided.

	**************************************************************************/

	bool HistoryList::AddItem(const vos::String& Text, 
							  UserData TheUserData, 
							  int* piItem)
	{
		// Do we have enough space to add another string?
		if (size() < m_cMaxItems)
		{
			// Yes, so we add it
			HistoryItem NewItem;
			NewItem.m_Text = Text;
			NewItem.m_UserData = TheUserData;
			push_back(NewItem);
			if (piItem != NULL)
				*piItem = size() - 1;
			return true;
		}
		else
		{
			// History list is full
			return false;
		}
	}

	/**************************************************************************

	>	bool HistoryList::InsertItemAt(int iInsertAt, 
									   const vos::String& Text, 
									   UserData TheUserData, 
									   int* piItem)


		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		iInsertAt - insert the item at this position (0 = first item)
					Text - text of the new item
					TheUserData - user-defined data for this item, if required.
		Outputs:	piItem - the (0-based) index of the new item (if not NULL)
		Returns:	true => item inserted ok
					false => item not added; index out of range

		Purpose:	Insert an item into a specific place in the history list.

	**************************************************************************/

	bool HistoryList::InsertItemAt(int iInsertAt, 
								   const vos::String& Text, 
								   UserData TheUserData, 
								   int* piItem)
	{
		if ((iInsertAt >= m_cMaxItems) || (iInsertAt > size()))
		{
			// Cannot insert past the max size of our history list, or
			// past the end of the existing list (although note that we allow
			// insertion as the last element, if the limit allows us)
			COMPLAIN("HistoryList::InsertItemAt() - index out of range!");
			return false;
		}

		// Ok, so insert the new item
		//
		// NB. FindItem() may return end() if inserting at the end of
		//	   the list, but this still works ok with insert().
		HistoryItem NewItem;
		NewItem.m_Text = Text;
		NewItem.m_UserData = TheUserData;
		insert(FindItem(iInsertAt), NewItem);
		return true;
	}

	/**************************************************************************
	
	>	HistoryList::const_iterator HistoryList::FindItem(int iItem) const
	
		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		iItem - the item to find (0 = first item in list)
		Returns:	iterator that points to the item (or end() if iItem is
					out of range)
	
		Purpose:	Obtain a const iterator that references the specified item.
					
	
	**************************************************************************/
	
	HistoryList::const_iterator HistoryList::FindItem(int iItem) const
	{
		HistoryList::const_iterator it = begin();

		while ((it != end()) && (iItem > 0))
		{
			++it;
			--iItem;
		}

		return it;
	}

	/**************************************************************************
	
	>	HistoryList::iterator HistoryList::FindItem(int iItem) const
	
		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		iItem - the item to find (0 = first item in list)
		Returns:	iterator that points to the item (or end() if iItem is
					out of range)
	
		Purpose:	Obtain a iterator that references the specified item.
					
	
	**************************************************************************/
	
	HistoryList::iterator HistoryList::FindItem(int iItem)
	{
		HistoryList::iterator it = begin();

		while ((it != end()) && (iItem > 0))
		{
			++it;
			--iItem;
		}

		return it;
	}

	/**************************************************************************
	
	>	bool HistoryList::DeleteItem(int iItemToDelete)
	
		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		iItemToDelete - index of the item to delete (0 = first item)
		Returns:	true => item was deleted
					false => index was out of range
	
		Purpose:	Delete the specified item from the list
	
	**************************************************************************/
	
	bool HistoryList::DeleteItem(int iItemToDelete)
	{
		HistoryList::iterator it = FindItem(iItemToDelete);

		if (it != end())
		{
			// Ok, delete the item
			erase(it);
			return true;
		}
		else
		{
			COMPLAIN("HistoryList::DeleteItem() - index out of range!");
			return false;
		}
	}

	/**************************************************************************
	
	>	bool HistoryList::SetItemText(int iItem, const String& Text)
	
		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		iItem - the item to change (0 = first item)
					Text - new text of the item
		Returns:	true => text changed ok
					false => index out of range
	
		Purpose:	Change the text of an item
	
	**************************************************************************/
	
	bool HistoryList::SetItemText(int iItem, const String& Text)
	{
		HistoryList::iterator it = FindItem(iItem);

		if (it != end())
		{
			// update item text
			it->m_Text = Text;
			return true;
		}
		else
		{
			COMPLAIN("HistoryList::SetItemText() - index out of range!");
			return false;
		}
	}

	/**************************************************************************
	
	>	bool HistoryList::SetItemData(int iItem, UserData TheUserData)
	
		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		iItem - the item to change (0 = first item)
					TheUserData - new user data to associate with this item
		Returns:	true => user data changed ok
					false => index out of range
	
		Purpose:	Change the user data associated with an item.
	
	**************************************************************************/
	
	bool HistoryList::SetItemData(int iItem, UserData TheUserData)
	{
		HistoryList::iterator it = FindItem(iItem);

		if (it != end())
		{
			// update item data
			it->m_UserData = TheUserData;
			return true;
		}
		else
		{
			COMPLAIN("HistoryList::SetItemData() - index out of range!");
			return false;
		}
	}

	/**************************************************************************
	
	>	uint HistoryList::GetCount() const
	
		Author:		Tim Browse
		Date:		27/7/2002
		Returns:	The number of items in the list
	
		Purpose:	Find out how many items are in the history list.
	
	**************************************************************************/
	
	uint HistoryList::GetCount() const
	{
		return size();
	}

	/**************************************************************************
	
	>	bool HistoryList::GetItemText(int iItem, String* pItemName) const
	
		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		iItem - the index of the item to get the text from
							(0 = first item)
		Outputs:	pItemName - the text of the item
		Returns:	true => text obtained successfully
					false => index out of range
	
		Purpose:	Get the text associated with the specified item.
	
	**************************************************************************/
	
	bool HistoryList::GetItemText(int iItem, String* pItemName) const
	{
		HistoryList::const_iterator it = FindItem(iItem);

		if (it != end())
		{
			// return item text
			*pItemName = it->m_Text;
			return true;
		}
		else
		{
			COMPLAIN("HistoryList::GetItemText() - index out of range!");
			return false;
		}
	}

	/**************************************************************************
	
	>	bool HistoryList::GetItemUserData(int iItem, UserData* pTheUserData) const
	
		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		iItem - the index of the item to get the user data from
							(0 = first item)
		Outputs:	pTheUserData - the user data from the item
		Returns:	true => user data obtained successfully
					false => index out of range
	
		Purpose:	Get the user data associated with the specified item.
	
	**************************************************************************/
	
	bool HistoryList::GetItemUserData(int iItem, UserData* pTheUserData) const
	{
		HistoryList::const_iterator it = FindItem(iItem);

		if (it != end())
		{
			// return item data
			*pTheUserData = it->m_UserData;
			return true;
		}
		else
		{
			COMPLAIN("HistoryList::GetItemData() - index out of range!");
			return false;
		}
	}

	/**************************************************************************
	
	>	bool HistoryList::FindItem(const String& ItemName, int* piFoundItem)
	
		Author:		Tim Browse
		Date:		27/7/2002
		Inputs:		ItemName - text to search for
		Outputs:	piFoundItem - index of the item that was found (if it was)
								  (may be NULL)
		Returns:	true => item text was found
					false => item text not found
	
		Purpose:	Find the specified string in the list.
	
	**************************************************************************/
	
	bool HistoryList::FindItem(const String& ItemName, int* piFoundItem)
	{
		HistoryList::const_iterator it = begin();

		int iItem = 0;

		while (it != end())
		{
			if (it->m_Text == ItemName)
			{
				// Found the text
				if (piFoundItem != NULL)
					*piFoundItem = iItem;
				return true;
			}

			// Try next item
			++it;
			++iItem;
		}

		// Text not found
		return false;
	}
};
