//
// HistoryList.h - simple class to provide a historical list of strings
//				   (and optional 32-bit data for each string)
//				   For use mainly by parts of the vGUI library, 
//				   but you can use it if you like.
//
//				   The interface is a replica of parts of the ListControl 
//				   interface.
//

#ifndef vgui_HistoryList_h
#define vgui_HistoryList_h

#include "vgui.h"

#include <list>

namespace vgui
{
	/**************************************************************************
	
	>	struct HistoryItem
	
		Author:		Tim Browse
		Date:		27/7/2002
	
		Purpose:	Holds information about an item in a HistoryList object.
	
	**************************************************************************/
	
	struct HistoryItem
	{
		vos::String m_Text;
		UserData m_UserData;
	};

	/**************************************************************************
	
	>	class HistoryList : public std::list<HistoryItem>
	
		Author:		Tim Browse
		Date:		26/7/2002
	
		Purpose:	Provide a simple list of strings, oriented towards
					storing a history array.
					
					e.g. this is used by Win32 DropDownList objects to
						 implement the optional history behaviour.

					NB. Some of these functions are specifically here to make
						it easy for a ListControl object to use this class, so
						some of the public functions may not be that useful in
						a more general sense - just use the functions that make
						the most sense - the best ones to use are:

							* SetHistoryLimit()
							* AddToHistory()
							* GetCount()
							* GetItemText()
							* GetItemUserData()

	**************************************************************************/
	
	class HistoryList : public std::list<HistoryItem>
	{
		public:
			HistoryList();

			//
			// General usage
			//

			// Set/Get the maximum number of items allowed
			void SetHistoryLimit(uint cMaxItems);
			uint GetHistoryLimit() const;

			// Add item to the head of the history list, removing any
			// existing copies of it from the list.
			void AddToHistory(const vos::String& Text, UserData TheUserData = 0);

			// Access to list data
			uint GetCount() const;
			bool GetItemText(int iItem, String* pItemName) const;
			bool GetItemUserData(int iItem, UserData* pTheUserData) const;


			//
			// ListControl interface subset - less generic
			//
			bool AddItem(const vos::String& Text, 
						 UserData TheUserData, 
						 int* piItem = NULL);

			bool InsertItemAt(int iInsertAt, 
							  const vos::String& Text, 
							  UserData TheUserData, 
							  int* piItem = NULL);

			bool DeleteItem(int iItemToDelete);

			bool SetItemText(int iItem, const String& Text);
			bool SetItemData(int iItem, UserData TheUserData);

			bool FindItem(const String& ItemName, int* piFoundItem = NULL);

		private:
			// Ensure history list is not too long
			void LimitNumberOfItems();

			// Finding particular items by index (as STL lists do not
			// support this directly)
			HistoryList::const_iterator FindItem(int iItem) const;
			HistoryList::iterator FindItem(int iItem);
			
		private:
			// Maximum number of items allowed in history list
			uint m_cMaxItems;
	};
};

#endif // !vgui_HistoryList_h
