//
// IconStore.cpp - repository for icons
//

#include "stdhdr.h"

#include "IconStore.h"

#include "Icon.h"

#include "vos\Errors.h"

namespace vgui
{

	//
	// The default icon (used when asked for an icon we don't have)
	//
	Icon IconStore::m_DefaultIcon;

	/**************************************************************************

	>	IconStore::~IconStore()

		Author:		Tim Browse
		Date:		24/03/2001

		Purpose:	Destructor - cleans up all the icons we've loaded.

	**************************************************************************/

	IconStore::~IconStore()
	{
		// Dispose all the icons we've loaded
		Icons::const_iterator it = m_Icons.begin();

		while (it != m_Icons.end())
		{
			delete it->second;
			++it;
		}
	}


	/**************************************************************************

	>	bool IconStore::AddIcon(TIconID ID, const & ResID)

		Author:		Tim Browse
		Date:		24/03/2001
		Inputs:		ID - the ID to associate with the icon - this is the ID
						 to pass to GetIcon() when you need to access the icon.
					ResID - resource ID of the icon to add.
		Returns:	true => icon add to store successfully
					false => unable to add icon

		Purpose:	Add an icon to the icon store.

					Notes: you can add the same physical icon many times with
					different IDs, and the icon will only be stored once.

	**************************************************************************/

	bool IconStore::AddIcon(TIconID ID, const IconResourceID& ResID)
	{
		// Check we don't already have this icon loaded
		Icons::iterator it = m_Icons.find(ResID.GetID());

		if (it != m_Icons.end())
		{
			// Icon is already loaded - just add an entry to the map
			m_IconMap[ID] = it->second;
			return true;
		}

		// Load the icon
		Icon* pNewIcon = new Icon;
		if (!ResID.Load(pNewIcon))
		{
			// Unable to find/load icon
			return false;
		}

		// Add entry to maps for this icon
		m_Icons[ResID.GetID()] = pNewIcon;
		m_IconMap[ID] = pNewIcon;
		return true;
	}


	/**************************************************************************

	>	Icon& IconStore::GetIcon(TIconID ID) const

		Author:		Tim Browse
		Date:		24/03/2001
		Inputs:		ID - the ID of the icon to access
		Returns:	The icon, if found - if not found, a null icon object is
					returned, but in debug builds this will cause a error.

		Purpose:	Access an icon in the icon store.

	**************************************************************************/

	Icon& IconStore::GetIcon(TIconID ID) const
	{
		// Look up ID in our map
		IconMap::const_iterator it = m_IconMap.find(ID);

		if (it != m_IconMap.end())
		{
			// The icon is in our store
			return *(it->second);
		}

		// No icon found
		COMPLAIN("IconStore::GetIcon() - icon not found!");
		return m_DefaultIcon;
	}
}
