//
// JobWorker.h
//

#ifndef inc_vgui_JobWorker_h
#define inc_vgui_JobWorker_h

#include "vgui.h"

#include "JobController.h"

namespace vgui
{
	//
	// TJobWorkerStatus
	//
	// Used by the JobWorker interface class to signify
	// whether or not the work has been completed.
	//
	typedef enum
	{
		Job_Finished,
		Job_NotFinished
	}
	TJobStatus;

	/**************************************************************************

	>	class JobWorker

		Author:		Tim Browse
		Date:		2001/10/25

		Purpose:	A callback interface to use to carry out work while a
					progress dialog is open.  The worker must carry out the
					work a piece at a time (in a timesliced manner).

					The ProgressDialog will call DoWork() periodically
					(driven by idle events) until the handler signifies that
					the job is complete (by returning JobWorker_Finished).
					The amount of time the worker spends before returning
					directly affects the responsiveness of the UI - the worker
					should try to return within <200ms to keep the UI
					working smoothly.

					If the user presses Cancel, the dialog calls OnCancel()
					to inform the worker object, which is expected to have
					performed all necessary cleanup by the time it returns.

	**************************************************************************/

	class JobWorker
	{
		public:
			virtual ~JobWorker();

			// Do a block of work, and let the ProgressDialog know how much of
			// the work has been completed so far, and whether the work has
			// been finished or not.
			virtual TJobStatus DoWork(JobController* pJobController,
									  int* pPercentComplete) = 0;

			// The user has cancelled the job, so the worker object
			// should perform any shutdown necessary before returning.  The
			// object will not be called by the JobController at all after
			// this function has been called.
			virtual void OnCancel(JobController* pJobController);

			// Convenience function to do the work without involving a vgui
			// ProgressDialog.  This allows you to do the same work with the
			// same function, with or without a ProgressDialog.
			void DoAllWork(JobController* pJobController = NULL);
	};
	
	/**************************************************************************
	
	>	class JobWorkerWrapper : public JobWorker
	
		Author:		Tim Browse
		Date:		8/8/2002
	
		Purpose:	Allows one worked to be wrapped in another.  This is
					useful when you have a number of workers, and want to
					dynamically allocate some of them and delete them when
					done.

					For the static workers, just wrap them using this class
					and you can delete them without worrying about deleting
					the underlying JobWorker object - this class just vectors
					the interface to the underlying object.
	
	**************************************************************************/
	
	class JobWorkerWrapper : public JobWorker
	{
		public:
			JobWorkerWrapper(JobWorker* pWorker);

			virtual TJobStatus DoWork(JobController* pJobController,
									  int* pPercentComplete);

			virtual void OnCancel(JobController* pJobController);

		private:
			JobWorker* m_pWorker;
	};

};

#endif // !inc_JobWorker_h
