//
// ListControl.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "ListControl.h"

#include "ListControlImp.h"

#include "vgui\Errors.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	ListControl::ListControl(ListControlImp *pImp) : Window(pImp),
													 m_pOwnerDrawHandler(NULL),
													 m_pSortFunctions(NULL),
													 m_cSortFunctions(0),
													 m_SortOrder(SortOrder_Ascending)
	{
		// Link back to this ListControl object
		m_pWindowImp->SetWindowLink(this);
	}

	ListControl::~ListControl()
	{
		delete [] m_pSortFunctions;
	}
	
	// Remove all strings
	void ListControl::Clear()
	{
		GetImp()->Clear();
	}

	// Find out how many there are.
	int ListControl::GetCount() const
	{
		return GetImp()->GetCount();
	}

	// Initialise from 0-terminated array of string resources
	bool ListControl::Initialise(const StringResourceID* StringIDs)
	{
		const StringResourceID *pID = StringIDs;

		// Add each specified element to the list
		while (!pID->IsNull())
		{
			// Load the string
			String ItemName;
			pID->Load(&ItemName);

			// Add this string to list
			int iDummy;
			if (!GetImp()->AddItem(ItemName, &iDummy, 0))
			{
				// Failed to add string
				return false;
			}

			// Try next ID
			++pID;
		}

		// Worked ok
		return true;
	}

	// Add and delete items
	bool ListControl::AddItem(const String& ItemName, 
							  int* piItem,
							  UserData TheUserData)
	{
		return GetImp()->AddItem(ItemName, piItem, TheUserData);
	}

	bool ListControl::InsertItemAfter(int iAfterItem,
									  const String& ItemName, 
									  int* piItem,
									  UserData TheUserData)
	{
		return GetImp()->InsertItemAfter(iAfterItem, ItemName, piItem, TheUserData);
	}

	bool ListControl::InsertItemBefore(int iBeforeItem,
									   const String& ItemName, 
									   int* piItem,
									   UserData TheUserData)
	{
		return GetImp()->InsertItemBefore(iBeforeItem, ItemName, piItem, TheUserData);
	}

	bool ListControl::DeleteItem(int iItem)
	{
		return GetImp()->DeleteItem(iItem);
	}

	bool ListControl::SetItemData(int iItem, UserData TheUserData)
	{
		return GetImp()->SetItemData(iItem, TheUserData);
	}

	bool ListControl::SetItemText(int iItem, const String& Text, int iColumn)
	{
		return GetImp()->SetItemText(iItem, Text, iColumn);
	}

	// Find info about an item
	bool ListControl::GetItemText(int iItem, String* pItemText, int iColumn) const
	{
		return GetImp()->GetItemText(iItem, pItemText, iColumn);
	}

	bool ListControl::GetItemUserData(int iItem, UserData* pTheUserData) const
	{
		return GetImp()->GetItemUserData(iItem, pTheUserData);
	}

	ListControl& ListControl::operator=(const String& ItemName)
	{
		SelectByString(ItemName);
		return *this;
	}

	ListControl& ListControl::operator=(const StringResourceID& ItemID)
	{
		SelectByResID(ItemID);
		return *this;
	}

	bool ListControl::SelectByIndex(int iItem)
	{
		return GetImp()->SetSelectedItem(iItem);
	}

	bool ListControl::SelectByString(const String& ItemName)
	{
		// Find the item
		int iItem;
		if (GetImp()->FindItem(ItemName, &iItem))
		{
			// if we found an item, then set it to be selected
			return GetImp()->SetSelectedItem(iItem);
		}

		// Could not find item
		return false;
	}

	bool ListControl::SelectByResID(const StringResourceID& ItemID)
	{
		// Validate resource ID as best we can
		ASSERT(!ItemID.IsNull());

		// Get that string!
		String ItemName;
		ItemID.Load(&ItemName);

		// Set the item
		return SelectByString(ItemName);
	}

	// Access currently selected item
	bool ListControl::GetSelectedItem(int* piSelectedItem) const
	{
		return GetImp()->GetSelectedItem(piSelectedItem);
	}

	bool ListControl::operator==(int iItem) const
	{
		ASSERT(iItem >= 0);

		int nSelectedItem;

		if (GetImp()->GetSelectedItem(&nSelectedItem))
		{
			// Is it a match
			return (nSelectedItem == iItem);
		}

		// No selected item
		return false;
	}

	bool ListControl::operator==(const String& ItemName) const
	{
		// String should not be empty
		ASSERT(!ItemName.IsEmpty());

		// Get name of selected item
		int nSelectedItem;
		if (GetImp()->GetSelectedItem(&nSelectedItem))
		{
			String SelectedItemName;
			
			if (GetImp()->GetItemText(nSelectedItem, &SelectedItemName))
			{
				if (SelectedItemName.IsEmpty())
				{
					// If item is empty, don't match - but also, there shouldn't
					// be any empty names, so complain.
					vguiERROR();
					return false;
				}

				// Does the string match?
				return ItemName == SelectedItemName;
			}
			else
			{
				// Caller should be told if they try to access an illegal
				// item
				vguiERROR();
			}
		}

		// Problem accessing selected item/name
		return false;
	}

	bool ListControl::operator==(const StringResourceID& ItemID) const
	{
		// Validate resource ID as best we can
		ASSERT(!ItemID.IsNull());

		// Load the string
		String ItemName;
		ItemID.Load(&ItemName);

		// Compare with the selected item
		return (*this == ItemName);
	}


	bool ListControl::operator!=(int iItem) const
	{
		ASSERT(iItem >= 0);

		return !(*this == iItem);
	}

	bool ListControl::operator!=(const String& ItemName) const
	{
		// String should not be empty
		ASSERT(!ItemName.IsEmpty());

		return !(*this == ItemName);
	}

	bool ListControl::operator!=(const StringResourceID& ItemID) const
	{
		// Validate resource ID as best we can
		ASSERT(!ItemID.IsNull());

		return !(*this == ItemID);
	}

	void ListControl::RemeasureItems()
	{
		GetImp()->RemeasureItems();
	}

	void ListControl::RemeasureItems(int NewHeight)
	{
		GetImp()->RemeasureItems(NewHeight);
	}

	/**************************************************************************
	
	>	void ListControl::EnableSorting()
	
		Author:		Tim Browse
		Date:		2/4/2002
	
		Purpose:	Enable sorting on this list control - items will be
					kept sorted.
	
	**************************************************************************/
	
	void ListControl::EnableSorting()
	{
		GetImp()->EnableSorting();
	}

	/**************************************************************************
	
	>	void ListControl::DisableSorting()
	
		Author:		Tim Browse
		Date:		2/4/2002

		Purpose:	Disable sorting - items will not be sorted as changes
					are made.
	
	**************************************************************************/
	
	void ListControl::DisableSorting()
	{
		GetImp()->DisableSorting();
	}

	/**************************************************************************
	
	>	void ListControl::SetSortOrder(TListSortOrder Order)
	
		Author:		Tim Browse
		Date:		9/4/2002
		Inputs:		Order - the sort order to use - one of:
								ListSortOrder_Ascending
								ListSortOrder_Descending
	
		Purpose:	Set the sort order for a list control
	
	**************************************************************************/
	
	void ListControl::SetSortOrder(TSortOrder Order)
	{
		m_SortOrder = Order;
	}

	/**************************************************************************
	
	>	TListSortOrder ListControl::GetSortOrder() const
	
		Author:		Tim Browse
		Date:		9/4/2002
		Returns:	The sort order - SortOrder_Ascending or
									 SortOrder_Descending
	
		Purpose:	Get the current sort order (ascending or descending) for
					this list control.
	
	**************************************************************************/
	
	ListControl::TSortOrder ListControl::GetSortOrder() const
	{
		return m_SortOrder;
	}

	/**************************************************************************
	
	>	void ListControl::SetSortFunction(pfnCompareItems pNewSortFunction,
										  UserData TheUserData = 0, 
										  int iColumn = 0);
	
		Author:		Tim Browse
		Date:		7/4/2002
		Inputs:		pNewSortFunction - the sort function to use.
					TheUserData - user data to pass to the sort function.
					iColumn - the column to use this sort function on
							  (ignored by single-column list controls)
							  if AllColumns, pNewSortFunction is used to
							  sort all columns that don't have their own
							  sort function specified.
	
		Purpose:	Set up the sort function to use when sorting the
					items in this list control.
	
	**************************************************************************/
	
	void ListControl::SetSortFunction(pfnCompareItems pNewSortFunction,
									  UserData TheUserData,
									  int iColumn)
	{
		if (iColumn == AllColumns)
		{
			m_DefaultSortFunction.m_pSortFunction = pNewSortFunction;
			m_DefaultSortFunction.m_UserData = TheUserData;
		}
		else
		{
			// Make sure we have enough room to store this function
			EnsureSortFunctionArraySize(iColumn + 1);

			// And store it
			m_pSortFunctions[iColumn].m_pSortFunction = pNewSortFunction;
			m_pSortFunctions[iColumn].m_UserData = TheUserData;
		}
	}

	/**************************************************************************
	
	>	ListControl::pfnCompareItems ListControl::GetSortFunction(int iColumn = 0) const
	
		Author:		Tim Browse
		Date:		7/4/2002
		Inputs:		iColumn - the column to get the this sort function for
							  (ignored by single-column list controls)
		Returns:	The sort function currently in use for this list control.
					(Never returns NULL)
	
		Purpose:	Get the sort function for this list control (so that, e.g.
					you can call it!)
	
	**************************************************************************/
	
	ListControl::pfnCompareItems ListControl::GetSortFunction(int iColumn) const
	{
		if (iColumn < m_cSortFunctions)
		{
			return m_pSortFunctions[iColumn].m_pSortFunction;
		}
		else
		{
			// Sort function not installed for this column - use default functions
			if (m_DefaultSortFunction.m_pSortFunction != NULL)
				return m_DefaultSortFunction.m_pSortFunction;
			else
				return CompareItems;
		}
	}

	UserData ListControl::GetSortUserData(int iColumn) const
	{
		if (iColumn < m_cSortFunctions)
		{
			return m_pSortFunctions[iColumn].m_UserData;
		}
		else
		{
			// (may be 0)
			return m_DefaultSortFunction.m_UserData;
		}
	}

	/**************************************************************************
	
	>	void ListControl::EnsureSortFunctionArraySize(int MinimumSize)
	
		Author:		Tim Browse
		Date:		8/4/2002
		Inputs:		MinimumSize - minimum number of elements required in
								  the array.
	
		Purpose:	Ensure the array of sort functions is at least the
					minimum size specified.
	
	**************************************************************************/
	
	void ListControl::EnsureSortFunctionArraySize(int MinimumSize)
	{
		if (MinimumSize <= m_cSortFunctions)
		{
			// Array is already big enough
			return;
		}

		// Quantise the array size to avoid realloc every time
		MinimumSize = (MinimumSize + 3) & ~3;
		
		// Make a new array, initialise unused entries, copy old data into 
		// it, and dispose of old one.
		ListControlSortInfo* pNewArray = new ListControlSortInfo[MinimumSize];

		for (int iIndex = 0; iIndex < MinimumSize; iIndex++)
		{
			if (iIndex < m_cSortFunctions)
			{
				// Copy existing entry
				pNewArray[iIndex] = m_pSortFunctions[iIndex];
			}
			else
			{
				// Initialise new entry
				pNewArray[iIndex].m_pSortFunction = ListControl::CompareItems;
				pNewArray[iIndex].m_UserData = 0;
			}
		}

		delete [] m_pSortFunctions;

		m_pSortFunctions = pNewArray;
		m_cSortFunctions = MinimumSize;
	}


	int ListControl::CompareItems(int iItem1, int iItem2, int iColumn) const
	{
		pfnCompareItems pSortFunction = GetSortFunction(iColumn);

		return (*pSortFunction)(this, GetSortUserData(iColumn), iItem1, iItem2, iColumn);
	}

	/**************************************************************************
	
	>	int ListControl::CompareItems(const ListControl* pListControl, 
									  UserData TheUserData,
									  int iItem1, int iItem2) const
	
		Author:		Tim Browse
		Date:		2/4/2002

		Inputs:		pListControl - the list control being sorted
					TheUserData - the user data passed to ListControl::SetSortFunction()
					iItem1/iItem2 - the indices of the list items to make
									a comparison of.
					iColumn - the column that the sorting should be done
							  with (for single-column list controls, this
							  will always be zero).

		Returns:	-1 => iItem1 precedes iItem2 in sort order
					 0 => iItem1 and iItem2 are equivalent in sort order
					 1 => iItem2 precedes iItem1 in sort order
	
		Purpose:	Sort function for ordering items in a list control.

					This default implementation does a simple textual
					comparison.

					The point of the function is to provide an ordering
					for the two list items defined by the index values
					passed in the iItem1 and iItem2 parameters.
	
	**************************************************************************/
	
	int ListControl::CompareItems(const ListControl* pListControl, 
								  UserData TheUserData,
								  int iItem1, int iItem2,
								  int iColumn)
	{
		// Default sort implementation - simple textual ordering
		// (ignoring UserData)
		vos::String Item1Text, Item2Text;
		if (pListControl->GetItemText(iItem1, &Item1Text, iColumn) &&
			pListControl->GetItemText(iItem2, &Item2Text, iColumn))
		{
			return Item1Text.CompareOrder(Item2Text);
		}
		else
		{
			// Can't get full info - assume equivalence
			return 0;
		}
	}

};

