//
// ListControl.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_ListControl_h
#define vgui_ListControl_h

#include "vgui.h"

#include "Window.h"
#include "String.h"
#include "ResourceID.h"

#include <vector>

namespace vgui
{
	class ListControl;
	class ListControlImp;
	class OwnerDrawListControl;

	//
	// This class is never instantiated directly.
	//
	class ListControl : public Window
	{
		public:
			//
			// Function prototype for custom list box sort routines
			//
			typedef int (*pfnCompareItems)(const ListControl* pListControl, 
										   UserData TheUserData,
										   int iItem1, int iItem2, 
										   int iColumn);

			//
			// Sort direction for list controls
			//
			typedef enum
			{
				SortOrder_Ascending,
				SortOrder_Descending,
			}
			TSortOrder;
		

		private:
			class ListControlSortInfo;

		public:
			~ListControl();
			
			// Remove all strings
			virtual void Clear();

			// Find out how many there are.
			virtual int GetCount() const;

			// Initialise from 0-terminated array of string resources
			virtual bool Initialise(const StringResourceID* StringIDs);

			// Add and delete items
			virtual bool AddItem(const String& ItemName, 
								 int* piItem = NULL,
								 UserData TheUserData = NULL);

			virtual bool InsertItemAfter(int iAfterItem,
										 const String& ItemName, 
										 int* piItem = NULL,
										 UserData TheUserData = NULL);

			virtual bool InsertItemBefore(int iBeforeItem,
										  const String& ItemName, 
										  int* piItem = NULL,
										  UserData TheUserData = NULL);

			virtual bool DeleteItem(int iItem);

			//
			// Get/set item text
			//
			// NB. Column index of 0 should be used for list controls that
			//	   do not support multiple columns.

			virtual bool SetItemText(int iItem, const String& Text, int iColumn = 0);
			virtual bool GetItemText(int iItem, String* pItemText, int iColumn = 0) const;

			// Get/set user data for an item
			virtual bool SetItemData(int iItem, UserData TheUserData);
			virtual bool GetItemUserData(int iItem, UserData* pTheUserData) const;


			// Change currently selected item
			//
			// NB.  it is intentional that there is no assignment operator
			//	    that takes an int, because then the following would
			//		not behave as expected:
			//
			//			m_List = IDS_ITEM3;
			//
			//		where IDS_ITEM3 is a string resource ID.  This is
			//		unlikely to have the desired effect, as it would
			//		use the resource ID value as the index of the item
			//		to be selected.  The intended effect is extremely
			//		likely to be that the item that matches the
			//		string specified by IDS_ITEM3 should be selected.
			ListControl& operator=(const String& ItemName);
			ListControl& operator=(const StringResourceID& ItemName);

			virtual bool SelectByIndex(int iItem);
			virtual bool SelectByString(const String& ItemName);
			virtual bool SelectByResID(const StringResourceID& ItemID);

			// Access currently selected item
			virtual bool GetSelectedItem(int* piSelectedItem) const;

			virtual bool operator==(int iItem) const;
			virtual bool operator==(const String& ItemName) const;
			virtual bool operator==(const StringResourceID& ItemID) const;

			virtual bool operator!=(int iItem) const;
			virtual bool operator!=(const String& ItemName) const;
			virtual bool operator!=(const StringResourceID& ItemName) const;

			////////////////////////////////////////////////////////////////////
			//
			// Sorting control
			//
			////////////////////////////////////////////////////////////////////

			virtual void EnableSorting();
			virtual void DisableSorting();

			virtual void SetSortOrder(TSortOrder Order);
			virtual TSortOrder GetSortOrder() const;

			// Sort functions - if control is not multi-column, column parameters are
			// ignored.
			enum { AllColumns = -1 };
			virtual void SetSortFunction(pfnCompareItems pfn, 
									     UserData TheUserData = 0, 
										 int iColumn = AllColumns);

			// Used by ListControlImp derived classes to access user-defined
			// sorting functions - you do not need to call them directly.
			virtual pfnCompareItems GetSortFunction(int iColumn = 0) const;
			virtual UserData GetSortUserData(int iColumn = 0) const;

			// Use appropriate sort function to compare items
			// (do not call directly - used by framework)
			virtual int CompareItems(int iItem1, int iItem2, int iColumn) const;

			////////////////////////////////////////////////////////////////////
			//
			// Owner draw list controls
			//
			////////////////////////////////////////////////////////////////////

			// Access the owner draw handler, which for most list controls will
			// be NULL (i.e. they don't need owner draw functionality).
			OwnerDrawListControl* GetOwnerDrawHandler() { return m_pOwnerDrawHandler; }

			// Call when metrics have changed, and you want to re-measure all
			// the list items (OnMeasureItem will be called for each item).
			// (NB. see next function)
			void RemeasureItems();

			// Faster alternative to above, if you know the height in advance, 
			// and it is the same for each item.
			void RemeasureItems(int NewHeight);

			////////////////////////////////////////////////////////////////////


		protected:
			// Only derived classes may access the constructor.
			// => This class is never instantiated directly
			ListControl(ListControlImp *pImp);

			void EnsureSortFunctionArraySize(int MinimumSize);

			// Default sort function
			static int CompareItems(const ListControl* pListControl, 
									UserData TheUserData,
									int iItem1, int iItem2,
									int iColumn);

		private:
			ListControlImp* GetImp() const { return (ListControlImp*) m_pWindowImp; }

		private:
			// Copying objects of this class is prohibited
			ListControl& operator=(const ListControl& rhs);

		private:
			/**************************************************************************
			
			>	class ListControlSortInfo
			
				Author:		Tim Browse
				Date:		8/4/2002
			
				Purpose:	Hold information about how a list column should be sorted.
			
			**************************************************************************/
			
			class ListControlSortInfo
			{
				public:
					// Constructor
					ListControlSortInfo() : m_pSortFunction(NULL), m_UserData(0) {}

				public:
					// The function to sort with
					pfnCompareItems m_pSortFunction;

					// User data to pass to sort function
					UserData m_UserData;
			};
			
		protected:
			// The pointer to the owner draw handler, if this object supports
			// owner draw methods.
			OwnerDrawListControl* m_pOwnerDrawHandler;

			// Function(s) used for sorting, if any.
			ListControlSortInfo m_DefaultSortFunction;
			ListControlSortInfo* m_pSortFunctions;
			int m_cSortFunctions;

			// Sort order for this list control
			TSortOrder m_SortOrder;
	};
}

#endif // !vgui_ListControl_h
