//
// MRUFileMenu.cpp - class for handling a MRU file list, and updating the
//				     application's menu accordingly.
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "MRUFileMenu.h"

#include "vos\FileSpec.h"
#include "vos\Preferences.h"
#include "vos\Errors.h"

#include "MessageMap.h"

#include "Menu.h"

namespace vgui
{
	//
	// Message map for menu entries.
	//
	// The command is set to 0, because we update the table when we
	// initialise the MRU file menu.
	//
	// The handler function is NULL, as we just trap this in our own
	// OnCommand function; we don't dispatch it in the usual way.
	//
	BEGIN_VGUI_MESSAGE_MAP(MRUFileMenu, MessageTarget)
		ON_VGUI_COMMAND(0, NULL)
	END_VGUI_MESSAGE_MAP()

	MRUFileMenu* MRUFileMenu::m_pInstance = NULL;

	MRUFileMenu* MRUFileMenu::Instance()
	{
		if (m_pInstance == NULL)
		{
			// Create and initialise instance
			m_pInstance = new MRUFileMenu();

			// Register with application as a shutdown handler.
			GetApp()->RegisterShutdownHandler(m_pInstance);
		}

		return m_pInstance;
	}

	MRUFileMenu::MRUFileMenu() : m_MRUID(0)
	{
		// To start with, there is one (dummy entry on the MRU file list)
		m_cFileItems = 1;
	}

	// Set up the file menu
	void MRUFileMenu::Initialise(MenuCommandID MRUMenuItem, uint cFiles)
	{
		// Set file list size
		m_Files.SetMaxSize(cFiles);

		// Read MRU file list from preferences
		m_pInstance->ReadMRUFromPrefs();

		// Remember MRU ID
		m_MRUID = MRUMenuItem;

		// Get main app menu
		Menu *pMainMenu = GetApp()->GetMainMenu();

		// If that didn't work, we've got no chance
		if (pMainMenu == NULL)
		{
			COMPLAIN("Unable to find main app menu in MRUFileMenu::Initialise()!");
			return;
		}

		// Ok, find the first sub-menu from this...
		if (!pMainMenu->GetSubMenu(0, &m_FileMenu))
		{
			COMPLAIN("Unable to find File menu in MRUFileMenu::Initialise()!");
			return;
		}

		uint nPos;
		if (!m_FileMenu.FindItem(MRUMenuItem, &nPos))
		{
			// Unable to find MRU item
			COMPLAIN("Unable to find the MRU menu item in MRUFileMenu::Initialise()!");
			return;
		}

		// OK, remember initial text label
		MenuItem MRUItem;
		m_FileMenu.GetItem(nPos, &MRUItem);
		m_MRULabel = MRUItem.m_Text;

		// Now update file menu
		UpdateFileMenu();

		// All set up - tell the app that we want menu messages
		GetApp()->InstallMRU(this);
	}

	void MRUFileMenu::UpdateFileMenu()
	{
		// Look for the ID
		MenuCommandID ID = m_MRUID;

		uint nPos;
		if (!m_FileMenu.FindItem(ID, &nPos))
		{
			// Unable to find MRU item
			COMPLAIN("Unable to find the MRU menu item in MRUFileMenu::UpdateFileMenu()!");
			return;
		}

		// Delete the MRU items
		for (int i = 0; i < m_cFileItems; i++)
		{
			m_FileMenu.DeleteItem(nPos);
		}

		// All gone now
		m_cFileItems = 0;

		// Add all the files in the list to the menu.
		int cFiles = m_Files.GetSize();

		if (cFiles > 0)
		{
			int iFile = 0;

			while (iFile < cFiles)
			{
				// Construct new menu item
				MenuItem Item;
				vos::FileSpec FileName;
				if (m_Files.GetItem(iFile,&FileName))
				{
					ConvertFileSpecToMenuText(iFile, FileName, &Item.m_Text);
					Item.m_ID = ID;

					// Add to menu in the correct place
					if (m_FileMenu.InsertItem(Item, nPos))
					{
						// Item inserted correctly
						m_cFileItems++;

						// Put next item below it
						nPos++;
					}
				}

				// Do next file
				iFile++;

				// Use next ID up for the next file
				ID++;
			}
		}
		else
		{
			// Just put a disabled marker in the menu
			MenuItem NewItem;
			NewItem.m_Text = m_MRULabel;
			NewItem.m_bEnabled = false;
			NewItem.m_ID = m_MRUID;
			m_FileMenu.InsertItem(NewItem,nPos);
			m_cFileItems = 1;
		}

		// Update the command ID ranges in our message map, so that we
		// get notified of any MRU file items that get chosen by the user.
		vgui::MessageMapEntry *MMap = (vgui::MessageMapEntry *) _vgui_MessageMapEntries;

		MMap[0].m_nMsg1 = m_MRUID;
		MMap[0].m_nMsg2 = ID - 1;
	}

	// Tell object that the specified file has just been used.
	void MRUFileMenu::NotifyUsed(const vos::FileSpec& FS)
	{
		// Re-arrange file list
		m_Files.NotifyUsed(FS);

		// Update menu accordingly
		UpdateFileMenu();
	}

	void MRUFileMenu::OnShutdown(bool bNormalShutdown)
	{
		// On shutdown, save the MRU data to persistent storage
		WriteMRUToPrefs();
	}

	void MRUFileMenu::ReadMRUFromPrefs()
	{
		// Read MRU data from preferences
		vos::Preferences *pMRUSection = vos::GetPreferences().FindSection("MRUFileList");

		if (pMRUSection == NULL)
		{
			// oops - can't find prefs section
			return;
		}

		// First, read in the number of files
		int cFiles;
		if (!pMRUSection->Read("NumFiles", &cFiles))
		{
			// Can't read number of files - this probably means this is the first time
			// this app has run, so just exit now.
			return;
		}

		// Now, read in details of each file
		int iFile = 0;
		vos::FileSpec FileName;

		while (iFile < cFiles)
		{
			// Create file pref name
			String PrefName;
			PrefName.Printf("File%d", iFile);

			// Read in filename...
			vos::String szFileName;
			if (pMRUSection->Read(PrefName, &szFileName))
			{
				// Convert string to filespec and add to list
				FileName.SetFromString(szFileName);
				m_Files.Append(FileName);
			}

			// Try next one
			++iFile;
		}
	}

	void MRUFileMenu::WriteMRUToPrefs()
	{
		// Save MRU data to preferences
		vos::Preferences *pMRUSection = vos::GetPreferences().FindSection("MRUFileList");

		if (pMRUSection == NULL)
		{
			// oops - can't find prefs section - just die quietly then...probably
			// an abnormal shutdown.
			return;
		}

		// First, save out the number of files
		int cFiles = (int) m_Files.GetSize();
		pMRUSection->Write("NumFiles", cFiles);

		// Now, write out details of each file
		int iFile = 0;
		vos::FileSpec FileName;

		while (iFile < cFiles)
		{
			// Create file pref name
			String PrefName;
			PrefName.Printf("File%d", iFile);

			// Create flattened filename...
			//
			// NB. We should really have a FileSpec-specific write primitive for prefs, 
			//     so that Macs can do clever things like storing links/aliases in prefs.
			if (m_Files.GetItem(iFile, &FileName))
			{
				// Write out the filespec
				pMRUSection->Write(PrefName, FileName.GetString());
			}
			else
			{
				COMPLAIN("Logic error in WriteMRUToPrefs! Not enough files!");
			}

			// Try next one
			++iFile;
		}
	}

	bool MRUFileMenu::DispatchCommand(int nCommand,
									  const MessageMapEntry *pEntry)
	{
		// If this function is called, we have a match on one of our
		// MRU file entries, so we work out the filename being 
		// referenced, and tell the app about it.

		// Convert command id into file index
		int iFile = nCommand - m_MRUID;

		// Tell app that the file should be opened
		vos::FileSpec FS;
		if ((iFile >= 0) && m_Files.GetItem(iFile, &FS))
		{
			GetApp()->OnOpenFile(FS);
		}
		else
		{
			// This ID should not be routed here - it is out of range.
			COMPLAIN("Illegal MRU command ID received!");
		}

		// All done
		return true;
	}

}

