//
// MRUList.h - template class for managing fixed size MRU lists
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_MRUList_h
#define vgui_MRUList_h

#include <vector>
#include <algorithm>

#include "vos\vos.h"

template <class T>
class MRUList
{
	public:
		MRUList();

		void SetMaxSize(uint MaxSize);

	public:
		// Tell the list that an item has just been used, so that
		// it can re-order the list accordingly.
		void NotifyUsed(const T& Item);

		// Get the size
		uint GetSize() { return (uint) m_Items.size(); }

		// Get the nth item
		bool GetItem(uint n, T* pItem);

		// Make sure list is not too long - remove items from the end if
		// it is.
		void TrimList();

		// Append item to list - used when loading list from persistent storage.
		void Append(const T& Item);

	private:
		// Maximum size of list
		uint m_MaxSize;

		// List of items
		std::vector<T> m_Items;
};


// ----------------------------------------------------------------------

template <class T>
MRUList<T>::MRUList() : m_MaxSize(0)
{
}

template <class T>
void MRUList<T>::NotifyUsed(const T& Item)
{
	// First, do we have this item in the list yet?
	std::vector<T>::iterator itFound = std::find(m_Items.begin(), m_Items.end(), Item);

	if (itFound != m_Items.end())
	{
		// Found the item - remove it, as we want to move it to the head of the list.
		m_Items.erase(itFound);
	}

	// Ok, place newly used item at head of list
	m_Items.insert(m_Items.begin(), Item);
	TrimList();
}

template <class T>
void MRUList<T>::SetMaxSize(uint MaxSize)
{
	m_MaxSize = MaxSize;
	TrimList();
}

template <class T>
void MRUList<T>::TrimList()
{
	// Make sure list is not too big.
	while (m_Items.size() > m_MaxSize)
	{
		// Remove last item
		m_Items.erase(m_Items.end() - 1);
	}
}

template <class T>
bool MRUList<T>::GetItem(uint n, T* pItem)
{
	if (n >= m_Items.size())
	{
		// No such item
		return false;
	}
	else
	{
		// Return the item
		*pItem = m_Items[n];
		return true;
	}
}

template <class T>
void MRUList<T>::Append(const T& Item)
{
	// Place item at end of list.
	m_Items.push_back(Item);
	TrimList();
}

#endif // !vgui_MRUList_h
