//
// MessageMap.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_MessageMap_h
#define vgui_MessageMap_h

#include "vgui.h"

#include "Messages.h"

namespace vgui
{

	//
	// MessageTarget Member Function typedef
	//
	class MessageTarget;

	typedef void (MessageTarget::*MessageTargetFn)(void);
	typedef bool (MessageTarget::*MessageTargetFn_bool)(void);
	typedef bool (MessageTarget::*MessageTargetFn_bool_int)(int);

	typedef bool (MessageTarget::*MessageTargetFn_Full)(TMessageType MsgType, int nMsg, int nCtrlID);

	//
	// All possible function signatures for message handlers.
	//
	typedef enum
	{
		MMFnSig_none,				// Sentinel - not a real function signature.

		MMFnSig_v_v,				// void Fn(void)
		MMFnSig_b_v,				// bool Fn(void)
		MMFnSig_b_i,				// bool Fn(int)

		MMFnSig_full,				// bool Fn(TMessageType MsgType, int nMsg, int nCtrlID)
	}
	MMFnSig;

	//
	// MessageMapEntry
	//
	class MessageMapEntry
	{
	public:
		// Type of message
		TMessageType m_Type;

		// Message code range
		int	m_nMsg1,
			m_nMsg2;

		// Range of control IDs (both NOCTL if controls irrelevant)
		int m_nFirstID, 
			m_nLastID;

		// Function to call if we match this entry
		MessageTargetFn m_pHandler;

		// Function signature of the handler
		MMFnSig m_FnSig;
	};

	class MessageMap
	{
	public:
		// Message map of the base class, to allow hierarchical chaining.
		const MessageMap*		m_pBaseMap;

		// Entries for this message map.
		const MessageMapEntry*	m_pEntries;
	};


	//
	// Macro to declare the message map in a class.
	//

	//
	// Macros to begin and end the definition of a class's message map.
	//
#define BEGIN_VGUI_MESSAGE_MAP(theClass, baseClass)							\
		const vgui::MessageMap* theClass::GetMessageMap() const				\
			{ return &theClass::_vgui_MessageMap; }							\
		const vgui::MessageMap theClass::_vgui_MessageMap =					\
		{																	\
			&baseClass::_vgui_MessageMap,									\
			&theClass::_vgui_MessageMapEntries[0]							\
		};																	\
		vgui::MessageMapEntry theClass::_vgui_MessageMapEntries[] =			\
		{


#define END_VGUI_MESSAGE_MAP()														\
			{vgui::mt_NONE, 0, 0, 0, 0, (vgui::MessageTargetFn) NULL, vgui::MMFnSig_none }	\
		};

	//
	// Macros to define entries in a class's message map.
	//

#define ON_VGUI_COMMAND(nCommandID, CommandFn)								\
			{vgui::mt_COMMAND, nCommandID, nCommandID, 0, 0, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) CommandFn), vgui::MMFnSig_b_v},

#define ON_VGUI_COMMAND_EX(nCommandID, CommandFn)								\
			{vgui::mt_COMMAND, nCommandID, nCommandID, 0, 0, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool_int) (CommandFn)), vgui::MMFnSig_b_i},

#define ON_VGUI_COMMAND_RANGE(nCommandID1, nCommandID2, CommandFn)								\
			{vgui::mt_COMMAND, nCommandID1, nCommandID2, 0, 0, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool_int) (CommandFn)), vgui::MMFnSig_b_i},

#define ON_VGUI_REDRAW()								\
			{vgui::mt_REPAINT, 0, 0, 0, 0, (vgui::MessageTargetFn) (bool (Window::*)(void))&OnRedraw, vgui::MMFnSig_b_v},


//
// EditBox notification messages
//

#define ON_VGUI_EBN_TEXTCHANGED(CtrlID, HandlerFn)								\
			{vgui::mt_EBN_TEXTCHANGED, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},

#define ON_VGUI_EBN_TEXTCOMMIT(CtrlID, HandlerFn)								\
			{vgui::mt_EBN_TEXTCOMMIT, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},

//
// LisBox notification messages
//

#define ON_VGUI_LBN_SELCHANGED(CtrlID, HandlerFn)								\
			{vgui::mt_LBN_SELCHANGED, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},

#define ON_VGUI_LBN_DOUBLECLICK(CtrlID, HandlerFn)								\
			{vgui::mt_LBN_DOUBLECLICK, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool_int) (HandlerFn)), vgui::MMFnSig_b_i},

#define ON_VGUI_LBN_CHECKBOXCHANGED(CtrlID, HandlerFn)								\
			{vgui::mt_LBN_CHECKBOXCHANGED, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool_int) (HandlerFn)), vgui::MMFnSig_b_i},

	
//
// DropDownList notification messages
//

#define ON_VGUI_DLN_SELCHANGED(CtrlID, HandlerFn)								\
			{vgui::mt_DLN_SELCHANGED, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},


//
// Slider notification messages
//

#define ON_VGUI_SLN_POSCHANGE(CtrlID, HandlerFn)								\
			{vgui::mt_SLN_POSCHANGE, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},

#define ON_VGUI_SLN_ENDTRACK(CtrlID, HandlerFn)								\
			{vgui::mt_SLN_ENDTRACK, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},

#define ON_VGUI_SLN_ANYCHANGE(CtrlID, HandlerFn)								\
			{vgui::mt_SLN_TRACKPOS, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},	\
			{vgui::mt_SLN_POSCHANGE, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},	\
			{vgui::mt_SLN_ENDTRACK, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},

	
//
// Button notification messages
//
#define ON_VGUI_BN_CLICKED(CtrlID, HandlerFn)								\
			{vgui::mt_BN_CLICKED, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},

#define ON_VGUI_BN_CLICKED_ALL(HandlerFn)								\
			{vgui::mt_BN_CLICKED, 0, 0, vgui::NOCTL, vgui::NOCTL, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_full},

//
// Tab control notification messages
//

#define ON_VGUI_TAB_CHANGING(CtrlID, HandlerFn)	\
			{vgui::mt_TAB_CHANGING, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},

#define ON_VGUI_TAB_SELECTED(CtrlID, HandlerFn)	\
			{vgui::mt_TAB_SELECTED, 0, 0, CtrlID, CtrlID, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_bool) (HandlerFn)), vgui::MMFnSig_b_v},

//
// User defined messages
//
#define ON_VGUI_USERMSG(HandlerFn, Msg)								\
			{vgui::mt_USER, (Msg), (Msg), vgui::NOCTL, vgui::NOCTL, (vgui::MessageTargetFn) ((vgui::MessageTargetFn_Full) (HandlerFn)), vgui::MMFnSig_full},
};

#endif // !vgui_MessageMap_h
