//
// MessageTarget.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "MessageTarget.h"

#include "MessageMap.h"
#include "MessageParams.h"

#include "Errors.h"

#include <algorithm>

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{

	union MessageMapFunctions
	{
		// Generic member function pointer
		MessageTargetFn m_pRawFn;

		// Specific type safe variants for command routing
		bool (MessageTarget::*m_pCommandFn)();

		// Message handlers
		void (MessageTarget::*m_pHandlerFn_v_v)();								// MMFnSig_v_v
		bool (MessageTarget::*m_pHandlerFn_b_v)();								// MMFnSig_b_v
		bool (MessageTarget::*m_pHandlerFn_b_i)(int);							// MMFnSig_b_i

		bool (MessageTarget::*m_pfullHandlerFn)(TMessageType MsgType, 
												int nMsg, 
												int nCtrlID);				// MMFnSig_full
	};


	MessageTarget::MessageTarget()
	{
	}

	MessageTarget::~MessageTarget()
	{
	}

	// Message map for MessageTarget base class is the end of the line
	// for all message map routing.

	const MessageMap* MessageTarget::GetMessageMap() const
	{
		return &_vgui_MessageMap;
	}

	const MessageMap MessageTarget::_vgui_MessageMap =
	{
		NULL,							// We have no base class to look at.
		&_vgui_MessageMapEntries[0]		// Ours is the final message map (and it's empty).
	};

	//
	// The base message map - it's empty; the last one to look at.
	//

	vgui::MessageMapEntry MessageTarget::_vgui_MessageMapEntries[] =
	{
		{mt_NONE, 0, 0, 0, 0, (MessageTargetFn) NULL }
	};


	bool MessageTarget::OnMessage(TMessageType MsgType,
								  int nMsg,
								  int nCtrlID,
								  MessageParams* pParams)
	{
		// Try any dynamic message maps first
		if (!m_DynamicMaps.empty())
		{
			DynamicMessageMaps::iterator itMap = m_DynamicMaps.begin();

			while (itMap != m_DynamicMaps.end())
			{
				if ((*itMap)->OnMessage(MsgType, nMsg, nCtrlID, pParams))
				{
					// Message was processed
					return true;
				}

				// Try next target object
				++itMap;
			}
		}

		// Ok, process using this target's normal message map.
		return OnMessage(GetMessageMap(), MsgType, nMsg, nCtrlID, pParams);
	}

	bool MessageTarget::OnMessage(const MessageMap* pMap,
								  TMessageType MsgType,
								  int nMsg,
								  int nCtrlID,
								  MessageParams* pParams)
	{
		while (pMap != NULL)
		{
			// See if there is a match in this map
			const MessageMapEntry *pEntry;

			if (FindMatchingEntry(pMap->m_pEntries,
								  MsgType,
								  nMsg,
								  nCtrlID,
								  pEntry))
			{
				// Found a match - see if the handler wants it...
				if (CallMessageHandler(pEntry, MsgType, nMsg, nCtrlID, pParams))
				{
					// Message handled
					return true;
				}
			}

			// Message not handled by this map, so try next message map
			pMap = pMap->m_pBaseMap;
		}

		// Message not handled.
		return false;
	}

	bool MessageTarget::FindMessageHandler(TMessageType MsgType,
										   int nMsg,
										   int nCtrlID,
										   const MessageMapEntry*& pFoundEntry)
	{
		// Get this target's message map.
		const MessageMap* pMap = GetMessageMap();

		// Find message handler
		while (pMap != NULL)
		{
			if (FindMatchingEntry(pMap->m_pEntries,
								  MsgType,
								  nMsg,
								  nCtrlID,
								  pFoundEntry))
			{
				// Found a match - tell caller
				return true;
			}

			// Message not handled by this map, so try next message map
			pMap = pMap->m_pBaseMap;
		}

		// Message handler not found.
		return false;
	}


	bool MessageTarget::CallMessageHandler(const MessageMapEntry* pEntry,
										   TMessageType MsgType,
										   int nMsg,
										   int nCtrlID,
										   MessageParams* pParams)
	{
		//
		// To call the handler, we put the pointer into the 
		// function union structure...
		//
		MessageMapFunctions mmf;
		mmf.m_pRawFn = pEntry->m_pHandler;

		switch (pEntry->m_FnSig)
		{
			case MMFnSig_b_v:
			{
				if ((this->*mmf.m_pHandlerFn_b_v)())
				{
					// Message handled
					return true;
				}

				break;
			}

			case MMFnSig_b_i:
			{
				if (pParams != NULL)
				{
					if ((this->*mmf.m_pHandlerFn_b_i)(pParams->m_Int1))
					{
						// Message handled
						return true;
					}
				}
				else
				{
					// Oops - someone's forgotten to give us some parameters, and
					// we need them for this type of handler.
					COMPLAIN("CallMessageHandler() - MessageParams pointer is NULL!");
					if ((this->*mmf.m_pHandlerFn_b_i)(0))
					{
						// Message handled
						return true;
					}
				}

				break;
			}

			case MMFnSig_full:
			{
				if ((this->*mmf.m_pfullHandlerFn)(MsgType, nMsg, nCtrlID))
				{
					// Message handled
					return true;
				}

				break;
			}
		}

		// If we get here, it wasn't handled.
		return false;
	}

	bool MessageTarget::OnCommand(int nCommand)
	{
		// Try any dynamic message maps first
		if (!m_DynamicMaps.empty())
		{
			DynamicMessageMaps::iterator itMap = m_DynamicMaps.begin();

			while (itMap != m_DynamicMaps.end())
			{
				if ((*itMap)->OnCommand(nCommand))
				{
					// Message (command) was processed
					return true;
				}

				// Try next target object
				++itMap;
			}
		}

		// Ok, process using this target's normal message map.
		return OnCommand(GetMessageMap(), nCommand);
	}

	bool MessageTarget::OnCommand(const MessageMap* pMap, int nCommand)
	{
		// Keep searching until command is handled or we run 
		// out of message maps to look at.
		while (pMap != NULL)
		{
			// See if there is a match in this map
			const MessageMapEntry *pEntry;

			if (FindCommandEntry(pMap->m_pEntries, nCommand, pEntry))
			{
				// Found a match - see if the handler wants it...
				if (DispatchCommand(nCommand, pEntry))
				{
					// Message handled
					return true;
				}
			}

			// Message not handled by this map, so try next message map
			pMap = pMap->m_pBaseMap;
		}

		// Message not handled.
		return false;
	}


	bool MessageTarget::DispatchCommand(int nCommand,
										const MessageMapEntry *pEntry)
	{
		//
		// Scary syntax alert!
		//
		// To call the handler, we put the pointer into the 
		// function union structure...
		//
		MessageMapFunctions mmf;
		mmf.m_pRawFn = pEntry->m_pHandler;

		// Work out what type of command handler it is (i.e. does it take
		// a parameter specifying the command being sent?), and use the
		// bizarre member function syntaxt to call the handler
		switch (pEntry->m_FnSig)
		{
			case MMFnSig_b_v:
			{
				return (this->*mmf.m_pCommandFn)();
			}

			case MMFnSig_b_i:
			{
				return (this->*mmf.m_pHandlerFn_b_i)(nCommand);
			}
		}

		// If we get here, it's gone a bit wrong
		COMPLAIN("DispatchCommand() - Wrong type of function handler for a command message!");
		return false;
	}

	bool MessageTarget::FindCommandEntry(const MessageMapEntry *pEntries,
										 int nCommand,
										 const MessageMapEntry*& pFoundEntry)
	{
		const MessageMapEntry* pEntry = pEntries;

		while (pEntry->m_Type != mt_NONE)
		{
			//
			// Is this a match?
			//
			// NB. we support ranges of commands for a handler, hence this check...
			//
			if ((pEntry->m_Type == mt_COMMAND) && 
				(nCommand >= pEntry->m_nMsg1) &&
				(nCommand <= pEntry->m_nMsg2))
			{
				// Yes
				pFoundEntry = pEntry;
				return true;
			}

			// No, so try next entry
			++pEntry;
		}

		// No entry found
		return false;
	}

	bool MessageTarget::FindMatchingEntry(const MessageMapEntry *pEntries,
										  TMessageType MsgType,
										  int nMsg,
										  int nCtrlID, 
										  const MessageMapEntry*& pFoundEntry)
	{
		const MessageMapEntry* pEntry = pEntries;

		const MessageMapEntry* pCatchAllEntry = NULL;

		while (pEntry->m_Type != mt_NONE)
		{
			// Check for general match
			if ((pEntry->m_Type == MsgType) && 
				(pEntry->m_nMsg1 == nMsg))
			{
				// Is this a positive match?
				if ((nCtrlID >= pEntry->m_nFirstID) &&
					(nCtrlID <= pEntry->m_nLastID))
				{
					// Yes
					pFoundEntry = pEntry;
					return true;
				}

				// Ok, not a full match, but is this a matching catch all handler?
				if ((pEntry->m_nFirstID == NOCTL) &&
					(pEntry->m_nLastID == NOCTL))
				{
					// Yes - remember it in case this is the closest match.
					pCatchAllEntry = pEntry;
				}
			}

			// No, so try next entry
			++pEntry;
		}

		// No entry found - did we find a catch all entry?
		if (pCatchAllEntry != NULL)
		{
			// Yes, so tell caller about it
			pFoundEntry = pCatchAllEntry;
			return true;
		}

		// No entry found at all.
		return false;
	}

	void MessageTarget::AddMessageTarget(MessageTarget* pMessageTarget)
	{
		m_DynamicMaps.push_back(pMessageTarget);
	}

	void MessageTarget::RemoveMessageTarget(MessageTarget* pMessageTarget)
	{
		// Remove any instances of this message target from our list
		m_DynamicMaps.remove(pMessageTarget);
	}

};
