//
// Metrics.cpp - metric information about GUI elements
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Metrics.h"

#ifdef VGUI_MSWINDOWS
#include "Win32\Win32_WindowImp.h"
#include <zmouse.h>
#endif

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	using namespace vos;

#ifdef VGUI_MSWINDOWS
	//
	// Support function for Windows - converts DLU (Dialog units) to
	// pixel values.
	//
	// THe default control sizes are defined in terms of DLUs, but we need
	// to be able to convert them to pixels.
	//
	// Note that DLUs are anisotropic, so we have different functions for
	// X and Y dimensions.
	//
	static int ConvertDLUToWidth(int DLU)
	{
		static int ScaleDLU = 0;

		if (ScaleDLU == 0)
		{
			// Get the DLU scale factor for x dimensions
			ScaleDLU = LOWORD(GetDialogBaseUnits());
		}

		return ::MulDiv(DLU, ScaleDLU, 4);
	}

	static int ConvertDLUToHeight(int DLU)
	{
		static int ScaleDLU = 0;

		if (ScaleDLU == 0)
		{
			// Get the DLU scale factor for y dimensions
			ScaleDLU = HIWORD(GetDialogBaseUnits());
		}

		return ::MulDiv(DLU, ScaleDLU, 8);
	}

#endif

	uint Metrics::GetTextMargin()
	{
		#if defined VGUI_MSWINDOWS
			// Standard margin of 8 pixels around text.
			return 8;
		#elif defined VGUI_MACOS
			// Will this change for Aqua?
			// 6 for items, but 5 between text and item it defines
			return 5;	
		#else
			#error GetTextMargin() needs porting.
		#endif
	}

	uint Metrics::GetButtonWidth(Dialog* pParent)
	{
		#ifdef VGUI_MSWINDOWS
			// Get the handle to the dialog
			HWND hDlg = Win32_WindowImp::GetHWND(pParent);
			RECT ButtonSize = { 0, 0, 50, 14};

			// Map the button area, and return the mapped width.
			::MapDialogRect(hDlg, &ButtonSize);
			return ButtonSize.right;
		#elif defined (VGUI_MACOS)
			return 70;
		#else
			#error GetButtonWidth() needs porting.
		#endif
	}

	uint Metrics::GetButtonHeight(Dialog* pParent)
	{
		#ifdef VGUI_MSWINDOWS
			// Get the handle to the dialog
			HWND hDlg = Win32_WindowImp::GetHWND(pParent);
			RECT ButtonSize = { 0, 0, 50, 14};

			// Map the button area, and return the mapped height
			::MapDialogRect(hDlg, &ButtonSize);
			return ButtonSize.bottom;
		#elif defined (VGUI_MACOS)
			return 20;
		#else
			#error GetButtonHeight() needs porting.
		#endif
	}

	uint Metrics::GetProgressBarHeight(Dialog* pParent)
	{
		#ifdef VGUI_MSWINDOWS
			// Get the handle to the dialog
			HWND hDlg = Win32_WindowImp::GetHWND(pParent);
			RECT ProgressBarSize = { 0, 0, 50, 14};

			// Map the button area, and return the mapped height
			::MapDialogRect(hDlg, &ProgressBarSize);
			return ProgressBarSize.bottom;
		#elif defined (VGUI_MACOS)
			return 16;
		#else
			#error GetProgressBarHeight() needs porting.
		#endif
	}

	uint Metrics::GetControl3DBorderSize()
	{
		#ifdef VGUI_MSWINDOWS
			// 3D borders are 2 pixels wide.
			return 2;
		#elif defined (VGUI_MACOS)
			return 1;
		#else
			#error GetControl3DBorderSize() needs porting.
		#endif
	}

	uint Metrics::GetMouseWheelDelta()
	{
		#ifdef VGUI_MSWINDOWS
			return WHEEL_DELTA;
		#else
			#error GetMouseWheelDelta() needs porting.
		#endif
	}

	uint Metrics::GetListBoxColumnGap()
	{
		#ifdef VGUI_MSWINDOWS
			return 1;
		#else
			#error GetListBoxColumnGap() needs porting.
		#endif
	}

	uint Metrics::GetHorizontalScrollBarHeight()
	{
		return ::GetSystemMetrics(SM_CYHSCROLL);
	}

	uint Metrics::GetVerticalScrollBarWidth()
	{
		return ::GetSystemMetrics(SM_CXVSCROLL);
	}

	vgui::Rect Metrics::GetPrimaryDisplayPosition()
	{
    	// Get size of screen
    	// TODO: Multi-monitor support
    	HWND hDesktop = ::GetDesktopWindow();

    	Rect rcDesktop;
    	::GetClientRect(hDesktop, rcDesktop.GetPointer());
		return rcDesktop;
	}

};

