//
// MsgDialog.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "MsgDialog.h"

#include "Dialog.h"
#include "Application.h"
#include "RedrawGC.h"
#include "Button.h"
#include "Metrics.h"

#include "vos\Errors.h"
#include "vos\Trace.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	MsgDialog::MsgDialog(const DialogResourceID& DialogID,
						 Window* pParent,
						 ControlID MessageBox) : Dialog(DialogID, pParent),
												 m_MessageBoxID(MessageBox)
	{
	}

	MsgDialog::~MsgDialog()
	{
		// Virtual destructor
	}

	void MsgDialog::OnInitDialog()
	{
		// Set the alert title
		SetText(m_Title);

		// Attach to the message control
		m_MessageBox.Attach(this, m_MessageBoxID);

		// Get its size
		m_MessageBox.GetPosition(&m_MessageArea);

		// Now hide it as we don't want to see it.
		m_MessageBox.ShowWindow(sw_HIDE);

		// Initialise derived classes
		InitTheDialog();

		// Make sure the textbox is big enough for the message.
		ResizeMessage();
	}

	bool MsgDialog::OnRedraw()
	{
		RedrawGC gc(this);

		// Create a font
		Font* pFont = GetFont();
		gc.SetFont(pFont);

		// Get position of the text
		Rect TextArea = m_MessageArea;

		// Draw a border around it
		gc.Draw3DEdge(TextArea, edge3d_ThinSunken);

		// Draw the text in the border
		uint Margin = Metrics::GetTextMargin();
		TextArea.Deflate(Margin, Margin);

		gc.DrawText(m_Message, 
					TextArea, 
					tf_VCentre|tf_HCentre|tf_Multiline);

		// Clean up GC
		gc.RestoreDefaultFont();

		return true;
	}

	void MsgDialog::ResizeMessage()
	{
		// We need a GC to get font size information
		GC* pGC = GetGC();
		pGC->SetFont(GetFont());

		uint Margin = Metrics::GetTextMargin();

		// Is the message area big enough?
		Rect NewTextRect = m_MessageArea;

		pGC->CalcTextArea(m_Message, tf_Left|tf_Multiline, &NewTextRect);

		// Add a border
		NewTextRect.right += (2 * Margin);
		NewTextRect.bottom += (2 * Margin);

		// Resize to fit this text
		int HeightDiff = NewTextRect.GetHeight() - m_MessageArea.GetHeight();
		int WidthDiff = NewTextRect.GetWidth() - m_MessageArea.GetWidth();

		// Reposition any controls
		RepositionControls(&WidthDiff, &HeightDiff);

		// Resize the dialog too
		ResizeDialog(WidthDiff, HeightDiff);

		// Clean up GC
		pGC->RestoreDefaultFont();
		ReleaseGC(pGC);
	}

	void MsgDialog::ResizeDialog(int DeltaX, int DeltaY)
	{
		// Resize the dialog too
		Rect DialogArea;
		GetClientArea(&DialogArea);
		DialogArea.right += DeltaX;
		DialogArea.bottom += DeltaY;
		SetClientArea(DialogArea);

		// Update the message box.
		m_MessageArea.right += DeltaX;
		m_MessageArea.bottom += DeltaY;
	}

	void MsgDialog::AdjustMessageArea(int DeltaX, int DeltaY)
	{
		m_MessageArea.right += DeltaX;
		m_MessageArea.bottom += DeltaY;
	}

};

