//
// vgui/Point.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_Point_h
#define vgui_Point_h

#include <vector>

namespace vgui
{
	class Point
	{
		public:

			//
			// Constructors
			//

			Point() : x(0), y(0) {}

			// Construct from two integer co-ordinates.
			Point(int newx, int newy) : x(newx), y(newy)
				{}

			// Copy constructor
			Point(Point const& pt)
			{ 
				Set(pt); 
			}

			//
			// Comparison operators
			//

			// Same as calling IsEqual() to determine if an existing Point object 
			// has the same coordinates as this point.
			bool operator==(Point const& pt) const
			{
				return IsEqual(pt); 
			}

			// Same as calling !IsEqual() to determine if an existing Point object 
			// does not have the same coordinates as this point.
			bool operator!=(Point const& pt) const
			{
				return !IsEqual(pt); 
			}

			// Returns true if an existing Point object has the same coordinates as 
			// this point, false otherwise.
			bool IsEqual(Point const& pt) const
			{
				return ((pt.x == x) && 
				 	    (pt.y == y)); 
			}

			// Returns true if the coordinates of this point are 0, 0, false otherwise.
			bool IsEmpty() const
			{
				return ((x != 0) && (y != 0)); 
			}

			//
			// Assignment
			//

			Point& operator=(const Point& pt)
			{
				x = pt.x;
				y = pt.y;
				return *this;
			}

			// Set the coordinates of this point to be the values nXCoordinate and 
			// nYCoordinate.
			void Set(int newx, int newy)
			{
				x = newx; 
				y = newy;
			}

			// Set the coordinates of this point to be the values from an existing 
			// Point object.
			void Set(Point const& pt)
			{
				x = pt.x;
				y = pt.y;
			}

			// Set the coordinates of this point to 0, 0.
			void Empty()
			{
				x = y = 0; 
			}

			//
			// Operations
			//

			// Offset the X coordinate of this point by nDistanceX, and the 
			// Y coordinate by nDistanceY. 
			//
			// Either can be negative, which will subtract.
			void Offset(int nDistanceX, int nDistanceY)
			{ 
				x += nDistanceX; 
				y += nDistanceY; 
			}

#if defined VGUI_MSWINDOWS
			//
			// Win32 specific functions
			//

			// Construct by converting coordinates from an existing POINT structure.
			Point(POINT const& pt)
			{
				Set(pt); 
			}

			// Set the coordinates of this point to be the values from an 
			// existing POINT structure.
			void Set(POINT const& pt)
			{
				Set(pt.x, pt.y); 
			}
			
			// Return a POINT structure reference to this object.
			POINT& GetPOINT() const
			{
				return (POINT&) *this; 
			}

			// Return the a pointer to the internal POINT structure.
			POINT* GetPointer() const
			{
				return (POINT*) this; 
			}
			
#elif defined VGUI_MACOS
			// Set from a Mac Point
			void SetPoint( const ::Point& MacPoint )
			{
				x = MacPoint.h;
				y = MacPoint.v;
			}

			// Return a Mac Point
			::Point GetPoint() const
			{
				::Point p;
				
				// This may overflow. There's not a lot we can do about this.
				// Look at PowerPlant for inspiration
				p.h = y;
				p.v = x;
				
				// Two shorts packed for 68k alignment = 4 bytes, the size of a pointer
				// So this is NOT a speed hit, and it allows us to feed directly into QD
				return p;
			}
			
#endif

		public:
			// The data
			int x, y;
	};

	//
	// Name useful STL type
	//
	typedef std::vector<Point> Points;
};

#endif // !vgui_Point_h


