//
// ProgressDialog.cpp
//

#include "StdHdr.h"

#include "ProgressDialog.h"

#include "Metrics.h"
#include "Timer.h"

#include "vos\Trace.h"
#include "vos\Errors.h"
#include "vos\MonotonicTime.h"

namespace vgui
{
	/**************************************************************************

	>	ProgressDialog::ProgressDialog(JobWorker* pWorker, 
									   const vos::String& strProgressMessage,
									   Window* pParent = NULL)

		Author:		Tim Browse
		Date:		24/07/2001
		Inputs:		pWorker - the object doing the actual job which we
							  are displaying the progress UI for.
					strProgressMessage - the progress message to display in
										 the dialog.
					pParent - the parent window of this dialog

		Purpose:	Constructor

	**************************************************************************/

	ProgressDialog::ProgressDialog(JobWorker* pWorker, 
								   const vos::String& strProgressMessage,
								   Window* pParent)
								   
 								      : Dialog(DialogResourceID::s_DynamicDialog, pParent),
										m_strProgressMessage(strProgressMessage),
										m_pJobWorker(pWorker),
										m_bCancelled(false),
										m_TimerID(NullTimer),
										m_bShowPercentage(false),
										m_bShowCancelButton(true),
										m_bDialogInitialised(false)
	{
		// TCILB
	}



	/**************************************************************************

	>	void ProgressDialog::OnInitDialog()

		Author:		Tim Browse
		Date:		25/7/2001

		Purpose:	Set up the progress bar etc in the dialog.

	**************************************************************************/

	void ProgressDialog::OnInitDialog()
	{
		// Set the title text
		SetText(m_strProgressMessage);

		// Layout metrics
		const int kGap = 6;
		const int kDialogWidth = 300;
		const int kButtonWidth = Metrics::GetButtonWidth(this);
		const int kButtonHeight = Metrics::GetButtonHeight(this);
	
		// By using -kGap as ProgressHeight, we automatically remove the gap 
		// normally allowed beneath the progress slider.
		// (This only matters if we don't show a slider)
		int ProgressBarHeight = -kGap;

		if (m_bShowPercentage)
		{
			ProgressBarHeight = Metrics::GetProgressBarHeight(this);

			// Create progress bar control
			Rect ProgressPosition(kGap, kGap, kDialogWidth - kGap, kGap + ProgressBarHeight);
			m_ProgressBar.Create(this, ProgressPosition, NOCTL);

			// Control is always a percentage - 0% to 100%
			m_ProgressBar.SetRange(0, 100);
		}

		int ButtonHeight = -kGap;

		if (m_bShowCancelButton)
		{
			ButtonHeight = kButtonHeight;

			// Create 'cancel' button
			Rect CancelPosition;
			CancelPosition.left = (kDialogWidth / 2) - (kButtonWidth / 2);
			CancelPosition.top  = kGap + ProgressBarHeight + kGap;
			CancelPosition.right = CancelPosition.left + kButtonWidth;
			CancelPosition.bottom = CancelPosition.top + kButtonHeight;
			
			m_Cancel.Create(this, CancelPosition, IDCANCEL);
			m_Cancel.SetText("Cancel");
		}

		// Resize dialog to desired size
		Rect NewClientArea(0,0, 
						   kDialogWidth, 
						   kGap + ProgressBarHeight + ButtonHeight + kGap);
		SetClientArea(NewClientArea);

		if (m_pJobWorker != NULL)
		{
			// Set timer callback periodically (every 50ms) so we can do chunks of work
			m_TimerID = Timer::SetTimer(TimerType_Periodic, 50, this);
		}
		else
		{
			// Set timer callback once so we can trigger the job to start.
			m_TimerID = Timer::SetTimer(TimerType_OneShot, 0, this);
		}

		// Set progress control to show how far we are through the job
		// (the dialog is not always shown immediately, so some of the work
		// may be done already - See DoTheWork()).
		SetPosition(m_PercentComplete);

		// All done
		m_bDialogInitialised = true;
	}

	/**************************************************************************

	>	bool ProgressDialog::OnTimer(TimerID ID)

		Author:		Tim Browse
		Date:		2001/10/25
		Inputs:		ID - ID of the timer we are processing
		Returns:	true (always)

		Purpose:	Handler timer events - this is when the actual 'work' 
					gets done - we call the worker object to do a bit of work.

	**************************************************************************/

	bool ProgressDialog::OnTimer(TimerID ID)
	{
		// If user's super quick and has already cancelled, then check out of here!
		if (m_bCancelled)
		{
			EndDialog(IDOK);
			return true;
		}

		// Sanity check
		COMPLAIN_IF(ID != m_TimerID, 
					"ProgressDialog::OnTimer(): Wrong timer ID in timer event!");

		// Let's call our worker object to see if it wants to do any more work.
		if (m_pJobWorker->DoWork(this, &m_PercentComplete) == Job_Finished)
		{
			// All done, so get rid of this dialog.
			EndDialog(IDOK);
			return true;
		}

		// Ok, not finished, so update the progress bar position
		if (m_bShowPercentage)
			SetPosition(m_PercentComplete);

		return true;
	}

	/**************************************************************************

	>	void ProgressDialog::OnCancel()

		Author:		Tim Browse
		Date:		2001/10/25

		Purpose:	Handle the user cancelling the operation - we tell the 
					worker object that the operation has been cancelled, and
					then get rid of this dialog.

	**************************************************************************/

	void ProgressDialog::OnCancel()
	{
		// User has pressed cancel
		m_bCancelled = true;

		// First, kill our timer event so we don't get any more timer events.
		// NB. This is particularly important if the worker object displays
		//	   any dialogs during OnCancel().
		Timer::KillTimer(m_TimerID);

		// Tell the worker object that the user has cancelled
		m_pJobWorker->OnCancel(this);

		// And shut this dialog down using default behaviour
		Dialog::OnCancel();
	}

	/**************************************************************************

	>	bool ProgressDialog::IsJobCancelled() const

		Author:		Tim Browse
		Date:		2001/10/27
		Returns:	true => user has clicked the cancel button in the dialog.
					false => operation should continue.

		Purpose:	Find out if the user has cancelled the operation.

	**************************************************************************/

	bool ProgressDialog::IsJobCancelled() const
	{
		return m_bCancelled;
	}

	/**************************************************************************

	>	void ProgressDialog::UpdateJobProgressDisplay(uint PercentComplete)

		Author:		Tim Browse
		Date:		2001/10/27
		Inputs:		PercentComplete - new position of progress slider; must
									  be at least 0, and no more than 100.

		Purpose:	For use by JobWorker objects to update the progress display.
					(historic - probably not needed anymore.  TODO: check this!)

	**************************************************************************/

	void ProgressDialog::UpdateJobProgressDisplay(uint PercentComplete)
	{
		// Update the progress bar, force the window to be updated, and service
		// the UI.
		if (m_bShowPercentage && m_bDialogInitialised)
		{
			SetPosition(PercentComplete);
			UpdateWindow();
			GetApp()->ProcessMessages();
		}
	}

	/**************************************************************************

	>	void ProgressDialog::SetPosition(int PercentComplete)

		Author:		Tim Browse
		Date:		2001/10/27
		Inputs:		PercentComplete - new position of progress slider; must
									  be at least 0, and no more than 100.

		Purpose:	Central function for setting the slider position and
					doing range-checking.

	**************************************************************************/

	void ProgressDialog::SetPosition(int PercentComplete)
	{
		COMPLAIN_IF(!m_bShowPercentage,
					"ProgressDialog::SetPosition called when percentage display not enabled.");

		// Range sanity checks
		if (PercentComplete < 0)
		{
			PercentComplete = 0;
			COMPLAIN("Percentage value passed to ProgressDialog out of range!");
		}

		if (PercentComplete > 100)
		{
			PercentComplete = 100;
			COMPLAIN("Percentage value passed to ProgressDialog out of range!");
		}

		// Ok, update the progress bar position, if it's available
		if (m_bDialogInitialised)
			m_ProgressBar.SetPosition(PercentComplete);
	}

	/**************************************************************************
	
	>	bool ProgressDialog::ShowProgress()
	
		Author:		Tim Browse
		Date:		30/4/2002
		Returns:	true => work finished
					false => user cancelled the job
	
		Purpose:	Execute the JobWorker object attached to this object, 
					and show a progress indicator for jobs which take an unknown
					amount of time (i.e. no precentage complete is displayed).
	
	**************************************************************************/
	
	bool ProgressDialog::ShowProgress()
	{
		m_bShowPercentage = false;
		return DoTheWork();
	}

	/**************************************************************************
	
	>	bool ProgressDialog::ShowProgressWithPercentage()
	
		Author:		Tim Browse
		Date:		30/4/2002
		Returns:	true => work finished
					false => user cancelled the job
	
		Purpose:	Execute the JobWorker object attached to this object, 
					and show a progress indicator for jobs which take an known
					amount of time (i.e a precentage complete value/slider 
					is displayed).
	
	**************************************************************************/
	
	bool ProgressDialog::ShowProgressWithPercentage()
	{
		m_bShowPercentage = true;
		return DoTheWork();
	}

	/**************************************************************************
	
	>	bool ProgressDialog::ShowProgressWithoutCancel()
	
		Author:		Tim Browse
		Date:		4/5/2002
		Returns:	true => work finished
					false => user cancelled the job
	
		Purpose:	Execute the JobWorker object attached to this object, 
					and show a progress indicator for jobs which take an known
					amount of time, but do not include a Cancel button on the
					dialog.
	
	**************************************************************************/
	
	bool ProgressDialog::ShowProgressWithoutCancel()
	{
		m_bShowPercentage = true;
		m_bShowCancelButton = false;
		return DoTheWork();
	}

	/**************************************************************************
	
	>	bool ProgressDialog::DoTheWork(uint msInitialDelay = 300)
	
		Author:		Tim Browse
		Date:		4/5/2002
		Inputs:		msInitialDelay - how long to execute the job before
									 showing the progress dialog.
		Returns:	true => job completed ok
					false => job was cancelled.
	
		Purpose:	Show the dialog, and execute the job using the worker
					object.

					There is a delay before showing the dialog for jobs, in
					case the job finishes very quickly.  e.g. if a job only lasts
					1/5th of a second, it's not worth showing a dialog, because
					it will just flash up quickly and be annoying.

					This can obviously still happen, but for quick jobs it
					won't.

	**************************************************************************/
	
	bool ProgressDialog::DoTheWork(uint msInitialDelay)
	{
		m_PercentComplete = 0;

		// To start with, do work without a progress display for a short
		// period of time, because the job might complete quickly.
		uint32 StartTime = vos::GetMonotonicTime();
		uint32 TimeNow = StartTime;

		while (TimeNow < (StartTime + msInitialDelay))
		{
			// Get worker to do some work.
			if (m_pJobWorker->DoWork(this, &m_PercentComplete) == Job_Finished)
			{
				// All done!
				return true;
			}

			TimeNow = vos::GetMonotonicTime();
		}

		// Ok, we've been going longer than the initial delay, so let's
		// get the dialog going
		return (Dialog::DoModal() == IDOK);
	}

	/**************************************************************************
	
	>	ControlID ProgressDialog::DoModal()
	
		Author:		Tim Browse
		Date:		30/4/2002
	
		Purpose:	Don't call this.  Call the ShowProgressXXX() functions
					instead.
	
	**************************************************************************/
	
	ControlID ProgressDialog::DoModal()
	{
		// Make sure client does not call this in error.
		COMPLAIN("Do not call ProgressDialog::DoModal() - call the "
				 "ShowProgressXXX() functions instead.");
		return IDCANCEL;
	}

};

