//
// Question.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Question.h"

#include "QuestionDialog.h"

#include "Errors.h"
#include "Msgs.h"

#include "vos\Preferences.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	//
	// Template ID of the dialog to use when asking the user a question.
	//
	DialogResourceID Question::s_QuestionDialogID = 0;

	// IDs of important controls in the above dialog template.
	ControlID Question::s_QuestionTextControl = NOCTL;
	ControlID Question::s_QuestionDefButtonControl = NOCTL;
	ControlID Question::s_QuestionButtonControl = NOCTL;
	ControlID Question::s_QuestionDontAskMeAgainCheckBox = NOCTL;

	/**************************************************************************

	>	void Question::Initialise(const DialogResourceID& DialogID,
								  ControlID TextControl,
								  ControlID DefButtonControl,
								  ControlID ButtonControl,
								  ControlID DontAskMeAgainControl = NOCTL)


		Author:		Tim Browse
		Date:		7/5/2002
		Inputs:		DialogID - resource ID of the dialog template
							   to use when asking questions of the 
							   user.
					TextControl - the control ID of the control that
								  indicates where to display the text
								  of the question being asked.
					DefButtonControl - control ID of the button which
									   should be the default button
									   in the question dialog.
					ButtonControl - control ID of the button to use as
									a template when creating new buttons.
					DontAskMeAgainControl - control ID of the checkbox to use
											as the "Don't ask me again" option.

		Purpose:	Initialise the Question system, so that the user can be
					asked questions by calling one of the Ask() functions.

					You only need to call this once.

					This functions is used to tell the Question system which
					dialog to use for asking questions, and how to refer to
					the various controls in that dialog.

	**************************************************************************/

	void Question::Initialise(const DialogResourceID& DialogID,
							  ControlID TextControl,
							  ControlID DefButtonControl,
							  ControlID ButtonControl,
							  ControlID DontAskMeAgainControl)
	{
		s_QuestionDialogID				 = DialogID;
		s_QuestionTextControl			 = TextControl;
		s_QuestionDefButtonControl		 = DefButtonControl;
		s_QuestionButtonControl			 = ButtonControl;
		s_QuestionDontAskMeAgainCheckBox = DontAskMeAgainControl;
	}

	/**************************************************************************

	>	Question::Question(const StringResourceID& Title,
						   const StringResourceID& Question
						   Window* pParent = NULL)


		Author:		Tim Browse
		Date:		11/5/2002
		Inputs:		Title - the title of the question dialog
					Question - the question to ask the user
					pParent - the parent of the question dialog (if NULL,
							  the main app window is used)

		Purpose:	Constructor

	**************************************************************************/

	Question::Question(const StringResourceID& Title,
					   const StringResourceID& Question,
					   Window* pParent) : m_bShowCheckBox(false)
	{
		// Sanity checks
		COMPLAIN_IF(s_QuestionDialogID.IsNull(), 
					"No dialog ID for Question class - have you called Question::Initialise()?");
		COMPLAIN_IF(s_QuestionTextControl == NOCTL, 
					"No control ID for Question class - have you called Question::Initialise()?");

		// If no parent passed in, use main app window as the owner/parent
		if (pParent == NULL)
		{
			pParent = GetApp()->GetMainWindow();
		}

		// Create the question dialog
		m_pDialog = new QuestionDialog(Title, Question, pParent);
	}

	/**************************************************************************

	>	Question::Question(const char* szTitle,
						   const char* szQuestion,
						   Window* pParent = NULL)


		Author:		Tim Browse
		Date:		11/5/2002
		Inputs:		szTitle - the title of the question dialog
					szQuestion - the question to ask the user
					pParent - the parent of the question dialog (if NULL,
							  the main app window is used)

		Purpose:	Constructor

	**************************************************************************/

	Question::Question(const char* szTitle,
					   const char* szQuestion,
					   Window* pParent) : m_bShowCheckBox(false)
	{
		// Sanity checks
		COMPLAIN_IF(s_QuestionDialogID.IsNull(), 
					"No dialog ID for Question class - have you called Question::Initialise()?");
		COMPLAIN_IF(s_QuestionTextControl == NOCTL, 
					"No control ID for Question class - have you called Question::Initialise()?");

		// If no parent passed in, use main app window as the owner/parent
		if (pParent == NULL)
		{
			pParent = GetApp()->GetMainWindow();
		}

		// Create the question dialog
		m_pDialog = new QuestionDialog(szTitle, szQuestion, pParent);
	}

	/**************************************************************************
	
	>	Question::~Question()
	
		Author:		Tim Browse
		Date:		11/5/2002
	
		Purpose:	Destructor (cleans up dialog)
	
	**************************************************************************/
	
	Question::~Question()
	{
		delete m_pDialog;
	}

	/**************************************************************************
	
	>	void Question::SetCheckBoxPreference(const char* szPreferenceName)
	
		Author:		Tim Browse
		Date:		11/5/2002
		Inputs:		szPreference - name of preference that controls the display
								   of this dialog, e.g. "UI.AskAbout.SearchFailures"
	
		Purpose:	Set the name of the preference to use for this Question
					object, so that the vGUI library can implement "Don't
					ask me this again" checkbox functionality.

					The named preference is used to control whether the
					question dialog is actually shown to the user.
	
	**************************************************************************/
	
	void Question::SetCheckBoxPreference(const char* szPreferenceName)
	{
		m_PrefName = szPreferenceName;
		m_bShowCheckBox = true;
	}

	/**************************************************************************
	
	>	void Question::EnableCheckBox()
	
		Author:		Tim Browse
		Date:		11/5/2002
	
		Purpose:	Show the "Don't ask me again" checkbox, but don't
					update any preferences - this just lets the caller
					handle this functionality themselves.

					See AskUserNextTime().
	
	**************************************************************************/
	
	void Question::EnableCheckBox()
	{
		m_bShowCheckBox = true;
	}

	/**************************************************************************
	
	>	bool Question::AskUserNextTime()
	
		Author:		Tim Browse
		Date:		11/5/2002
		Returns:	true => user checked the "Don't ask again" checkbox.
					false => otherwise
	
		Purpose:	Find out if the user checked the "Don't ask me again"
					checkbox - the checkbox must be enabled with
					SetCheckBoxPreference() or EnableCheckBox().
	
	**************************************************************************/
	
	bool Question::AskUserNextTime()
	{
		return m_pDialog->AskUserNextTime();
	}
	
	/**************************************************************************

	>	StringResourceID Question::Ask(const StringResourceID* Buttons)


		Author:		Tim Browse
		Date:		7/5/2002
		Inputs:		Buttons - array of string resource IDs containing the
							  labels of the buttons to display in the question
							  dialog.  (Terminated by ID 0)
		Returns:	One of the resource IDs listed in 'Buttons', which indicates
					which button the user clicked to answer the question.

		Purpose:	Ask the user a question, and report which button they chose
					to answer the question with.

	**************************************************************************/

	StringResourceID Question::Ask(const StringResourceID* Buttons)
	{
		// Initialise dialog buttons (we need to do this first, even
		// if we don't actually show the dialog to the user).
		m_pDialog->SetButtons(Buttons);

		// Find out how this question should behave...
		bool bShowCheckBox;
		bool bAskUser;
		int iLastChoice;

		GetQuestionOptions(&bShowCheckBox, &bAskUser, 
						   &iLastChoice);

		// If we shouldn't ask the user this question again, just return
		// the answer they gave us last time.
		if (!bAskUser)
		{
			return Buttons[iLastChoice];
		}

		// Show check box if required
		if (bShowCheckBox)
		{
			m_pDialog->ShowCheckBox();
		}

		// Ask the question
		int iResult = m_pDialog->DoModal();

		// Check to see if user asked not to be bothered again
		if (!m_pDialog->AskUserNextTime())
		{
			// User does not want to be asked again
			vos::GetPreferences().Write(m_PrefName, false);

			// Remember which choice they made this time so we can make
			// it for them.
			vos::String ButtonName(MakeButtonName(Buttons[iResult].Load()));
			vos::GetPreferences().Write(GetButtonPrefName(), ButtonName);
		}

		// Tell caller which button was pressed
		return Buttons[iResult];
	}

	/**************************************************************************

	>	const char* Question::Ask(const char* szButtons[])


		Author:		Tim Browse
		Date:		7/5/2002
		Inputs:		szButtons - array of strings containing the labels of the 
								buttons to display in the question dialog.  
								(Terminated by NULL)
		Returns:	An index into the 'szButtons' array, which corresponds to
					the button the user answered the question with.

		Purpose:	Ask the user a question, and report which button they chose
					to answer the question with.

	**************************************************************************/

	int Question::Ask(const char* szButtons[])
	{
		// Initialise dialog buttons (we need to do this first, even
		// if we don't actually show the dialog to the user).
		m_pDialog->SetButtons(szButtons);

		// Find out how this question should behave...
		bool bShowCheckBox;
		bool bAskUser = true;
		int iLastChoice;

		GetQuestionOptions(&bShowCheckBox, &bAskUser, 
						   &iLastChoice);

		// If we shouldn't ask the user this question again, just return
		// the answer they gave us last time.
		if (!bAskUser)
		{
			return iLastChoice;
		}

		// Show check box if required
		if (bShowCheckBox)
		{
			m_pDialog->ShowCheckBox();
		}

		// Ask the question
		int iResult = m_pDialog->DoModal();

		// Check to see if user asked not to be bothered again
		if (!m_pDialog->AskUserNextTime())
		{
			// User does not want to be asked again
			vos::GetPreferences().Write(m_PrefName, false);

			// Remember which choice they made this time so we can make
			// it for them.
			vos::String ButtonName(MakeButtonName(szButtons[iResult]));
			vos::GetPreferences().Write(GetButtonPrefName(), ButtonName);
		}

		// Tell caller which button was pressed
		return iResult;
	}

	//
	// These really need to be localisable and platform dependent
	const char* kYesNoButtons[] = { msg_ButtonYes, msg_ButtonNo, NULL };
	const char* kOKCancelButtons[] = { msg_ButtonOK, msg_ButtonCancel, NULL };

	/**************************************************************************
	
	>	bool Question::AskYesNo()
	
		Author:		Tim Browse
		Date:		15/7/2002
		Returns:	true => the user answered 'Yes' to the question
					false => the user answered 'No' to the question
	
		Purpose:	Ask the user a question with simple Yes/No options.
	
	**************************************************************************/
	
	bool Question::AskYesNo()
	{
		// Ask question and tell caller if the answer was yes
		return (Ask(kYesNoButtons) == 0);
	}

	/**************************************************************************
	
	>	bool Question::AskOKCancel()
	
		Author:		Tim Browse
		Date:		15/7/2002
		Returns:	true => the user answered 'Yes' to the question
					false => the user answered 'No' to the question
	
		Purpose:	Ask the user a question with simple OK/Cancel options.
	
	**************************************************************************/
	
	bool Question::AskOKCancel()
	{
		// Ask question and tell caller if the answer was OK
		return (Ask(kOKCancelButtons) == 0);
	}
	
	/**************************************************************************

	>	void Question::GetQuestionOptions(bool *pbShowCheckBox, 
										  bool *pbAskUser, 
										  int  *piLastChoice)


		Author:		Tim Browse
		Date:		11/5/2002
		Outputs:	pbShowCheckBox - true => show the "Don't ask again" checkbox
					pbAskUser - true => Ask the user the question
								false => don't ask; use their last response
					piLastChoice - the button the user chose the last time
								   they were asked this question.

		Purpose:	Get options for the particular preference (as named by
					szPreferenceName).  If szPreferenceName is NULL, the
					question is always asked (and the check box is not shown)

	**************************************************************************/

	void Question::GetQuestionOptions(bool *pbShowCheckBox, 
									  bool *pbAskUser, 
									  int  *piLastChoice)
	{
		// Default values
		*pbAskUser = true;
		*piLastChoice = 0;
		*pbShowCheckBox = m_bShowCheckBox;

		// Deal with the easy case first
		if (m_PrefName.IsEmpty())
		{
			// Options already correct
			return;
		}

		// Ok, we have a preference name, so show the "don't ask again"
		// checkbox
		*pbShowCheckBox = true;

		// Read the preference value (true => ask the user the question)
		vos::GetPreferences().Read(m_PrefName, pbAskUser);

		// Ok, see if we know what they chose last time
		vos::String strLastChoice;
		vos::GetPreferences().Read(GetButtonPrefName(), &strLastChoice);

		if (strLastChoice.IsEmpty())
		{
			// No, so assume default is the first button
			*piLastChoice = 0;

			// And if we can't find the last choice, we'd better ask
			// the question anyway
			*pbAskUser = true;
		}
		else
		{
			// Ok, we have a button name - find the match in the
			// question dialog
			*piLastChoice = m_pDialog->FindButton(strLastChoice);
		}
	}


	/**************************************************************************
	
	>	vos::String Question::MakeButtonName(const vos::String& ButtonText)
	
		Author:		Tim Browse
		Date:		11/5/2002
		Inputs:		ButtonText - the text used for a button, e.g. "No", or "&No"
		Returns:	The normalised button name, e.g. "&No All" turns into "NoAll"
	
		Purpose:	Make a plain-text button name, from the text used for a
					button.  This strips out any non-alphanumeric characters,
					so we ignore accelerator characters, etc.
	
	**************************************************************************/
	
	vos::String Question::MakeButtonName(const vos::String& ButtonText)
	{
		vos::String ButtonName;

		int Length = ButtonText.GetLength();

		for (int i = 0; i < Length; i++)
		{
			if (isalnum(ButtonText[i]))
				ButtonName += ButtonText[i];
		}

		return ButtonName;
	}

	/**************************************************************************
	
	>	vos::String Question::GetButtonPrefName(const vos::String& PrefName)
	
		Author:		Tim Browse
		Date:		11/5/2002
		Returns:	Preference name to be used to remember the user's last button choice.
	
		Purpose:	Using the preference name for the "Don't ask me again" 
					preference, this function constructs a preference name
					for the button value that should be returned when
					the question is not asked.

					(i.e. we use this along with MakeButtonName() to store
					the user's last button choice, and return it the next
					time if the user told us not to ask next time)
	
	**************************************************************************/
	
	vos::String Question::GetButtonPrefName()
	{
		return m_PrefName + "_LastChoice";
	}

};

