//
// QuestionDialog.cpp
//

#include "StdHdr.h"

#include "QuestionDialog.h"

#include "Question.h"
#include "Button.h"
#include "GC.h"
#include "Errors.h"
#include "Metrics.h"

namespace vgui
{
	//
	// QuestionDialog message map
	//
	BEGIN_VGUI_MESSAGE_MAP(QuestionDialog, Dialog)
		ON_VGUI_BN_CLICKED_ALL(OnButtonClicked)
	END_VGUI_MESSAGE_MAP()

	/**************************************************************************

	>	QuestionDialog::QuestionDialog(const StringResourceID& QuestionTitle,
									   const StringResourceID& QuestionText,
									   Window* pParent) 


		Author:		Tim Browse
		Date:		7/5/2002
		Inputs:		QuestionTitle - the string resource to show as the title
									of the question dialog.
					QuestionText - the string resource containing the text of
								   the question to ask.
					pParent - the parent window of the question dialog 
							  (may be NULL)

		Purpose:	Constructor

	**************************************************************************/

	QuestionDialog::QuestionDialog(const StringResourceID& QuestionTitle,
								   const StringResourceID& QuestionText,
								   Window* pParent) 
								   
									  : MsgDialog(Question::s_QuestionDialogID, 
												  pParent,
												  Question::s_QuestionTextControl),
										m_pButtons(NULL),
										m_pButtonIDs(NULL),
										m_pButtonStrings(NULL),
										m_bShowDontAskMeAgainCheckBox(false)
	{								
		// Load the strings
		QuestionTitle.Load(&m_Title);
		QuestionText.Load(&m_Question);

		// Set the dialog up with them
		SetTitle(m_Title);
		SetMessage(m_Question);
	}

	/**************************************************************************

	>	QuestionDialog::QuestionDialog(const char* szQuestionTitle,
									   const char* szQuestionText,
									   Window* pParent) 


		Author:		Tim Browse
		Date:		7/5/2002
		Inputs:		szQuestionTitle - the string to show as the title of the 
									  question dialog.
					szQuestionText - the text of the question to ask.
					pParent - the parent window of the question dialog 
							  (may be NULL)

		Purpose:	Constructor

	**************************************************************************/

	QuestionDialog::QuestionDialog(const char* szQuestionTitle,
								   const char* szQuestionText,
								   Window* pParent) 
								   
									  : MsgDialog(Question::s_QuestionDialogID, 
												  pParent,
												  Question::s_QuestionTextControl),
										m_Title(szQuestionTitle),
										m_Question(szQuestionText),
										m_pButtons(NULL),
										m_pButtonIDs(NULL),
										m_pButtonStrings(NULL),
										m_bShowDontAskMeAgainCheckBox(false)
	{
		// Set the dialog up with the title/message
		SetTitle(szQuestionTitle);
		SetMessage(szQuestionText);
	}

	/**************************************************************************
	
	>	QuestionDialog::~QuestionDialog()
	
		Author:		Tim Browse
		Date:		7/5/2002
	
		Purpose:	Destructor
	
	**************************************************************************/
	
	QuestionDialog::~QuestionDialog()
	{
		// Clean up

		if (m_pButtons != NULL)
		{
			// Delete the button objects
			for (int i = 0; i < m_cButtons; i++)
			{
				delete m_pButtons[i];
			}
		}

		delete [] m_pButtons;
		delete [] m_pButtonIDs;
		delete [] m_pButtonStrings;
	}

	void QuestionDialog::SetButtons(const StringResourceID Buttons[])
	{
		// Find out how many buttons there are.
		m_cButtons = 0;
		while (!Buttons[m_cButtons].IsNull())
		{
			// Try next ID
			m_cButtons++;
		}

		// Load the button strings
		m_pButtonStrings = new String[m_cButtons];

		for (int i = 0; i < m_cButtons; i++)
		{
			Buttons[i].Load(&m_pButtonStrings[i]);
		}
	}

	void QuestionDialog::SetButtons(const char *szButtons[])
	{
		// Find out how many buttons there are.
		m_cButtons = 0;
		while (szButtons[m_cButtons] != NULL)
		{
			// Try next ID
			m_cButtons++;
		}

		// Copy the button strings
		m_pButtonStrings = new String[m_cButtons];

		for (int i = 0; i < m_cButtons; i++)
		{
			m_pButtonStrings[i] = szButtons[i];
		}
	}

	void QuestionDialog::ShowCheckBox()
	{
		if (Question::s_QuestionDontAskMeAgainCheckBox != NOCTL)
		{
			m_bShowDontAskMeAgainCheckBox = true;
		}
	}

	/**************************************************************************
	
	>	void QuestionDialog::InitTheDialog()
	
		Author:		Tim Browse
		Date:		7/5/2002
	
		Purpose:	Perform any initialisation for the dialog when it is
					created.  In this case, we create/position the buttons 
					that the client has asked for.
	
	**************************************************************************/
	
	void QuestionDialog::InitTheDialog()
	{
		// Create the buttons according to the array of strings passed in
		CreateButtons();

		// Attach checkbox if necessary
		if (m_bShowDontAskMeAgainCheckBox)
		{
			m_CheckBox.Attach(this, Question::s_QuestionDontAskMeAgainCheckBox);
		}
	}


	/**************************************************************************
	
	>	void QuestionDialog::RepositionControls(int* pDeltaX, int* pDeltaY)
	
		Author:		Tim Browse
		Date:		7/5/2002
		Inputs:		pDeltaX,pDeltaY - the change in size of the dialog required
									  to make the question text fit.

		Outputs:	pDeltaX, pDeltaY - the possibly adjusted deltas, in case this
									   function needs to add/remove any controls.
	
		Purpose:	Reposition any controls as necessary when the dialog is
					resized to make the question text fit.
	
	**************************************************************************/
	
	void QuestionDialog::RepositionControls(int* pDeltaX, int* pDeltaY)
	{
		ControlID CheckBoxID = Question::s_QuestionDontAskMeAgainCheckBox;

		// Does the dialog have a "Don't ask me again" checkbox?
		if (CheckBoxID != NOCTL)
		{
			// Yes - do we actually want a "Don't tell me again" checkbox?
			if (m_bShowDontAskMeAgainCheckBox)
			{
				// Yes, so centre it
				MoveControlRelative(CheckBoxID, (*pDeltaX)/2, *pDeltaY);
			}
			else
			{
				// No, so hide the control, and adjust DeltaY accordingly
				HideControl(CheckBoxID);

				Rect CheckboxPosition;
				GetControlPosition(CheckBoxID, &CheckboxPosition);

				int Diff = CheckboxPosition.bottom - m_MessageArea.bottom;

				*pDeltaY -= Diff;

				// Grow message area to cover checkbox area.
				AdjustMessageArea(0, Diff);
			}
		}

		// Move all the buttons
		if (m_pButtonIDs != NULL)
		{
			for (int i = 0; i < m_cButtons; i++)
			{
				// Move this button
				// We only move half the X distance so we keep the buttons 
				// centred in the dialog.
				MoveControlRelative(m_pButtonIDs[i], (*pDeltaX)/2, *pDeltaY);
			}
		}
	}

	/**************************************************************************
	
	>	void QuestionDialog::CreateButtons()
	
		Author:		Tim Browse
		Date:		7/5/2002
	
		Purpose:	Create the required buttons in this dialog, according to
					the button labels we were supplied with.
	
	**************************************************************************/
	
	void QuestionDialog::CreateButtons()
	{
		// Create the array of button objects
		m_pButtonIDs = new ControlID[m_cButtons];

		Rect DefaultButtonArea;
		Rect NormalButtonArea;

		m_pButtons = new Button*[m_cButtons];

		if (m_cButtons >= 1)
		{
			// Attach to default button
			m_pButtons[0] = new Button;
			m_pButtons[0]->Attach(this, Question::s_QuestionDefButtonControl);

			// Get the position of this button
			m_pButtons[0]->GetPosition(&DefaultButtonArea);

			// Remember its control ID
			m_pButtonIDs[0] = m_pButtons[0]->GetControlID();
		}

		// Attach to normal button if necessary
		if (m_cButtons >= 2)
		{
			// Attach to default button
			m_pButtons[1] = new Button;
			m_pButtons[1]->Attach(this, Question::s_QuestionButtonControl);

			// Get the position of this button
			m_pButtons[1]->GetPosition(&NormalButtonArea);

			// Remember its control ID
			m_pButtonIDs[1] = m_pButtons[1]->GetControlID();
		}

		// Create other buttons as needed
		if (m_cButtons > 2)
		{
			for (int i = 2; i < m_cButtons; i++)
			{
				// Create the button
				m_pButtons[i] = new Button;
				m_pButtons[i]->Create(this, NormalButtonArea);

				// Remember its control ID
				m_pButtonIDs[i] = m_pButtons[i]->GetControlID();
			}
		}

		//
		// Reposition the buttons
		//
		uint *pWidths = new uint[m_cButtons];

		// Use the width if the default button as a guide, unless the 
		// text will not fit - in which case, the button is made wider.

		// We need a GC to get font size information
		GC* pGC = GetGC();
		pGC->SetFont(GetFont());

		Rect TextRect;

		int i;

		for (i = 0; i < m_cButtons; i++)
		{
			// Find the width of the text
			pGC->CalcTextArea(m_pButtonStrings[i], tf_Left, &TextRect);

			// Adjust width accordingly.
			int Width = TextRect.GetWidth() + (2 * Metrics::GetTextMargin());

			// Use the biggest width
			pWidths[i] = max(Width, DefaultButtonArea.GetWidth());
		}

		// Clean up GC
		pGC->RestoreDefaultFont();
		ReleaseGC(pGC);

		// Get the gap size between buttons.
		int ButtonGap;
		if (NormalButtonArea.left < DefaultButtonArea.left)
		{
			ButtonGap = DefaultButtonArea.left - NormalButtonArea.right;
		}
		else
		{
			ButtonGap = NormalButtonArea.left - DefaultButtonArea.right;
		}

		// Work out total width needed.
		int TotalWidth = (m_cButtons -1) * ButtonGap;

		for (i = 0; i < m_cButtons; i++)
		{
			TotalWidth += pWidths[i];
		}

		// Make sure the dialog is wide enough to cope with the buttons
		Rect TextArea = m_MessageArea;

		// Get size/position of dialog
		Rect ClientArea;
		GetClientArea(&ClientArea);

		if (TextArea.GetWidth() < TotalWidth)
		{
			int WidthDiff = TotalWidth - TextArea.GetWidth();

			// Resize dialog to cope with wide buttons
			ResizeDialog(WidthDiff, 0);

			// Keep client area up to date.
			ClientArea.right += WidthDiff;

			// Reposition the checkbox if necessary
			if (m_bShowDontAskMeAgainCheckBox)
			{
				// Yes, so centre it
				MoveControlRelative(Question::s_QuestionDontAskMeAgainCheckBox,
									WidthDiff/2, 0);
			}
		}

		// Ok, work out where the first button goes, and reposition the 
		// others accordingly.
		int X = ClientArea.GetHalfWidth() - (TotalWidth / 2);

		Rect ButtonPosition;
		ButtonPosition.top = NormalButtonArea.top;
		ButtonPosition.bottom = NormalButtonArea.bottom;

		for (i = 0; i < m_cButtons; i++)
		{
			// Calculate position for this button
			ButtonPosition.left = X;
			ButtonPosition.right = ButtonPosition.left + pWidths[i];

			// Move the button
			m_pButtons[i]->Move(ButtonPosition);

			// Get ready for next button
			X = ButtonPosition.right + ButtonGap;
		}

		// Finished with widths
		delete [] pWidths;

		//
		// Set the text to be displayed by each button.
		//
		for (i = 0; i < m_cButtons; i++)
		{
			m_pButtons[i]->SetText(m_pButtonStrings[i]);
		}
	}

	/**************************************************************************
	
	>	bool QuestionDialog::AskUserNextTime()
	
		Author:		Tim Browse
		Date:		5/5/2002
		Returns:	true => ask the user this question the next time
					false => user has requested not to be asked again.
	
		Purpose:	Find out if we should ask the user about this question
					next time it happens.  This is controlled by an (optional)
					"Don't ask me again" checkbox in the question dialog.
	
	**************************************************************************/
	
	bool QuestionDialog::AskUserNextTime()
	{
		return m_bAskUserNextTime;
	}


	/**************************************************************************

	>	bool QuestionDialog::OnButtonClicked(TMessageType MsgType, 
											 int nMsg, 
											 int ButtonPressed)


		Author:		Tim Browse
		Date:		7/5/2002
		Inputs:		MsgType, nMsg - ignored (we only register for mt_BN_CLICKED
									events)
					ButtonPressed - the control ID of the button that was just
									clicked.
		Returns:	true => message handled (always)

		Purpose:	Handle a click on one of the response buttons, as it means
					the user is answering the question.

	**************************************************************************/

	bool QuestionDialog::OnButtonClicked(TMessageType MsgType, 
										 int nMsg, 
										 int ButtonPressed)
	{
		// Check for 'don't ask me again checkbox'
		if (ButtonPressed == Question::s_QuestionDontAskMeAgainCheckBox)
		{
			// Ignore these clicks - let standard checkbox processing happen
			return false;
		}

		// Find out which button was pressed
		int iButton = -1;

		for (int i = 0; i < m_cButtons; i++)
		{
			if (m_pButtonIDs[i] == ButtonPressed)
			{
				// Found the button that was pressed.
				iButton = i;
				break;
			}
		}

		// Check to see if we found a button
		if (iButton >= 0)
		{
			if ((Question::s_QuestionDontAskMeAgainCheckBox != NOCTL) &&
				m_bShowDontAskMeAgainCheckBox)
			{
				// Logic is inverted
				if (m_CheckBox)
				{
					// User does not want to be asked again
					m_bAskUserNextTime = false;
				}
				else
				{
					// They didn't check the box
					m_bAskUserNextTime = true;
				}
			}

			// Found the button that was pressed.
			EndDialog(iButton);
			return true;
		}
		else
		{
			// We really ought to have found a button
			COMPLAIN("Could not find button that was pressed in Question Dialog!");

			// Failsafe handling
			EndDialog(0);
			return true;
		}
	}

	int QuestionDialog::FindButton(const vos::String& ButtonName)
	{
		if (m_pButtonStrings == NULL)
		{
			COMPLAIN("No strings in FindButton()!");
			return 0;
		}

		// Search array of button names to find match
		vos::String ThisButtonName;

		for (int iButton = 0; iButton < m_cButtons; iButton++)
		{
			ThisButtonName = Question::MakeButtonName(m_pButtonStrings[iButton]);

			if (ThisButtonName.CompareText(ButtonName))
			{
				// Found a match
				return iButton;
			}
		}

		// No match found
		COMPLAIN("FindButton() failed to find a button!");
		return 0;
	}
	
};
