//
// Rect.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_Rect_h
#define vgui_Rect_h

#include "vgui.h"

#include "Point.h"

namespace vgui
{
	class Rect
	{
		public:

			//
			// Constructors
			//

			Rect() : left(0), top(0), right(0), bottom(0)
			{
			}

			Rect(int newleft, int newtop, int newright, int newbottom) : left(newleft),
																		 top(newtop),
																		 right(newright),
																		 bottom(newbottom)
			{
			}

			// Construct with existing Rect object.
			Rect(const Rect& rhs)
			{
				left = rhs.left;
				top = rhs.top;
				right = rhs.right;
				bottom = rhs.bottom;
			}

			//
			// Comparison operators
			//

			// Same as calling IsEqual() to determine if an existing Rect object 
			// has the same coordinates as this object.
			bool operator==(const Rect& r)
			{
				return IsEqual(r); 
			}

			// Same as calling !IsEqual() to determine if an existing Rect object 
			// does not have the same coordinates as this object.
			bool operator!=(const const Rect& r)
			{
				return !IsEqual(r); 
			}

			// Returns true if an existing Rect object has the same coordinates 
			// as this rectangle, false otherwise.
			bool IsEqual(const Rect& rcCompare) const;

			// Returns true if the coordinates of this rectangle are 0, 0, 0, 0, 
			// false otherwise.
			bool IsEmpty() const
			{ 
				return ((left == 0) && (right == 0) && (top == 0) && (bottom == 0)); 
			}

			//
			// Assignment
			//

			// Same as calling Set() to set the coordinates of this object to 
			// those of an existing Rect object.
			Rect& operator=(const Rect& rhs)
			{
				left = rhs.left;
				top = rhs.top;
				right = rhs.right;
				bottom = rhs.bottom;

				return *this;
			}

			// Set the coordinates of this rectangle to 0, 0, 0, 0.
			void Empty()
			{
				left = top = right = bottom = 0; 
			}

			// Set the rectangle coordinates to those passed to function.
			void Set(int x, int y, int width, int height)
			{ 
				left = x; 
				top = y; 
				right = left + width; 
				bottom = top + height; 
			}

			// Set the coordinates of this rectangle to be the values from 
			// an existing Rect object.
			void Set(const Rect& r)
			{ 
				left = r.left;
				top = r.top;
				right = r.right;
				bottom = r.bottom;
			}

			void SetWidth(int Width)
			{
				right = left + Width;
			}

			void SetHeight(int Height)
			{
				bottom = top + Height;
			}

			//
			// Operations
			//

			// Enlarge the rectangle by deltaX left to right, and deltaY top to bottom.
			void Inflate(int deltaX, int deltaY)
			{
				left -= deltaX; right  += deltaX;
				top  -= deltaY; bottom += deltaY;
			}

			// Shrink the rectangle by deltaX left to right, and deltaY top to bottom.
			void Deflate(int deltaX, int deltaY)
			{
				left += deltaX; right  -= deltaX;
				top  += deltaY; bottom -= deltaY;
			}

			//
			// Corner, centre-point, height access
			//

			// Return a Point object which contains the coordinates of the top left 
			// corner of this rectangle.
			Point GetTopLeft() const
			{
				return Point(left, top); 
			}

			// Return a Point object which contains the coordinates of the top right 
			// corner of this rectangle.
			Point GetTopRight() const
			{ 
				return Point(right, top); 
			}

			// Return a Point object which contains the coordinates of the bottom left 
			// corner of this rectangle.
			Point GetBottomLeft() const
			{ 
				return Point(left, bottom); 
			}

			// Return a Point object which contains the coordinates of the bottom right 
			// corner of this rectangle.
			Point GetBottomRight() const
			{
				return Point(right, bottom); 
			}


			// Return a Point object which contains the coordinates of the centre of 
			// this rectangle.
			Point GetCentrePoint() const
			{
				return Point((left + right) / 2, (top + bottom) / 2); 
			}


			// Return half of the height of this rectangle.
			int GetHalfHeight() const
			{
				return GetHeight() / 2; 
			}

			// Return half of the width of this rectangle.
			int GetHalfWidth() const
			{
				return GetWidth() / 2; 
			}


			// Return the height of this rectangle.
			int GetHeight() const
			{
				return bottom - top; 
			}

			// Return the width of this rectangle.
			int	GetWidth() const
			{
				return right - left; 
			}


			//
			// Calculates intersection - if no intersection, then empty rect is used.
			// true returned if intersection found, false if no intersection.
			//
			bool Intersect(const Rect& r1, const Rect& r2);

			// Calculate union of two rectangles.
			void Union(const Rect& r1, const Rect& r2);

			// Returns true if the point specified by the Point object is 
			// inside the rectangle, false otherwise.
			inline bool IsPointInside(Point const& pt) const
			{
				return (pt.x >= left) && (pt.x <= right) &&
					   (pt.y >= top)  && (pt.y <= bottom);
			}

			// Call to verify rectangle left and top coordinates are LESS than right 
			// and bottom coordinates.
			void Normalize()
			{
				if (left > right)
				{
					int temp = left;
					left =	right;
					right =	temp;
				}
				
				if (top > bottom)
				{
					int temp = top;
					top = bottom;
					bottom = temp;
				}
			}

			// Offset (move) the rectangle by nDistanceX and nDistanceY. 
			// Either can be negative, which will subtract.
			void Translate(int nDistanceX, int nDistanceY);


#if defined VGUI_MSWINDOWS
			//
			// Win32 specific functions
			//
			
			// Construct by converting from an existing RECT structure.
			Rect(const RECT& rhs)
			{ 
				left = rhs.left;
				top = rhs.top;
				right = rhs.right;
				bottom = rhs.bottom;
			}

			// Return the a pointer to the internal RECT structure.
			RECT* GetPointer() const
			{ 
				// NB. Although we don't actually use a RECT, we rely on the fact
				//     that vGui defines basic types like these to match the memory
				//	   layout of the platform-dependent type (in this case, RECT).
				return (RECT*) this; 
			}

			// Return a RECT structure reference to this object.
			RECT& GetRECT() const
			{ 
				// NB. Although we don't actually use a RECT, we rely on the fact
				//     that vGui defines basic types like these to match the memory
				//	   layout of the platform-dependent type (in this case, RECT).
				return (RECT&) *this; 
			}

			// Set the coordinates of this rectangle to be the values from an 
			// existing RECT structure.
			void Set(const RECT& r)
			{
				Set(r.left, r.top, r.right, r.bottom); 
			}
#elif defined VGUI_MACOS
			//
			// Carbon specific functions
			//
			
			// Fill out and return a Carbon Rect
			::Rect& GetRect() const
			{ 
				// This may overflow. We can't do anything about this
				// See PowerPlant for inspiration on how to handle this
				m_MacRect.left = left;
				m_MacRect.right = right; 
				m_MacRect.top = top; 
				m_MacRect.bottom = bottom;
				return m_MacRect;
			}

			// Set the coordinates of this rectangle to be the values from an 
			// existing Rect structure.
			void Set(::const Rect& r)
			{
				Set(r.left, r.top, r.right, r.bottom); 
			}
			
		private:
			mutable ::Rect	m_MacRect;
#endif

		public:
			// The data
			int left;
			int top;
			int right;
			int bottom;
	};
};

#endif // !vgui_Rect_h


