//
// Timer.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Timer.h"

#include "vOS\Trace.h"

#ifdef VGUI_MSWINDOWS

#include "Win32\Win32_WindowImp.h"
#include "Win32\Win32_Headers.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	//
	// TimerIDs start at 1000 (arbitrary)
	//
	TimerID Timer::s_NextID = 1000;

	//
	// Timer maps (one of each - static data)
	//
	WindowTimerMap Timer::m_WindowTimers;
	CallbackTimerMap Timer::m_CallbackTimers;

	/**************************************************************************

	>	TimerID Timer::SetTimer(TimerType Type,
								int MillisecondDelay,
								Window *pWindow)


		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		Type - the type of the timer:
							+ Periodic - timer events are delivered every
										 'MillisecondDelay' milliseconds
							+ OneShot  - one timer event is delivered after
										 'MillisecondDelay' milliseconds, and
										 then the timer is automatically killed.

					MillisecondDelay - the number of milliseconds between timer
									   events.
					pWindow - the window to receive the timer events (via its
							  OnTimer() method).
		Returns:	A unique ID that identifies this (and only this) timer.
					This ID is passed to the OnTimer() method of your window.

					OR: NullTimer if the timer could not be set up.

		Purpose:	Set up a timer to deliver events to a particular Window
					object.

					Over-ride the OnTimer() function in your window to process
					the timer messages.

	**************************************************************************/

	TimerID Timer::SetTimer(TimerType Type,
							int MillisecondDelay,
							Window *pWindow)
	{
		// Get the Win32 window handle for the specified window, as we need
		// it to set the WM_TIMER events up.
		HWND hWnd = Win32_WindowImp::GetHWND(pWindow);

		if (hWnd == NULL)
		{
			// Must have a valid window handle
			vguiERROR();
			return NullTimer;
		}

		// Allocate a timer ID for window-based timers.
		UINT ID = GetNextTimerID();
		
		// Set the timer.
		if (::SetTimer(hWnd, ID, MillisecondDelay, NULL) == 0)
		{
			// Failed to create timer
			return NullTimer;
		}

		// Remember this timer.
		WindowTimer Timer = {Type, pWindow};
		m_WindowTimers[ID] = Timer;

		// Return the timer ID to the caller
		return (TimerID) ID;
	}

	/**************************************************************************

	>	TimerID Timer::SetTimer(TimerType Type,
								int MillisecondDelay,
								TimerProc Handler,
								UserData TheUserData = NULL)


		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		Type - the type of the timer:
							+ Periodic - timer events are delivered every
										 'MillisecondDelay' milliseconds
							+ OneShot  - one timer event is delivered after
										 'MillisecondDelay' milliseconds, and
										 then the timer is automatically killed.

					MillisecondDelay - the number of milliseconds between timer
									   events.
					Handler - the callback function that is to receive the
							  timer events.
				
					TheUserData - instance data to be passed to Handler().

		Returns:	A unique ID that identifies this (and only this) timer.
					This ID is passed back to your callback function.

					OR: NullTimer if the timer could not be set up.

		Purpose:	Set up a timer to deliver events to a particular callback
					function.

	**************************************************************************/

	TimerID Timer::SetTimer(TimerType Type,
							int MillisecondDelay,
							TimerProc Handler,
							UserData TheUserData)
	{
		// Set the timer.
		UINT ID = ::SetTimer(NULL, 0, MillisecondDelay, Win32TimerProc);

		if (ID == 0)
		{
			// Failed to create timer
			return NullTimer;
		}

		// Remember this timer.
		CallbackTimer Timer = {Type, Handler, TheUserData};
		m_CallbackTimers[ID] = Timer;

		// Return the timer ID to the caller
		return (TimerID) ID;
	}

	/**************************************************************************
	
	>	void Timer::KillTimer(TimerID ID)
	
		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		ID - the ID of the timer to kill.
	
		Purpose:	Kill the specified timer, so that it does not deliver any
					more events.

					Do NOT try to kill off OneShot timers, unless they have
					not yet delivered the first event, as otherwise KillTimer()
					will ASSERT (because one shot timers are automatically
					killed off after the first event, not surprisingly).
	
	**************************************************************************/
	
	void Timer::KillTimer(TimerID ID)
	{
		// is this a window-based timer?
		WindowTimerMap::iterator itWindow = m_WindowTimers.find(ID);

		if (itWindow != m_WindowTimers.end())
		{
			// Found window timer, so kill it and exit
			Window *pWindow = (*itWindow).second.m_pWindow;
			HWND hWnd = Win32_WindowImp::GetHWND(pWindow);

			// Kill the Win32 timer
			BOOL bSuccess = ::KillTimer(hWnd, ID);

			// Make sure we killed the timer successfully
			ASSERT(bSuccess);

			// Remove timer from the map
			m_WindowTimers.erase(itWindow);

			// All done
			return;
		}

		// Is it a callback timer?
		CallbackTimerMap::iterator itCallback = m_CallbackTimers.find(ID);

		if (itCallback != m_CallbackTimers.end())
		{
			// Found callback timer, so kill it and exit

			// Kill the Win32 timer
			BOOL bSuccess = ::KillTimer(NULL, ID);

			// Make sure we killed the timer successfully
			ASSERT(bSuccess);

			// Remove timer from the map
			m_CallbackTimers.erase(itCallback);

			// All done
			return;
		}

		// Request to kill unregistered timer
		vguiERROR();
	}

	/**************************************************************************
	
	>	bool Timer::ServiceTimer(TimerID ID)
	
		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		ID - the ID of the timer in question.
		Returns:	true => timer registered by the Timer class, and was processed.
					false => timer not known; not processed.
	
		Purpose:	Called by the vGUI framework when the host OS delivers timer
					events to a window.
	
	**************************************************************************/
	
	bool Timer::ServiceTimer(TimerID ID)
	{
		// is this a window-based timer?
		WindowTimerMap::iterator itWindow = m_WindowTimers.find(ID);

		if (itWindow != m_WindowTimers.end())
		{
			// Found window timer...
			Window* pWindow = (*itWindow).second.m_pWindow;

			// If it's a one-shot timer, then kill it off now, so
			// we don't get any more events.
			if (itWindow->second.m_Type == TimerType_OneShot)
			{
				KillTimer(ID);
			}

			// Now service timer and exit
			pWindow->OnTimer(ID);
			return true;
		}

		// Timer not known
		return false;
	}


	/**************************************************************************

	>	VOID CALLBACK Timer::Win32TimerProc(HWND hwnd,
											UINT uMsg,
											UINT idEvent,
											DWORD dwTime)


		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		hwnd - handle of the window (not used by us)
					uMsg - WM_TIMER
					idEvent - the ID of the timer.
					dwTime - current system time (ignored)

		Purpose:	Standard Win32 timer callback procedure - it looks up
					the timer ID in our map, and services the registered timer 
					callback.

	**************************************************************************/

	VOID CALLBACK Timer::Win32TimerProc(HWND hwnd,
									    UINT uMsg,
									    UINT idEvent,
									    DWORD dwTime)
	{
		TimerID ID = (TimerID) idEvent;

		// Look for TimerProc timer
		CallbackTimerMap::iterator itTimerProc = m_CallbackTimers.find(ID);

		if (itTimerProc != m_CallbackTimers.end())
		{
			// Found callback timer...
			TimerProc Proc = itTimerProc->second.m_TimerProc;
			UserData TimerUserData = itTimerProc->second.m_UserData;

			// If it's a one-shot timer, then kill it off first so we don't
			// get another event.
			if (itTimerProc->second.m_Type == TimerType_OneShot)
			{
				// NB. after this call, itTimerProc will be invalid, so
				//	   don't use it!
				KillTimer(ID);
			}

			// Now service timer and exit
			(*Proc)(ID, TimerUserData);
			return;
		}
		else
		{
			// Unknown WM_TIMER event - complain.
			vguiERROR();
		}
	}

};

#endif


