//
// Timer.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui__Timer_h
#define vgui__Timer_h

#include "vgui.h"

#include <map>

namespace vgui
{
	typedef unsigned int TimerID;

	enum { NullTimer = 0 };

	// Timer callback function
	typedef void (VGUI_OSCALLBACK *TimerProc)(TimerID, UserData);

	// Type of timer to use
	typedef enum
	{
		TimerType_OneShot,
		TimerType_Periodic
	}
	TimerType;


	/**************************************************************************
	
		class WindowTimer 
	
		Author:		Tim Browse
	
		Date:		2001/10/28
	
		Purpose:	Describes a window-based timer.

	**************************************************************************/
	
	class WindowTimer 
	{
		public:
			TimerType	m_Type;
			Window*		m_pWindow;
	};

	//
	// Convenient name for STL type.
	//
	typedef std::map<TimerID, WindowTimer> WindowTimerMap;

	/**************************************************************************
	
		class CallbackTimer
	
		Author:		Tim Browse
	
		Date:		2001/10/28
	
		Purpose:	Describes a callback-based timer.
	
	**************************************************************************/
	
	class CallbackTimer
	{
		public:
			TimerType m_Type;
			TimerProc m_TimerProc;
			UserData  m_UserData;
	};

	//
	// Convenient name for STL type.
	//
	typedef std::map<TimerID, CallbackTimer> CallbackTimerMap;

	/**************************************************************************
	
		class Timer
	
		Author:		Tim Browse
	
		Date:		2001/10/28
	
		Purpose:	Provide timer facilities to vGUI based applications.

					Timers can either be periodic (repeating) or 'one-shot'.

					Timer events can be delivered to a particular Window
					object, or to a specified timer callback function.

					Usage
					~~~~~

					Use SetTimer() to set up a timer event, which will result
					in a unique timer ID.  This ID will be used to identify
					the timer event to your window/callback function.

					For window based timer events, over-ride the OnTimer()
					function in your vgui::Window-derived class to handle
					the timer event.

					Use KillTimer() to kill a timer when you don't need it
					anymore.  OneShot timers are automatically killed after
					the first event, and you should not try to kill them
					yourself (unless you do it before the first event has
					even fired), as KillTimer() will ASSERT if it the timer
					to be killed it not currently 'live'.
  
	**************************************************************************/
	
	class Timer
	{
		public:
			// Set timer events up
			static TimerID SetTimer(TimerType Type,
									int MillisecondDelay,
									Window *pWindow);

			static TimerID SetTimer(TimerType Type,
									int MillisecondDelay,
									TimerProc Handler,
									UserData TheUserData = NULL);

			// Cancel a timer event.
			static void KillTimer(TimerID TimerToKill);

			// Called by the vGUI framework when the host OS delivers
			// timer events.
			static bool ServiceTimer(TimerID ID);

#ifdef VGUI_MSWINDOWS
			// Timer callback for Win32 timers
		private:
			static VOID CALLBACK Win32TimerProc(HWND hwnd,
											    UINT uMsg,
											    UINT idEvent,
											    DWORD dwTime);
#endif
 
		private:
			// Active timers
			static WindowTimerMap m_WindowTimers;
			static CallbackTimerMap m_CallbackTimers;

			// Timer ID source
			static TimerID s_NextID;

			// Allocate a new timer ID.
			static TimerID GetNextTimerID() { return s_NextID++; }
	};
};

#endif // !vgui__Timer_h
