//
// WaitCursor.cpp - the WaitCursor class
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "WaitCursor.h"

#include "vos\Errors.h"

namespace vgui
{
	//
	// Semaphore for wait cursors - if >0 then we should be showing
	// a wait cursor.
	//
	int WaitCursor::s_cWaitCursors = 0;

	/**************************************************************************

	>	WaitCursor::WaitCursor()


		Author:		Tim Browse
		Date:		2001/10/28

		Purpose:	Constructor - initiates display of a WaitCursor.

	**************************************************************************/

	WaitCursor::WaitCursor() : m_bActive(false)
	{
		AddWaitCursor();
	}

	/**************************************************************************
	
	>	WaitCursor::~WaitCursor()
	
		Author:		Tim Browse
		Date:		2001/10/28
	
		Purpose:	Destructor - gets rid of wait cursor display - unless client
					has already called Stop().
	
	**************************************************************************/
	
	WaitCursor::~WaitCursor()
	{
		if (m_bActive)
		{
			RemoveWaitCursor();
		}
	}

	/**************************************************************************
	
	>	void WaitCursor::Stop()
	
		Author:		Tim Browse
		Date:		2001/10/28
	
		Purpose:	Stop the wait cursor being displayed any more.
	
	**************************************************************************/
	
	void WaitCursor::Stop()
	{
		if (m_bActive)
		{
			RemoveWaitCursor();
		}
		else
		{
			// Error - cursor has already been stopped!
			vguiERROR();
		}
	}

	/**************************************************************************
	
	>	void WaitCursor::AddWaitCursor()
	
		Author:		Tim Browse
		Date:		2001/10/28
	
		Purpose:	Increase wait cursor semaphore, so the vGUI system knows
					when to display a wait cursor.
	
	**************************************************************************/
	
	void WaitCursor::AddWaitCursor()
	{
		if (m_bActive)
		{
			// Error - already active!
			vguiERROR();
			return;
		}
		else
		{
			// Wait cursor becoming active, so increase semaphore count.
			m_bActive = true;
			s_cWaitCursors++;
		}
	}

	/**************************************************************************
	
	>	void WaitCursor::RemoveWaitCursor()
	
		Author:		Tim Browse
		Date:		2001/10/28
	
		Purpose:	Decrease wait cursor semaphore, so the vGUI system knows
					when to display a wait cursor.
	
	**************************************************************************/
	
	void WaitCursor::RemoveWaitCursor()
	{
		if (m_bActive)
		{
			// Wait cursor becoming inactive, so decrease semaphore count.
			m_bActive = false;
			s_cWaitCursors--;
			COMPLAIN_IF(s_cWaitCursors < 0, "Illegal wait cursor semaphore value");
		}
		else
		{
			// Error - already inactive!
			vguiERROR();
			return;
		}
	}

	/**************************************************************************
	
	>	bool WaitCursor::IsActive()
	
		Author:		Tim Browse
		Date:		2001/10/28
		Returns:	true => wait cursor should be displayed
					false => normal cursor should be displayed
	
		Purpose:	Find out if any wait cursors are active - used by the
					vGUI application to control display of wait cursors.
	
	**************************************************************************/
	
	bool WaitCursor::IsActive()
	{
		return (s_cWaitCursors > 0);
	}
};


