//
// Window.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Window.h"

#include "Application.h"
#include "WindowImp.h"
#include "Rect.h"
#include "GC.h"
#include "Errors.h"

#include <memory>


//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{

	Window::Window() : m_pWindowImp(NULL),
					   m_pParent(NULL),
					   m_ScrollDeltaX(1),
					   m_ScrollDeltaY(1)
	{
		// Create WindowImp object
		m_pWindowImp = TheFactory().MakeWindow();

		// Link back to this one
		m_pWindowImp->SetWindowLink(this);
	}

	Window::Window(WindowImp *pImp) : m_pWindowImp(pImp),
									  m_pParent(NULL),
									  m_ScrollDeltaX(1),
									  m_ScrollDeltaY(1)
	{
		// We've already got an implementation object.
	}

	Window::~Window()
	{
		// Kill the OS window
		Destroy();

		// Lose the window implementation object.
		delete m_pWindowImp;
	}

	bool Window::Create(Window* pParent,
						TWindowStyle Style, 
						const StringResourceID& resWindowTitle)
	{
		m_pParent = pParent;
		return m_pWindowImp->Create(this, pParent, Style, &resWindowTitle);
	}

	bool Window::Create(Window* pParent,
						TWindowStyle Style, 
						const String& strWindowTitle)
	{
		m_pParent = pParent;
		return m_pWindowImp->Create(this, pParent, Style, &strWindowTitle);
	}

	void Window::Destroy()
	{
		m_pWindowImp->Destroy();
	}

	void Window::Attach(Window *pParent,
						ControlID ID)
	{
		m_pParent = pParent;
		m_pWindowImp->Attach(pParent, ID);
	}

	void Window::AttachByHandle(Window *pParent,
								WindowHandle hWindow)
	{
		m_pParent = pParent;
		m_pWindowImp->AttachByHandle(pParent, hWindow);
	}

	void Window::Detach()
	{
		m_pParent = NULL;
		m_pWindowImp->Detach();
	}

	bool Window::OnCreate()
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnDestroy()
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnCommand(int nCommand)
	{
		// No window specific stuff yet - just do message map handling
		return MessageTarget::OnCommand(nCommand);

	}

	bool Window::OnRedraw()
	{
		// Null redraw handler - don't claim the redraw message
		return false;
	}

	bool Window::OnMove(int, int)
	{
		// Null size handler - don't claim the message
		return false;
	}

	bool Window::OnResizing(vgui::Size* pNewSize)
	{
		// Null size handler - don't claim the message
		return false;
	}

	bool Window::OnResize(int, int)
	{
		// Null size handler - don't claim the message
		return false;
	}

	bool Window::OnGainFocus()
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnLoseFocus()
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnLoseMouseCapture()
	{
		// Don't claim this message
		return false;
	}

	
	/**************************************************************************

	>	bool Window::OnChar(uint16 Char)

		Author:		Tim Browse
		Date:		11/03/2001

		Inputs:		Char - the Unicode character.

		Returns:	true => message processed;
					false otherwise (message should be passed on to other
									 handlers)

		Purpose:	Handle simple keyboard character events (ASCII and/or
					Unicode).

					Be careful what you claim - especially in dialogs.  On some
					platforms, claiming keys like Tab, cursor keys in a dialog
					control can prevent the standard navigation mechanisms 
					working, so don't blindly claim everything; claim only what
					you need to process.

	**************************************************************************/

	bool Window::OnChar(uint16 Char)
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnKeyDown(TKeyCode Key, int Modifiers)
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnKeyUp(TKeyCode Key, int Modifiers)
	{
		// Don't claim this message
		return false;
	}

	bool Window::IsCharWanted(uint16 Char)
	{
		// Do not claim input by default
		return false;
	}

	bool Window::IsKeyWanted(TKeyCode Key)
	{
		// Do not claim input by default
		return false;
	}

	bool Window::OnTimer(TimerID ID)
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnClose()
	{
		// Over-ride and return true when you don't want the view to close.
		// (true => message handled, so no further close processing is done.)
		return false;
	}

	bool Window::OnMouseClick(TMouseButton Button, const Point& Pos)
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnMouseDoubleClick(TMouseButton Button, const Point& Pos)
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnMouseRelease(TMouseButton Button, const Point& Pos)
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnMouseMove(const Point& Pos)
	{
		// Don't claim this message
		return false;
	}

	bool Window::OnMouseWheel(int WheelDelta)
	{
		// Don't claim this message
		return false;
	}

	void Window::ShowWindow(TShowWindow Command)
	{
		m_pWindowImp->ShowWindow(Command);
	}


	void Window::UpdateWindow()
	{
		m_pWindowImp->UpdateWindow();
	}

	void Window::CentreWindow()
	{
		m_pWindowImp->CentreWindow();
	}

	bool Window::SetText(const String& NewText)
	{
		return m_pWindowImp->SetText(NewText);
	}

	bool Window::SetText(const StringResourceID& NewTextID)
	{
		String Text;
		if (NewTextID.Load(&Text))
		{
			return SetText(Text);
		}
		else
		{
			return false;
		}
	}

	bool Window::GetText(String* pText) const
	{
		return m_pWindowImp->GetText(*pText);
	}

	bool Window::TextIsEmpty() const
	{
		vos::String Text;
		if (m_pWindowImp->GetText(Text) && !Text.IsEmpty())
		{
			// We have some text
			return false;
		}
		else
		{
			// The text is empty
			return true;
		}
	}

	GC* Window::GetGC()
	{
		try
		{
			// Create GC and attach to this window
			return new GC(this);
		}
		catch (...)
		{
			// Failed to create GC - return no GC.
			return NULL;
		}
	}

	void Window::ReleaseGC(GC* pGC)
	{
		// Ok, delete this GC
		delete pGC;
	}

	WindowHandle Window::GetWindowHandle() const
	{
		return m_pWindowImp->GetWindowHandle();
	}

	void Window::GetArea(Rect* pArea)
	{
		m_pWindowImp->GetArea(pArea);
	}

	void Window::GetClientArea(Rect* pClientArea)
	{
		m_pWindowImp->GetClientArea(pClientArea);
	}

	void Window::ClientToWindow(Window *pWindow, Point* pPoint)
	{
		m_pWindowImp->ClientToWindow(pWindow, pPoint, 1);
	}
	
	void Window::ClientToWindow(Window *pWindow, Rect* pRect)
	{
		m_pWindowImp->ClientToWindow(pWindow, (Point*) pRect, 2);
	}

	void Window::ForceRedraw()
	{
		// No rectangle => complete redraw
		m_pWindowImp->ForceRedraw(NULL);
	}

	void Window::ForceRedraw(const Rect& Area)
	{
		m_pWindowImp->ForceRedraw(&Area);
	}

	void Window::SetWindowMode(TWindowMode Mode)
	{
		m_pWindowImp->SetWindowMode(Mode);
	}

	void Window::GetPosition(Rect* pWindowPosition)
	{
		m_pWindowImp->GetPosition(pWindowPosition);
	}

	void Window::Move(const Rect& NewPosition, bool bForceRedraw)
	{
		m_pWindowImp->Move(NewPosition, bForceRedraw);
	}

	void Window::SetPosition(const Rect& NewPosition, bool bForceRedraw)
	{
		m_pWindowImp->Move(NewPosition, bForceRedraw);
	}

	void Window::SetSafePosition(const Rect& NewPosition, bool bForceRedraw)
	{
		m_pWindowImp->SetSafePosition(NewPosition, bForceRedraw);
	}
	
	void Window::MoveRelative(int DeltaX, int DeltaY, bool bForceRedraw)
	{
		// Get the current position of the window
		Rect WindowPosition;
		m_pWindowImp->GetPosition(&WindowPosition);

		// Translate the window position
		WindowPosition.Translate(DeltaX, DeltaY);

		// Write the position back to the window
		m_pWindowImp->Move(WindowPosition, bForceRedraw);
	}

	void Window::SetClientArea(const Rect& NewClientArea)
	{
		m_pWindowImp->SetClientArea(NewClientArea);
	}

	void Window::SetWorkArea(const Rect& WorkArea)
	{
		m_pWindowImp->SetWorkArea(WorkArea);
	}

	void Window::GetWorkArea(Rect* pWorkArea)
	{
		m_pWindowImp->GetWorkArea(pWorkArea);
	}

	void Window::Resize(uint NewWidth, uint NewHeight)
	{
		m_pWindowImp->Resize(NewWidth, NewHeight);
	}

	ControlID Window::GetControlID()
	{
		return m_pWindowImp->GetControlID();
	}

	ControlID Window::FindUniqueControlID()
	{
		return m_pWindowImp->FindUniqueControlID();
	}

	void Window::ShowScrollBar(TScrollBar Bar)
	{
		m_pWindowImp->ShowScrollBar(Bar);
	}

	void Window::HideScrollBar(TScrollBar Bar)
	{
		m_pWindowImp->HideScrollBar(Bar);
	}

	void Window::EnableScrollBar(TScrollBar Bar, bool bEnable)
	{
		m_pWindowImp->EnableScrollBar(Bar, bEnable);
	}

	bool Window::IsScrollBarVisible(TScrollBar Bar)
	{
		return m_pWindowImp->IsScrollBarVisible(Bar);
	}

	void Window::SetScrollDelta(TScrollBar Bar, int Delta)
	{
		if (Bar & sb_Horizontal)
			m_ScrollDeltaX = Delta;

		if (Bar & sb_Vertical)
			m_ScrollDeltaY = Delta;
	}

	int Window::GetScrollDelta(TScrollBar Bar)
	{
		if (Bar & sb_Horizontal)
			return m_ScrollDeltaX;
		else
			return m_ScrollDeltaY;
	}

	Font* Window::GetFont() const
	{
		return m_pWindowImp->GetFont();
	}

	int Window::GetScrollPos(TScrollBar Bar)
	{
		return m_pWindowImp->GetScrollPos(Bar);
	}

	void Window::SetScrollPos(TScrollBar Bar, int Pos)
	{
		m_pWindowImp->SetScrollPos(Bar, Pos);
	}

	bool Window::IsWindowActive() const
	{
		return (GetWindowHandle() != (WindowHandle) NullWindowHandle);
	}

	void Window::SetFocus()
	{
		m_pWindowImp->SetFocus();
	}

	bool Window::HasFocus()
	{
		return m_pWindowImp->HasFocus();
	}

	bool Window::CaptureMouse()
	{
		return m_pWindowImp->CaptureMouse();
	}

	void Window::ReleaseMouse()
	{
		m_pWindowImp->ReleaseMouse();
	}

	void Window::LockRedraw()
	{
		GetImp()->LockRedraw();
	}

	void Window::UnlockRedraw()
	{
		GetImp()->UnlockRedraw();
	}

	TWindowPosture Window::GetPosture()
	{
		return GetImp()->GetPosture();
	}

	void Window::SetPosture(TWindowPosture Posture)
	{
		GetImp()->SetPosture(Posture);
	}

	void Window::OnControlResizing()
	{
		GetImp()->OnControlResizing();
	}

	void Window::OnControlResize()
	{
		GetImp()->OnControlResize();
	}
	
	/**************************************************************************
	
	>	bool Window::GetDialog(Dialog** ppDialog)
	
		Author:		Tim Browse
		Date:		25/7/2002
		Outputs:	ppDialog - the dialog pointer
		Returns:	true => the window is a Dialog object, and ppDialog contains
							a pointer to the dialog.
					false => this window is not a dialog.
	
		Purpose:	Access this window object as a dialog, if it is actually
					a dialog.

					This saves you from doing fiddly checks and casts etc.
	
	**************************************************************************/
	
	bool Window::GetDialog(Dialog** ppDialog)
	{
		// This is a normal window
		return false;
	}

	/**************************************************************************
	
	>	void Window::EnableFloatingBehaviour(bool bFloat = true)
	
		Author:		Tim Browse
		Date:		11/9/2002
		Inputs:		bFloat - true => make this window appear above all other
									 windows on screen, even if it does not
									 have focus.
							false => make window behave normally, i.e. other
									 windows can appear on top of it
	
		Purpose:	Control the behaviour of a window with respect to whether
					it appears above all 'normal' windows.
	
	**************************************************************************/
	
	void Window::EnableFloatingBehaviour(bool bFloat)
	{
		GetImp()->EnableFloatingBehaviour(bFloat);
	}
};
