//
// Window.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_Window_h
#define vgui_Window_h

#include "vgui.h"

#include "String.h"
#include "ResourceID.h"
#include "MessageTarget.h"
#include "Timer.h"
#include "ControlID.h"
#include "Mouse.h"
#include "Keyboard.h"
#include "Size.h"

#include <memory>

namespace vgui
{
	// Low-level opaque OS-dependent window handle
#if defined VGUI_MSWINDOWS
	typedef HWND WindowHandle;
#elif defined VGUI_MACOS
	typedef WindowRef WindowHandle;
#endif

	// Null window handle constant
	enum { NullWindowHandle = 0 };

	class WindowImp;
	class GC;
	class Rect;
	class Point;
	class Font;

	typedef enum
	{
		wp_Default,
		wp_Centred
	}
	WindowPos;

	//
	// Window manipulation
	//
	typedef enum
	{
		sw_SHOW,
		sw_HIDE
	}
	TShowWindow;

	typedef enum
	{
		sb_Horizontal = BIT(0),
		sb_Vertical   = BIT(1)
	}
	TScrollBar;


	//
	// Window styles
	//

	typedef enum
	{
		ws_MainAppWindow,
		ws_DocumentView,
		ws_GenericChild
	}
	TWindowStyle;

	//
	// Window Posture - describes how a window is positioned
	// (minimised, normal, maximised - where these states make any
	//  sense on the host GUI platform).
	//
	typedef enum
	{
		WindowPosture_Normal,
		WindowPosture_Minimised,
		WindowPosture_Maximised
	}
	TWindowPosture;

	//
	// Used for windows that take over the whole screen
	// (see SetWindowMode()).
	//
	typedef enum
	{
		WindowMode_Normal,
		WindowMode_FullScreen
	}
	TWindowMode;

	class Dialog;

	class Window : public MessageTarget
	{
		public:

			//
			// Constructor/Destructor
			//

			Window();
			virtual ~Window();

			//
			// Window creation/destruction
			//
			bool Create(Window* pParent,
						TWindowStyle Style, 
						const StringResourceID& resWindowTitle);

			bool Create(Window* pParent,
						TWindowStyle Style, 
						const vos::String& strWindowTitle);

			void Destroy();

			void ShowWindow(TShowWindow Command);

			void ShowScrollBar(TScrollBar Bar);
			void HideScrollBar(TScrollBar Bar);

			void EnableScrollBar(TScrollBar Bar, bool bEnable);

			bool IsScrollBarVisible(TScrollBar Bar);

			void SetScrollDelta(TScrollBar Bar, int Delta);
			int  GetScrollDelta(TScrollBar Bar);

			int GetScrollPos(TScrollBar Bar);
			void SetScrollPos(TScrollBar Bar, int Pos);

			TWindowPosture GetPosture();
			void SetPosture(TWindowPosture Posture);

			void UpdateWindow();

			void CentreWindow();

			// Get position of window relative to parent window
			void GetPosition(Rect* pWindowPosition);

			// Move window - SetSafePosition() ensures the window is visible
			// on screen.
			void SetPosition(const Rect& NewPosition, bool bForceRedraw = true);
			void SetSafePosition(const Rect& NewPosition, bool bForceRedraw = true);

			// Alias for SetPosition (deprecated)
			void Move(const Rect& NewPosition, bool bForceRedraw = true);

			void MoveRelative(int DeltaX, int DeltaY, bool bForceRedraw = true);

			void SetClientArea(const Rect& NewClientArea);
			void Resize(uint NewWidth, uint NewHeight);

			void SetWorkArea(const Rect& WorkArea);
			void GetWorkArea(Rect* pWorkArea);

			void EnableFloatingBehaviour(bool bFloat = true);

			bool SetText(const vos::String& NewText);
			bool SetText(const StringResourceID& NewTextID);
			bool GetText(vos::String* pText) const;
			bool TextIsEmpty() const;


			void SetFocus();
			bool HasFocus();

			bool CaptureMouse();
			void ReleaseMouse();

			//
			// Get information
			//

			Window* GetParent() const { return m_pParent; }

			void GetArea(Rect* pArea);
			void GetClientArea(Rect* pClientArea);

			ControlID GetControlID();

			ControlID FindUniqueControlID();

			WindowHandle GetWindowHandle() const;

			bool IsWindowActive() const;

			Font* GetFont() const;

			// If this window is actually a Dialog, get the Dialog
			// object.
			virtual bool GetDialog(Dialog** ppDialog);

			//
			// GC functions
			//

			GC* GetGC();
			void ReleaseGC(GC*);

			// Control related functions
			virtual void Attach(Window *pParent,
								ControlID ID);

			virtual void AttachByHandle(Window *pParent,
										WindowHandle hWindow);

			void Detach();

			virtual void OnControlResizing();
			virtual void OnControlResize();

			//
			// Message handling
			//

			// Sent during window creation
			virtual bool OnCreate();

			// Sent when window is being destroyed.
			virtual bool OnDestroy();

			virtual bool OnCommand(int nCommand);

			virtual bool OnRedraw();

			virtual bool OnMove(int NewX, int NewY);

			virtual bool OnResizing(vgui::Size* pNewSize);
			virtual bool OnResize(int NewWidth, int NewHeight);

			virtual bool OnGainFocus();
			virtual bool OnLoseFocus();

			virtual bool OnLoseMouseCapture();

			virtual bool OnTimer(TimerID ID);

			virtual bool OnClose();

			//
			// Input functions
			//

			// Handle 'cooked' input - i.e. the standard characters that
			// result from key presses or key press combinations.
			// 
			// See function header for detailed spec.
			virtual bool OnChar(uint16 Char);

			// Key up/down events for any keys pressed on keyboard, including
			// those that do not map to characters
			virtual bool OnKeyDown(TKeyCode Key, int Modifiers);
			virtual bool OnKeyUp(TKeyCode Key, int Modifiers);

			// Input masking - by default a window claims all key events
			virtual bool IsCharWanted(uint16 Char);
			virtual bool IsKeyWanted(TKeyCode Key);

			// Mouse events
			virtual bool OnMouseClick(TMouseButton Button, const Point& Pos);
			virtual bool OnMouseDoubleClick(TMouseButton Button, const Point& Pos);
			virtual bool OnMouseRelease(TMouseButton Button, const Point& Pos);
			virtual bool OnMouseMove(const Point& Pos);
			virtual bool OnMouseWheel(int WheelDelta);

			// Co-ordinate functions
			void ClientToWindow(Window *pWindow, Point* pPoint);
			void ClientToWindow(Window *pWindow, Rect* pRect);

			// Redraw functions
			void ForceRedraw();
			void ForceRedraw(const Rect& Area);

			// Redraw lock/unlock
			void LockRedraw();
			void UnlockRedraw();

			// Full-screen support
			void SetWindowMode(TWindowMode Mode);

		protected:
			Window(WindowImp *pImp);

		private:
			// We don't ever want to let anyone copy these objects.
			Window(const Window& rhs);

		public:
			WindowImp *GetImp() const { return m_pWindowImp; }

		protected:
			// The implementation of this Window.
			// (Pattern: Bridge)
			WindowImp* m_pWindowImp;

			// Parent of this window
			Window *m_pParent;

			// Scroll bar increments
			int m_ScrollDeltaX;
			int m_ScrollDeltaY;
	};
};

#endif // !vgui_Window_h


