//
// FileOpenDialog.cpp - dialog for specifying where to open a file from.
// (Win32 version)
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "..\FileOpenDialog.h"
#include "Win32_WindowImp.h"

#include "Win32_Headers.h"

#include <commdlg.h>
#include <shlobj.h>
#include <cderr.h>

#include "..\Errors.h"
#include "..\Msgs.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	static void InitialiseFilters(FileDialogParams *pParams,
								  OPENFILENAME* pOFN)
	{
		// Create filters from document classes.
		DocumentClasses::iterator it = pParams->m_FileClasses.begin();

		if (it == pParams->m_FileClasses.end())
		{
			// No file classes specified - use standard filter (all files)
			pParams->m_DefaultExtension.Clear();

			// Copy standard filter into the string
			strcpy(pParams->m_szFileFilters, msgFileDialog_DefaultFilter);
		}
		else
		{
			char *pszFilter = pParams->m_szFileFilters;

			while (it != pParams->m_FileClasses.end())
			{
				// Get default extension from document class, if 
				// we don't have one already.
				if (pParams->m_DefaultExtension.IsEmpty())
				{
					pParams->m_DefaultExtension = (*it).m_ClassType.m_FileExtension;
				}

				// Construct filter - filters are formatted as follows:
				// "File description (*.ext)|*.ext|"
				// where | is a null terminator (two null terminators are used
				// to terminate the filter string).
				sprintf(pszFilter,
						msgDocument_FilterDescription,
						(*it).m_ClassName.c_str(),
						(*it).m_ClassType.m_FileExtension.c_str(),
						(*it).m_ClassType.m_FileExtension.c_str());

				// Move past end of string, ready for next filter.
				pszFilter += strlen(pszFilter);

				// Look for next class of file
				++it;
			}
		}

		// Go through the filter string and convert all '|' characters to nulls,
		// so we get a proper filter string.
		char *s = pParams->m_szFileFilters;

		while (*s != 0)
		{
			if (*s == '|')
				*s = 0;

			s++;
		}

		// Set up the file filters data for save dialog.
		pOFN->lpstrFilter = pParams->m_szFileFilters;

		// Select the appropriate filter.
		pOFN->nFilterIndex = pParams->m_nFilterIndex;
	}


	FileOpenDialog::FileOpenDialog(FileDialogParams* pParams,
								   const Window* pParent) : m_pParams(pParams),
															m_pParent(pParent)
	{
		// TCILB
	}

	FileOpenDialog::~FileOpenDialog()
	{
	}

	bool FileOpenDialog::GetFilename(vos::FileSpec* pFilename)
	{
		// Create a Win32 open file dialog - first, fill in the OPENFILENAME structure.
		OPENFILENAME ofn;
		ZeroMemory(&ofn, sizeof(ofn));
		ofn.lStructSize = sizeof(ofn);

		// Fill in parent window handle - if we have a parent.
		if (m_pParent != NULL)
		{
			ofn.hwndOwner = Win32_WindowImp::GetHWND(m_pParent);
		}

		// Set up buffers to remember the filter chosen by the user.
		InitialiseFilters(m_pParams, &ofn);

		// Create buffer for storing filename in.
		const int kFileBufSize = 512;
		char *pFileNameBuf = new char[kFileBufSize];

		// If the caller provided a filename, fill it in
		if (!pFilename->IsEmpty())
		{
			strcpy(pFileNameBuf, pFilename->GetString().c_str());
		}
		else
		{
			// No filename - clear to empty string
			pFileNameBuf[0] = 0;
		}

		ofn.lpstrFile = pFileNameBuf;
		ofn.nMaxFile = kFileBufSize - 1;

		// We want to fill in the initial folder - if none provided, then 
		// choose an appropriate folder.
		if (m_pParams->m_InitialFolder.IsEmpty())
		{
			// Get location of user's personal files
			LPITEMIDLIST pidl;
			if (SUCCEEDED(SHGetSpecialFolderLocation(ofn.hwndOwner, CSIDL_PERSONAL, &pidl)))
			{
				// Extract the path
				char szPath[MAX_PATH+1];
				if (SHGetPathFromIDList(pidl, szPath))
				{
					// Found a usable pathname - store it in the params object
					m_pParams->m_InitialFolder.SetFromString(szPath);

				}

				// Delete the string buffer returned by the Shell
				LPMALLOC pMalloc;
				if (SUCCEEDED(SHGetMalloc(&pMalloc)))
				{
					// Got IMalloc pointer - use it to delete PIDL
					pMalloc->Free(pidl);

					// Finished with IMalloc interface
					pMalloc->Release();
				}
				else
				{
					// Unable to get the shell's memory allocator - complain about this,
					// as it should not happen.
					COMPLAIN("Unable to get Shell's IMalloc interface while "
							 "cleaning up data in FileOpenDialog::GetFilename().");
				}
			}
		}

		// Ok, if we have an initial path, then use it.
		String InitialFolder = m_pParams->m_InitialFolder.GetString();
		if (!InitialFolder.IsEmpty())
		{
			ofn.lpstrInitialDir = InitialFolder.c_str();
		}

		// Fill in title of dialog
		if (!m_pParams->m_DialogTitle.IsEmpty())
		{
			ofn.lpstrTitle = m_pParams->m_DialogTitle.c_str();
		}
		else
		{
			// Default to standard English "Open..."
			ofn.lpstrTitle = "Open...";
		}

		// Ok, ask the Dialog library to execute the Open File dialog.
		// (GetFilename() returns zero on failure)
		bool bSuccess = ::GetOpenFileName(&ofn) != 0;

		// If it worked, copy back any useful data
		if (bSuccess)
		{
			// Remember the filename!
			pFilename->SetFromString(ofn.lpstrFile);

			// Remember what filter the user chose.
			m_pParams->m_nFilterIndex = ofn.nFilterIndex;

			// Remember the folder
			m_pParams->m_InitialFolder.SetFromString(ofn.lpstrInitialDir);
		}
		else
		{
			// Something went wrong - find out what...
			DWORD dwError = CommDlgExtendedError();
			if (dwError == CDERR_INITIALIZATION)
			{
				::MessageBox(NULL, "Init failure", "Open File error", MB_OK);
			}
		}

		// Free up any buffers we allocated.
		delete [] pFileNameBuf;

		// Tell caller if it worked
		return bSuccess;
	}


};
