//
// Win32\Menu.cpp - Win32 code for accessing menus
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "..\Menu.h"
#include "..\MenuItem.h"

#include "Win32_Headers.h"

#include "vos\Errors.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	/**************************************************************************

		void FillInMenuItemInfo(MENUITEMINFO* pmio, const MenuItem& Item)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		Item - the item to extract the information from.
		Outputs:	pmio - the structure to fill in according to the contents
						   of 'Item'.

		Purpose:	Convert the menu item data in a vgui format MenuItem to
					a Win32 structure.

	**************************************************************************/

	void FillInMenuItemInfo(MENUITEMINFO* pmio, const MenuItem& Item)
	{
		// Set up structure
		::ZeroMemory(pmio, sizeof(MENUITEMINFO));
		pmio->cbSize = sizeof(MENUITEMINFO);

		pmio->fMask = MIIM_CHECKMARKS|MIIM_TYPE|MIIM_STATE|MIIM_ID;
		pmio->fType = MFT_STRING;

		// Fill in text string
		pmio->dwTypeData = (LPSTR) Item.m_Text.c_str();
		pmio->cch = Item.m_Text.GetLength();

		// Check if item is enabled/disabled
		if (Item.m_bEnabled)
			pmio->fState = MFS_ENABLED;
		else
			pmio->fState = MFS_DISABLED;

		if (Item.m_bChecked)
			pmio->fState |= MFS_CHECKED;

		// Set command ID
		pmio->wID = Item.m_ID;
	}

	/**************************************************************************

		Menu::Menu()

		Author:		Tim Browse
		Date:		29/8/00

		Purpose:	Constructor

	**************************************************************************/

	Menu::Menu() : m_rMenu(NULL)
	{
	}

	/**************************************************************************

		Menu::Menu(MenuRef rMenu)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		rMenu - reference to the menu this object will operate on.

		Purpose:	Constructor

	**************************************************************************/

	Menu::Menu(MenuRef rMenu) : m_rMenu(rMenu)
	{
	}

	/**************************************************************************
	
	>	void Menu::SetMenuRef(MenuRef rMenu)
	
		Author:		Tim Browse
		Date:		8/8/2002
		Inputs:		rMenu - new menu reference to use
	
		Purpose:	Set the menu reference used by this menu object.
	
	**************************************************************************/
	
	void Menu::SetMenuRef(MenuRef rMenu)
	{
		m_rMenu = rMenu;
	}

	/**************************************************************************
	
	>	MenuRef Menu::GetMenuRef()
	
		Author:		Tim Browse
		Date:		8/8/2002
		Returns:	The MenuRef of this object.
	
		Purpose:	Get the menu reference handle used by this Menu object.
	
	**************************************************************************/
	
	MenuRef Menu::GetMenuRef()
	{
		return m_rMenu;
	}

	/**************************************************************************

		uint Menu::GetItemCount()

		Author:		Tim Browse
		Date:		29/8/00
		Returns:	Number of items in the menu.

		Purpose:	Find out how many items are in the menu.

	**************************************************************************/

	uint Menu::GetItemCount()
	{
		int cItems = ::GetMenuItemCount(m_rMenu);

		if (cItems == -1)
		{
			// Failed to get menu item count
			cItems = 0;
		}

		return cItems;
	}


	/**************************************************************************

		bool Menu::InsertItem(const MenuItem& NewItem, uint nPos)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		NewItem - details of the item to add.
					nPos - where to add the item (0-based)
		Returns:	true => item inserted successfully
					false otherwise

		Purpose:	Insert a new menu item into the menu at the specified 
					position.  (The item is inserted before that position)

	**************************************************************************/

	bool Menu::InsertItem(const MenuItem& NewItem, uint nPos)
	{
		// Translate menu info to Win32 format.
		MENUITEMINFO mio;
		FillInMenuItemInfo(&mio, NewItem);

		// Insert the item
		return (::InsertMenuItem(m_rMenu, nPos, TRUE, &mio) != 0);
	}

	/**************************************************************************

		bool Menu::AppendItem(const MenuItem& NewItem)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		NewItem - details of the item to add.
		Returns:	true => item inserted successfully
					false otherwise

		Purpose:	Append an item to the end of the menu.

	**************************************************************************/

	bool Menu::AppendItem(const MenuItem& NewItem)
	{
		return InsertItem(NewItem, GetItemCount());
	}

	/**************************************************************************

		bool Menu::DeleteItem(uint nPos)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		nPos - index of the item to delete (0-based index)
		Returns:	true => item deleted successfully
					false otherwise

		Purpose:	Delete a menu item from the menu.

	**************************************************************************/

	bool Menu::DeleteItem(uint nPos)
	{
		return (::DeleteMenu(m_rMenu, nPos, MF_BYPOSITION) != 0);
	}

	/**************************************************************************

		void Menu::UpdateItem(uint nPos, const MenuItem& NewItem)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		nPos - zero-based index of the menu item to change.
					NewItem - details of the new item data.

		Purpose:	Change/update a menu item's text or state.

	**************************************************************************/

	void Menu::UpdateItem(uint nPos, const MenuItem& NewItem)
	{
		// Translate menu info to Win32 format.
		MENUITEMINFO mio;
		FillInMenuItemInfo(&mio, NewItem);

		// Insert the item
		::SetMenuItemInfo(m_rMenu, nPos, TRUE, &mio);
	}

	/**************************************************************************

		void Menu::TickItem(MenuCommandID ID, bool bTick)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		ID - the command ID of the item to tick
					bTick - true => tick the item
						   false => untick the item

		Purpose:	Change/update a menu item tick/checkmark

	**************************************************************************/

	void Menu::TickItem(MenuCommandID ID, bool bTick)
	{
		// Set up flags - we want to identify the menu item by its command id
		UINT uCheck = MF_BYCOMMAND;
		if (bTick)
			uCheck |= MF_CHECKED;
		else
			uCheck |= MF_UNCHECKED;

		::CheckMenuItem(m_rMenu, ID, uCheck);
	}


	/**************************************************************************

		bool Menu::GetItem(uint nPos, MenuItem* pItem)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		nPos - zero-based index of the menu item to interrogate.
		Outputs:	pItem - the data for this menu item.
		Returns:	true => Menu item found and data retrieved ok
					false => Menu index out of range

		Purpose:	Retrieve the data for a specified menu item.

	**************************************************************************/

	bool Menu::GetItem(uint nPos, MenuItem* pItem)
	{
		// Range check...
		if (nPos >= GetItemCount())
		{
			// No such item in the menu.
			return false;
		}

		// Extract menu item info into Win32 structure
		MENUITEMINFO mio;
		::ZeroMemory(&mio, sizeof(MENUITEMINFO));
		mio.cbSize = sizeof(MENUITEMINFO);

		mio.fMask = MIIM_CHECKMARKS|MIIM_TYPE|MIIM_STATE|MIIM_ID;
		mio.fType = MFT_STRING;

		::GetMenuItemInfo(m_rMenu, nPos, TRUE, &mio);

		// Add 1 for 0 terminator
		mio.cch++;

		// Ok, now we know how long the text buffer needs to be
		char *pBuffer;
		try
		{
			pBuffer = new char[mio.cch];
		}
		catch (...)
		{
			// Failed to allocate buffer - abort operation
			return false;
		}

		mio.dwTypeData = (LPSTR) pBuffer;

		// Re-request menu info so we get the text label this time.
		::GetMenuItemInfo(m_rMenu, nPos, TRUE, &mio);

		// Decode to vgui format...

		// Get textual label
		pItem->m_Text = pBuffer;
		delete [] pBuffer;

		// Check item state
		pItem->m_bEnabled = (mio.fState & MFS_ENABLED) ? true : false;
		pItem->m_bChecked = (mio.fState & MFS_CHECKED) ? true : false;

		// Check item type
		pItem->m_bSeparator = (mio.fType & MFT_SEPARATOR) ? true : false;

		// Get command ID
		pItem->m_ID = mio.wID;

		return true;
	}

	/**************************************************************************

		bool Menu::FindItem(MenuCommandID ID, uint* pPos)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		ID - the ID of the menu item to search for.
		Outputs:	pPos - where the item was found (if it was)
		Returns:	true => menu item was found
					false => othrwise

		Purpose:	Search for a menu item with a command ID matching that
					passed in.

	**************************************************************************/

	bool Menu::FindItem(MenuCommandID ID, uint* pPos)
	{
		// Search this menu for this command ID
		uint cItems = GetItemCount();

		// Set up win32 structure ready to receive command ID
		MENUITEMINFO mio;
		::ZeroMemory(&mio, sizeof(MENUITEMINFO));
		mio.cbSize = sizeof(MENUITEMINFO);
		mio.fMask = MIIM_ID;

		uint i = 0;

		while (i < cItems)
		{
			// Get info on this item's command ID
			::GetMenuItemInfo(m_rMenu, i, TRUE, &mio);

			if (mio.wID == ID)
			{
				// Positive match
				*pPos = i;
				return true;
			}

			// Otherwise, try next item
			i++;
		}

		// Failed to match
		return false;
	}


	/**************************************************************************

		void Menu::InsertSeparator(uint nPos)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		nPos - where to insert the separator (zero-based index)

		Purpose:	Insert a menu separator bar at the specified point.

	**************************************************************************/

	void Menu::InsertSeparator(uint nPos)
	{
		// Set up structure to indicate a separator
		MENUITEMINFO mio;
		::ZeroMemory(&mio, sizeof(MENUITEMINFO));
		mio.cbSize = sizeof(MENUITEMINFO);
		mio.fType = MFT_SEPARATOR;

		// Insert the separator item
		::InsertMenuItem(m_rMenu, nPos, TRUE, &mio);
	}

	/**************************************************************************

		void Menu::AppendSeparator()

		Author:		Tim Browse
		Date:		29/8/00

		Purpose:	Append a menu separator bar to the end of the menu.

	**************************************************************************/

	void Menu::AppendSeparator()
	{
		// Insert separator at end of menu
		InsertSeparator(GetItemCount());
	}

	/**************************************************************************

		bool Menu::GetSubMenu(uint nPos, Menu* pSubMenu)

		Author:		Tim Browse
		Date:		29/8/00
		Inputs:		nPos - the zero-based index of the sub-menu item
		Outputs:	pMenu - the sub-menu that was asked for.
		Returns:	true => found sub-menu successfully
					false => index out of bounds, or item is not a sub-menu.

		Purpose:	Get a Menu object for the specified sub-menu of this menu.

	**************************************************************************/

	bool Menu::GetSubMenu(uint nPos, Menu* pSubMenu)
	{
		if (nPos >= GetItemCount())
		{
			COMPLAIN("GetSubMenu() - index out of range!");
			return false;
		}

		HMENU hMenu = ::GetSubMenu(m_rMenu, nPos);

		if (hMenu == NULL)
		{
			// item is not a sub-menu
			return false;
		}
		
		// Install handle into menu provided
		pSubMenu->m_rMenu = hMenu;
		return true;
	}


};

