//
// Win32_ApplicationImp.cpp - Application Implementation Interface 
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_ApplicationImp.h"

#include "..\Application.h"
#include "..\GC.h"
#include "..\Dialog.h"
#include "..\Menu.h"
#include "..\Keyboard.h"

#include "Win32_GCImp.h"
#include "Win32_WindowImp.h"
#include "Win32_Theme.h"
#include "Win32_WindowInfo.h"

#include "Win32_Headers.h"
#include "commctrl.h"
#include "shellapi.h"
#include "objbase.h"

#include "..\Errors.h"

#include "vos\Trace.h"

#ifdef _MSC_VER
#include <crtdbg.h>
#endif

static char *szCommandLine = NULL;

//
// The delay (in ms) that we wait for before kicking off idle processing.
// i.e. the user has to be idle for this many seconds before
//      we start processing idle messages.
//
//		At the moment, the delay is 2 seconds.
//

const DWORD kmsIdleDelay = 2000;

//
// Function typedef for InitCOmmonControlsEx(), which is used to
// initialise COMCTL32.DLL.
//
typedef BOOL WINAPI InitCommonControlsExFn(LPINITCOMMONCONTROLSEX);

//
// Function typedef for CoInitialize(), which is used to
// initialise OLE/COM support (from OLE32.DLL).
//
typedef HRESULT STDAPICALLTYPE CoInitializeFn(LPVOID);
 
namespace vgui
{
	// Return code for WNDPROCs to return when they haven't processed the
	// message.
	const int kWndProc_NotHandled = 0x42424242;

	//////////////////////////////////////////////////////////////////////
	//
	// Constants for system tray usage
	//

	// kSysTrayID is the ID of our app's system tray icon.  We can
	// have multiple icons by using multiple IDs, but so far we only
	// support one icon in the tray area.
	const int kSysTray_ID = 1000;


	//////////////////////////////////////////////////////////////////////


	//
	// The singleton pointer to the Win32 app object.
	//
	Win32_ApplicationImp *Win32_ApplicationImp::s_pWin32App = NULL;


	/**************************************************************************
	
	>	Win32_ApplicationImp::Win32_ApplicationImp()
	
		Author:		Tim Browse
		Date:		15/6/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	Win32_ApplicationImp::Win32_ApplicationImp() : m_hInstance(NULL),
												   m_hAccelTable(NULL),
												   m_bSysTrayIcon(false),
												   m_cmdDefault(NullCommand),
												   m_hAccessoryMenu(NULL),
												   m_cmdNotify(NullCommand),
												   m_bNotifyActive(false),
												   m_bQuitReceived(false),
												   m_hMainWindow(NULL),
												   m_hComCtl32(NULL),
												   m_hOLE32(NULL),
												   m_MsgWParam(0),
												   m_pMainMenu(NULL),
												   m_CurrentModelessDlg(NULL),
												   m_hSysTrayWindow(NULL),
												   m_dwTimeOfLastInputMessage(0)
	{
		// Only one application should be created, so we make sure we don't
		// create two for any strange reasons.
		ASSERT(s_pWin32App == NULL);

		// Initialise singleton pointer
		s_pWin32App = this;

		// Initialise window classes
		m_szWindowClass[0] = 0;
		m_szChildWindowClass[0] = 0;
	}

	/**************************************************************************
	
	>	Win32_ApplicationImp::~Win32_ApplicationImp()
	
		Author:		Tim Browse
		Date:		15/6/2002
	
		Purpose:	Destructor
	
	**************************************************************************/
	
	Win32_ApplicationImp::~Win32_ApplicationImp()
	{
		// Clean up the accessory menu if it was created.
		if (m_hAccessoryMenu != NULL)
		{
			::DestroyMenu(m_hAccessoryMenu);
			m_hAccessoryMenu = NULL;
		}

		// Clean up the systray window if it was created
		if (m_hSysTrayWindow != NULL)
		{
			::DestroyWindow(m_hSysTrayWindow);
			m_hSysTrayWindow = NULL;
		}

		// Unload any libraries
		if (m_hComCtl32 != NULL)
		{
			::FreeLibrary(m_hComCtl32);
			m_hComCtl32 = NULL;
		}

		if (m_hOLE32 != NULL)
		{
			::FreeLibrary(m_hOLE32);
			m_hOLE32 = NULL;
		}

		// General clean up
		delete m_pMainMenu;
	}

	/**************************************************************************
	
	>	HWND Win32_ApplicationImp::GetSysTrayWindow()
	
		Author:		Tim Browse
		Date:		17/6/2002
		Returns:	The systray routing window
	
		Purpose:	Get the Window used to route system tray notification
					messages.  This is an invisible window that is always
					available when we need it.
	
	**************************************************************************/
	
	HWND Win32_ApplicationImp::GetSysTrayWindow()
	{
		// Create sys tray window on demand
		if (m_hSysTrayWindow == NULL)
		{
			static ATOM SysTrayClassAtom = RegisterSysTrayWindowClass();

			m_hSysTrayWindow = ::CreateWindow(GetWindowClassName(wc_SysTrayWindow),
											  "Invisible SysTray Window", 
											  WS_OVERLAPPEDWINDOW,
											  CW_USEDEFAULT,
											  CW_USEDEFAULT,
											  CW_USEDEFAULT,
											  CW_USEDEFAULT,
											  NULL,
											  NULL,
											  m_hInstance,
											  NULL);
			
			if (m_hSysTrayWindow != NULL)
			{
				::ShowWindow(m_hSysTrayWindow, SW_HIDE);
				::UpdateWindow(m_hSysTrayWindow);
			}
		}

		return m_hSysTrayWindow;
	}

	/**************************************************************************
	
	>	void Win32_ApplicationImp::SetModelessDialog(HWND hDlg)
	
		Author:		Tim Browse
		Date:		15/6/2002
		Inputs:		hDlg - the modeless dialog that has just been activated; or
						   NULL if a modeless dialog just lost focus.
	
		Purpose:	Keep the app object up to date with focus changes on 
					modeless dialogs.  The app needs to know if a modeless
					dialog currently has focus, so that messages can be
					processed properly.

					Currently, this function is called by Win32_DialogImp
					when WM_ACTIVATE messages are received for modeless
					dialogs.
	
	**************************************************************************/
	
	void Win32_ApplicationImp::SetModelessDialog(HWND hDlg)
	{
		m_CurrentModelessDlg = hDlg;
	}

	/**************************************************************************
	
	>	char *Win32_ApplicationImp::GetWindowClassName(WindowClassType Type)
	
		Author:		Tim Browse
		Date:		15/6/2002
		Inputs:		Type - wc_MainWindow => get the main window class
						   wc_ChildWindow => get window class for child windows.
		Returns:	C string for the window class requested.
					(or NULL if unknown window class requested)
	
		Purpose:	Get the textual name of the Win32 window class we have
					registered for the requested window type.

					This is used for the boilerplate Win32 window class
					code - nothing fancy going on here.
	
	**************************************************************************/
	
	char *Win32_ApplicationImp::GetWindowClassName(WindowClassType Type)
	{
		// Register main window class the first time we're called
		static ATOM MainClassAtom = RegisterMainWindowClass();
		static ATOM ChildClassAtom = RegisterChildWindowClass();
		static ATOM SysTrayClassAtom = RegisterSysTrayWindowClass();
			
		switch (Type)
		{
			// Return name of registered window class.
			case wc_MainWindow:
			{
				return m_szWindowClass; 
			}

			case wc_ChildWindow:
			{
				return m_szChildWindowClass; 
			}
			
			case wc_SysTrayWindow:
			{
				return m_szSysTrayWindowClass;
			}
		}

		// Illegal window class
		COMPLAIN("Unknown window class requested.");
		return NULL;
	}


	/**************************************************************************

	>	bool Win32_ApplicationImp::Init(HINSTANCE hInstance, 
										LPCTSTR pszCmdLine, 
										int nCmdShow)


		Author:		Tim Browse
		Date:		15/6/2002
		Inputs:		hInstance - handle to instance of Win32 app
					pszCmdLine - command line string used to start the app
					nCmdShow - controls how to show the app (see WinMain in the
							   Win32 SDK docs)
		Outputs:	
		Returns:	true => app initialised ok
					false => unable to initialise the app

		Purpose:	Initialise the application object, including creating
					the main window object, initialising system libraries,
					and so on.

	**************************************************************************/

	bool Win32_ApplicationImp::Init(HINSTANCE hInstance, 
									LPCTSTR pszCmdLine, 
									int nCmdShow)
	{
		// Remember application instance
		m_hInstance = hInstance;

		// Initialise common controls
		if (m_pApp->GetOptions().m_Win32_bInitCommonControls)
		{
			InitControls();
		}

		// Initialise the COM library
		if (m_pApp->GetOptions().m_Win32_bUseCOM)
		{
			InitCOM();
		}

		// Initialise theme support
		Win32_Theme::Initialise();

		// Create the window
		if (!GetApp()->CreateMainWindow())
		{
			// initialisation failed
			return false;
		}

		// Set up the Application's main Menu object
		HWND hMainWindow = GetMainHWND();
		if (hMainWindow != NULL)
		{
			m_pMainMenu = new Menu(::GetMenu(hMainWindow));
		}
		else
		{
			// No main menu
			m_pMainMenu = NULL;
		}

		return true;
	}


	/**************************************************************************
	
	>	int Win32_ApplicationImp::Run()
	
		Author:		Tim Browse
		Date:		15/6/2002
		Returns:	Exit code passed by the PostQuitMessage() function.
	
		Purpose:	Run the app, until the user (or the app) decides to quit.

					This involves basically running the event loop for the app,
					and doing any last minute preparations on the window.

					In this case, it means showing the main window if it's a
					dialog, and loading the accelerator table.
	
	**************************************************************************/
	
	int Win32_ApplicationImp::Run()
	{
		// Two ways of running an app - normal window or a dialog-based app
		Window* pWindow = GetApp()->GetMainWindow();

		// Hello!  if you're staring at this next line of code because an 
		// exception is being thrown when it is executed, then you probably 
		// need to enable RTTI on your C++ project :-)
		Dialog* pDialog = dynamic_cast<Dialog*>(pWindow);

		if (pDialog != NULL)
		{
			// Install dialog as the main window.
			m_hMainWindow = pDialog->GetWindowHandle();
		}

		m_hAccelTable = ::LoadAccelerators(m_hInstance, GetApp()->GetAcceleratorsID().GetID());

		return DoMessageLoop();
	}

	/**************************************************************************
	
	>	void Win32_ApplicationImp::NotifyUserInputMessageReceived(DWORD dwTimeOfMessage = 0)
	
		Author:		Tim Browse
		Date:		25/7/2002
		Inputs:		dwTimeOfMessage - the timestamp of the input message - if 0,
									  then the current time is used.
	
		Purpose:	Let the application know when an input message was last
					received.  This is chiefly used to control idle processing,
					so that whenever the user is trying to do something (type, 
					etc) then we don't do idle processing at the same time and
					slow them down.
	
	**************************************************************************/
	
	void Win32_ApplicationImp::NotifyUserInputMessageReceived(DWORD dwTimeOfMessage)
	{
		if (dwTimeOfMessage == 0)
		{
			dwTimeOfMessage = ::GetTickCount();
		}

		if (dwTimeOfMessage > m_dwTimeOfLastInputMessage)
		{
			m_dwTimeOfLastInputMessage = dwTimeOfMessage;
		}
	}

	/**************************************************************************
	
	>	WPARAM Win32_ApplicationImp::DoMessageLoop()
	
		Author:		Tim Browse
		Date:		15/6/2002
		Returns:	The exit code of the app, as passed by PostQuitMessage(),
					and received via the WM_QUIT message.
	
		Purpose:	Start the app's message loop, and keep processing
					messages until the app quits.  This includes the servicing
					of any registered idle handlers.
	
	**************************************************************************/
	
	WPARAM Win32_ApplicationImp::DoMessageLoop()
	{
		while (!m_bQuitReceived)
		{
			// Do normal messages first
			ServiceMessageQueue();

			// How long have we been waiting for input messages?
			if (::GetTickCount() > (m_dwTimeOfLastInputMessage + kmsIdleDelay))
			{
				// And now do idle processing
				if (!m_bQuitReceived && !ServiceIdleHandlers())
				{
					// Idle handlers have finished, so we wait until
					// user wakes us up.
					::WaitMessage();
				}
			}
			else
			{
				// Go to sleep for a bit to avoid sucking up cpu time while
				// we wait for idle timeout to occur.
				::Sleep(50);
			}
		}

		// m_MsgWParam holds the last value of WPARAM field of the WM_QUIT
		// message we receive (which is the exit code of the app).
		return m_MsgWParam;
	}

	/**************************************************************************
	
	>	void Win32_ApplicationImp::ServiceMessageQueue()
	
		Author:		Tim Browse
		Date:		15/6/2002
	
		Purpose:	Process any outstanding messages in the app's message 
					queue.  This is the heart of the Win32 message processing
					loop for vGUI apps.

					The function exits when no more messages are pending.
	
	**************************************************************************/
	
	void Win32_ApplicationImp::ServiceMessageQueue()
	{
		// Service outstanding messages
		MSG msg;

		while (::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) 
		{
			// Check for dialog messages first...
			if ((m_CurrentModelessDlg == NULL) ||
				!::IsDialogMessage(m_CurrentModelessDlg, &msg))
			{
				// ...then accelerators...
				if ((m_hAccelTable != NULL) &&
				    ::TranslateAccelerator(m_hMainWindow, m_hAccelTable, &msg))
				{
					// It's an accelerator (input) message
					NotifyUserInputMessageReceived(msg.time);
				}
				else
				{
					// ...otherwise, it's a normal message
					::TranslateMessage(&msg);
					::DispatchMessage(&msg);

					// Was it an input message?
					if (((msg.message >= WM_MOUSEFIRST) && (msg.message <= WM_MOUSELAST)) ||
						((msg.message >= WM_KEYFIRST) && (msg.message <= WM_KEYLAST)))
					{
						// We don't count mouse movement as worthy of holding up
						// idle processing
						if (msg.message != WM_MOUSEMOVE)
						{
							NotifyUserInputMessageReceived(msg.time);
						}
					}
				}

				// Watch out for WM_QUIT messages sneaking in
				if (msg.message == WM_QUIT)
				{
					// Tell the main loop that a quit has been received.
					m_bQuitReceived = true;
				}
			}

			m_MsgWParam = msg.wParam;
		}
	}

	/**************************************************************************
	
	>	bool Win32_ApplicationImp::ServiceIdleHandlers()
	
		Author:		Tim Browse
		Date:		1/5/2002
		Returns:	true => idle handlers were processed, and want more idle
							time.
					false => no idle handlers installed, or they have all 
							 finished their idle processing.
	
		Purpose:	Service any idle handlers waiting for idle notifications.
	
	**************************************************************************/
	
	bool Win32_ApplicationImp::ServiceIdleHandlers()
	{
		// Defer to main app object
		return GetApp()->ServiceIdleHandlers();
	}

	/**************************************************************************
	
	>	void Win32_ApplicationImp::OnDestroyMainWindow()
	
		Author:		Tim Browse
		Date:		17/6/2002

		Purpose:	Inform the application object that the main window is
					about to be destroyed, so anything that needs the
					main window object in order to be cleaned up/shutdown
					had better do it now, before the window dies.
	
	**************************************************************************/
	
	void Win32_ApplicationImp::OnDestroyMainWindow()
	{
		// Delete system tray icon
		if (m_bSysTrayIcon)
		{
			DeleteTrayIcon();
			m_bSysTrayIcon = false;
		}

		// Main window destroyed => app is quitting
		::PostQuitMessage(0);
	}

	/**************************************************************************
	
	>	bool Win32_ApplicationImp::Exit()
	
		Author:		Tim Browse
		Date:		16/6/2002
		Returns:	true => app shutdown ok (always)
	
		Purpose:	Clean up and shut down the application, ready for exit.
	
	**************************************************************************/
	
	bool Win32_ApplicationImp::Exit()
	{
		return true;
	}

	/**************************************************************************

	>	bool Win32_ApplicationImp::InstallAccessory(const IconResourceID& IconID,
													const MenuResourceID& MenuID,
													const StringResourceID& ToolTipID,
													int DefaultCommand)


		Author:		Tim Browse
		Date:		16/6/2002
		Inputs:		IconID - the icon to use for the accessory application in
							 the system tray.
					MenuID - ID of the menu to attach to the icon - Win32 programs
							 don't use this here, but use the GetAccessoryMenu()
							 function on demand.
					ToolTipID - the text to use for tooltip text for the icon
					DefaultCommand - the command message to send when the user
									 issues the default command on the icon
									 (e.g. under Win32, left-clicking on the icon)
		Returns:	true - accessory installed ok
					false - failed to install (e.g. icon failed to load, etc)

		Purpose:	Set up an accessory application - for Win32, this means
					adding an icon to the system tray.

	**************************************************************************/

	bool Win32_ApplicationImp::InstallAccessory(const IconResourceID& IconID,
											    const MenuResourceID& MenuID,
											    const StringResourceID& ToolTipID,
												int DefaultCommand)
	{
		// We ignore the menu ID under Win32, as we don't need it now
		// (we use it later, on demand)

		// Remember default command 
		m_cmdDefault = DefaultCommand;

		// Load tooltip text
		ToolTipID.Load(&m_ToolTip);

		// Load the icon
		if (!IconID.Load(&m_AppIcon))
		{
			// Failed to load icon
			vguiERROR();
			return false;
		}

		// Install the icon
		bool bSuccess = SetTrayIcon(m_AppIcon, m_ToolTip);

		// Remember whether this worked, so we can remove it on shutdown
		m_bSysTrayIcon = bSuccess;

		return bSuccess;
	}

	/**************************************************************************

	>	bool Win32_ApplicationImp::SetTrayIcon(const Icon& TrayIcon,
											   const String& ToolTip)


		Author:		Tim Browse
		Date:		16/6/2002
		Inputs:		TrayIcon - the icon to use in the system tray.
					ToolTip - the string to use to display as a tooltip
							  for the tray icon.
		Returns:	true => icon updated ok
					false => unable to update icon

		Purpose:	Add/change the tray icon used for this application.

	**************************************************************************/

	bool Win32_ApplicationImp::SetTrayIcon(const Icon& TrayIcon,
										   const String& ToolTip)
	{
		// Initialise system tray notification structure
		NOTIFYICONDATA nid;
		ZeroMemory(&nid, sizeof(nid));
		nid.cbSize = sizeof(nid);

		// We want to install an icon + callback;
		nid.uFlags = NIF_MESSAGE | NIF_ICON;

		// Install tooltip if it's valid
		if (!ToolTip.IsEmpty())
		{
			// Copy first 64 chars (it's all the system tray allows for tooltips)
			strncpy(nid.szTip, ToolTip.c_str(), 63);
			nid.szTip[63] = 0;

			// Tell system we're installing tooltip text
			nid.uFlags |= NIF_TIP;
		}

		nid.hIcon = TrayIcon.GetIconHandle();

		// Pass messages to systray routing window
		nid.hWnd = GetSysTrayWindow();

		// We must have a real window handle.
		ASSERT(nid.hWnd != NULL);

		// Set up unique IDs
		nid.uID = kSysTray_ID;
		nid.uCallbackMessage = WM_SYSTRAYCALLBACK;

		// Install/modify the icon
		DWORD dwMessage = NIM_ADD;

		if (m_bSysTrayIcon)
		{
			// Icon already installed; modify it.
			dwMessage = NIM_MODIFY;
		}

		// Add new icon to tray
		bool bSuccess = (Shell_NotifyIcon(dwMessage, &nid) != 0);

		// Remember whether this worked, so we can remove it on shutdown, etc.
		m_bSysTrayIcon = bSuccess;

		return bSuccess;
	}

	/**************************************************************************
	
	>	void Win32_ApplicationImp::DeleteTrayIcon()
	
		Author:		Tim Browse
		Date:		17/6/2002
	
		Purpose:	Remove the tray icon for this application.
	
	**************************************************************************/
	
	void Win32_ApplicationImp::DeleteTrayIcon()
	{
		// Initialise system tray notification structure
		NOTIFYICONDATA nid;
		ZeroMemory(&nid, sizeof(nid));
		nid.cbSize = sizeof(nid);

		nid.hIcon = m_AppIcon.GetIconHandle();
		nid.hWnd = GetSysTrayWindow();

		// We must have a real window handle.
		ASSERT(nid.hWnd != NULL);

		// Set up unique IDs
		nid.uID = kSysTray_ID;
		nid.uCallbackMessage = WM_SYSTRAYCALLBACK;

		// Delete the icon
		Shell_NotifyIcon(NIM_DELETE, &nid);
	}

	/**************************************************************************

	>	LRESULT CALLBACK Win32_ApplicationImp::SysTrayWndProc(HWND hWnd, 
															  UINT message, 
															  WPARAM wParam, 
															  LPARAM lParam)


		Author:		Tim Browse
		Date:		17/6/2002

		Purpose:	Handle notification messages from the system tray icon.

	**************************************************************************/

	LRESULT CALLBACK Win32_ApplicationImp::SysTrayWndProc(HWND hWnd, 
														  UINT message, 
													      WPARAM wParam, 
													      LPARAM lParam)
	{
		switch (message) 
		{
			case WM_SYSTRAYCALLBACK:
			{
				// Callback for system tray icon
				if (GetWin32App()->TrayCallback((UINT) wParam, (UINT) lParam))
				{
					// Message handled
					return 0;
				}

				// Note that if the app wasn't expecting a tray callback (because
				// it's not an accessory app) then we don't respond to the message,
				// and leave it for further processing.  This is because our
				// WM_SYSTRAYCALLBACK message is user-defined, so we reduce
				// interference with other user-defined messages as much as 
				// possible.
				break;
			}
		}

		// Fall back to default behaviour
		return ::DefWindowProc(hWnd, message, wParam, lParam);
	}

	/**************************************************************************

	>	LRESULT CALLBACK Win32_ApplicationImp::vGuiWndProc(HWND hWnd, 
														   UINT message, 
														   WPARAM wParam, 
														   LPARAM lParam)


		Author:		Tim Browse
		Date:		17/6/2002
		Inputs:		hWnd - handle to window
					message - message identifier
					wParam - first message parameter
					lParam - second message parameter

		Returns:	The return value is the result of the message processing 
					and depends on the message sent.

		Purpose:	The WNDPROC for vGUI application windows.  Messages
					are routed after arriving at this function.

	**************************************************************************/

	LRESULT CALLBACK Win32_ApplicationImp::vGuiWndProc(HWND hWnd, 
													   UINT message, 
													   WPARAM wParam, 
													   LPARAM lParam)
	{
		// Is this message handled?
		LRESULT lResult = vGuiRawWndProc(hWnd, message, wParam, lParam);

		if (lResult != kWndProc_NotHandled)
		{
			// Yes, so we're done
			return lResult;
		}
		else
		{
			// Fall back to default behaviour
			return ::DefWindowProc(hWnd, message, wParam, lParam);
		}
	}

		

	/**************************************************************************

	>	LRESULT CALLBACK Win32_ApplicationImp::vGuiRawWndProc(HWND hWnd, 
															  UINT message, 
															  WPARAM wParam, 
															  LPARAM lParam)

		Author:		Tim Browse
		Date:		24/02/2001
		Inputs:		Usual WndProc parameters
		Returns:	0 => message handled
					kWndProc_NotHandled => message not handled - caller must
										   supply default handling.

		Purpose:	The basic WndProc functionality for vGUI.

					This is just the vGUI-specific WndProc functionality - if
					a message is not understood, then the function just does
					nothing and returns kWndProc_NotHandled.

					This allows the function to be used by different types
					of windows (normal windows, frame windows, MDI windows)
					even though they may have different default processing.
					(This is because this function does *no* default processing;
					the caller must take care of that - see vGuiWndProc() for
					an example).


	**************************************************************************/

	LRESULT CALLBACK Win32_ApplicationImp::vGuiRawWndProc(HWND hWnd, 
														  UINT message, 
													      WPARAM wParam, 
													      LPARAM lParam)
	{
		switch (message) 
		{
			case WM_COMMAND:
			case WM_SYSCOMMAND:
			{
				//
				// Command message - use message routing system.
				//

				// Extract command ID
				int nCommand = LOWORD(wParam);

				// Route the message
				if (GetApp()->OnCommand(nCommand))
				{
					// Message handled ok.
					return 0;
				}

				break;
			}
		}

		// Extract vGui window object from this window
		Window *pWindow;
		if (Win32_WindowInfo::GetAttachedWindow(hWnd, &pWindow))
		{
			// Window has been properly initialised...
			Win32_WindowImp *pWindowImp = Win32_WindowImp::GetImp(pWindow);

			return pWindowImp->WindowProc(message, wParam, lParam);
		}

		// Message not processed
		return kWndProc_NotHandled;
	}

	//
	//  PURPOSE: Registers the window class.
	//
	//  COMMENTS:
	//
	//    This function and its usage is only necessary if you want this code
	//    to be compatible with Win32 systems prior to the 'RegisterClassEx'
	//    function that was added to Windows 95. It is important to call this function
	//    so that the application will get 'well formed' small icons associated
	//    with it.
	//


	/**************************************************************************
	
	>	ATOM Win32_ApplicationImp::RegisterMainWindowClass()
	
		Author:		Tim Browse
		Date:		17/6/2002
		Returns:	The atom for the registered window class
	
		Purpose:	Registers the window class for the main window.

	**************************************************************************/
	
	ATOM Win32_ApplicationImp::RegisterMainWindowClass()
	{
		// Get application object
		Application *pApp = GetApp();

		// Make a unique name for the class
		wsprintf(m_szWindowClass, "vguiWindow_%x", (int) ::GetModuleHandle(NULL));

		WNDCLASSEX wcex;

		wcex.cbSize = sizeof(WNDCLASSEX); 

		wcex.style			= CS_HREDRAW | CS_VREDRAW;
		wcex.lpfnWndProc	= (WNDPROC) vGuiWndProc;
		wcex.cbClsExtra		= 0;
		wcex.cbWndExtra		= 0;
		wcex.hInstance		= m_hInstance;
		wcex.hIcon			= LoadIcon(wcex.hInstance, pApp->GetIconID().GetID());
		wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
		wcex.lpszMenuName	= pApp->GetMenuID().GetID();
		wcex.lpszClassName	= m_szWindowClass;
		wcex.hIconSm		= LoadIcon(wcex.hInstance, pApp->GetSmallIconID().GetID());

		// Find out what behaviour the client wants from the window background re: repainting
		if (m_pApp->GetOptions().m_Win32_bRedrawAppWindowBackground)
		{
			// Default to redraw window with standard window background colour
			wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
		}
		else
		{
			// Don't redraw window background automatically
			wcex.hbrBackground	= (HBRUSH) ::GetStockObject(NULL_BRUSH);
		}

		return RegisterClassEx(&wcex);
	}


	/**************************************************************************
	
	>	ATOM Win32_ApplicationImp::RegisterChildWindowClass()
	
		Author:		Tim Browse
		Date:		17/6/2002
		Returns:	The atom for the registered window class
	
		Purpose:	Register the window class for child windows.
	
	**************************************************************************/
	
	ATOM Win32_ApplicationImp::RegisterChildWindowClass()
	{
		// Get application object
		Application *pApp = GetApp();

		// Make a unique name for the class
		wsprintf(m_szChildWindowClass, "vguiChildWindow_%x", (int) ::GetModuleHandle(NULL));

		WNDCLASSEX wcex;

		wcex.cbSize = sizeof(WNDCLASSEX); 

		wcex.style			= CS_HREDRAW | CS_VREDRAW;
		wcex.lpfnWndProc	= (WNDPROC) vGuiWndProc;
		wcex.cbClsExtra		= 0;
		wcex.cbWndExtra		= 0;
		wcex.hInstance		= m_hInstance;
		wcex.hIcon			= NULL;
		wcex.hCursor		= NULL;
		wcex.hbrBackground	= NULL;
		wcex.lpszMenuName	= NULL;
		wcex.lpszClassName	= m_szChildWindowClass;
		wcex.hIconSm		= NULL;

		return RegisterClassEx(&wcex);
	}


	/**************************************************************************
	
	>	ATOM Win32_ApplicationImp::RegisterSysTrayWindowClass()
	
		Author:		Tim Browse
		Date:		17/6/2002
		Returns:	The registered window class.
	
		Purpose:	Register the window class used for the hidden window that
					we use to route system tray notification messages.
	
	**************************************************************************/
	
	ATOM Win32_ApplicationImp::RegisterSysTrayWindowClass()
	{
		// Make a unique name for the class
		wsprintf(m_szSysTrayWindowClass, "vguiSysTrayWindow_%x", (int) ::GetModuleHandle(NULL));

		WNDCLASSEX wcex;

		wcex.cbSize = sizeof(WNDCLASSEX); 

		wcex.style			= CS_HREDRAW | CS_VREDRAW;
		wcex.lpfnWndProc	= (WNDPROC) SysTrayWndProc;
		wcex.cbClsExtra		= 0;
		wcex.cbWndExtra		= 0;
		wcex.hInstance		= m_hInstance;
		wcex.hIcon			= NULL;
		wcex.hCursor		= NULL;
		wcex.hbrBackground	= NULL;
		wcex.lpszMenuName	= NULL;
		wcex.lpszClassName	= m_szSysTrayWindowClass;
		wcex.hIconSm		= NULL;

		return RegisterClassEx(&wcex);
	}

	/**************************************************************************
	
	>	bool Win32_ApplicationImp::TrayCallback(UINT nID, UINT uMouseMsg)
	
		Author:		Tim Browse
		Date:		17/6/2002
		Inputs:		nID - the user-defined ID for the tray icon
					nMouseMsg - the mouse message passed to the icon
		Returns:	true => message handled ok
					false => message not expected (or handled either)
	
		Purpose:	Handle any callback messages for the system tray icon.

					If an unexpected message is passed in, this function
					will COMPLAIN.

	**************************************************************************/
	
	bool Win32_ApplicationImp::TrayCallback(UINT nID, UINT uMouseMsg)
	{
		// If we have no tray icon installed, complain and return that
		// we don't want this message
		if (!m_bSysTrayIcon)
		{
			// We may get this happening if another WM_USER based message is
			// defined as the same as our WM_SYSTRAYCALLBACK constant.
			COMPLAIN("Received a message for the system tray icon when there "
					 "is no tray icon installed.  Check WM_SYSTRAYCALLBACK for "
					 "possible interference with other WM_USER based messages.");
			return false;
		}

		// Make sure we have the right message
		if (nID != kSysTray_ID)
		{
			// Unknown sys tray callback message - complain.
			vguiERROR();
			return false;
		}

		switch (uMouseMsg)
		{
			case WM_LBUTTONDOWN:
			{
				// If we have a notify active, send the notify command, otherwise
				// try the default command for an accessory.
				if (m_bNotifyActive && (m_cmdNotify != NullCommand))
				{
					GetApp()->OnCommand(m_cmdNotify);
				}
				else
				if (m_cmdDefault != 0)
				{
					// Send the default command to the application
					GetApp()->OnCommand(m_cmdDefault);
				}
				break;
			}

			case WM_RBUTTONUP:
			{
				// Open the accessory menu
				HMENU hMenu = GetAccessoryMenu();

				if (hMenu != NULL)
				{
					HWND hParent = GetSysTrayWindow();

					if (hParent != NULL)
					{
						// Make the parent window the foreground window.
						// See KB Article ID: Q135788
						::SetForegroundWindow(hParent);

						// Show the menu at the cursor position
						POINT point;
						::GetCursorPos(&point);

						int nCommand = Win32_WindowImp::TrackPopupMenuEx(hMenu,
																		 TPM_RIGHTALIGN|
																		 TPM_BOTTOMALIGN|
																		 TPM_RETURNCMD|
																		 TPM_RIGHTBUTTON,
																		 point.x, 
																		 point.y,
																		 hParent,
																		 NULL);

						// If user selected a command, then do it
						if (nCommand != 0)
						{
							GetApp()->OnCommand(nCommand);
						}
					}
				}
				break;
			}
		}

		// Message handled ok
		return true;
	}

	/**************************************************************************
	
	>	HMENU Win32_ApplicationImp::GetAccessoryMenu()
	
		Author:		Tim Browse
		Date:		17/6/2002
		Returns:	The handle to the accessory menu (or NULL if there is a
					problem loading the menu).
	
		Purpose:	Access the menu handle for the accessory menu.

					(The menu will be created if necessary - i.e. if this
					is the first time it has been accessed)
	
	**************************************************************************/
	
	HMENU Win32_ApplicationImp::GetAccessoryMenu()
	{
		// If we have already created the menu, just return it to the caller.
		if (m_hAccessoryMenu != NULL)
		{
			// Ok, return menu to the caller - we have to take the first submenu
			// as TrackPopupMenuEx() can't handle menu bars, only popups,
			// and we can't create a popup directly from a resource without
			// a load of grief.
			return ::GetSubMenu(m_hAccessoryMenu, 0);
		}

		// Get the application object
		Application *pApp = GetApp();

		if ((pApp != NULL) && (pApp->IsAccessory()))
		{
			// Create menu
			m_hAccessoryMenu = ::LoadMenu(GetWin32App()->GetInstance(),
										  pApp->GetAccessoryMenuID().GetID());

			// If that worked, make the default command display in bold
			if (m_hAccessoryMenu != NULL)
			{
				HMENU hPopupMenu = ::GetSubMenu(m_hAccessoryMenu, 0);

				if (m_cmdDefault != NullCommand)
				{
					::SetMenuDefaultItem(hPopupMenu, m_cmdDefault, FALSE);
				}

				return hPopupMenu;
			}
		}
		else
		{
			// Logic error - calling accessory function on an application that
			// is not an accessory application.
			vguiERROR();
		}

		// Failed to load menu
		return NULL;
	}

	/**************************************************************************
	
	>	void Win32_ApplicationImp::InitControls()
	
		Author:		Tim Browse
		Date:		17/6/2002
	
		Purpose:	Load the Win32 common control library.  
		
					We don't link directly to the common control, to allow apps 
					to be as robust as possible (i.e. they don't fail if the 
					lib is not available, as they might not use it).
	
	**************************************************************************/
	
	void Win32_ApplicationImp::InitControls()
	{
		// Try to load common control library - we don't bind explicitly so we make
		// vGUI apps as robust as possible.
		m_hComCtl32 = ::LoadLibrary("comctl32.dll");

		if (m_hComCtl32 != NULL)
		{
			// Find init function
			InitCommonControlsExFn* pInitFn;
			pInitFn = (InitCommonControlsExFn*) ::GetProcAddress(m_hComCtl32, "InitCommonControlsEx");

			if (pInitFn != NULL)
			{
				// Found function - initialise control library
				INITCOMMONCONTROLSEX ctls;
				ZeroMemory(&ctls, sizeof(ctls));
				ctls.dwSize = sizeof(ctls);
				ctls.dwICC = ICC_WIN95_CLASSES | ICC_DATE_CLASSES;

				// Call init function
				(*pInitFn)(&ctls);
			}
			else
			{
				// Init function not found...discard library
				::FreeLibrary(m_hComCtl32);
				m_hComCtl32 = NULL;

				// TODO: Report library failure to user....
			}
		}
		else
		{
			// TODO: Report library failure to user....
		}
	}

	/**************************************************************************
	
	>	void Win32_ApplicationImp::InitCOM()
	
		Author:		Tim Browse
		Date:		17/6/2002
	
		Purpose:	Initialise/load the COM library for this application.
	
	**************************************************************************/
	
	void Win32_ApplicationImp::InitCOM()
	{
		// Try to load COM/OLE library - we don't bind explicitly so we make
		// vGUI apps as robust as possible.
		m_hOLE32 = ::LoadLibrary("ole32.dll");

		if (m_hOLE32 != NULL)
		{
			// Find init function
			CoInitializeFn* pInitFn;
			pInitFn = (CoInitializeFn*) ::GetProcAddress(m_hOLE32, "CoInitialize");

			if (pInitFn != NULL)
			{
				// Found function - initialise COM library
				(*pInitFn)(NULL);
			}
			else
			{
				// Init function not found...discard library
				::FreeLibrary(m_hOLE32);
				m_hOLE32 = NULL;

				// TODO: Report library failure to user....
			}
		}
		else
		{
			// TODO: Report library failure to user....
		}
	}

	/**************************************************************************
	
	>	Menu* Win32_ApplicationImp::GetMainMenu()
	
		Author:		Tim Browse
		Date:		17/6/2002
		Returns:	The main menu (or NULL if not available)
	
		Purpose:	Access the main menu of the application.
	
	**************************************************************************/
	
	Menu* Win32_ApplicationImp::GetMainMenu()
	{
		return m_pMainMenu;
	}

	/**************************************************************************

	>	void Win32_ApplicationImp::SetNotifyIcon(const Icon& NotifyIcon, 
												 const String& ToolTip,
												 CommandID NotifyCommand)


		Author:		Tim Browse
		Date:		17/6/2002
		Inputs:		NotifyIcon - the notification icon to use.
					ToolTip - the text to show as tooltip for the icon
					NotifyCommand - the command to send to the app when
									the user 'activates' the notify icon.
									(under Win32, this means left-clicking 
									the icon)

		Purpose:	Display a notification icon for the application.  This
					is a way of getting the user's attention.

					For Win32 apps, this is displayed in the system tray
					(notification area).

					See also SetFlashingNotifyIcon().

	**************************************************************************/

	void Win32_ApplicationImp::SetNotifyIcon(const Icon& NotifyIcon, 
											 const String& ToolTip,
											 CommandID NotifyCommand)
	{
		// Remember notify command 
		m_cmdNotify = NotifyCommand;

		// Install the icon
		m_bNotifyActive = SetTrayIcon(NotifyIcon, ToolTip);
	}

	/**************************************************************************

	>	void Win32_ApplicationImp::SetFlashingNotifyIcon(const Icon& NotifyIcon, 
														 const String& ToolTip,
														 CommandID NotifyCommand)


		Author:		Tim Browse
		Date:		17/6/2002
		Inputs:		NotifyIcon - the notification icon to use.
					ToolTip - the text to show as tooltip for the icon
					NotifyCommand - the command to send to the app when
									the user 'activates' the notify icon.
									(under Win32, this means left-clicking 
									the icon)

		Purpose:	Display a flashing notification icon for the application.  This
					is a way of getting the user's attention.

					For Win32 apps, this is displayed in the system tray
					(notification area).

					See also SetNotifyIcon().

	**************************************************************************/

	void Win32_ApplicationImp::SetFlashingNotifyIcon(const Icon& NotifyIcon, 
													 const String& ToolTip,
													 CommandID NotifyCommand)
	{
		// Remember notify command 
		m_cmdNotify = NotifyCommand;

		// Install the icon
		m_bNotifyActive = SetTrayIcon(NotifyIcon, ToolTip);
	}

	/**************************************************************************
	
	>	void Win32_ApplicationImp::ClearNotifyIcon()
	
		Author:		Tim Browse
		Date:		17/6/2002
	
		Purpose:	Remove a notify icon previously installed by SetNotifyIcon()
					or SetFlashingNotifyIcon().
	
	**************************************************************************/
	
	void Win32_ApplicationImp::ClearNotifyIcon()
	{
		// If notification is not active, ignore this, but complain in
		// debug builds.
		if (!m_bNotifyActive)
		{
			vguiERROR();
			return;
		}

		// Do we need to revert to the accessory icon, or just remove the notify icon?
		if (m_bSysTrayIcon)
		{
			SetTrayIcon(m_AppIcon, m_ToolTip);
		}
		else
		{
			DeleteTrayIcon();
		}

		// Notification has gone
		m_bNotifyActive = false;
	}

	/**************************************************************************

	>	HWND Win32_ApplicationImp::CreateWin32Window(Window *pWindow,
													 const Window* pParent,
													 TWindowStyle Style, 
													 const String* psWindowTitle)


		Author:		Tim Browse
		Date:		17/6/2002

		Inputs:		pWindow - the vGUI Window object to link to the new
							  Win32 window.

					pParent - the parent of the new window (may be NULL)

					Style - controls the type of window to create:

							ws_MainAppWindow
								- create a window of a type suitable for a
								  main window (and m_hMainWindow is updated
								  accordingly).

							ws_GenericChild
							    - create a window suitable for use as a 
								  generic child window.

					psWindowTitle - the text to use as the window title
									(may be NULL if no title needed)
		Returns:	Handle of the new window (or NULL if creation failed)

		Purpose:	Convenience function for creating Win32 windows of
					a certain type.

					This function will be over-ridden by other types of
					apps - e.g. MDI apps - to provide other types of
					windows, or slightly different behaviour.

	**************************************************************************/

	HWND Win32_ApplicationImp::CreateWin32Window(Window *pWindow,
												 const Window* pParent,
												 TWindowStyle Style, 
												 const String* psWindowTitle)
	{
		// Work out what Window class/style to use.
		LPCTSTR szWindowClass = NULL;
		DWORD dwStyle = 0;

		// Default to NULL menu.
		HMENU hMenu = NULL;

		switch (Style)
		{
			case ws_MainAppWindow:
			{
				// Get window class etc.
				szWindowClass = GetWindowClassName(wc_MainWindow);
				dwStyle = WS_OVERLAPPEDWINDOW;

				// Leave hMenu as NULL to use the window class's menu.
				break;
			}

			case ws_GenericChild:
			{
				// Just a raw window - no frame, no title bar.
				szWindowClass = GetWindowClassName(wc_ChildWindow);
				dwStyle = WS_CHILDWINDOW;

				// Don't use 0 for menu - it will clash with IDCANCEL.
				hMenu = (HMENU) 1200;
				break;
			}

			default:
			{
				// Unknown window style
				COMPLAIN("Unknown window style in CreateWindow().");
				return false;
			}
		}

		HWND hWndParent = NULL;

		if (pParent != NULL)
		{
			// Extract parent's window handle.
			hWndParent = Win32_WindowImp::GetImp(pParent)->GetHWND();
		}

		LPCTSTR szWindowTitle = NULL;

		if (psWindowTitle != NULL)
		{
			szWindowTitle = psWindowTitle->c_str();
		}

		// Create the window
		HWND hWnd = ::CreateWindow(szWindowClass, 
								   szWindowTitle,
								   dwStyle, 
								   100, 100, 
								   300, 300,
								   hWndParent, 
								   hMenu, 
								   GetInstance(), 
								   NULL);

		if (hWnd != NULL)
		{
			// Add pointer to vGui object to the window.
			Win32_WindowInfo::Attach(hWnd, pWindow);

			// If we're making the main window, remember the handle to use for
			// translating accelerators.
			if (Style == ws_MainAppWindow)
			{
				m_hMainWindow = hWnd;
			}
		}

		return hWnd;
	}


	/**************************************************************************
	
	>	void Win32_ApplicationImp::GetCommandLine(String* pCommandLine)
	
		Author:		Tim Browse
		Date:		17/6/2002
		Outputs:	The command line string
	
		Purpose:	Get the command line string that was used to launch the
					application.
	
	**************************************************************************/
	
	void Win32_ApplicationImp::GetCommandLine(String* pCommandLine)
	{
		// Return command line string
		if (szCommandLine != NULL)
			*pCommandLine = szCommandLine;
	}

	/**************************************************************************

	>	int WINAPI Win32_ApplicationImp::WinMain(HINSTANCE hInstance, 
												 HINSTANCE hPrevInstance, 
												 LPSTR pszCmdLine, 
												 int nCmdShow)


		Author:		Tim Browse
		Date:		17/6/2002

		Purpose:	Start the application.  This is where the vGUI library
					is started up, and the application objects are created and
					executed.

	**************************************************************************/

	int WINAPI Win32_ApplicationImp::WinMain(HINSTANCE hInstance, 
											 HINSTANCE hPrevInstance, 
											 LPSTR pszCmdLine, 
											 int nCmdShow)
	{
		using vgui::Application;
		using vgui::ApplicationImp;
		using vgui::Win32_ApplicationImp;

		// Initialize the application (and vGui library).
		Application::CreateApplication();

		// Get the application object we just created.
		Application* pApp = Application::GetApp();

		// Cast it down to Win32 implementation object
		ApplicationImp *pImp = pApp->GetImp();
		Win32_ApplicationImp *pWin32App = dynamic_cast<Win32_ApplicationImp*>(pImp);

		// That damn well ought to work!
		if (pWin32App == NULL)
		{
			// But it didn't - catastrophic error.
			ASSERT(FALSE);
			exit(-1);
		}

		// Remember command line
		szCommandLine = pszCmdLine;

		// Initialise...
		if (pWin32App->Init(hInstance, pszCmdLine, nCmdShow))
		{
			// ...do app-specific start up...
			pApp->OnStartUp();

			// ...and run the program
			pApp->Run();

			// and finally delete the main Application object.
			delete pApp;
		}

		return 0;
	}

	/**************************************************************************
	
	>	Win32_ApplicationImp *GetWin32App()
	
		Author:		Tim Browse
		Date:		17/6/2002
		Returns:	The Win32_ApplicationImp object.
	
		Purpose:	Get the main Win32 specific application object.
	
	**************************************************************************/
	
	Win32_ApplicationImp *GetWin32App()
	{
		return Win32_ApplicationImp::GetWin32App();
	}
};

//
// Memory debugging facility for Win32/MSVC platform - if you have memory
// leaks that are hard to track down from the allocation point (e.g. they
// are String objects, so are allocated in String.cpp - not helpful) then
// check the allocation index reported in the debug output, and set
// CRTDBG_BREAKONALLOC to be this value.  e.g. set it to 37 to trap the 37th
// block allocated.
//

#ifdef _DEBUG

//#define CRTDBG_BREAKONALLOC 300

#ifdef _MSC_VER

#ifdef CRTDBG_BREAKONALLOC

static _CRT_ALLOC_HOOK s_SystemAllocHook = NULL;

int vGUIAllocHook(int allocType, void *userData, size_t size, 
				  int blockType, long requestNumber, 
				  const unsigned char *filename, int lineNumber)
{
	// If this is the request we're interested in, then break into
	// the debugger
	if (requestNumber == CRTDBG_BREAKONALLOC)
	{
		_asm int 3;
	}

	// Pass on to system hook
	return s_SystemAllocHook(allocType, userData, size, blockType,
							 requestNumber, filename, lineNumber);
}



#endif // CRTDBG_BREAKONALLOC

#endif // _MSC_VER

#endif // _DEBUG


/**************************************************************************

>	int WINAPI WinMain(HINSTANCE hInstance, 
					   HINSTANCE hPrevInstance, 
					   LPSTR pszCmdLine, 
					   int nCmdShow)


	Author:		Tim Browse
	Date:		17/6/2002

	Purpose:	Standard entry point for all MS Windows exectuables.

				This is just a vector to kick-start the vGUI library.

**************************************************************************/

int WINAPI WinMain(HINSTANCE hInstance, 
				   HINSTANCE hPrevInstance, 
				   LPSTR pszCmdLine, 
				   int nCmdShow)
{
#ifdef _DEBUG

	//
	// Report any memory leaks...
	//

#ifdef _MSC_VER
	// Get the current state of the flag, turn
	// on memory dumps at exit, and set the flag.
	int DbgFlag = _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG);
	DbgFlag |= _CRTDBG_LEAK_CHECK_DF;
	_CrtSetDbgFlag(DbgFlag);

	// Do we want to watch for specific allocations?
#ifdef CRTDBG_BREAKONALLOC
	// Yes, we do
	::s_SystemAllocHook = _CrtSetAllocHook(vGUIAllocHook);
#endif // CRTDBG_BREAKONALLOC

#endif // _MSC_VER

#endif // _DEBUG

	return vgui::Win32_ApplicationImp::WinMain(hInstance,
											   hPrevInstance,
											   pszCmdLine,
											   nCmdShow);
}


