//
// Win32_ApplicationImp.h - Application Implementation Interface 
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_Win32_ApplicationImp_h
#define vgui_Win32_ApplicationImp_h

#include "..\vgui.h"

#include "..\ApplicationImp.h"

#include "..\Icon.h"

#include "..\Window.h"

#include "String.h"



namespace vgui
{
	typedef enum
	{
		wc_MainWindow,
		wc_ChildWindow,
		wc_SysTrayWindow
	}
	WindowClassType;

	//
	// Value to return when a WNDPROC does not handle a message.
	//
	extern const int kWndProc_NotHandled;

	//
	// Constants used for System tray handling
	//
	extern const int kSysTray_ID;

	// WM_SYSTRAYCALLBACK is our own app-defined window message that we
	// ask the system tray manager to send to us for events that occur
	// on our icon.  It's essentially a 'made-up' message from the
	// WM_APP message address space.
	#define WM_SYSTRAYCALLBACK ((WM_APP) + 0x1001)

	/**************************************************************************
	
	>	class Win32_ApplicationImp : public ApplicationImp
	
		Author:		Tim Browse
		Date:		15/6/2002
	
		Purpose:	Class to implement a Win32 application.

					Technical Details
					=================

					Here is a description of the main points of interest
					of this class, which has to deal with a fair amount of
					weirdness...

					Message Loop
					~~~~~~~~~~~~
					
					blah blah blah...
					
					Modeless dialogs
					~~~~~~~~~~~~~~~~

					For various reasons, the app has to keep track of which
					(if any) modeless dialog currently has focus.  This is
					because we need to be able to call IsDialogMessage()
					when processing the message queue, otherwise lots of
					things don't work.  The only way to do this is to make
					sure we know which is the currently active modeless dialog
					at any time.

					Therefore the Win32_DialogImp class traps any WM_ACTIVATE
					messages, and calls our SetModelessDialog() function, to 
					keep us up to date.

					That's pretty much it - it's fairly simple, but a bit
					yucky.  See ServiceMessageQueue() to see where this
					actually has an effect.
					
					Accessory/Notification Icons/System Tray etc
					~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	
	**************************************************************************/
	
	class Win32_ApplicationImp : public ApplicationImp
	{
		public:
			Win32_ApplicationImp();
			~Win32_ApplicationImp();

		public:
			bool Init(HINSTANCE hInstance, LPCTSTR pszCmdLine, int nCmdShow);

			// ApplicationImp interface
			int Run();
			bool Exit();

			void ServiceMessageQueue();

			// Control of notification icons
			void SetNotifyIcon(const Icon& NotifyIcon, 
							   const String& ToolTip,
							   CommandID NotifyCommand);

			void SetFlashingNotifyIcon(const Icon& NotifyIcon, 
									   const String& ToolTip,
									   CommandID NotifyCommand);
			void ClearNotifyIcon();

			// Install an app as an 'accessory'
			bool InstallAccessory(const IconResourceID& IconID,
								  const MenuResourceID& MenuID,
								  const StringResourceID& ToolTipID,
								  int DefaultCommand);

			// Get the command line used to launch the app
			void GetCommandLine(String* pCommandLine);
			
			// Access to Win32 application singleton.
			static Win32_ApplicationImp *GetWin32App() { return s_pWin32App; }

			// Get the generic App handle for this app
			OSAppHandle GetOSAppHandle() { return m_hInstance; }

			// Get the main menu used for the main app window
			Menu* GetMainMenu();

			// 
			// Low-level Win32 handle/info access...
			//
			HINSTANCE GetInstance() const { return m_hInstance; }
			HWND GetMainHWND() const { return m_hMainWindow; }
			char *GetWindowClassName(WindowClassType Type);

			// Call this if implementing your own WinMain with special start-up handling.
			// (It will not return until the program is quit.)
			static int WINAPI WinMain(HINSTANCE hInstance, 
									  HINSTANCE hPrevInstance, 
									  LPSTR pszCmdLine, 
									  int nCmdShow);

			// Message loop - can be over-ridden
			virtual WPARAM DoMessageLoop();

			// Creating windows - all window creation is vectored 
			// through the Win32 app object.
			virtual HWND CreateWin32Window(Window *pWindow,
										   const Window* pParent,
										   TWindowStyle Style, 
										   const String* psWindowTitle);

			// Keep the application up to date with which modeless dialog
			// currently has input focus (if any)
			void SetModelessDialog(HWND hDlg);

			// Notify app that the main window is being destroyed (hence
			// the app is about to quit)
			void OnDestroyMainWindow();
			
		protected:
			//
			// WNDPROCs...
			//
			static LRESULT CALLBACK vGuiWndProc(HWND hWnd, 
												UINT message, 
												WPARAM wParam, 
												LPARAM lParam);

			static LRESULT CALLBACK vGuiRawWndProc(HWND hWnd, 
												   UINT message, 
												   WPARAM wParam, 
												   LPARAM lParam);

			static LRESULT CALLBACK SysTrayWndProc(HWND hWnd, 
												   UINT message, 
												   WPARAM wParam, 
												   LPARAM lParam);

			// These window class registration functions may be over-ridden
			// by different types of Win32 applications (e.g. MDI apps)
			virtual ATOM RegisterMainWindowClass();
			virtual ATOM RegisterChildWindowClass();
			virtual ATOM RegisterSysTrayWindowClass();

			// Service all the registered idle handlers that are active.
			bool ServiceIdleHandlers();

			// Notify the app that an input message has been received
			// (usually in a dialog message loop), so idle processing
			// should be temporarily halted.
			void NotifyUserInputMessageReceived(DWORD dwTimeOfMessage = 0);

			// Functions to encapsulate system tray operations
			bool SetTrayIcon(const Icon& TrayIcon, const String& ToolTip);
			void DeleteTrayIcon();
			bool TrayCallback(UINT nID, UINT uMouseMsg);

			// Get the menu that has been attached as the 'accessory'
			// menu.
			HMENU GetAccessoryMenu();

			// Get window used for routine system tray messages
			HWND GetSysTrayWindow();

			// Loading Win32 platform-specific libraries.
			void InitControls();
			void InitCOM();

		protected:
			// Instance handle of this app
			HINSTANCE m_hInstance;

			// Handle to accelerator table
			HACCEL m_hAccelTable;

			// Application icon
			Icon m_AppIcon;

			// Name of window classes
			char m_szWindowClass[80];
			char m_szChildWindowClass[80];
			char m_szSysTrayWindowClass[80];

			// Flag to indicate if a WM_QUIT message has been received.
			bool m_bQuitReceived;

			// Timestamp of the last message recieved (used to calculate
			// when to start servicing idle handlers)
			DWORD m_dwTimeOfLastInputMessage;

			// Handle of main window
			HWND m_hMainWindow;

			// Main menu
			Menu* m_pMainMenu;

		private:
			// Command to execute for notification UI events.
			CommandID m_cmdNotify;

			// Flag to indicate if a notification is active
			bool m_bNotifyActive;

			// Win32 application singleton.
			static Win32_ApplicationImp *s_pWin32App;

			// Library handles
			HMODULE m_hComCtl32;
			HMODULE m_hOLE32;

			// Final mesg handling status
			WPARAM m_MsgWParam;

			// Modeless dialog handling - we keep track of the currently
			// active modeless dialog by monitoring WM_ACTIVATE messages.
			HWND m_CurrentModelessDlg;

		private:
			//
			// Data for accessory apps
			//

			// Flag to indicate if we have installed a system tray icon
			bool m_bSysTrayIcon;

			// The default command to issue, e.g. if they click on the
			// accessory icon under Windows.
			CommandID m_cmdDefault;

			// Tooltip text for accessory
			String m_ToolTip;

			// The accessory menu
			HMENU m_hAccessoryMenu;

			// The systray window (if we have one)
			HWND m_hSysTrayWindow;

			//
			// Friends
			//

			// For access to vGuiWndProc
			friend class Win32_WindowImp;
			friend class Win32_DialogImp;
	};

	extern Win32_ApplicationImp *GetWin32App();

};

#endif // !vgui_Win32_ApplicationImp_h
