//
// Win32_CustomControl.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_CustomControl.h"

#include "Win32_WindowClasses.h"
#include "Win32_ApplicationImp.h"
#include "Win32_Theme.h"
#include "Win32_WindowInfo.h"
#include "Win32_Debug.h"

#include "..\Errors.h"

#include "vos\Trace.h"

#include <tmschema.h>

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	/**************************************************************************
	
	>	Win32_CustomControl::Win32_CustomControl()
	
		Author:		Tim Browse
		Date:		20/8/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	Win32_CustomControl::Win32_CustomControl() : m_WndProc(NULL)
	{
		// TCILB
	}

	/**************************************************************************
	
	>	Win32_CustomControl::~Win32_CustomControl()
	
		Author:		Tim Browse
		Date:		10/9/2002
	
		Purpose:	Destructor - ensures the object is detached from the
					Win32 window.
	
	**************************************************************************/
	
	Win32_CustomControl::~Win32_CustomControl()
	{
		Detach();
	}

	/**************************************************************************
	
	>	void Win32_CustomControl::Attach(HWND hControl)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hControl - window handle of the control to attach to.
	
		Purpose:	Attach this object to a particular active control.
				
					All messages to this control are then routed through the
					ControlWndProc() function, which you can over-ride to 
					provide custom message handling.
	
	**************************************************************************/
	
	void Win32_CustomControl::Attach(HWND hControl)
	{
		if (hControl == NULL)
		{
			COMPLAIN("Win32_CustomControl::Attach() - NULL control handle!");
			return;
		}
		
		// Subclass the control, so we intercept WM_PAINT messages
		// and mouse clicks, etc.
		m_WndProc = (WNDPROC) ::GetWindowLong(hControl, GWL_WNDPROC);
		::SetWindowLong(hControl, GWL_WNDPROC, (LONG) RawControlWndProc);

		// Put pointer to this object into the control's user data
		Win32_WindowInfo::Attach(hControl, this);

		// Keep window handle
		m_hWnd = hControl;
	}

	/**************************************************************************
	
	>	void Win32_CustomControl::Detach()
	
		Author:		Tim Browse
		Date:		20/8/2002
	
		Purpose:	Detach this object from its control.
	
	**************************************************************************/
	
	void Win32_CustomControl::Detach()
	{
		if (m_hWnd != NULL)
		{
			::SetWindowLong(m_hWnd, GWL_WNDPROC, (LONG) m_WndProc);
			Win32_WindowInfo::Detach(m_hWnd);
			m_hWnd = NULL;
		}
		else
		{
			COMPLAIN("Win32_CustomControl::Detach() - control is not attached!");
		}
	}
	/**************************************************************************

	>	LRESULT CALLBACK Win32_CustomControl::RawControlWndProc(HWND hWnd, UINT message, 
																WPARAM wParam, LPARAM lParam)


		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		Usual WNDPROC inputs
		Returns:	Usual WNDPROC return value.

		Purpose:	Redirect control messages to the ControlWndProc() virtual
					function.

	**************************************************************************/

	LRESULT CALLBACK Win32_CustomControl::RawControlWndProc(HWND hWnd, UINT message, 
														    WPARAM wParam, LPARAM lParam)
	{
		// Extract Win32_CustomControl object from window user data
		Win32_CustomControl *pControl;
		LRESULT lResult;

		if (Win32_WindowInfo::GetAttachedControl(hWnd, &pControl))
		{
			lResult = pControl->ControlWndProc(hWnd, message, wParam, lParam);

			// If window is being destroyed, then clean up
			if (message == WM_DESTROY)
				pControl->Detach();
		}
		else
		{
			// Control object not found (the C++ object may have been
			// deleted and detached by now) so we use default processing
			lResult = ::DefWindowProc(hWnd, message, wParam, lParam);
		}

		return lResult;
	}

	/**************************************************************************

	>	LRESULT Win32_CustomControl::ControlWndProc(HWND hWnd, UINT message, 
													WPARAM wParam, LPARAM lParam)


		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		Usual WNDPROC inputs
		Returns:	Usual WNDPROC return value.

		Purpose:	Provide simple way of processing control messages, and
					over-riding/customising behaviour of a standard control.

					Over-ride this function to implement custom message
					handling/filtering.  
					
					NB. To provide *additional* redraw functionality, just
						over-ride OnRedraw().  If you want to completely
						replace redraw functionality, then over-ride this
						function and process WM_PAINT messages.

					If you do not handle the message, or you want the default
					processing to be done in addition to your processing,
					then call Win32_CustomControl::ControlWndProc() at the
					end of your version of ControlWndProc().

	**************************************************************************/

	LRESULT Win32_CustomControl::ControlWndProc(HWND hWnd, UINT message, 
												WPARAM wParam, LPARAM lParam)
	{
		WIN32_LOGMESSAGE(hWnd, message, wParam, lParam);

		switch (message)
		{
			case WM_ERASEBKGND:
			{
				return 1;
			}

			case WM_PAINT:
			{
				// Does this control provide a rendering function?

				// NB. this test doesn't work - and I think it should!
				// TODO: investigate this.
//				if (this->OnRedraw != &Win32_CustomControl::OnRedraw)
				if (true)
				{
					// ...then do any custom painting required
					PAINTSTRUCT ps;
					::BeginPaint(hWnd, &ps);
					OnRedraw(hWnd, &ps);
					::EndPaint(hWnd, &ps);
					return 0;
				}
				else
				{
					// No custom redraw - get control to draw itself...
					return ::CallWindowProc(m_WndProc, hWnd, message, wParam, lParam);
				}
			}

			case WM_THEMECHANGED:
			{
				InvalidateTheme();
				break;
			}
		}

    	// Message not handled - revert to default handling
		return ::CallWindowProc(m_WndProc, hWnd, message, wParam, lParam);
	}

	/**************************************************************************
	
	>	void Win32_CustomControl::OnRedraw(HWND hWnd, PAINTSTRUCT* pPS)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hWnd - handle of the window to redraw
					pPS - the PAINTSTRUCT to use to redraw with, as filled
						  in by BeginPaint().  Use the hdc member to draw into
						  the control.
	
		Purpose:	Provide custom redraw facilities.

					NB. These are *additional* redraw facilities - first, the
						control itself is asked to redraw itself, then
						OnRedraw() is called, so you may draw extra elements
						on top of the control.

						If you want to *replace* redraw functionality, then
						over-ride ControlWndProc() and handle WM_PAINT yourself.
						(See Win32_CustomControl::ControlWndProc() for an example)

					Also, note that you do not need to call BeginPaint() or
					EndPaint() - this is done automatically.

	**************************************************************************/
	
	void Win32_CustomControl::OnRedraw(HWND hWnd, PAINTSTRUCT* pPS)
	{
		// By default, do nothing
	}
};



