//
// Win32_DateTimeBoxImp.cpp - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_DateTimeBoxImp.h"

#include "vos\Time.h"
#include "vos\TimeOfDay.h"
#include "vos\Date.h"

#include "Win32_headers.h"

#include <commctrl.h>
#include <winnls.h>

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	void Win32_DateTimeBoxImp::Attach(HWND hWnd)
	{
		// We have to check that the class name is correct.
		if (!IsWindowOfClass(hWnd, "SysDateTimePick32"))
		{
			// Trying to attach to the wrong kind of control - it must be a date/time picker
			vguiERROR();
			return;
		}

		// Ok, we allow it, so pass down to base class.
		DateTimeBoxImp::Attach(hWnd);
	}

	void Win32_DateTimeBoxImp::SetTime(const Time& NewTime)
	{
		// Convert time to SYSTEMTIME format
		SYSTEMTIME SysTime;
		NewTime.ToSystemTime(&SysTime);

		// Tell control the new time to use
		DateTime_SetSystemtime(m_hWnd, GDT_VALID, &SysTime);
	}

	void Win32_DateTimeBoxImp::GetTime(Time* pCurrentTime)
	{
		// Read time into a Win32 time structure
		SYSTEMTIME SysTime;
		DateTime_GetSystemtime(m_hWnd, &SysTime);

		// Convert to vGui time
		pCurrentTime->SetFromSystemTime(SysTime);
	}

	void Win32_DateTimeBoxImp::SetTimeOfDay(const TimeOfDay& NewTime)
	{
		// Get current time/date from control
		SYSTEMTIME SysTime;
		DateTime_GetSystemtime(m_hWnd, &SysTime);

		// Modify time
		SysTime.wHour = NewTime.GetHours();
		SysTime.wMinute = NewTime.GetMinutes(); 
		SysTime.wSecond = NewTime.GetSeconds(); 
		SysTime.wMilliseconds = 0;

		// Tell control the new time to use
		DateTime_SetSystemtime(m_hWnd, GDT_VALID, &SysTime);
	}

	void Win32_DateTimeBoxImp::GetTimeOfDay(TimeOfDay* pCurrentTime)
	{
		// Read time into a Win32 time structure
		SYSTEMTIME SysTime;
		DateTime_GetSystemtime(m_hWnd, &SysTime);

		// Copy to caller's variable
		pCurrentTime->Set(SysTime.wHour, 
						  SysTime.wMinute, 
						  SysTime.wSecond);
	}

	void Win32_DateTimeBoxImp::SetDate(const Date& NewDate)
	{
		// Get current time/date from control
		SYSTEMTIME SysTime;
		DateTime_GetSystemtime(m_hWnd, &SysTime);

		// Modify date
		SysTime.wYear = NewDate.GetYear();
		SysTime.wMonth = NewDate.GetCardinalMonth();
		SysTime.wDay = NewDate.GetDay();

		// Tell control the new time to use
		DateTime_SetSystemtime(m_hWnd, GDT_VALID, &SysTime);
	}

	void Win32_DateTimeBoxImp::GetDate(Date* pCurrentDate)
	{
		// Read time into a Win32 time structure
		SYSTEMTIME SysTime;
		DateTime_GetSystemtime(m_hWnd, &SysTime);

		// Copy to caller's variable
		pCurrentDate->Set(SysTime.wDay,
						  SysTime.wMonth,
						  SysTime.wYear);
	}

	void Win32_DateTimeBoxImp::SetFormat(DateTimeFormat NewFormat)
	{
		// Get localised form of time separator.
		String TimeSep = Time::GetTimeSeparator();

		String Format;

		switch (NewFormat)
		{
			case dtf_Time_12Hour:
			{
				// hh:mm tt
				Format += "hh'";
				Format += TimeSep;
				Format += "'mm tt";
				break;
			}

			case dtf_TimeWithSeconds_12Hour:
			{
				// hh:mm:ss tt
				Format += "hh'";
				Format += TimeSep;
				Format += "'mm'";
				Format += TimeSep;
				Format += "'ss tt";
				break;
			}

			case dtf_Time_24Hour:
			{
				// HH:mm
				Format += "HH'";
				Format += TimeSep;
				Format += "'mm";
				break;
			}

			case dtf_TimeWithSeconds_24Hour:
			{
				// HH:mm:ss
				Format += "HH'";
				Format += TimeSep;
				Format += "'mm'";
				Format += TimeSep;
				Format += "'ss";
				break;
			}

			case dtf_Date:
			{
				//
				// Interrogate the user locale to find out the date string
				// to use to format the control...
				//

				// Find out how big the buffer should be for the locale
				// string
				int cBufSize = ::GetLocaleInfo(LOCALE_USER_DEFAULT,
											   LOCALE_SSHORTDATE,
											   NULL, 0);

				// Allocate the buffer accordingly
				LPTSTR szDateString = new TCHAR[cBufSize+1];

				// Get the locale date string
				int cChars = ::GetLocaleInfo(LOCALE_USER_DEFAULT,
											 LOCALE_SSHORTDATE,
											 szDateString, cBufSize);

				// Check it worked before using the result
				if (cChars == cBufSize)
				{
					// Ok, use the string as a format string
					Format = szDateString;
				}
				else
				{
					// Failed to get date locale information - complain!
					vguiERROR();
				}
		
				// De-allocate the buffer
				delete [] szDateString;

				break;
			}

			default:
			{
				// Error - illegal enum type
				vguiERROR();
				break;
			}
		}

		// Set the new format if we constructed one without error
		if (!Format.IsEmpty())
		{
			DateTime_SetFormat(m_hWnd, Format.c_str());
		}
	}

	void Win32_DateTimeBoxImp::SetCustomFormat(const String& NewFormat)
	{
		// Set the new format if it's not an empty string
		if (!NewFormat.IsEmpty())
		{
			DateTime_SetFormat(m_hWnd, NewFormat.c_str());
		}
	}
};