//
// Win32_Debug.cpp - debugging facilities
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#ifdef _DEBUG

#include "Win32_Debug.h"

#include "..\Errors.h"

#include "vos\Trace.h"

#include "Win32_WindowImp.h"
#include "Win32_Messages.h"

#include <vector>
#include <algorithm>

namespace vgui
{
	// We need this adaptor because you can't store a type_info
	// directly in an STL container, as it has no copy ctor.
	struct type_info_key
	{
		type_info_key(const type_info& type) : type(&type) { }

		bool operator==(const type_info_key& other) const 
		{
			return !!type->operator==(*other.type);
		}

		bool operator<(const type_info_key& other) const 
		{
			return !!type->before(*other.type);
		} const type_info* type;
	};

	//
	// Tables which define which messages/events should be logged
	//

	static bool s_bLoggingEnabled = false;

	typedef std::pair<UINT, UINT> TMessageRange;
	typedef std::vector<TMessageRange> TMessageRanges;
	typedef std::list<type_info_key> TWindowTypes;
	typedef std::list<WindowHandle> TWindowHandles;

	static TMessageRanges  s_LogMessages;
	static TWindowTypes	s_LogWindowTypes;
	static TWindowHandles  s_LogWindows;

	void Win32_Debug::EnableMessageLogging()
	{
		s_bLoggingEnabled = true;
	}

	void Win32_Debug::DisableMessageLogging()
	{
		s_bLoggingEnabled = false;
	}

	void Win32_Debug::EnableLoggingOfAllMessages()
	{
		// Make sure all messages are logged
		s_LogMessages.clear();
		s_LogMessages.push_back(TMessageRange(0, UINT_MAX));
	}

	void Win32_Debug::EnableLoggingByMessage(UINT Msg)
	{
		Win32_Debug::EnableLoggingByMessage(Msg, Msg);
	}

	void Win32_Debug::EnableLoggingByMessage(UINT msgFirst, UINT msgLast)
	{
		if (msgFirst > msgLast)
		{
			COMPLAIN("Win32_Debug::EnableLoggingByMessage() - illegal message range!");
			return;
		}

		// Find out if this range is already enabled.
		TMessageRanges::iterator it = s_LogMessages.begin();

		bool bUpdatedRanges = false;

		// NB. This algorithm can result in more than one range
		//	   containing a particular message, but we stick with
		//	   it for simplicity's sake.
		while (!bUpdatedRanges && (it != s_LogMessages.end()))
		{
			if ((msgFirst >= it->first) && (msgFirst <= it->second))
			{
				// Range overlaps...make it include this range if it
				// doesn't already
				if (msgLast > it->second)
					it->second = msgLast;

				bUpdatedRanges = true;
			}
			else if ((msgLast >= it->first) && (msgLast <= it->second))
			{
				// Range overlaps...make it include this range
				it->first = msgFirst;
				bUpdatedRanges = true;
			}

			++it;
		}

		// If we get here, then we need to add a new range
		s_LogMessages.push_back(TMessageRange(msgFirst, msgLast));
	}

	void Win32_Debug::EnableLoggingByWindow(WindowHandle hWindow)
	{
		// Find the window handle...
		if (std::find(s_LogWindows.begin(), s_LogWindows.end(), hWindow) != s_LogWindows.end())
		{
			// It's not there, so add it
			s_LogWindows.push_back(hWindow);
		}
	}

	void Win32_Debug::EnableLoggingByWindowType(const type_info& Type)
	{
		// Find the window type...
		if (std::find(s_LogWindowTypes.begin(), 
					  s_LogWindowTypes.end(), 
					  type_info_key(Type)) != s_LogWindowTypes.end())
		{
			// It's not there, so add it
			s_LogWindowTypes.push_back(type_info_key(Type));
		}
	}

	void Win32_Debug::DisableLoggingByMessage(UINT Msg)
	{
		Win32_Debug::DisableLoggingByMessage(Msg, Msg);
	}

	void Win32_Debug::DisableLoggingByMessage(UINT msgFirst, UINT msgLast)
	{
		if (msgFirst > msgLast)
		{
			COMPLAIN("Win32_Debug::DisableLoggingByMessage() - illegal message range!");
			return;
		}

		// Find out if this range (or part of it) is enabled.
		//

		// Here are the possible cases:
		
		/*

		Range to examine:			|-----------------------|
									.						.
		Range to disable:			.						.
									.						.
		A	-	-	-	-	|---|	.						.
		B						  |------|					.
		C	-	-	-	-	-	-   |--------|				.
		D							.      |----------|		.
		E	-	-	-	-	-	-	.		        |-------|
		F							.				    |------|
		G	-	-	-	-	-	-	.					    .    |----|
		H							|-----------------------|
		I	-	-	-	-	-	|---------------------------|
		J							|----------------------------|

		*/
		TMessageRanges::iterator it = s_LogMessages.begin();

		while (it != s_LogMessages.end())
		{
			// Check for ranges outside this one (cases A and G)
			if ((msgLast < it->first) || (msgFirst > it->second))
			{
				// no overlap
				++it;
				continue;
			}

			// Check for complete overlap (cases H, I, and J)
			if ((msgFirst <= it->first) && (msgLast >= it->second))
			{
				// Simple - just remove the range
				it = s_LogMessages.erase(it);

				// Avoid incrementing 'it', as it's already at the right 
				// position for the next iteration
				continue;
			}
			// Check for cases B and C
			else if (msgFirst <= it->first)
			{
				// Alter this range to exclude the range we're disabling
				it->first = msgLast + 1;
			}
			// Check for cases E and F
			else if (msgLast >= it->second)
			{
				// Alter this range to exclude the range we're disabling
				it->second = msgFirst-1;
			}
			else
			{
				// Range to disable is a subset of this range, so 
				// we split the range into two (case D)
				UINT NewFirst = msgLast+1;
				UINT NewLast  = it->second;

				// Alter the existing range...
				it->second = msgFirst-1;

				// And add the new one
				s_LogMessages.push_back(TMessageRange(NewFirst, NewLast));
				
				// Start from beginning again, as that could have invalidated
				// the iterator.
				it = s_LogMessages.begin();
			}

			++it;
		}
	}

	void Win32_Debug::DisableLoggingByWindow(WindowHandle hWindow)
	{
		// Remove this window from the handle list
		s_LogWindows.remove(hWindow);
	}

	void Win32_Debug::DisableLoggingByWindowType(const type_info& Type)
	{
		// Remove this window type from the list		
		s_LogWindowTypes.remove(type_info_key(Type));
	}

	const char* Win32_Debug::GetWindowName(Win32_WindowImp* pWindowImp)
	{
		if (pWindowImp != NULL)
		{
			const type_info& ti = typeid(*pWindowImp);

			const char* szName = ti.name();

			static const char szClass[] = "class ";

			if (strncmp(szName, szClass, strlen(szClass)) == 0)
			{
				szName += strlen(szClass);
			}

			static const char szVGUI[] = "vgui::";

			if (strncmp(szName, szVGUI, strlen(szVGUI)) == 0)
			{
				szName += strlen(szVGUI);
			}

			return szName;
		}
		else
		{
			COMPLAIN("GetWindowName() - NULL window pointer!");
			return "<UNKNOWN>";
		}
	}

	static int s_iLog = 1;

	void Win32_Debug::OutputMessageLog(HWND hwnd, UINT message,
									   WPARAM wParam, LPARAM lParam)
	{
		// Translate message name
		const char* szMessageName = Win32_Message_Decode(message);

		Win32_WindowImp* pWindowImp = Win32_WindowImp::GetImpFromHWND(hwnd);

		if (pWindowImp != NULL)
		{
			const char* szName = GetWindowName(pWindowImp);

			TRACE("S %.8X %s:\t%s(0x%.8X),\twParam=0x%.8X,\tlParam=0x%.8X\n",
				  s_iLog, szMessageName, szName, hwnd, wParam, lParam);
		}
		else
		{
			TRACE("S %.8X %s:\thwnd=0x%.8X,\twParam=0x%.8X,\tlParam=0x%.8X\n",
				  s_iLog, szMessageName, hwnd, wParam, lParam);
		}

		s_iLog++;
	}

	void Win32_Debug::OutputMessageHandledLog(HWND hwnd, UINT message,
											  LRESULT lResult)
	{
		// Translate message name
		const char* szMessageName = Win32_Message_Decode(message);

		Win32_WindowImp* pWindowImp = Win32_WindowImp::GetImpFromHWND(hwnd);

		if (pWindowImp != NULL)
		{
			const char* szName = GetWindowName(pWindowImp);

			TRACE("R %.8X %s:\t%s(0x%.8X),\tlResult = 0x%.8X\n",
				  s_iLog, szMessageName, szName, hwnd, lResult);
		}
		else
		{
			TRACE("R %.8X %s:\thwnd=0x%.8X,\tlResult = 0x%.8X\n",
				  s_iLog, szMessageName, hwnd, lResult);
		}

		s_iLog++;
	}

	bool Win32_Debug::MessageShouldBeLogged(HWND hwnd, UINT message)
	{
		// Ok, check the window handle first...
		if (std::find(s_LogWindows.begin(), s_LogWindows.end(), hwnd) != s_LogWindows.end())
		{
			// This message should be logged...
			return true;
		}
		else
		{
			// Find the window type attached to this window (if any)
			Win32_WindowImp* pWindowImp = Win32_WindowImp::GetImpFromHWND(hwnd);

			const type_info& ti = typeid(pWindowImp);

			if (pWindowImp != NULL)
			{
				if (std::find(s_LogWindowTypes.begin(), 
							  s_LogWindowTypes.end(), 
							  type_info_key(ti)) != s_LogWindowTypes.end())
				{
					// This message should be logged...
					return true;
				}
			}

			// Check to see if this message type should be logged
			TMessageRanges::iterator it = s_LogMessages.begin();

			while (it != s_LogMessages.end())
			{
				if ((message >= it->first) && (message <= it->second))
				{
					// This message should be logged...
					return true;
				}

				++it;
			}
		}

		// Message should not be logged
		return false;
	}

	void Win32_Debug::LogMessage(HWND hwnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
	{
		// Ok, log the message if required
		if (MessageShouldBeLogged(hwnd, message))
			OutputMessageLog(hwnd, message, wParam, lParam);
	}

	void Win32_Debug::LogMessageHandled(HWND hwnd, UINT message,
										LRESULT lResult)
	{
		// Ok, log the message if required
		if (MessageShouldBeLogged(hwnd, message))
			OutputMessageHandledLog(hwnd, message, lResult);
	}

};

#endif // !_DEBUG
