//
// Win32_DialogImp.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_DialogImp.h"

#include "Win32_WindowImp.h"
#include "Win32_ApplicationImp.h"
#include "Win32_FontImp.h"
#include "Win32_Headers.h"
#include "Win32_StaticControl.h"
#include "Win32_Debug.h"
#include "Win32_WindowInfo.h"
#include "Win32_MessageInfo.h"

#include "..\Font.h"
#include "..\Errors.h"
#include "..\Rect.h"
#include "..\WaitCursor.h"

#include "vos\Trace.h"

#define ASSERT_VALID_WINDOW()							\
		if (!IsValid())									\
		{												\
			COMPLAIN("Invalid window handle used!");	\
			return;										\
		}

#define ASSERT_VALID_WINDOW_RETURN(RetVal)				\
		if (!IsValid())									\
		{												\
			COMPLAIN("Invalid window handle used!");	\
			return (RetVal);							\
		}


namespace vgui
{
	// Special control ID for the gripper resize control
	const int kGripperID = 13042;

	Win32_DialogImp::Win32_DialogImp() : m_OpenPos(wp_Default),
										 m_bModal(true),
										 m_cUpdateLock(0),
										 m_InitialDialogVisibility(sw_SHOW),
										 m_pResizeGripper(NULL),
										 m_bUseAutoResizeGripper(true)
	{
	}

	Win32_DialogImp::~Win32_DialogImp()
	{
		delete m_pResizeGripper;
	}

	DLGTEMPLATE* Win32_DialogImp::CreateEmptyTemplate()
	{
		// Fill in the dialog template for a dialog with no controls
		static const wchar_t wszDialogTitle[] = L"vGUI Dynamic Dialog - call SetText() to change this text!";

		// Standard default font for Dialogs - 8pt MS Sans Serif
		static const wchar_t wszDialogFont[] = L"MS Sans Serif";
		static const int DialogFontSize = 8;

		// Work out how big the dialog template should be
		int cTemplateSize = sizeof(DLGTEMPLATE);

		// Allow space for empty menu array
		cTemplateSize += sizeof(WORD);

		// Allow space for empty class array
		cTemplateSize += sizeof(WORD);

		// Allow space for title.
		cTemplateSize += (wcslen(wszDialogTitle) + 1) * sizeof(wchar_t);

		// Allow space for font size and name
		cTemplateSize += sizeof(WORD);
		cTemplateSize += (wcslen(wszDialogFont) + 1) * sizeof(wchar_t);

		// Create the dialog template
		DLGTEMPLATE* pDT = (DLGTEMPLATE*) malloc(cTemplateSize);

		pDT->style = DS_CENTER | DS_SETFONT | DS_MODALFRAME;
		pDT->dwExtendedStyle = 0;
		pDT->cdit = 0;  // No controls in our dialog
		pDT->x = 100;
		pDT->y = 100;
		pDT->cx = 100;
		pDT->cy = 100;

		// Fill in the extra data
		WORD* pExtraData = (WORD*) (pDT+1);

		// Menu array (empty)
		*pExtraData++ = 0;

		// Class array (empty)
		*pExtraData++ = 0;

		// Dialog title (Unicode, null-terminated).
		wcscpy(pExtraData, wszDialogTitle);

		// Font size and typface name
		pExtraData += wcslen(wszDialogTitle) + 1;
		*pExtraData++ = DialogFontSize;
		wcscpy(pExtraData, wszDialogFont);

		// Pass template back to caller.
		return pDT;
	}

	ControlID Win32_DialogImp::DoModalDialog(const DialogResourceID DialogID,
											 const Window *pParent,
											 WindowPos Pos)
	{
		// Remember desired position
		m_OpenPos = Pos;

		// Get parent window handle
		HWND hParentWnd = GetHWND(pParent);

		// Create the dialog from the template and execute it modally.
		//
		// NB We pass a pointer to this Win32_DialogImp object so that
		//    the WM_INITDIALOG handler can patch this WHND to this
		//	  object.
		int Result;
		
		// Check for dynamic dialogs - if one is requested, we build
		// the DLGTEMPLATE object here
		if (DialogID == DialogResourceID::s_DynamicDialog)
		{
			// Create dialog from dynamic template
			DLGTEMPLATE* pDT = CreateEmptyTemplate();

			Result = ::DialogBoxIndirectParam(GetWin32App()->GetInstance(),
											  pDT, 
											  hParentWnd, 
											  vGuiDialogProc,
											  (LPARAM) this);

			// We've finished with the dialog template
			free(pDT);
		}
		else
		{
			// use dialog resource
			Result = ::DialogBoxParam(GetWin32App()->GetInstance(),
									  DialogID.GetID(),
									  hParentWnd,
									  vGuiDialogProc,
									  (LPARAM) this);
		}

		// Modal dialog
		m_bModal = true;

		// Tell caller the result
		return (ControlID) Result;
	}

	void Win32_DialogImp::DoModelessDialog(const DialogResourceID DialogID,
										   const Window *pParent,
										   WindowPos Pos,
										   TShowWindow Show)
	{
		// Remember desired position
		m_OpenPos = Pos;

		// Get parent window handle
		HWND hParentWnd = GetHWND(pParent);

		// Modeless dialog
		m_bModal = false;

		// Create the dialog from the template and execute it in a modeless way.
		// NB We pass a pointer to this Win32_DialogImp object so that
		//    the WM_INITDIALOG handler can patch this WHND to this
		//	  object.
		if (::CreateDialogParam(GetWin32App()->GetInstance(),
								DialogID.GetID(),
								hParentWnd,
								vGuiDialogProc,
								(LPARAM) this) == NULL)
		{
			// Failed to create dialog...
			DWORD dwError = ::GetLastError();
			COMPLAIN("DoModelessDialog() - Failed to create dialog!");
		}

		// The WM_INITDIALOG handling code takes care of showing/hiding the
		// modeless dialog...
		m_InitialDialogVisibility = sw_SHOW;
	}


	/**************************************************************************

	>	bool Win32_DialogImp::Create(const DialogResourceID DialogID,
									 const Window *pParent)


		Author:		Tim Browse
		Date:		13/7/2002
		Inputs:		DialogID - the dialog template to create the dialog from
					pParent - parent window of the dialog
		Returns:	true => dialog created successfully
					false => otherwise

		Purpose:	Create (and show) a dialog direct from a dialog resource.

					This is useful mainly for child dialogs (see ChildDialog
					class).

	**************************************************************************/

	bool Win32_DialogImp::Create(const DialogResourceID DialogID,
								 const Window *pParent)
	{
		// Ok, try to lock down the resource
		// (this resource is unloaded automatically when app exits)
		DLGTEMPLATE* pTemplate = LockDialogResource(DialogID);

		if (pTemplate == NULL)
		{
			COMPLAIN("Win32_ChildDialog::Init() - cannot find/load the specified dialog!");
			return false;
		}

		// Create the new child dialog box. 
		HWND hChildDialog = ::CreateDialogIndirectParam(GetWin32App()->GetInstance(), 
													    pTemplate,
													    pParent->GetWindowHandle(), 
													    vGuiDialogProc,
														(LPARAM) this); 

		if (hChildDialog == NULL)
		{
			COMPLAIN("Win32_ChildDialog::Init() - Unable to create dialog!");
			return false;
		}

		// All done
		return true;
	}

	/**************************************************************************

	>	BOOL CALLBACK Win32_DialogImp::vGuiDialogProc(HWND hwndDlg,
													  UINT uMsg,
													  WPARAM wParam,
													  LPARAM lParam)


		Author:		Tim Browse
		Date:		22/05/2001
		Inputs:		See DialogProc in Win32 SDK.
		Returns:	See notes below, and DialogProc in Win32 SDK.

		Purpose:	The basic DialogProc for vGUI dialogs.


					Tech Notes
					==========


					Extract from Win32SDK docs...

					Return Values
					~~~~~~~~~~~~~

					Typically, the dialog box procedure should return TRUE if it 
					processed the message, and FALSE if it did not. If the dialog 
					box procedure returns FALSE, the dialog manager performs the 
					default dialog operation in response to the message.

					If the dialog box procedure processes a message that requires 
					a specific return value, the dialog box procedure should set 
					the desired return value by calling 
					SetWindowLong(hwndDlg, DWL_MSGRESULT, lResult) immediately 
					before returning TRUE. 
					
					Note that you must call SetWindowLong immediately before 
					returning TRUE; doing so earlier may result in the DWL_MSGRESULT 
					value being overwritten by a nested dialog box message.

					The following messages are exceptions to the general rules 
					stated above. Consult the documentation for the specific message 
					for details on the semantics of the return value.

						WM_CHARTOITEM
						WM_COMPAREITEM
						WM_CTLCOLORBTN
						WM_CTLCOLORDLG
						WM_CTLCOLOREDIT
						WM_CTLCOLORLISTBOX
						WM_CTLCOLORSCROLLBAR
						WM_CTLCOLORSTATIC
						WM_INITDIALOG
						WM_QUERYDRAGICON
						WM_VKEYTOITEM



	**************************************************************************/

	BOOL CALLBACK Win32_DialogImp::vGuiDialogProc(HWND hwndDlg,
												  UINT uMsg,
												  WPARAM wParam,
												  LPARAM lParam)
	{
		WIN32_LOGMESSAGE(hwndDlg, uMsg, wParam, lParam);
			
		switch (uMsg)
		{
			// Child controls (such as scroll bars, sliders etc) send
			// notification messages to the parent via WM_[HV]SCROLL
			// messages, so we map notification messages here.
			case WM_HSCROLL:
			case WM_VSCROLL:
			{
				// This is almost always the result of a user action
				NotifyUserInputMessageReceived();

				// Get the handle of the scroll bar control
				HWND hControl = (HWND) lParam;
				if (hControl != NULL)
				{
					int ControlID = ::GetDlgCtrlID(hControl);

					Win32_DialogImp* pWin32Dialog = GetImpFromHWND(hwndDlg);

					if ((pWin32Dialog  != NULL) && 
						(pWin32Dialog->DispatchControlNotification(uMsg, ControlID, lParam, wParam) == 0))
					{
						// Processed message
						return TRUE;
					}
				}
				break;
			}

			//
			// Messages that we let the normal Win32_Window wndproc handle...
			//
			case WM_SYSCOMMAND:
			{
				Win32_DialogImp* pWin32Dialog = GetImpFromHWND(hwndDlg);

				if (pWin32Dialog != NULL)
				{
					if (pWin32Dialog->WindowProc(uMsg, wParam, lParam) == 0)
					{
						// processed message
						return TRUE;
					}
				}

				break;
			}

#if 0
			case WM_CTLCOLOREDIT:
			{
				COLORREF clr = RGB(224,224,160);
				static HBRUSH hBrush = ::CreateSolidBrush(clr);
				HDC hDC = (HDC) wParam;
				::SetBkColor(hDC, clr);
				::SetTextColor(hDC, RGB(255,255,40));
				return (BOOL) hBrush;
			}
#endif

			case WM_ACTIVATE:
			{
				Win32_DialogImp* pWin32Dialog = GetImpFromHWND(hwndDlg);

				if (!pWin32Dialog->m_bModal)
				{
					// Modeless dialog - we need to keep the app up to 
					// date with which modeless dialog is active, so it
					// can call IsDialogMessage() correctly.
					//
					// Gorgeous...

					if (wParam == 0)
					{
						// becoming inactive
						GetWin32App()->SetModelessDialog(NULL);
					}
					else
					{
						// becoming active
						GetWin32App()->SetModelessDialog(hwndDlg);
					}
				}

				// We don't need to claim the message
				break;
			}

			case WM_DESTROY:
			{
				// Window being destroyed - let the OS-independent window object know.
				Win32_DialogImp* pDialogImp = GetImpFromHWND(hwndDlg);

				Win32_ApplicationImp* pWin32App = GetWin32App();
				bool bIsMainWindow = pDialogImp->m_hWnd == pWin32App->GetMainHWND();

    			if (bIsMainWindow)
    			{
    				// Main window is being destroyed, so we're done.
					pWin32App->OnDestroyMainWindow();
				}

				// OnDestroy() handler is the last time the window is
				// considered to be 'alive'
				pDialogImp->m_pWindow->OnDestroy();
				pDialogImp->m_bAlive = false;

    			if (!bIsMainWindow && !pDialogImp->m_bModal)
				{
					//
					// Modeless dialogs delete themselves, so they can be used
					// in a sort of fire-and-forget manner.
					//
					// NB. Note we studiously avoid deleting the main app window!
					//
					delete pDialogImp->m_pWindow;
				}
				else
				{
					// Clean up window info (modeless dialogs will already have
					// done this when they are deleted by the above line of code.
					Win32_WindowInfo::Detach(hwndDlg);
				}

				// Message handled
				return TRUE;
			}

			case WM_INITDIALOG:
			{
				// Get Dialog and DialogImp pointers.
				Win32_DialogImp *pDialogImp = reinterpret_cast<Win32_DialogImp*>(lParam);
				Dialog *pDialog = pDialogImp->GetDialog();

				// Initialise our window handle
				pDialogImp->m_hWnd = hwndDlg;

				// Window should be marked as being alive.
				ASSERT(!pDialogImp->m_bAlive);
				pDialogImp->m_bAlive = true;

				// Store WindowImp pointer for this object in the window.
				Win32_WindowInfo::Attach(hwndDlg, pDialog);

				// If this dialog is the main window, set up the dialog to
				// use the App icon (special case, because for dialogs we don't
				// register a private dialog class - because this way is easier).
				Application* pApp = GetApp();
				Win32_ApplicationImp* pWin32App = GetWin32App();
				HINSTANCE hInstance = pWin32App->GetInstance();

				bool bIsMainWindow = pApp->GetMainWindow() == pDialog;

				if (bIsMainWindow)
				{
					// Load the app icon (Windows will dispose of it for us at program exit)
					SimpleResourceID IconID = pApp->GetIconID().GetID();
					HICON hIcon = ::LoadIcon(hInstance, MAKEINTRESOURCE(IconID));
					pDialogImp->SendMessage(WM_SETICON, ICON_BIG, (LPARAM) hIcon);

					IconID = pApp->GetSmallIconID().GetID();
					if (IconID != 0)
					{
						hIcon = ::LoadIcon(hInstance, MAKEINTRESOURCE(IconID));
						pDialogImp->SendMessage(WM_SETICON, ICON_SMALL, (LPARAM) hIcon);
					}

					// Ok, if this is an accessory app, should the main window
					// appear immediately?
					const vgui::AppOptions& Options = pApp->GetOptions();

					if (Options.m_bAccessory && !Options.m_bShowWindowOnAppStart)
					{
						// No, main dialog should be hidden
						pDialogImp->m_InitialDialogVisibility = sw_HIDE;
					}
				}

				// Call dialog handler
				pDialog->OnPreInitDialog();
					pDialog->OnInitDialog();
				pDialog->OnPostInitDialog();

				// Is this dialog resizable?
				if (::GetWindowLong(hwndDlg, GWL_STYLE) & WS_THICKFRAME)
				{
					// Yes, so add a resize gripper.
					pDialogImp->CreateResizeGripper();
				}
					
				if (bIsMainWindow)
				{
					// Is there a menu for the main dialog window?
					//
					// TODO: This means that in OnInitDialog(), the menu is not
					//		 attached, so the client can't update the menu.
					//		 But if we call SetMenu() before calling OnInitDialog(),
					//		 it causes a resize/repaint event, and the dialog
					//       may not be able to handle this before the dialog
					//		 controls are initialised (e.g. not Attach'ed yet).
					//
					//		 Should really find a way to fix this, or document
					//		 the behaviour...
					MenuResourceID Menu = GetApp()->GetMenuID();
					if (!Menu.IsNull())
					{
						SimpleResourceID MenuID = Menu.GetID();
						HMENU hMenu = ::LoadMenu(hInstance, MAKEINTRESOURCE(MenuID));

						if (hMenu != NULL)
						{
							::SetMenu(hwndDlg, hMenu);
						}
						else
						{
							COMPLAIN("Unable to load main menu to attach to main dialog!");
						}
					}
				}

				// Should we reposition the dialog?
				if (pDialogImp->m_OpenPos == wp_Centred)
				{
					pDialogImp->CentreWindow();
				}

				// Hide or show the dialog as appropriate
				//
				// (Windows will only show the dialog automatically if
				//  the WS_VISIBLE flag is set in the dialog template, so
				//  we don't rely on this).
				//
				if (pDialogImp->m_InitialDialogVisibility == sw_HIDE)
				{
					::ShowWindow(hwndDlg, SW_HIDE);
				}
				else
				{
					::ShowWindow(hwndDlg, SW_SHOW);
				}

				// Allow DialogProc to set default keyboard focus
				return TRUE;
			}

			case WM_COMMAND:
			{
				// Decode parameters
				UINT uNotifyCode = HIWORD(wParam);

				// Very annoying Windows bug here - if uNotifyCode is 0, then
				// this is supposed to be a menu command - however: BN_CLICKED
				// is also zero, so you can't tell if it's a button being
				// clicked or a menu item being selected, so you need to be
				// careful that your menu command IDs and control IDs do not
				// overlap.
				// TODO: Tell the user about this issue in the docs.
				if (uNotifyCode == 0)
				{
					// It's a menu command
					int nMenuCommand = LOWORD(wParam);

					// See if we should process this menu item
					// (use routing handler if it exists)
					Win32_WindowImp* pWindowImp = GetImpFromHWND(hwndDlg);

					Window *pHandlerWindow = pWindowImp->GetControlHandler();

					if (pHandlerWindow == NULL)
						pHandlerWindow = pWindowImp->GetWindowLink();

					if (pHandlerWindow->OnCommand(nMenuCommand))
					{
						// processed
						return TRUE;
					}
				}

				// Extract item,control or accelerator ID
				ControlID nCtrlID = LOWORD(wParam);

				// handle of control 
				HWND hwndCtl = (HWND) lParam;

				// Try control notification messages
				//
				// (We don't bother passing in lParam or wParam, as there's 
				// nothing useful in them that we haven't already used).
				if (GetImpFromHWND(hwndDlg)->DispatchControlNotification(uNotifyCode, nCtrlID, 0, 0) == 0)
				{
					// Processed message
					return TRUE;
				}
				
				break;
			}

			case WM_NOTIFY:
			{
				// Decode parameters
				NMHDR* pnmh = (NMHDR*) lParam;
				UINT uNotifyCode = pnmh->code;

				// Extract IF of control that sent this message
				ControlID nCtrlID = (int) wParam;

				// Try control notification messages
				Win32_DialogImp* pDialogImp = GetImpFromHWND(hwndDlg);

				if (pDialogImp != NULL)
				{
					if (pDialogImp->DispatchControlNotification(uNotifyCode, nCtrlID, lParam, wParam) == 0)
					{
						// Processed message
						return TRUE;
					}
				}
				break;
			}

			case WM_PAINT:
			{
				if (GetImpFromHWND(hwndDlg)->m_pWindow->OnRedraw())
				{
					// Message processed
					return TRUE;
				}
				break;
			}

			case WM_NCHITTEST:
			{
				// We over-ride non-client hit testing for dynamic dialogs, so
				// that the resize gripper we position in the bottom right
				// acts as a resize control using standard Windows logic.
				Win32_DialogImp* pWin32Dialog = GetImpFromHWND(hwndDlg);
				if (pWin32Dialog->GetDialog()->IsDynamicDialog())
				{
					Rect rcGripper;
					pWin32Dialog->GetGripperPos(&rcGripper);
					
					// The point to be tested is in screen coordinates
					::MapWindowPoints(hwndDlg, HWND_DESKTOP, (POINT*) &rcGripper, 2);

					// When in the size grip, pretend to be the diagonal resize
					// area of a resizable window
					Point HitPoint(LOWORD(lParam), HIWORD(lParam));
					if (rcGripper.IsPointInside(HitPoint))
					{
						::SetWindowLong(hwndDlg, DWL_MSGRESULT, HTBOTTOMRIGHT);
						return TRUE;
					}
				}

				// Message not processed
				break;
			}

			//
			// Owner-draw control messages
			//
			case WM_DRAWITEM:
			case WM_MEASUREITEM:
			{
				// Dispatch message to control using central method.
				Win32_DialogImp* pWin32Dialog = GetImpFromHWND(hwndDlg);
				
				if (pWin32Dialog != NULL)
				{
					if (pWin32Dialog->DispatchControlNotification(uMsg, (int) wParam, lParam, wParam) == 0)
					{
						// Processed message
						return TRUE;
					}
				}

				// Ok, try a standalone control instead
				Win32_CustomControl* pControl;
				DRAWITEMSTRUCT* pDIS = (DRAWITEMSTRUCT*) lParam;
				if (Win32_WindowInfo::GetAttachedControl(pDIS->hwndItem, &pControl))
				{
					// Bit of a cheat - we reflect the messages to the
					// control's WndProc so we can neatly provide owner-draw
					// controls without having to put the logic into the
					// dialog code.
					return pControl->ControlWndProc(pDIS->hwndItem, uMsg, wParam, lParam);
				}
				break;
			}

			case WM_WINDOWPOSCHANGING:
			{
				Win32_DialogImp* pWin32Dialog = GetImpFromHWND(hwndDlg);
				if (pWin32Dialog->HandleWindowPosChanging((WINDOWPOS*) lParam))
					return TRUE;

				break;
			}

			case WM_SIZE:
			{
				Win32_DialogImp* pDialogImp = GetImpFromHWND(hwndDlg);

				if (pDialogImp != NULL)
				{
					Dialog* pDialog = pDialogImp->GetDialog();

					if (pDialogImp->m_pResizeGripper != NULL)
					{
						// We have a gripper, so move it.
						Rect GripperPos;
						pDialogImp->GetGripperPos(&GripperPos);
						::MoveWindow(pDialogImp->m_pResizeGripper->GetHWND(),
									 GripperPos.left, GripperPos.top,
									 GripperPos.GetWidth(), GripperPos.GetHeight(),
									 TRUE);
					}

					if (pDialog->OnResize(LOWORD(lParam), HIWORD(lParam)))
					{
						// Message processed
						return TRUE;
					}
				}
				else
				{
					COMPLAIN("Unable to find dialog implementation object!");
				}
				break;
			}

    		case WM_TIMER:
    		{
    			// Decode parameters
    			TimerID ID = (TimerID) wParam;

    			if (Timer::ServiceTimer(ID))
    			{
    				// Message processed
    				return TRUE;
    			}

    			break;
    		}

			// Cursor appearance
			case WM_SETCURSOR:
			{
				// Should we display an hourglass?
				if (WaitCursor::IsActive())
				{
					// Yes...
					::SetCursor(::LoadCursor(NULL, IDC_WAIT));
				}
				break;
			}

			case WM_KEYDOWN:
			case WM_KEYUP:
			{
				NotifyUserInputMessageReceived();

				Win32_DialogImp* pWin32Dialog = GetImpFromHWND(hwndDlg);
				if (pWin32Dialog != NULL)
				{
					if (pWin32Dialog->HandleRawKeyEvent(uMsg, wParam))
					{
    					// Message processed
    					return TRUE;
					}
				}

				break;
			}

    		case WM_CHAR:
    		{
				NotifyUserInputMessageReceived();

    			// Check to see if window wants this character message
				Win32_DialogImp* pWin32Dialog = GetImpFromHWND(hwndDlg);
				if ((pWin32Dialog != NULL) && (pWin32Dialog->m_pWindow != NULL))
				{
    				if (pWin32Dialog->m_pWindow->OnChar((uint16) wParam))
					{
    					// Message processed
    					return TRUE;
					}
				}
				break;
			}


#if 0
			// NB. This code is disabled, because despite what the MS
			//	   docs say, it doesn't work.  If this code is enabled, 
			//     the loop never exits because PeekMessage() never reports
			//     that there are any more messages to process, despite me
			//     moving the mouse like crazy.  Go figure.
			case WM_ENTERIDLE:
			{
				if (wParam == MSGF_DIALOGBOX)
				{
					// Modal dialog is entering idle loop - service
					// our idle handlers until we get another message
					// in the queue
					MSG Msg;

					do
					{
						GetApp()->ServiceIdleHandlers();
					}
					while (::PeekMessage(&Msg, NULL, 0, 0, PM_NOREMOVE) == 0);
				}

				// Message processed
				return TRUE;
			}
#endif

		}

		// Didn't process the message
		return FALSE;
	}

	Win32_DialogImp* Win32_DialogImp::GetImpFromHWND(HWND hWnd)
	{
		return static_cast<Win32_DialogImp*>(Win32_WindowImp::GetImpFromHWND(hWnd));
	}

	void Win32_DialogImp::EndDialog(ControlID ID)
	{
		// End the dialog - method differs on what sort of dialog
		// it is.
		if (m_bModal)
		{
			::EndDialog(m_hWnd, (int) ID);
		}
		else
		{
			::DestroyWindow(m_hWnd);
		}
	}

	void Win32_DialogImp::EnableDialogControl(ControlID ID, bool bEnable)
	{
		// Find window handle of the control
		HWND hwndCtl = FindControl(m_hWnd, (int) ID);

		if (hwndCtl != NULL)
		{
			// Enable/disable as appropriate.
			::EnableWindow(hwndCtl, bEnable ? TRUE : FALSE);
		}
	}

	Dialog* Win32_DialogImp::GetDialog(HWND hwnd)
	{
		Dialog *pDialog = (Dialog*) GetWindowFromHWND(hwnd);

		ASSERT(pDialog != NULL);

		return pDialog;
	}

	void Win32_DialogImp::Detach()
	{
		ASSERT_VALID_WINDOW();

		m_hWnd = NULL;
		m_bAlive = false;
	}

	void Win32_DialogImp::MoveWindow(HWND hCtrl, int x, int y, 
									 int Width, int Height,
									 BOOL bRepaint)
	{
		// If LockUpdate() is still active, queue up this window change.
		if (m_cUpdateLock > 0)
		{
			// Get the DeferWindowPos structure for this control
			int iEntry = GetHWDPEntry(hCtrl);

			// Don't change the Z-order of the window
			UINT flags = SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOZORDER;

			// Don't repaint the window if the caller doesn't want repainting
			if (!bRepaint)
				flags |= SWP_NOREDRAW;
	
			m_HDWPEntries[iEntry].m_hDWP = ::DeferWindowPos(m_HDWPEntries[iEntry].m_hDWP, hCtrl, 0, 
															x, y, Width, Height, flags);
			
			ASSERT(m_HDWPEntries[iEntry].m_hDWP!= NULL);
		}
		else
		{	
			// No active update lock, so just move the window in the normal way.
			::MoveWindow(hCtrl, x, y, Width, Height, TRUE);
		}
	}

	void Win32_DialogImp::MoveControl(ControlID ID, int x, int y)
	{
		ASSERT_VALID_WINDOW();

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		COMPLAIN_IF(hCtrl == NULL, "Control not found in MoveControl().");

		// Get the control's position
		Rect rcCtrlPos;
		::GetWindowRect(hCtrl, (RECT*) &rcCtrlPos);

		// Move control
		MoveWindow(hCtrl, x, y, rcCtrlPos.GetWidth(), rcCtrlPos.GetHeight(), TRUE);
	}

	void Win32_DialogImp::MoveControlRelative(ControlID ID, int DeltaX, int DeltaY)
	{
		ASSERT_VALID_WINDOW();

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		COMPLAIN_IF(hCtrl == NULL, "Control not found in MoveControl().");

		// Get the control's position
		Rect rcCtrlPos;
		::GetWindowRect(hCtrl, (RECT*) &rcCtrlPos);

		// Convert to parent window co-ordinates
		::ScreenToClient(m_hWnd, (POINT*) &rcCtrlPos);
		::ScreenToClient(m_hWnd, (POINT*) &rcCtrlPos.right);

		// Move control
		MoveWindow(hCtrl, 
				   rcCtrlPos.left + DeltaX,
				   rcCtrlPos.top + DeltaY,
				   rcCtrlPos.GetWidth(), 
				   rcCtrlPos.GetHeight(), 
				   TRUE);
	}

	void Win32_DialogImp::ResizeControl(ControlID ID, int Width, int Height)
	{
		ASSERT_VALID_WINDOW();

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		COMPLAIN_IF(hCtrl == NULL, "Control not found in ResizeControl().");

		// Get the control's position
		Rect rcCtrlPos;

#if 0
		::GetWindowRect(hCtrl, (RECT*) &rcCtrlPos);

		// Convert (top, left) to parent window co-ordinates
		::ScreenToClient(m_hWnd, (POINT*) &rcCtrlPos);

		// Resize control
		MoveWindow(hCtrl, rcCtrlPos.left, rcCtrlPos.top, Width, Height, TRUE);
#else
		GetControlPosition(ID, &rcCtrlPos);
		rcCtrlPos.SetWidth(Width);
		rcCtrlPos.SetHeight(Height);
		SetControlPosition(ID, rcCtrlPos);
#endif
	}

	void Win32_DialogImp::EnableControl(ControlID ID)
	{
		ASSERT_VALID_WINDOW();

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		COMPLAIN_IF(hCtrl == NULL, "Control not found in EnableControl().");

		// Enable the control
		::EnableWindow(hCtrl, TRUE);
	}

	void Win32_DialogImp::DisableControl(ControlID ID)
	{
		ASSERT_VALID_WINDOW();

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		COMPLAIN_IF(hCtrl == NULL, "Control not found in DisableControl().");

		// Disable the control
		::EnableWindow(hCtrl, FALSE);
	}

	void Win32_DialogImp::ShowControl(ControlID ID)
	{
		ASSERT_VALID_WINDOW();

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		COMPLAIN_IF(hCtrl == NULL, "Control not found in ShowControl().");

		::ShowWindow(hCtrl, SW_SHOW);
	}

	void Win32_DialogImp::HideControl(ControlID ID)
	{
		ASSERT_VALID_WINDOW();

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		COMPLAIN_IF(hCtrl == NULL, "Control not found in HideControl().");

		::ShowWindow(hCtrl, SW_HIDE);
	}

	void Win32_DialogImp::GetControlPosition(ControlID ID, Rect* pPosition)
	{
		ASSERT_VALID_WINDOW();

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		if (hCtrl != NULL)
		{
			// Get position in screen co-ordinates
			::GetWindowRect(hCtrl, (RECT*) pPosition);

			// Translate to parent window co-ordinates
			HWND hParent = ::GetParent(hCtrl);

			if (hParent != NULL)
			{
				::ScreenToClient(hParent, (POINT*) pPosition);
				::ScreenToClient(hParent, (POINT*) &(pPosition->right));
			}
		}
		else
		{
			COMPLAIN("Control not found in GetControlPosition().");
		}
	}

	void Win32_DialogImp::SetControlPosition(ControlID ID, const Rect& rcCtrlPos)
	{
		ASSERT_VALID_WINDOW();

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		COMPLAIN_IF(hCtrl == NULL, "Control not found in GetControlPosition().");

		// Reposition control
		MoveWindow(hCtrl, 
				   rcCtrlPos.left, rcCtrlPos.top, 
				   rcCtrlPos.GetWidth(), rcCtrlPos.GetHeight(), TRUE);
	}

	void Win32_DialogImp::SetControlState(ControlID ID, ControlState NewState)
	{
		ASSERT_VALID_WINDOW();

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		COMPLAIN_IF(hCtrl == NULL, "Control not found in SetControlState().");

		// Send message to set the state
		::SendMessage(hCtrl, BM_SETCHECK, GetWin32State(NewState), 0);
	}

	ControlState Win32_DialogImp::GetControlState(ControlID ID)
	{
		ASSERT_VALID_WINDOW_RETURN(cs_Undefined);

		// Get control handle
		HWND hCtrl = FindControl(m_hWnd, ID);

		COMPLAIN_IF(hCtrl == NULL, "Control not found in GetControlState().");

		// Send message to set the state
		switch (::SendMessage(hCtrl, BM_GETCHECK, 0, 0))
		{
			case BST_CHECKED:		return cs_Checked;
			case BST_UNCHECKED:		return cs_Unchecked;
			default:				return cs_Undefined;
		}
	}


	bool Win32_DialogImp::FindDialogControl(ControlID ID, 
											Win32_WindowImp** ppControl)
	{
		HWND hControl = FindControl(m_hWnd, ID);

		if (hControl != NULL)
		{
			// Is there a vGUI window attached?
			Win32_WindowImp *pWindow = GetImpFromHWND(hControl);

			if (pWindow != NULL)
			{
				// Yes, so we have a match
				*ppControl = pWindow;
				return true;
			}
		}

		// Control not found
		return false;
	}

	/**************************************************************************

	>	ControlTranslation Win32_DialogImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																		 TMessageType* pMsgType,
																		 LRESULT* pResult,
																		 MessageParams* pParams)

		Author:		Tim Browse
		Date:		22/02/2001
		Inputs:		See Win32_WindowImp::TranslateControlNotification().
		Outputs:	See Win32_WindowImp::TranslateControlNotification().
		Returns:	See Win32_WindowImp::TranslateControlNotification().

		Purpose:	Translate Win32 control notification messages into
					standardised vgui messages.

	**************************************************************************/

	ControlTranslation Win32_DialogImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																	 TMessageType* pMsgType,
																	 LRESULT* pResult,
																	 MessageParams* pParams)
	{
		if (MsgInfo.m_ControlID != 0)
		{
			// If this relates to a control we know about, get it to translate
			// the message.
			Win32_WindowImp* pControlWindow;
			if (FindDialogControl(MsgInfo.m_ControlID, &pControlWindow))
			{
				ControlTranslation TranslateResult = pControlWindow->TranslateControlNotification(MsgInfo, pMsgType, pResult, pParams);

				// As this is a dialog, if a specific return code is required, we
				// need to set this in the special DWLP_MSGRESULT field on the
				// window.  I found this via Google, and the Microsoft documentation
				// is a bit quiet on this subject.
				if (TranslateResult == ControlTranslation_UseReturnCode)
				{
					::SetWindowLongPtr(m_hWnd, DWLP_MSGRESULT, (LONG_PTR)*pResult);

					// Also, we need to change the result to 0 to indicate that
					// we processed the dialog message...the caller will pick up
					// the actual result code from the DWLP_MSGRESULT field on 
					// the dialog window.
					*pResult = 0;
				}

				return TranslateResult;
			}
		}

		// Otherwise, fall back to base class
		return Win32_WindowImp::TranslateControlNotification(MsgInfo, pMsgType, pResult, pParams);
	}

	void Win32_DialogImp::LockControlUpdate()
	{
		m_cUpdateLock++;
	}

	void Win32_DialogImp::UnlockControlUpdate()
	{
		m_cUpdateLock--;

		if (m_cUpdateLock == 0)
		{
			// End deferred window updates
			int cEntries = m_HDWPEntries.size();

			if (cEntries > 0)
			{
				for (int iEntry = cEntries-1; iEntry >= 0; iEntry--)
				{
					::EndDeferWindowPos(m_HDWPEntries[iEntry].m_hDWP);
				}

				// And now get rid of these elements
				m_HDWPEntries.clear();
			}
		}

		COMPLAIN_IF(m_cUpdateLock < 0, "UnlockControlUpdate() called too many times!");
	}

	int Win32_DialogImp::GetHWDPEntry(HWND hWindowToMove)
	{
		HWND hParent = ::GetParent(hWindowToMove);

		// Find this entry...
		int cEntries = m_HDWPEntries.size();

		for (int iEntry = 0; iEntry < cEntries; iEntry++)
		{
			if (m_HDWPEntries[iEntry].m_hParent == hParent)
			{
				// We already have an entry for this parent window
				return iEntry;
			}
		}

		// No entry yet, so make one.
		HDWPEntry NewEntry;
		NewEntry.m_hParent = hParent;

		// Begin deferred window update (number 10 is arbitrary)
		NewEntry.m_hDWP = ::BeginDeferWindowPos(10);

		// Add new entry to array, and return index of new entry to caller
		m_HDWPEntries.push_back(NewEntry);
		return cEntries;
	}

	/**************************************************************************
	
	>	DLGTEMPLATE* Win32_DialogImp::LockDialogResource(const DialogResourceID& ID)
	
		Author:		Tim Browse
		Date:		21/6/2002
		Inputs:		ID - the dialog template to load
		Returns:	The DLGTEMPLATE data, or NULL if the template could not be
					loaded.
	
		Purpose:	Load and lock the specified dialog template resource.
	
	**************************************************************************/
	
	DLGTEMPLATE* Win32_DialogImp::LockDialogResource(const DialogResourceID& ID)
	{ 
		// Load the dialog resource
		HRSRC hResource = ::FindResource(NULL, MAKEINTRESOURCE(ID.GetID()), RT_DIALOG); 

		if (hResource == NULL)
		{
			COMPLAIN("TabControl - failed to find the specified dialog template!");
			return NULL;
		}

		
		HGLOBAL hGlobal = ::LoadResource(GetWin32App()->GetInstance(), hResource);

		if (hGlobal == NULL)
		{
			COMPLAIN("TabControl - unable to load the specified dialog template!");
			return NULL;
		}

		// Otherwise, we lock the resource
		return (DLGTEMPLATE *) ::LockResource(hGlobal); 
	} 

	bool Win32_DialogImp::FindControlWindow(ControlID ID, Window** ppWindow)
	{
		HWND hControl = FindControl(m_hWnd, ID);

		if (hControl == NULL)
		{
			// Control not found
			return false;
		}

		// Ok, try to get the window pointer
		*ppWindow = GetWindowFromHWND(hControl);
		return (*ppWindow != NULL);
	}

	void Win32_DialogImp::GetGripperPos(Rect* pGripperPos)
	{
		GetClientArea(pGripperPos);
		pGripperPos->left = pGripperPos->right - ::GetSystemMetrics(SM_CXHSCROLL);
		pGripperPos->top = pGripperPos->bottom - ::GetSystemMetrics(SM_CYVSCROLL);
	}

	void Win32_DialogImp::CreateResizeGripper()
	{
		// We only create this if the user wants us to
		if (!m_bUseAutoResizeGripper)
			return;

		m_pResizeGripper = new Win32_StaticControl;

		Rect GripperPos;
		GetGripperPos(&GripperPos);

		// Increase window to allow gripper to fit
		Rect ClientArea;
		GetClientArea(&ClientArea);
		Rect ControlArea;
		GetControlArea(&ControlArea);
		Rect NewClientArea(ClientArea);
		NewClientArea.bottom = max(ControlArea.bottom + GripperPos.GetHeight() + 1, NewClientArea.bottom);

		// YDelta is how much we've actually changed the height by
		int YDelta = NewClientArea.bottom - ClientArea.bottom;

		// Let dialog object know about this change in size, so that
		// resizes work correctly.
		GetDialog()->AlterInitialSizeBy(0, YDelta);

		// Ok, now actually resize the dialog
		SetClientArea(NewClientArea);

		GetGripperPos(&GripperPos);
		m_pResizeGripper->Create(m_hWnd, GripperPos, kGripperID);
		m_pResizeGripper->Display(StaticControlImage_ResizeGripper);

		// Make sure gripper is at the bottom wrt z-order.
		::SetWindowPos(m_pResizeGripper->GetHWND(), HWND_BOTTOM, 0, 0, 0, 0,
					   SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOSIZE);

		// Disable input for this control - we don't want to receive any
		// kind of mouse input.
		//
		// This is important, because we don't want to steal WM_NCHITTEST 
		// messages from the dialog window - we want this control to show 
		// up as a resize area of the dialog.
		//
		// Disabling the window ensures the WM_NCHITTEST messages go to
		// the parent window (i.e. the dialog)
		::EnableWindow(m_pResizeGripper->GetHWND(), FALSE);
	}

	/**************************************************************************
	
	>	void Win32_DialogImp::GetControlArea(Rect* pControlArea)
	
		Author:		Tim Browse
		Date:		25/8/2002
		Outputs:	pControlArea - the area occupied by controls in this
								   dialog.
	
		Purpose:	Find the combined area occupied by all controls (child
					windows) in this dialog.
	
	**************************************************************************/
	
	void Win32_DialogImp::GetControlArea(Rect* pControlArea)
	{
		pControlArea->Empty();

		HWND hChild = ::GetWindow(m_hWnd, GW_CHILD);

		Rect ControlPosition;

		while (hChild != NULL)
		{
			Win32_WindowImp::GetPosition(hChild, &ControlPosition);
			pControlArea->Union(*pControlArea, ControlPosition);

			hChild = ::GetWindow(hChild, GW_HWNDNEXT);
		}
	}

	/**************************************************************************
	
	>	void Win32_DialogImp::Move(const Rect& NewPosition, bool bForceRedraw)
	
		Author:		Tim Browse
		Date:		10/9/2002
		Inputs:		NewPosition - the new position to move the dialog to
					bForceRedraw - whether or not to force a redraw of the window
	
		Purpose:	Move the dialog to the specified position.

					This is over-ridden from Win32_WindowImp, to make sure
					if OnInitDialog() moves the dialog, then we don't try
					to centre it on screen when we show it (as centred is
					the default option).
	
	**************************************************************************/
	
	void Win32_DialogImp::Move(const Rect& NewPosition, bool bForceRedraw)
	{
		// Get base class to position the window...
		Win32_WindowImp::Move(NewPosition, bForceRedraw);

		// And make sure we don't try to move the dialog when it is shown
		// (if it's already shown, this won't have any effect)
		m_OpenPos = wp_Default;
	}

	/**************************************************************************
	
	>	void Win32_DialogImp::DisableAutoResizeGripper()
	
		Author:		Tim Browse
		Date:		11/9/2002
	
		Purpose:	Specified that the dialog should not add a resize gripper
					to this dialog, even if it is a resizable dialog.

					NB. Must be called in OnInitDialog(), or it will have
						no effect.

	**************************************************************************/
	
	void Win32_DialogImp::DisableAutoResizeGripper()
	{
		m_bUseAutoResizeGripper = false;
	}
};

