//
// Win32_DropDownListImp.cpp - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_DropDownListImp.h"

#include "Win32_Headers.h"
#include "Win32_MessageInfo.h"

#include "..\Errors.h"

#include "vos\Trace.h"

//
// Macro to check for valid window handle/state
//
#define ASSERT_VALID_WINDOW()							\
		if (!IsValid())									\
		{												\
			COMPLAIN("Invalid window handle used!");	\
			return;										\
		}

#define ASSERT_VALID_WINDOW_RETURN(RetVal)				\
		if (!IsValid())									\
		{												\
			COMPLAIN("Invalid window handle used!");	\
			return (RetVal);							\
		}


namespace vgui
{
	//
	// All Win32 comboboxes create their child edit controls
	// with an ID of 1001.
	//
	const int kComboBoxEditControlID = 1001;

	Win32_DropDownListImp::Win32_DropDownListImp()

		: m_bUseHistoryBehaviour(false),
		  m_bHasEditBox(false),
		  m_hEditBox(NULL),
		  m_bEditBoxContentsValid(false)
	{
		// TCILB
	}

	void Win32_DropDownListImp::Attach(HWND hWnd)
	{
		// We have to check that the class name is correct.
		if (!IsWindowOfClass(hWnd, "COMBOBOX"))
		{
			// Trying to attach to the wrong kind of control - it must be a combobox.
			vguiERROR();
			return;
		}

		// Ok, we allow it, so pass down to base class.
		DropDownListImp::Attach(hWnd);

		// Do we have an edit box?
		if (::GetWindowLong(hWnd, GWL_STYLE) & (CBS_DROPDOWN|CBS_SIMPLE))
		{
			// Yes, this type of combobox has an edit box in the control.
			m_bHasEditBox = true;

			// Find the edit box window
			m_hEditBox = ::GetDlgItem(hWnd, kComboBoxEditControlID);
			if (m_hEditBox != NULL)
			{
				// We have an edit box window handle, so make a new edit
				// box object, and attach it, so we can receive its 
				// notification messages.
				//
				// NB. We use AttachByHandle() rather than Attach(),
				//	   because we can't use the control ID to uniquely
				//	   identify the editbox, as all combo boxes create
				//	   their edit boxes with the same control ID (1001)
				m_EditBox.AttachByHandle(m_pParent, m_hEditBox);

				// Vector notification messages to our parent, otherwise
				// they'll never be processed by anything.
				// Note that we vector these messages using our own
				// control ID, so the messages appear to be from the
				// DropDownList, as any sane person would expect.
				m_EditBox.SendNotificationsTo(m_pParent, GetControlID());
			}
		}
	}

	void Win32_DropDownListImp::Clear()
	{
		// Special case for history mode
		if (m_bUseHistoryBehaviour)
		{
			m_History.clear();
		}
		else
		{
			RawClear();
		}
	}

	void Win32_DropDownListImp::RawClear()
	{
		// Keep deleting items until they're all gone
		while (SendMessage(CB_DELETESTRING, 0, 0) != CB_ERR)
		{
			// empty
		}
	}

	void Win32_DropDownListImp::ClearTextSelection()
	{
		SendMessage(CB_SETEDITSEL, 0, MAKELONG(-1, 0));
	}


	// Add and delete items
	bool Win32_DropDownListImp::AddItem(const String& ItemName, 
										int* piItem,
										UserData TheUserData)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Special case for history mode
		if (m_bUseHistoryBehaviour)
		{
			return m_History.AddItem(ItemName, TheUserData, piItem);
		}
		else
		{
			return AddRawItem(ItemName, piItem, TheUserData);
		}
	}

	bool Win32_DropDownListImp::AddRawItem(const String& ItemName, 
										   int* piItem,
										   UserData TheUserData)
	{
		// Add the string
		int iNewItem = SendMessage(CB_ADDSTRING, 0, (LPARAM) ItemName.c_str());

		// Check for problems
		if (iNewItem < 0)
		{
			// Negative index => error

		#ifdef _DEBUG
			// Make sure people check for errors by setting index to 
			// an outlandish value so they're more likely to notice it
			// in a debugger.  Step through your code (tm) :-)
			if (piItem != NULL)
				*piItem = 999999;
		#endif

			return false;
		}

		// Ok, we managed to insert the item, so set the user data
		if (SendMessage(CB_SETITEMDATA, iNewItem, (LPARAM) TheUserData) == CB_ERR)
		{
			// Something went wrong.
		#ifdef _DEBUG
			// See above.
			if (piItem != NULL)
				*piItem = 999999;
		#endif

			return false;
		}

		// If we get this far, it's all worked
		if (piItem != NULL)
			*piItem = iNewItem;
		return true;
	}

	bool Win32_DropDownListImp::InsertItemAfter(int iAfterItem,
												const String& ItemName, 
												int* piItem,
												UserData TheUserData)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Add the string - InsertItemAt takes the desired index of
		// the new string, so we add 1 to the value passed in to insert
		// it after the specified item.
		return InsertItemAt(iAfterItem + 1, ItemName, piItem, TheUserData);
	}

	bool Win32_DropDownListImp::InsertItemBefore(int iBeforeItem,
												 const String& ItemName, 
												 int* piItem,
												 UserData TheUserData)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		return InsertItemAt(iBeforeItem, ItemName, piItem, TheUserData);
	}

	bool Win32_DropDownListImp::InsertItemAt(int iInsertAt,
										     const String& ItemName, 
											 int* piItem,
											 UserData TheUserData)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Special case for history mode
		if (m_bUseHistoryBehaviour)
		{
			return m_History.InsertItemAt(iInsertAt, ItemName, TheUserData, piItem);
		}
		
		int iNewItem = SendMessage(CB_INSERTSTRING, 
								   iInsertAt,
								   (LPARAM) ItemName.c_str());

		// Check for problems
		if (iNewItem < 0)
		{
			// Negative index => error

		#ifdef _DEBUG
			// Make sure people check for errors by setting index to 
			// an outlandish value so they're more likely to notice it
			// in a debugger.  Step through your code (tm) :-)
			if (piItem != NULL)
				*piItem = 999999;
		#endif

			return false;
		}

		// Ok, we managed to insert the item, so set the user data
		if (SendMessage(CB_SETITEMDATA, iNewItem, (LPARAM) TheUserData) == CB_ERR)
		{
			// Something went wrong.
		#ifdef _DEBUG
			// See above.
			if (piItem != NULL)
				*piItem = 999999;
		#endif

			return false;
		}

		// If we get this far, it's all worked
		if (piItem != NULL)
			*piItem = iNewItem;
		return true;
	}

	bool Win32_DropDownListImp::DeleteItem(int iItem)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Special case for history mode
		if (m_bUseHistoryBehaviour)
		{
			return m_History.DeleteItem(iItem);
		}
		
		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		// Negative values (e.g. CB_ERR) mean there was a problem.
		bool bSuccess = (SendMessage(CB_DELETESTRING, iItem, 0) >= 0);

		// Deleting non-existent items is something we want to know about.
		ASSERT(bSuccess);

		return bSuccess;
	}

	bool Win32_DropDownListImp::SetItemText(int iItem, const String& Text, int iColumn)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		// Make sure no-one's trying any multi-column stuff
		ASSERT(iColumn == 0);

		// Special case for history mode
		if (m_bUseHistoryBehaviour)
		{
			return m_History.SetItemText(iItem, Text);
		}
		
		// List control does not support changing item text, so we add a new
		// item and delete the old one.

		// First, get the user data
		UserData TheUserData;
		if (!GetItemUserData(iItem, &TheUserData))
		{
			// Unable to access item
			vguiERROR();
			return false;
		}

		// Insert new item before old one
		int iNewItem;
		if (!InsertItemAt(iItem, Text, &iNewItem, TheUserData))
		{
			// Unable to insert new item
			vguiERROR();
			return false;
		}

		// New item should be inserted where the old one was
		ASSERT(iItem == iNewItem);

		// Delete the old one
		return DeleteItem(iItem + 1);
	}

	bool Win32_DropDownListImp::SetItemData(int iItem, UserData TheUserData)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Special case for history mode
		if (m_bUseHistoryBehaviour)
		{
			return m_History.SetItemData(iItem, TheUserData);
		}
		
		if (SendMessage(CB_SETITEMDATA, iItem, (LPARAM) TheUserData) == CB_ERR)
		{
			// Something went wrong.
			return false;
		}

		// All ok
		return false;
	}


	int Win32_DropDownListImp::GetCount() const
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Special case for history mode
		if (m_bUseHistoryBehaviour)
		{
			return m_History.GetCount();
		}
		else
		{
			return GetRawCount();
		}
	}

	int Win32_DropDownListImp::GetRawCount() const
	{
		// Nice simple case - but it can still error according to Win32 docs...
		int cItems = (int) SendMessage(CB_GETCOUNT, 0, 0);

		if (cItems < 0)
		{
			// Just tell them there aren't any, but shout about it.
			vguiERROR();
			return 0;
		}

		// Return actual count
		return cItems;
	}

	// Find info about an item
	bool Win32_DropDownListImp::GetItemText(int iItem, String* pItemName, int iColumn) const
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		// Make sure no-one's trying any multi-column stuff
		ASSERT(iColumn == 0);

		// Special case for history mode
		if (m_bUseHistoryBehaviour)
		{
			return m_History.GetItemText(iItem, pItemName);
		}
		else
		{
			return GetRawItemText(iItem, pItemName, iColumn);
		}
	}
	
	bool Win32_DropDownListImp::GetRawItemText(int iItem, String* pItemName, int iColumn) const
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetRawCount());

		// Make sure no-one's trying any multi-column stuff
		ASSERT(iColumn == 0);

		// Use a buffer on the stack for most purposes
		const int kBufferLength = 80;

		// Find out length of string (add 1 because the control doesn't
		// include the null terminator in the length calculation).
		int cTextLength = SendMessage(CB_GETLBTEXTLEN, iItem, 0) + 1;

		// Check for errors
		if (cTextLength < 0)
		{
			// Error - complain about this as it's probably an 
			// algorithmic error from the client.
			vguiERROR();
			return false;
		}

		bool bSuccess = false;

		// Is our on-stack buffer sufficient?
		if (cTextLength <= kBufferLength)
		{
			// Yes, so just get the text.
			TCHAR NameBuffer[kBufferLength];

			if (SendMessage(CB_GETLBTEXT, iItem, (LPARAM) NameBuffer) >= 0)
			{
				// It worked - assign name to caller's string
				*pItemName = NameBuffer;
				bSuccess = true;
			}
		}
		else
		{
			// We need a bigger buffer so dynamically allocate one
			// big enough to cope with exceptional cases.
			TCHAR *NameBuffer = new TCHAR[cTextLength];

			// Extract the item string into this buffer
			bSuccess = (SendMessage(CB_GETLBTEXT, iItem, (LPARAM) NameBuffer) >= 0);

			// And assign it to the caller's string.
			*pItemName = NameBuffer;

			// De-allocate the buffer
			delete [] NameBuffer;
		}

		// Tell caller if it worked
		return bSuccess;
	}

	bool Win32_DropDownListImp::GetItemUserData(int iItem, UserData* pTheUserData) const
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		// Special case for history mode
		if (m_bUseHistoryBehaviour)
		{
			return m_History.GetItemUserData(iItem, pTheUserData);
		}
		else
		{
			return GetRawItemUserData(iItem, pTheUserData);
		}
	}
	
	bool Win32_DropDownListImp::GetRawItemUserData(int iItem, UserData* pTheUserData) const
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetRawCount());

		// Ask control for the data
		*pTheUserData = (UserData) SendMessage(CB_GETITEMDATA, iItem, 0);

		// Win32 has no way of detecting an error for this function, except
		// by returning the value CB_ERR - however, this is a legal value
		// for user data, so it's no help at all.
		//
		// Bet those bool returns on all the functions in this class don't
		// look so stupid now, huh?
		return true;
	}

	// Change currently selected item
	bool Win32_DropDownListImp::SetSelectedItem(int iItem)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		int iSelectedItem = (int) SendMessage(CB_SETCURSEL, iItem, 0);

		if (iSelectedItem >= 0)
		{
			// Tell caller it worked.
			return true;
		}
		else
		{
			// Some kind of failure - illegal index? - so complain.
			vguiERROR();
			return false;
		}
	}

	// Access currently selected item
	bool Win32_DropDownListImp::GetSelectedItem(int* piSelectedItem) const
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Ask control for current selection index
		int iItem = (int) SendMessage(CB_GETCURSEL, 0, 0);

		if (iItem >= 0)
		{
			*piSelectedItem = iItem;
			return true;
		}
		else
		{
			// Some kind of failure, or no item is selected.
			return false;
		}
	}

	// Find an item
	bool Win32_DropDownListImp::FindItem(const String& ItemName, int* piFoundItem)
	{
		// Special case for history mode
		if (m_bUseHistoryBehaviour)
		{
			return m_History.FindItem(ItemName, piFoundItem);
		}
		else
		{
			// Start search from beginning
			return FindItem(ItemName, piFoundItem, -1);
		}
	}

	bool Win32_DropDownListImp::FindItem(const String& ItemName, 
										 int* piFoundItem,
										 int iStartFrom)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		int iItem = (int) SendMessage(CB_FINDSTRINGEXACT, 
									  iStartFrom,		
									  (LPARAM) ItemName.c_str());

		if (iItem >= 0)
		{
			// Success
			*piFoundItem = iItem;
			return true;
		}
		else
		{
			// Failure
			return false;
		}
	}

	/**************************************************************************

	>	void Win32_DropDownListImp::RemeasureItems()

		Author:		Tim Browse
		Date:		24/03/2001

		Purpose:	Force an update of all list items in an owner-draw list 
					box, so that they are sized correctly.

					(See below for faster version of this function)

	**************************************************************************/

	void Win32_DropDownListImp::RemeasureItems()
	{
		// UNIMPLEMENTED - currently we don't support owner-draw drop down lists.
		//
		// See Win32_ListBoxImp::RemeasureItems() for implementation code to
		// steal when we do.
		COMPLAIN("Win32_DropDownListImp::RemeasureItem() - UNIMPLEMENTED!");
	}

	/**************************************************************************

	>	void Win32_DropDownListImp::RemeasureItems(int NewHeight)

		Author:		Tim Browse
		Date:		24/03/2001
		Inputs:		NewHeight - the new height to use for all items in the list.

		Purpose:	Force an update of all list items in an owner-draw list 
					box, so that they are sized correctly.

					This version is faster, because the height is specified
					directly, rather than being determined by calling
					OnMeasureItem() for each item (unlike the function above).

	**************************************************************************/

	void Win32_DropDownListImp::RemeasureItems(int NewHeight)
	{
		// UNIMPLEMENTED - currently we don't support owner-draw drop down lists.
		//
		// See Win32_ListBoxImp::RemeasureItems() for implementation code to
		// steal when we do.
		COMPLAIN("Win32_DropDownListImp::RemeasureItem() - UNIMPLEMENTED!");
	}


	/**************************************************************************
	
	>	void Win32_DropDownListImp::EnableSorting()
	
		Author:		Tim Browse
		Date:		6/4/2002
	
		Purpose:	Enable sorting on this drop down list.  The items in the
					list will be displayed in a sorted order.
	
	**************************************************************************/
	
	void Win32_DropDownListImp::EnableSorting()
	{
		HWND hWnd = GetHWND();

		// Set the window style to specify sorting
		LONG lStyle = ::GetWindowLong(hWnd, GWL_STYLE);
		LONG lNewStyle = lStyle | CBS_SORT;
		if (lNewStyle != lStyle)
		{
			::SetWindowLong(hWnd, GWL_STYLE, lNewStyle);
		}
	}

	/**************************************************************************
	
	>	void Win32_DropDownListImp::DisableSorting()
	
		Author:		Tim Browse
		Date:		6/4/2002

		Purpose:	Disable sorting on this list - items will be displayed
					in the order they were added.
	
	**************************************************************************/
	
	void Win32_DropDownListImp::DisableSorting()
	{
		HWND hWnd = GetHWND();

		// Set the window style to specify no sorting
		LONG lStyle = ::GetWindowLong(hWnd, GWL_STYLE);
		LONG lNewStyle = lStyle & ~CBS_SORT;
		if (lNewStyle != lStyle)
		{
			::SetWindowLong(hWnd, GWL_STYLE, lNewStyle);
		}
	}	

	/**************************************************************************

	>	ControlTranslation Win32_DropDownListImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																			   TMessageType* pMsgType,
																			   LRESULT* pResult,
																			   MessageParams* pParams)

		Author:		Tim Browse
		Date:		22/02/2001
		Inputs:		See Win32_WindowImp::TranslateControlNotification().
		Outputs:	See Win32_WindowImp::TranslateControlNotification().
		Returns:	See Win32_WindowImp::TranslateControlNotification().

		Purpose:	Translate Win32 combobox notification messages into
					standardised vgui messages.

	**************************************************************************/

	ControlTranslation 
		Win32_DropDownListImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
															TMessageType* pMsgType,
															LRESULT* pResult,
															MessageParams* pParams)
	{
		switch (MsgInfo.m_uNotifyCode)
		{
			// User changes the edit box directly
			case CBN_EDITCHANGE:
			{
				// Force full window text retrieval the next time the
				// edit box contents are required.
				m_bEditBoxContentsValid = false;

				*pMsgType = mt_DLN_EDITBOXCHANGED;	

				// TODO: work out which of these is correct!
				// BODGE
				return ControlTranslation_Unknown;
				return ControlTranslation_OK;
			}

			case CBN_DROPDOWN:
			{
				if (m_bUseHistoryBehaviour)
				{
					// List is about to be dropped down, so we transfer
					// the up-to-date history list to the control.
					TransferHistoryToList();
				}
				break;
			}

			case CBN_SELCHANGE:
			{
				if (m_bHasEditBox)
				{
					// User has changed selection, so update our idea of
					// what the edit box contents are.
					//
					// NB. This is the key reason we need m_EditBoxContents,
					//	   because if you just call GetWindowText() here, you
					//	   won't get the right results.
					int iSelectedItem;
					if (GetSelectedItem(&iSelectedItem))
					{
						GetRawItemText(iSelectedItem, &m_EditBoxContents, 0);
						m_bEditBoxContentsValid = true;
					}

					// Ok, do we need to update the history?
					if (m_bUseHistoryBehaviour)
					{
						CommitToHistory(m_EditBoxContents);
					}

					// Translate to standard vGUI message
					*pMsgType = mt_DLN_SELCHANGED;	
					return ControlTranslation_OK;
				}

				break;
			}
		}

		// Message not recognised
		return ControlTranslation_Unknown;
	}

	/**************************************************************************
	
	>	void Win32_DropDownListImp::EnableHistoryBehaviour(bool bEnable)
	
		Author:		Tim Browse
		Date:		1/6/2002
		Inputs:		bEnable - true => enable history behaviour
							 false => disable history behaviour
	
		Purpose:	Determine whether or not this drop down list uses
					'history' behaviour (see DropDownList)
	
	**************************************************************************/
	
	void Win32_DropDownListImp::EnableHistoryBehaviour(bool bEnable)
	{
		// Ignore NOPs
		if (m_bUseHistoryBehaviour == bEnable)
			return;

		m_bUseHistoryBehaviour = bEnable;

		if (m_bUseHistoryBehaviour)
		{
			// If we have history behaviour, we want to have things
			// displayed in the explicit order we use, rather than
			// being sorted
			DisableSorting();

			// Ok, now transfer list contents to the history list
			TransferListToHistory();
		}
		else
		{
			// History behaviour has been disabled - copy the history
			// list to the list control, as we won't do this when the
			// list is dropped down anymore.
			TransferHistoryToList();
		}
	}

	/**************************************************************************
	
	>	void Win32_DropDownListImp::TransferListToHistory()
	
		Author:		Tim Browse
		Date:		27/7/2002
	
		Purpose:	Copy the contents of the list control to the history
					list.
	
	**************************************************************************/
	
	void Win32_DropDownListImp::TransferListToHistory()
	{
		m_History.clear();

		// Work out how many items to transfer
		int cItems = GetRawCount();

		int cMaxItems = m_History.GetHistoryLimit();

		if (cItems > cMaxItems)
			cItems = cMaxItems;

		for (int iItem = 0; iItem < cItems; iItem++)
		{
			vos::String Text;
			UserData TheUserData;

			if (GetRawItemText(iItem, &Text) && GetRawItemUserData(iItem, &TheUserData))
			{
				m_History.AddItem(Text, TheUserData);
			}
		}
	}

	/**************************************************************************
	
	>	void Win32_DropDownListImp::TransferHistoryToList()
	
		Author:		Tim Browse
		Date:		27/7/2002
	
		Purpose:	Copy the contents of the history list to the list control.
	
	**************************************************************************/
	
	void Win32_DropDownListImp::TransferHistoryToList()
	{
		// Get current contents of edit box (because the combobox is
		// about to delete it, for reasons best known to itself)
		GetText(m_EditBoxContents);

		// Clear the contents of the list
		RawClear();

		// Now copy each item from the history list to the list control
		int cItems = m_History.GetCount();

		for (int iItem = 0; iItem < cItems; iItem++)
		{
			vos::String Text;
			UserData TheUserData;

			if (m_History.GetItemText(iItem, &Text) && 
				m_History.GetItemUserData(iItem, &TheUserData))
			{
				AddRawItem(Text, NULL, TheUserData);
			}
		}

		if (m_History.GetCount() > 0)
		{
			// Now set text to be first entry in history list (this solves
			// all sorts of awkward problems)
			m_History.GetItemText(0, &m_EditBoxContents);
		}
		else
		{
			// If history is empty, leave the text alone
		}

		SetText(m_EditBoxContents);
	}

	/**************************************************************************
	
	>	void Win32_DropDownListImp::CommitToHistory(const vos::String& Text)
	
		Author:		Tim Browse
		Date:		1/6/2002
		Inputs:		Text - the text to add to the history
	
		Purpose:	Add the specified text to the control's history.

					What this actually does is remove any existing instance
					of this text, and then adds it to the head of the control.
	
	**************************************************************************/
	
	void Win32_DropDownListImp::CommitToHistory(const vos::String& Text)
	{
		if (!m_bUseHistoryBehaviour)
		{
			COMPLAIN("CommitToHistory(): history behaviour not enabled - "
					 "call EnableHistoryBehaviour()!");
			return;
		}

		m_History.AddToHistory(Text);
	}

	/**************************************************************************
	
	>	void Win32_DropDownListImp::SetHistoryLimit(int Limit)
	
		Author:		Tim Browse
		Date:		1/6/2002
		Inputs:		Limit - max number of items to show
	
		Purpose:	Set the maximum number of items to show in the list
					history.
	
	**************************************************************************/
	
	void Win32_DropDownListImp::SetHistoryLimit(int Limit)
	{
		if (!m_bUseHistoryBehaviour)
		{
			COMPLAIN("SetHistoryLimit(): history behaviour not enabled - "
					 "call EnableHistoryBehaviour()!");
			return;
		}

		m_History.SetHistoryLimit(Limit);
	}

	/**************************************************************************
	
	>	void Win32_DropDownListImp::EnableCommitNotification(bool bEnable)
	
		Author:		Tim Browse
		Date:		25/7/2002
		Inputs:		bEnable - true => when commit (return) is activated, then
									  send the message to the DropDownList.
							 false => let the control handle commits
	
		Purpose:	This function copes with the idiosyncratic design of
					the standard combo box - the edit box is part of the
					control, but it's also not; if you want to alter its
					behaviour, you have to extract it from the combobox
					and subclass it directly, rather than doing it all
					via the combobox interface.

					Hence we tell our embedded edit box about the change
					in commit behaviour here.  This function is called
					whenever the DropDownList object has 
					EnableCommitNotification() called on it (and we have 
					an editable combo box).
	
	**************************************************************************/
	
	void Win32_DropDownListImp::EnableCommitNotification(bool bEnable)
	{
		// If we have an edit box, keep it up to date
		if (m_hEditBox != NULL)
		{
			m_EditBox.EnableCommitNotification(bEnable);
		}
	}

	bool Win32_DropDownListImp::OnTextCommit()
	{
		// Propogate the text commit message from the combo's
		// edit box up to the parent of the list control
		return m_pParent->OnMessage(mt_EBN_TEXTCOMMIT, 0, GetControlID());
	}

	bool Win32_DropDownListImp::SetText(const String& Text)
	{
		// Just a wrapper for the base class, so we know when the edit
		// box contents change.
		if (DropDownListImp::SetText(Text))
		{
			if (m_bHasEditBox)
			{
				m_EditBoxContents = Text;
			}

			return true;
		}
		else
		{
			return false;
		}
	}

	bool Win32_DropDownListImp::GetText(String& Text)
	{
		// We keep our own idea of what the edit box contents are, because
		// the combo box is crap at this.  e.g. it will send you a CBN_SELCHANGE
		// message, and then you call GetWindowText() to get the edit box
		// text, and it returns the old value, as it hasn't been updated yet.
		//
		// Fscking Microsoft interns... :-(
		//
		if (m_bHasEditBox)
		{
			if (!m_bEditBoxContentsValid)
			{
				if (DropDownListImp::GetText(m_EditBoxContents))
				{
					m_bEditBoxContentsValid = true;
				}
				else
				{
					return false;
				}
			}

			Text = m_EditBoxContents;
			return true;
		}
		else
		{
			return DropDownListImp::GetText(Text);
		}
	}

	void Win32_DropDownListImp::OnControlResizing()
	{
		if (m_bHasEditBox)
		{
			m_EditBox.OnControlResizing();
		}
	}

	void Win32_DropDownListImp::OnControlResize()
	{
		if (m_bHasEditBox)
		{
			m_EditBox.OnControlResize();
		}
	}

	void Win32_DropDownListImp::InsertTextAtCursor(const String& Text,
												   EditBoxBehaviour Behaviour)
	{
		if (m_bHasEditBox)
		{
			m_EditBox.InsertTextAtCursor(Text, Behaviour);
		}
		else
		{
			COMPLAIN("InsertTextAtCursor() - list box does not have editable text field!");
		}
	}

};
