//
// Win32_EditBoxImp.cpp - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_EditBoxImp.h"

#include "Win32_MessageInfo.h"

#include "vgui\Errors.h"

#include "vos\Trace.h"

namespace vgui
{
	Win32_EditBoxImp::Win32_EditBoxImp() : m_bResizing(false),
										   m_iSelectionStart(0),
										   m_iSelectionEnd(0)
	{
		// TCILB
	}

	void Win32_EditBoxImp::Attach(HWND hWnd)
	{
		// We have to check that the class name is correct.
		if (!IsWindowOfClass(hWnd, "EDIT"))
		{
			// Trying to attach to the wrong kind of control - it must be an edit field.
			vguiERROR();
			return;
		}

		// Ok, we allow it, so pass down to base class.
		EditBoxImp::Attach(hWnd);
	}


	/**************************************************************************
	
	>	bool Win32_EditBoxImp::SetText(const String& Text)
	
		Author:		Tim Browse
		Date:		10/9/2002
		Inputs:		Text - the new text to display in the edit box
		Returns:	true => text was updated
					false => a problem occurred
	
		Purpose:	Set the text in a Win32 edit box - by default, edit boxes
					will select all the text, which is ugly and not usually
					what is desired, so we clear the selection and put the cursor
					at the end of the the text (or at the start of the text for
					multi-line edit boxes).
	
	**************************************************************************/
	
	bool Win32_EditBoxImp::SetText(const String& Text)
	{
		if (!Win32_WindowImp::SetText(Text))
		{
			return false;
		}

		// Edit boxes annoyingly select new text, so we clear the
		// selection and put the cursor at the end for single line
		// edit boxes, and at the start for multi-line ones.
		if (::GetWindowLong(m_hWnd, GWL_STYLE) & ES_MULTILINE)
		{
			SendMessage(EM_SETSEL, -1, 0);
		}
		else
		{
			int cChars = SendMessage(WM_GETTEXTLENGTH, 0, 0);
			SendMessage(EM_SETSEL, cChars, cChars);
		}

		return true;
	}

	/**************************************************************************

	>	void Win32_EditBoxImp::AppendText(const String& Text)

		Author:		Tim Browse
		Date:		23/02/2001
		Inputs:		Text - the text to append

		Purpose:	Allow text to be appended to an edit box's text.

	**************************************************************************/

	void Win32_EditBoxImp::AppendText(const String& Text)
	{
		// The most efficient way to append text is to make the selection
		// be the end of the current text, and replace the selection with the new
		// text.
		int cChars = SendMessage(WM_GETTEXTLENGTH, 0, 0);

		// Ok, set the selection to be at the end of the text
		SendMessage(EM_SETSEL, cChars, cChars);

		// Replace this with the new text
		SendMessage(EM_REPLACESEL, FALSE, (LPARAM) Text.c_str());
	}

	/**************************************************************************
	
	>	void Win32_EditBoxImp::InsertTextAtCursor(const String& Text,
												  EditBoxBehaviour Behaviour = eb_Normal)
	
		Author:		Tim Browse
		Date:		26/8/2002
		Inputs:		Text - the text to insert into the edit box.
					Behaviour - if eb_SetFocus, the control is also given focus
								and the cursor is positioned at the end of the
								new text
	
		Purpose:	Insert text at the current cursor position.  This will
					replace any currently selected text.
	
	**************************************************************************/
	
	void Win32_EditBoxImp::InsertTextAtCursor(const String& Text,
										      EditBoxBehaviour Behaviour)
	{
		// Replace selection with the new text

		// We check to see if the edit box has focus - if not, then the
		// selection has probably been removed, which we don't want, so
		// we restore it, using our saved selection which we got by
		// intercepting WM_KILLFOCUS - see WindowProc().

		if (::GetFocus() != m_hWnd)
		{
			SendMessage(EM_SETSEL, m_iSelectionStart, m_iSelectionEnd);
		}

		// Ok, replace the selection with the text
		SendMessage(EM_REPLACESEL, FALSE, (LPARAM) Text.c_str());

		if (Behaviour == eb_SetFocus)
		{
			// Ok, now give focus back to edit window and position cursor
			// after replaced text.
			SetFocus();
			int iCursorPos = m_iSelectionStart + Text.GetLength();
			SendMessage(EM_SETSEL, iCursorPos, iCursorPos);
		}
	}

	/**************************************************************************

	>	ControlTranslation Win32_EditBoxImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																		  TMessageType* pMsgType,
																		  LRESULT* pResult,
																		  MessageParams* pParams)

		Author:		Tim Browse
		Date:		22/02/2001
		Inputs:		uMsg - the Win32 notify code
					nCtrlID - the control that this message is for.
					lParam, wParam - the Win32 message data.
		Outputs:	The vgui notification message.

		Returns:	ControlTranslation_OK      - Message translated ok, and
												 it should be delivered.

					ControlTranslation_Unknown - Message not recognised 
												 by the control.

					ControlTranslation_Handled - Message recognised and 
												 automatically handled by 
												 control - do not pass the 
												 message on.

		Purpose:	Translate Win32 edit box notification messages into
					standardised vgui messages.

	**************************************************************************/

	ControlTranslation Win32_EditBoxImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																	  TMessageType* pMsgType,
																	  LRESULT* pResult,
																	  MessageParams* pParams)
	{
		switch (MsgInfo.m_uNotifyCode)
		{
			// Messages we understand...
			case EN_CHANGE:		*pMsgType = mt_EBN_TEXTCHANGED;		break;
			case EN_UPDATE:		*pMsgType = mt_EBN_TEXTCHANGING;	break;
			case EN_SETFOCUS:	*pMsgType = mt_EBN_GAINFOCUS;		break;
			case EN_KILLFOCUS:	*pMsgType = mt_EBN_LOSEFOCUS;		break;

			// Edit control notifications - untranslated as yet
			case EN_ERRSPACE:
			case EN_MAXTEXT:
			case EN_HSCROLL:
			case EN_VSCROLL:
			default:
				return ControlTranslation_Unknown;
				break;
		}

		// Message converted
		return ControlTranslation_OK;
	}

	bool Win32_EditBoxImp::Create(Window* pParent, 
								  const Rect& rcPosition,
								  ControlID ID)
	{
		// Access the Win32 object for the parent
		Win32_WindowImp* pParentImp = GetImp(pParent);

		if (pParentImp != NULL)
		{
			// Create an edit box in the specified window
			HWND hWnd = pParentImp->CreateChildControl(m_pWindow,
			 									    "EDIT", "",
												    WS_BORDER, 0,
													rcPosition,
												    NOCTL);
			if (hWnd != NULL)
			{
				Attach(hWnd);
				return true;
			}
		}

		// Failed to create window
		return false;
	}

	void Win32_EditBoxImp::OnControlResizing()
	{
		m_bResizing = true;
	}

	void Win32_EditBoxImp::OnControlResize()
	{
		m_bResizing = false;
	}

	LRESULT Win32_EditBoxImp::WindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		switch (uMsg)
		{
			case EM_SETSEL:
			{
				// We ignore text selection messages while resizing, unless
				// the control has focus.
				if (m_bResizing && (::GetFocus() != m_hWnd))
				{
					return 0;
				}

				break;
			}

			case WM_KILLFOCUS:
			{
				// Remember current selection (used by InsertTextAtCursor())
				SendMessage(EM_GETSEL, (WPARAM) &m_iSelectionStart, (LPARAM) &m_iSelectionEnd);
				break;
			}
		}

		return Win32_WindowImp::WindowProc(uMsg, wParam, lParam);
	}

};
