//
// Win32_FontImp.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_FontImp.h"

#include "Win32_Headers.h"

// Lose windows macro
#undef ERROR

#include "..\Errors.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	uint Win32_FontImp::m_LogPixelsY = 0;

	Win32_FontImp::Win32_FontImp() : m_hFont(NULL),
									 m_bOwnFont(false),
									 m_HeightInPixels(-1)
	{
	}

	Win32_FontImp::~Win32_FontImp()
	{
		if ((m_hFont != NULL) && (m_bOwnFont))
		{
			// Clean up the Win32 font object.
			::DeleteObject(m_hFont);
			m_hFont = NULL;
		}
	}

	bool Win32_FontImp::Create(const String& TypefaceName,
							   uint PointSize,
							   FontWeight Weight,
							   FontStyle Style,
							   FontFamily Family,
							   FontPitch Pitch)
	{
		// Fill in font definition
		LOGFONT lf;
		ZeroMemory(&lf, sizeof(lf));

		// 
		// Convert points to Win32 units
		//
		lf.lfHeight = ConvertPointsToUnits(PointSize);

		//
		// Convert font weight to Win32 values
		//
		lf.lfWeight = FW_NORMAL;

		switch (Weight)
		{
			case fw_Light:
				lf.lfWeight = FW_LIGHT;
				break;

			case fw_Bold:
				lf.lfWeight = FW_BOLD;
				break;
		}

		// 
		// Convert font style to Win32 flags
		//
		if (Style & fs_Italic)
		{
			lf.lfItalic = TRUE;
		}

		if (Style & fs_Underline)
		{
			lf.lfUnderline = TRUE;
		}

		//
		// Convert font family and pitch to Win32 values.
		//

		lf.lfPitchAndFamily = FF_DONTCARE;

		switch (Family)
		{
			case ff_Roman:
				lf.lfPitchAndFamily = FF_ROMAN;
				break;

			case ff_Swiss:
				lf.lfPitchAndFamily = FF_SWISS;
				break;
		}

		switch (Pitch)
		{
			case fp_DefaultPitch:
				lf.lfPitchAndFamily |= DEFAULT_PITCH;
				break;

			case fp_FixedPitch:
				lf.lfPitchAndFamily |= FIXED_PITCH;
				break;

			case fp_VariablePitch:
				lf.lfPitchAndFamily |= VARIABLE_PITCH;
				break;
		}

		//
		// Set up the typeface name
		strcpy(lf.lfFaceName, TypefaceName.c_str());

		//
		// Set up standard Win32 output parameters.
		//
		lf.lfCharSet = DEFAULT_CHARSET;
		lf.lfOutPrecision = OUT_TT_PRECIS;
		lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
		lf.lfQuality = PROOF_QUALITY;

		// Create the font
 		m_hFont = ::CreateFontIndirect(&lf);

		m_bOwnFont = true;

		// Tell caller if it worked
		return (m_hFont != NULL);
	}

	void Win32_FontImp::CalcLogPixelsY()
	{
		HDC hScreenDC = ::GetDC(NULL);

		m_LogPixelsY = ::GetDeviceCaps(hScreenDC, LOGPIXELSY);

		::ReleaseDC(NULL, hScreenDC);
	}

	void Win32_FontImp::CalcFontHeight()
	{
		if (m_hFont == NULL)
			return;

		// Get screen DC
		HDC hScreenDC = ::GetDC(NULL);

		// Select font, and ask how big it is
		HFONT hOldFont = (HFONT) ::SelectObject(hScreenDC, m_hFont);

		TEXTMETRIC tm;
		::GetTextMetrics(hScreenDC, &tm);

		// Restore DC
		::SelectObject(hScreenDC, hOldFont);
		::ReleaseDC(NULL, hScreenDC);

		// Remember font size
		m_HeightInPixels = tm.tmHeight;
	}

	uint Win32_FontImp::ConvertPointsToUnits(uint Points)
	{
		// If we haven't interrogated the display yet, find out its
		// pixel resolution.
		if (m_LogPixelsY == 0)
		{
			CalcLogPixelsY();
		}

		int Units = MulDiv(Points, m_LogPixelsY, 72);

		COMPLAIN_IF(Units < 0, "Illegal font size.");

		return (uint) Units;
	}


	void Win32_FontImp::Attach(HFONT hFont)
	{
		m_hFont = hFont;
		m_bOwnFont = false;
	}

	uint Win32_FontImp::GetHeightInPixels()
	{
		if (m_HeightInPixels < 0)
			CalcFontHeight();

		return m_HeightInPixels;
	}

	uint Win32_FontImp::GetHeightInPoints()
	{
		if (m_HeightInPixels < 0)
			CalcFontHeight();

		// Convert pixels to points
		return MulDiv(m_HeightInPixels, 72, m_LogPixelsY);
	}

};

