//
// Win32_GCImp.cpp - Win32 version of Graphics Context
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_GCImp.h"

#include "Win32_WindowImp.h"
#include "Win32_FontImp.h"

#include "Win32_Headers.h"

#include "..\Icon.h"

//
// Win32 GDI is required.
//
//#include <wingdi.h>

// Lose stupid Windows macro
#undef ERROR

#include "vOS\Errors.h"
#include "vOS\Trace.h"


namespace vgui
{
	//
	// class NullPen
	//
	// Simple class to encapsulate creation and destruction of the 
	// singleton null-pen object.
	//
	class NullPen
	{
		public:
			NullPen()
			{
				// Create NULL pen.
				m_hPen = ::CreatePen(PS_NULL, 0, RGB(0,0,0));
			}

			~NullPen()
			{
				::DeleteObject(m_hPen);
			}

		public:
			HPEN m_hPen;
	};

	//
	// Constructors
	//
	Win32_GCImp::Win32_GCImp() : m_hdc(NULL),
								 m_OwnDC(FALSE),
								 m_hWnd(NULL),
								 m_ScrollX(0),
								 m_ScrollY(0),
								 m_hPen(NULL),
								 m_hBrush(NULL),
								 m_hOldPen(NULL),
								 m_hOldBrush(NULL),
								 m_hOldFont(NULL),
								 m_bValidPen(false),
								 m_bValidBrush(false),
								 m_bValidFont(false),
								 m_bValidClipRegion(false),
								 m_bValidBgMode(false),
								 m_bValidText(false),
								 m_bOwnHDC(false)
	{
		ZeroMemory(&m_ScrollInfo, sizeof(m_ScrollInfo));
		m_ScrollInfo.cbSize = sizeof(m_ScrollInfo);
	}

	Win32_GCImp::~Win32_GCImp()
	{
		// If the GC is attached to a window, then release the Win32 DC.
		if ((m_hdc != NULL) && (m_hWnd != NULL))
		{
			Detach();
		}
		else
		{
			// Just clear the DC value
			m_hdc = NULL;
		}
	}

	void Win32_GCImp::Attach(HDC hdc, HWND hWnd)
	{
		// Attach to specific HDC, not a window.
		m_hdc = hdc;
		m_hWnd = hWnd;

		// We should simply detach from this HDC when requested, without
		// releasing it.
		m_bOwnHDC = false;
	}

	//
	// Drawing Attributes
	//

	void Win32_GCImp::EnsureValidGC()
	{
		// Work out what we need to change to get GDI up to speed.
		if (!m_bValidPen)
		{
			// Deselect & delete current pen if it exists
			if (m_hPen != NULL)
			{
				::SelectObject(m_hdc, m_hOldPen);
				::DeleteObject(m_hPen);
				m_hPen = NULL;
			}

			// Make new pen
			CreatePen();

			// Select it into GDI's context
			m_hOldPen = (HPEN) ::SelectObject(m_hdc, m_hPen);

			m_bValidPen = true;
		}

		if (!m_bValidBrush)
		{
			// Deselect & delete current brush if it exists
			if (m_hBrush != NULL)
			{
				::SelectObject(m_hdc, m_hOldBrush);
				::DeleteObject(m_hBrush);
				m_hBrush = NULL;
			}

			// Make new brush
			CreateBrush();

			// Select it into GDI's context
			m_hOldBrush = (HBRUSH) ::SelectObject(m_hdc, m_hBrush);

			m_bValidBrush = true;
		}

		if (!m_bValidFont)
		{
			if (m_State.m_pFont == NULL)
			{
				// Select the default font
				if (m_hOldFont != NULL)
				{
					// Restore the original font.
					::SelectObject(m_hdc, m_hOldFont);
				}
				else
				{
					// We haven't changed the font yet, so do nothing.
				}
			}
			else
			{
				// Select the specified font
				Win32_FontImp *pFontImp = Win32_FontImp::GetWin32_FontImp(m_State.m_pFont);

				HFONT hNewFont = pFontImp->GetHFONT();

				HFONT hFont = (HFONT) ::SelectObject(m_hdc, hNewFont);

				// If this is the first time we've selected a font, then remember
				// the default font object returned
				if (m_hOldFont == NULL)
				{
					m_hOldFont = hFont;
				}
			}

			m_bValidFont = true;
		}

		if (!m_bValidBgMode)
		{
			// Determine whether we need opaque or transparent mode.
			if (m_State.m_BgMode == GC::bg_Transparent)
			{
				::SetBkMode(m_hdc, TRANSPARENT);
			}
			else
			{
				::SetBkMode(m_hdc, OPAQUE);
			}

			// We now have a valid bg mode.
			m_bValidBgMode = true;
		}

		if (!m_bValidText)
		{
			// Set the text colour.
			::SetTextColor(m_hdc, RGB(m_State.m_TextColour.GetRed(),
									  m_State.m_TextColour.GetGreen(),
									  m_State.m_TextColour.GetBlue()));
		}
	}

	void Win32_GCImp::CreatePen()
	{
		// Convert vGui types to GDI pen parameters.
		int PenStyle = PS_SOLID;

		switch (m_State.m_StrokeStyle)
		{
			case GC::ss_SOLID:
				// Do nothing
				break;

			case GC::ss_DASHED:
				PenStyle = PS_DASH;
				break;

			case GC::ss_DOTTED:
				PenStyle = PS_DOT;
				break;

			case GC::ss_DASHDOT:
				PenStyle = PS_DASHDOT;
				break;

			default:
				// Complain loudly in debug builds
				ASSERT(FALSE);
				break;
		}

		m_hPen = ::CreatePen(PenStyle, 
							 m_State.m_StrokeWidth, 
							 RGB(m_State.m_StrokeColour.GetRed(),
								 m_State.m_StrokeColour.GetGreen(),
								 m_State.m_StrokeColour.GetBlue()));
	}

	void Win32_GCImp::CreateBrush()
	{
		// GDI brushes always created indirectly.
		LOGBRUSH lgBrush;
		lgBrush.lbStyle = BS_SOLID;
		lgBrush.lbColor = RGB(m_State.m_FillColour.GetRed(),
							  m_State.m_FillColour.GetGreen(),
							  m_State.m_FillColour.GetBlue());

		m_hBrush = ::CreateBrushIndirect(&lgBrush);
	}

	// Save/restore entire GC state
	void Win32_GCImp::Save()
	{
		// Copy this state to the stack
		m_StateStack.push(m_State);
	}

	void Win32_GCImp::Restore()
	{
		// Restore state from stack.
		GCState& NewState = m_StateStack.top();

		// Work out what is changing
		// TODO: do this!
		m_bValidPen = false;
		m_bValidBrush = false;
		m_bValidFont = false;

		// Transfer new state
		m_State = NewState;
		m_StateStack.pop();
	}
	
	void Win32_GCImp::SetStrokeStyle(GC::StrokeStyle Style)
	{
		m_State.m_StrokeStyle = Style;
		m_bValidPen = false;
	}

	void Win32_GCImp::SetStrokeColour(const Colour& StrokeColour)
	{
		m_State.m_StrokeColour = StrokeColour;
		m_bValidPen = false;
	}

	void Win32_GCImp::SetFillColour(const Colour& FillColour)
	{
		m_State.m_FillColour = FillColour;
		m_bValidBrush = false;
	}

	void Win32_GCImp::SetBgColour(const Colour& BgColour)
	{
		m_State.m_BgColour = BgColour;
		m_bValidPen = false;
		m_bValidBrush = false;
	}

	Colour Win32_GCImp::SetTextColour(const Colour& TextColour)
	{
		Colour OldColour = m_State.m_TextColour;
		m_State.m_TextColour = TextColour;
		m_bValidText = false;
		return OldColour;
	}

	void Win32_GCImp::SetBgMode(GC::BgMode Mode)
	{
		m_State.m_BgMode = Mode;
		m_bValidBgMode = false;
	}

	void Win32_GCImp::SetFillStyle(GC::FillStyle Style)
	{
		m_State.m_FillStyle = Style;
		m_bValidBrush = false;
	}

	void Win32_GCImp::SetStrokeWidth(int Width)
	{
		m_State.m_StrokeWidth = Width;
		m_bValidPen = false;
	}

	// Drawing lines
	void Win32_GCImp::MoveTo(int x, int y)
	{
		// Use GDI primitive
		::MoveToEx(m_hdc,
				   x - m_ScrollX, 
				   y - m_ScrollY,
				   NULL);
	}

	void Win32_GCImp::LineTo(int x, int y)
	{
		// Make sure GDI is up to date with where we are.
		EnsureValidGC();

		// Use GDI primitive
		::LineTo(m_hdc,
				 x - m_ScrollX, 
				 y - m_ScrollY);
	}

	// Simple shapes
	void Win32_GCImp::DrawRect(const Rect& rc, GC::DrawType Type)
	{
		// Make sure GDI is up to date with where we are.
		EnsureValidGC();

		// Convert co-ordinates
		Rect rcGDI = rc;
		RectToGDI(rcGDI);

		switch (Type)
		{
			case GC::dt_Stroke:
			{
				// Temporarily use null brush
				HBRUSH hOldBrush = (HBRUSH) ::SelectObject(m_hdc, TheNullBrush());

				// Use GDI primitive to stroke but not fill the rectangle
				::Rectangle(m_hdc, rcGDI.left, rcGDI.top, rcGDI.right, rcGDI.bottom);

				// Restore brush
				::SelectObject(m_hdc, hOldBrush);
				break;
			}

			case GC::dt_Fill:
			{
				// Use GDI primitive

				// NB.	Note to would-be optimisers - I've tried using the 
				//		ExtTextOut() trick instead, but it turned out to be
				//		slightly slower (~8% slower) - presumably due 
				//		to the need to change the background colour and
				//		restore it each time.
				//
				//		Even though MFC still uses ::ExtTextOut() to implement
				//		FillSolidRect(), it isn't actually any faster.  On my
				//		PC they perform at exactly the same speed on small rectangles,
				//		and on larger ones ::FillRect() is faster.
				//
				//		Executive summary: don't change this without confirming that
				//		it would be measurably faster under all conditions.

				::FillRect(m_hdc, rcGDI.GetPointer(), m_hBrush);
				break;
			}

			case GC::dt_FillAndStroke:
			{
				// Use GDI primitive
				::Rectangle(m_hdc, rcGDI.left, rcGDI.top, rcGDI.right, rcGDI.bottom);
				break;
			}

			case GC::dt_StrokeAndFill:
			{
				// Temporarily use null brush
				HBRUSH hOldBrush = (HBRUSH) ::SelectObject(m_hdc, TheNullBrush());

				// Use GDI primitive to stroke but not fill the rectangle
				::Rectangle(m_hdc, rcGDI.left, rcGDI.top, rcGDI.right, rcGDI.bottom);

				// Restore brush
				::SelectObject(m_hdc, hOldBrush);

				// Now fill (but don't stroke) the rectangle
				::FillRect(m_hdc, rcGDI.GetPointer(), m_hBrush);

				break;
			}

			default:
			{
				// Illegal mode - complain loudly in debug builds
				ASSERT(FALSE);

				// Fail silently in release builds
				return;
			}
		}
	}

	void Win32_GCImp::DrawCircle(int cx, int cy, int Radius, GC::DrawType Type)
	{
		// Make sure GDI is up to date with where we are.
		EnsureValidGC();

		// Convert co-ordinates
		cx -= m_ScrollX;
		cy -= m_ScrollY;

		// This translates down to a call to Ellipse(), so we need
		// a bounding rectangle.
		Rect rc(cx - Radius, cy - Radius, cx + Radius, cy + Radius);

		switch (Type)
		{
			case GC::dt_Stroke:
			{
				// Temporarily use null brush
				HBRUSH hOldBrush = (HBRUSH) ::SelectObject(m_hdc, TheNullBrush());

				// Use GDI primitive to stroke but not fill the circle
				::Ellipse(m_hdc, rc.left, rc.top, rc.right, rc.bottom);

				// Restore brush
				::SelectObject(m_hdc, hOldBrush);
				break;
			}

			case GC::dt_Fill:
			{
				// Temporarily use null pen
				HPEN hOldPen = (HPEN) ::SelectObject(m_hdc, TheNullPen());

				// Use GDI primitive to fill but not stroke the circle
				::Ellipse(m_hdc, rc.left, rc.top, rc.right, rc.bottom);

				// Restore pen
				::SelectObject(m_hdc, hOldPen);
				break;
			}

			case GC::dt_FillAndStroke:
			{
				// Use GDI primitive
				::Ellipse(m_hdc, rc.left, rc.top, rc.right, rc.bottom);
				break;
			}

			case GC::dt_StrokeAndFill:
			{
				// Temporarily use null brush
				HBRUSH hOldBrush = (HBRUSH) ::SelectObject(m_hdc, TheNullBrush());

				// Use GDI primitive to stroke but not fill the circle
				::Ellipse(m_hdc, rc.left, rc.top, rc.right, rc.bottom);

				// Restore brush
				::SelectObject(m_hdc, hOldBrush);

				// Temporarily use null pen
				HPEN hOldPen = (HPEN) ::SelectObject(m_hdc, TheNullPen());

				// Now fill (but don't stroke) the circle
				::Ellipse(m_hdc, rc.left, rc.top, rc.right, rc.bottom);

				// Restore pen
				::SelectObject(m_hdc, hOldPen);

				break;
			}

			default:
			{
				// Illegal mode - complain loudly in debug builds
				ASSERT(FALSE);

				// Fail silently in release builds
				return;
			}
		}
	}

	// Polygons
	void Win32_GCImp::DrawPolygon(const Point* pts,
								  int cPoints,
								  GC::DrawType Type)
	{
		// Make sure GDI is up to date with where we are.
		EnsureValidGC();

		// UNIMPLEMENTED
		ASSERT(FALSE);
	}

	void Win32_GCImp::DrawPolygon(const Points& pts,
								  GC::DrawType Type)
	{
		// Make sure GDI is up to date with where we are.
		EnsureValidGC();

		// UNIMPLEMENTED
		ASSERT(FALSE);
	}

	HPEN Win32_GCImp::TheNullPen()
	{
		// Singleton pen object
		static NullPen TheNullPen;

		return TheNullPen.m_hPen;
	}
	
	HBRUSH Win32_GCImp::TheNullBrush()
	{
		// Singleton brush object - use GDI's built-in null brush.
		static HBRUSH NullBrush = (HBRUSH) ::GetStockObject(NULL_BRUSH);

		return NullBrush;
	}

	bool Win32_GCImp::Attach(Window *pWindow)
	{
		// We should not attach if we already have a valid HDC.
		COMPLAIN_IF(m_hdc != NULL, "Win32_GCImp::Attach() - already attached");

		// Get the HDC from the Win32 object.
		Win32_WindowImp *pWin32Window = Win32_WindowImp::GetImp(pWindow);

		if (pWin32Window == NULL)
		{
			// Unable to access window
			return false;
		}

		// Get the DC from the window
		m_hdc = ::GetDC(pWin32Window->m_hWnd);

		if (m_hdc != NULL)
		{
			// DC obtained successfully
			m_OwnDC = TRUE;
			m_hWnd = pWin32Window->m_hWnd;

			// We must release this HDC when detaching from it
			m_bOwnHDC = true;

			return true;
		}
		else
		{
			// Failed to obtain GC
			return false;
		}
	}

	void Win32_GCImp::Detach()
	{
		// Can't detach if we don't have a valid DC.
		COMPLAIN_IF(m_hdc == NULL, "Win32_GCImp::Detach() - null HDC");

		if (m_hdc == NULL)
		{
			return;
		}

		// Deselect & delete current pen if it exists
		if (m_hPen != NULL)
		{
			::SelectObject(m_hdc, m_hOldPen);
			::DeleteObject(m_hPen);
			m_hPen = NULL;
		}

		// Deselect & delete current brush if it exists
		if (m_hBrush != NULL)
		{
			::SelectObject(m_hdc, m_hOldBrush);
			::DeleteObject(m_hBrush);
			m_hBrush = NULL;
		}

		if ((m_hWnd != NULL) && m_bOwnHDC)
		{
			// Release the Win32 object
			::ReleaseDC(m_hWnd, m_hdc);
			m_bOwnHDC = false;
		}

		m_hWnd = NULL;
		m_hdc = NULL;
	}

	void Win32_GCImp::SetFont(Font* pFont)
	{
		COMPLAIN_IF(pFont == NULL, "NULL Font");

		m_State.m_pFont = pFont;
		m_bValidFont = false;
	}

	void Win32_GCImp::RestoreDefaultFont()
	{
		m_State.m_pFont = NULL;
		m_bValidFont = false;
	}

	void Win32_GCImp::DrawText(const String& Text, int x, int y)
	{
		COMPLAIN_IF(m_hdc == NULL, "NULL HDC");

		// Make sure GDI is up to date with where we are.
		EnsureValidGC();

		::TextOut(m_hdc, x - m_ScrollX, y - m_ScrollY, Text.c_str(), Text.GetLength());
	}


	void Win32_GCImp::DrawText(const String& Text, 
 							   const Rect& TextArea,
							   int FormatFlags)
	{
		COMPLAIN_IF(m_hdc == NULL, "NULL HDC");

		// Make sure GDI is up to date with where we are.
		EnsureValidGC();

		// Translate area to cope with scroll position
		Rect GDIRect(TextArea);
		RectToGDI(GDIRect);

		// Construct format value
		int iFormat = FormatFlags;
		iFormat |= DT_NOPREFIX;

		if (FormatFlags & tf_VCentre)
		{
			if (FormatFlags & tf_Multiline)
			{
				// We have multiline and vertical centreing, and Win32 won't do that for
				// us, so we first ask DrawText to recalculate the rectangle, so we
				// can move it so that it is vertically centred.
				Rect NewRect = GDIRect;

				// DT_CALCRECT forces the rectangle to be calculated, and the
				// text is not drawn.
				::DrawText(m_hdc, 
						   Text.c_str(), 
						   Text.GetLength(), 
						   (RECT*) &NewRect, iFormat | DT_CALCRECT);

				// Ok, centre the rectangle vertically, if it is smaller than the 
				// area specified by the caller.
				if (NewRect.GetHeight() < GDIRect.GetHeight())
				{
					NewRect.Translate(0, (GDIRect.GetHeight() - NewRect.GetHeight()) / 2);
				}

				// Reset x co-ords of the rectangle (DrawText changes these when 
				// it calculates the rectangle).
				GDIRect.top = NewRect.top;
				GDIRect.bottom = NewRect.bottom;
			}
			else
			{
				// Single line text, vertically centred - Win32 requires that
				// we specify DT_SINGLELINE to get vertical centreing.
				iFormat |= DT_SINGLELINE;
			}
		}

		// Ok, now we actually draw the text
		::DrawText(m_hdc, Text.c_str(), Text.GetLength(), (RECT*) &GDIRect, iFormat);
	}

	void Win32_GCImp::CalcTextArea(const String& Text,
								   int FormatFlags,
								   Rect* pTextArea)
	{
		COMPLAIN_IF(m_hdc == NULL, "NULL HDC");

		// Make sure GDI is up to date with where we are.
		EnsureValidGC();

		// Construct format value
		int iFormat = FormatFlags;
		iFormat |= DT_NOPREFIX;

		Rect TextArea = *pTextArea;

		// DT_CALCRECT forces the rectangle to be calculated, and the
		// text is not drawn.
		::DrawText(m_hdc, 
				   Text.c_str(), 
				   Text.GetLength(), 
				   (RECT*) &TextArea, iFormat | DT_CALCRECT);

		if (iFormat & DT_WORDBREAK)
		{
			// For multiline text, only copy y-coords of the rectangle 
			// (Win32 DrawText changes y-coords when it calculates the rectangle).
			pTextArea->top = TextArea.top;
			pTextArea->bottom = TextArea.bottom;
		}
		else
		{
			*pTextArea = TextArea;
		}
	}

	void Win32_GCImp::Draw3DEdge(const Rect& TheRect, Edge3DType EdgeType)
	{
		// Decode vGUI flags into Win32 flags.
		UINT EdgeFlags = 0;

		if (EdgeType & edge3d_RaisedInner)
			EdgeFlags |= BDR_RAISEDINNER;

		if (EdgeType & edge3d_RaisedOuter)
			EdgeFlags |= BDR_RAISEDOUTER;

		if (EdgeType & edge3d_SunkenInner)
			EdgeFlags |= BDR_SUNKENINNER;

		if (EdgeType & edge3d_SunkenOuter)
			EdgeFlags |= BDR_SUNKENOUTER;

		// Translate area to cope with scroll position
		Rect Area = TheRect;
		RectToGDI(Area);

		// Draw the 3D edge - BF_RECT means draw the whole rectangle.
		::DrawEdge(m_hdc, (RECT*) &Area, EdgeFlags, BF_RECT);
	}

	void Win32_GCImp::DrawFocusRect(const Rect& TheRect)
	{
		// Make sure GDI is up to date with where we are.
		EnsureValidGC();

		// Translate area to cope with scroll position
		Rect Area = TheRect;
		RectToGDI(Area);

		// Draw a rectangle indicating focus
		::DrawFocusRect(m_hdc, (RECT*) &Area);
	}

	void Win32_GCImp::BeginDrawing()
	{
		// Read the scroll offsets, ready for drawing
		m_ScrollInfo.fMask = SIF_POS;
		::GetScrollInfo(m_hWnd, SB_HORZ, &m_ScrollInfo);
		m_ScrollX = m_ScrollInfo.nPos;
		::GetScrollInfo(m_hWnd, SB_VERT, &m_ScrollInfo);
		m_ScrollY = m_ScrollInfo.nPos;
	}

	void Win32_GCImp::GetClipRect(Rect* pClipRect)
	{
		// Default to full client rect
		RECT rcClient;
		::GetClientRect(m_hWnd, &rcClient);

		pClipRect->left   = rcClient.left   + m_ScrollX;
		pClipRect->top    = rcClient.top    + m_ScrollY;
		pClipRect->right  = rcClient.right  + m_ScrollX;
		pClipRect->bottom = rcClient.bottom + m_ScrollY;
	}

	void Win32_GCImp::DrawIcon(const Icon& Icon, int X, int Y)
	{
		// Get the icon handle
		::DrawIconEx(m_hdc, 
				     X - m_ScrollX, Y - m_ScrollY, 
				     Icon.GetIconHandle(),
					 0, 0,
					 0, NULL,
					 DI_NORMAL);
	}
};




