//
// Win32_GCImp.h - Win32 version of Graphics Context
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_Win32_GCImp_h
#define vgui_Win32_GCImp_h

#include "..\vgui.h"

#include "..\GCImp.h"
#include "..\GCState.h"

namespace vgui
{
	class Win32_GCImp : public GCImp
	{
		public:

			//
			// Constructors
			//
			Win32_GCImp();
			~Win32_GCImp();

			// Platform dependent Initialisation
			void Attach(HDC hdc, HWND hWnd = NULL);

			//
			// Attach/detach to/from window
			//
			bool Attach(Window *pWindow);
			void Detach();

			// Clip rectangle
			virtual void GetClipRect(Rect* pClipRect);

			//
			// Drawing Attributes
			//

			// Save/restore entire GC state
			virtual void Save();
			virtual void Restore();

			// Line/Fill style
			virtual void SetStrokeStyle(GC::StrokeStyle Style);
			virtual void SetFillStyle(GC::FillStyle Style);

			virtual void SetFillColour(const Colour& FillColour);
			virtual void SetStrokeColour(const Colour& StrokeColour);

			virtual void SetBgColour(const Colour& BgColour);
			virtual void SetBgMode(GC::BgMode Mode);

			virtual Colour SetTextColour(const Colour& TextColour);

			virtual void SetStrokeWidth(int Width);

			// Drawing lines
			virtual void MoveTo(int x, int y);
			virtual void LineTo(int x, int y);

			// Simple shapes
			virtual void DrawRect(const Rect& rc,
								  GC::DrawType Type = GC::dt_FillAndStroke);

			virtual void DrawCircle(int cx, int cy, int Radius,
									GC::DrawType Type = GC::dt_FillAndStroke);

			// Polygons
			virtual void DrawPolygon(const Point* pts,
									 int cPoints,
									 GC::DrawType Type = GC::dt_FillAndStroke);

			virtual void DrawPolygon(const Points& pts,
									 GC::DrawType Type = GC::dt_FillAndStroke);

			// Text rendering

			virtual void SetFont(Font* pFont);
			virtual void RestoreDefaultFont();

			virtual void DrawText(const String& Text, int x, int y);

			virtual void DrawText(const String& Text, 
								  const Rect& TextArea,
								  int FormatFlags);

			virtual void CalcTextArea(const String& Text,
									  int FormatFlags,
									  Rect* pTextArea);

			// Specialised rendering
			virtual void Draw3DEdge(const Rect& TheRect, Edge3DType EdgeType);
			virtual void DrawFocusRect(const Rect& TheRect);

			// Image/bitmap rendering
			virtual void DrawIcon(const Icon& Icon, int X, int Y);

			// GC Focus
			
			virtual void BeginDrawing();
			virtual void EndDrawing() {}	// Do nothing

			// Co-ordinate functions

			// Access to raw Win32 structures
			HDC GetHDC() { return m_hdc; }

		protected:
			void EnsureValidGC();

			inline void PointToGDI(Point& pt)
			{
				pt.x -= m_ScrollX;
				pt.y -= m_ScrollY;
			}

			inline void RectToGDI(Rect& rc)
			{
				rc.left   -= m_ScrollX;
				rc.right  -= m_ScrollX;
				rc.top    -= m_ScrollY;
				rc.bottom -= m_ScrollY;
			}

		private:
			void CreatePen();
			void CreateBrush();

		protected:
			// The Win32 implementation of this GCImp.
			HDC m_hdc;
			bool m_OwnDC;
			HWND m_hWnd;

			// true => we own/created this HDC, so we should release it when done.
			bool m_bOwnHDC;

			// Scroll offsets - Win32 still has 16-bit co-ordinates, so
			// we fake a 32-bit virtual area by using 32-bit scroll offsets,
			// and fixing up all co-ordinates accordingly.
			int m_ScrollX,
				m_ScrollY;

		private:
			// Current state of context
			GCState m_State;

			// Validity of elements in the context
			bool m_bValidPen;
			bool m_bValidBrush;
			bool m_bValidFont;
			bool m_bValidClipRegion;
			bool m_bValidBgMode;
			bool m_bValidText;

			// Context save stack
			GCStateStack m_StateStack;

			// GDI objects currently in use
			HPEN   m_hPen;
			HBRUSH m_hBrush;

			// GDI objects to restore when we're done
			HPEN   m_hOldPen;
			HBRUSH m_hOldBrush;
			HFONT  m_hOldFont;

			// Commonly used singleton GDI objects
			HPEN TheNullPen();
			HBRUSH TheNullBrush();

			// Scroll offsets
			SCROLLINFO m_ScrollInfo;

		private:
			// Prohibit copying/assignment of GCImps
			Win32_GCImp(const Win32_GCImp&);
			Win32_GCImp& operator=(const Win32_GCImp&);
	};

};

#endif // !vgui_Win32_GCImp_h


