//
// Win32_ImageButtonImp.cpp - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_ImageButtonImp.h"

#include "Win32_ApplicationImp.h"
#include "Win32_Headers.h"
#include "Win32_MessageInfo.h"

#include "..\Rect.h"

#include "vos\Errors.h"
#include "vos\Trace.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	Win32_ImageButtonImp::Win32_ImageButtonImp() : m_Type(ImageButton_None),
												   m_bHotTrack(false)
	{
		// TCILB
	}

	void Win32_ImageButtonImp::Attach(HWND hWnd)
	{
		// We have to check that the class name is correct.
		if (!IsWindowOfClass(hWnd, "BUTTON"))
		{
			// Trying to attach to the wrong kind of control - it must be a button.
			vguiERROR();
			return;
		}

		// Ok, we allow it, so pass down to base class.
		ImageButtonImp::Attach(hWnd);
	}

	bool Win32_ImageButtonImp::Create(Window* pParent, 
									  const Rect& Position,
									  ControlID ID)
	{
		COMPLAIN("UNIMPLEMENTED");

		// Ask parent window to make child control
		Win32_WindowImp* pParentImp = GetImp(pParent);

		if (pParentImp != NULL)
		{
			m_hWnd = pParentImp->CreateChildControl(m_pWindow, 
													"button", "",
													BS_PUSHBUTTON, 0,
													Position, 
													ID);
		}

		return (m_hWnd != NULL);
	}

	bool Win32_ImageButtonImp::GetState()
	{
		LRESULT State = SendMessage(BM_GETCHECK, 0, 0);

		// Translate result to cross-platform value
		return (State == BST_CHECKED);
	}

	/**************************************************************************

	>	ControlTranslation Win32_ImageButtonImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																			  TMessageType* pMsgType,
																			  LRESULT* pResult,
																			  MessageParams* pParams)

		Author:		Tim Browse
		Date:		22/02/2001
		Inputs:		uMsg - the Win32 notify code
					nCtrlID - the control that this message is for.
					lParam, wParam - the Win32 message data.
		Outputs:	The vgui notification message.

		Returns:	ControlTranslation_OK      - Message translated ok, and
												 it should be delivered.

					ControlTranslation_Unknown - Message not recognised 
												 by the control.

					ControlTranslation_Handled - Message recognised and 
												 automatically handled by 
												 control - do not pass the 
												 message on.

		Purpose:	Translate Win32 owner-draw button notification messages into
					standardised vgui messages, or handle them directly.

	**************************************************************************/

	ControlTranslation Win32_ImageButtonImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																		  TMessageType* pMsgType,
																		  LRESULT* pResult,
																		  MessageParams* pParams)
	{
		switch (MsgInfo.m_uNotifyCode)
		{
			// Translate to vgui message
			case BN_CLICKED:
			{
				*pMsgType = mt_BN_CLICKED;	
				return ControlTranslation_OK;
			}

			//
			// Owner draw
			//
			// TODO: Check into ramifications of receiving WM_NOTIFY and
			//       normal WM_XXX messages here - is there a clash?
			case WM_DRAWITEM:
			{
				// Should we be doing owner draw messages?
				COMPLAIN_IF(m_pWindow == NULL, "TranslateControlNotification(): NULL window");

				LPDRAWITEMSTRUCT pDIS = (LPDRAWITEMSTRUCT) MsgInfo.m_Msg.lParam;

				ButtonState State = bs_Normal;

				if (pDIS->itemState & ODS_SELECTED)
					State = bs_Pressed;

				// Draw edge of button first...

				// Now redraw the button's internal appearance.
				switch (m_Type)
				{
					case ImageButton_CheckBox_Unchecked:
					{
						DrawButtonEdge(pDIS->hDC, pDIS->rcItem, State, m_bHotTrack);
						DrawCheckBox(pDIS->hDC, pDIS->rcItem, cs_Unchecked);
						break;
					}

					case ImageButton_CheckBox_Checked:
					{
						DrawButtonEdge(pDIS->hDC, pDIS->rcItem, State, m_bHotTrack);
						DrawCheckBox(pDIS->hDC, pDIS->rcItem, cs_Checked);
						break;
					}
					
					case ImageButton_PopupMenu:
					{
						DrawPopupMenuIcon(pDIS->hDC, pDIS->rcItem, State, m_bHotTrack);
						break;
					}
				}

				// We have handled this message ourselves
				return ControlTranslation_Handled;
			}
		}

		// Message not recognised
		return ControlTranslation_Unknown;
	}

	LRESULT Win32_ImageButtonImp::WindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		switch (uMsg)
		{
			// We capture mouse move messages so that we can do hot-tracking
			// of the cursor over buttons, so that the button behaves like
			// a normal button (we rely on the theme API to draw the
			// hot-tracking appearance for us).
			case WM_MOUSEMOVE:
			{
				if (!m_bHotTrack)
				{
					// Mouse has just moved over the window...
					TRACKMOUSEEVENT tme;
					ZeroMemory(&tme, sizeof(tme));
					tme.cbSize = sizeof(tme);
					tme.hwndTrack = m_hWnd;
					tme.dwFlags = TME_LEAVE;

					if (TrackMouseEvent(&tme))
					{
						// If TrackMouseEvent is supported
						m_bHotTrack = true;
						ForceRedraw();
					}
				}
				break;
			}

			case WM_MOUSELEAVE:
			{
				// Mouse has left window - switch off hot tracking and
				// redraw button in its usual state.
				m_bHotTrack = false;
				ForceRedraw();
				break;
			}
		}

		// Not handled - defer to base class
		return ImageButtonImp::WindowProc(uMsg, wParam, lParam);
	}

	void Win32_ImageButtonImp::SetImage(TImageButton Type)
	{
		m_Type = Type;
		ForceRedraw();
	}

	/**************************************************************************
	
	>	LPCWSTR Win32_ThemedObject::GetThemeClass() const
	
		Author:		Tim Browse
		Date:		21/8/2002
		Returns:	Name of this object's theme class.
	
		Purpose:	Get the theme class for this object (see Win32 function
					OpenThemeData() for an explanation of theme classes).
	
	**************************************************************************/
	
	LPCWSTR Win32_ImageButtonImp::GetThemeClass() const
	{
		switch (m_Type)
		{
			case ImageButton_CheckBox_Unchecked:
			case ImageButton_CheckBox_Checked:
			{
				return L"BUTTON";
			}

			case ImageButton_PopupMenu:
			{
				// We use the scrollbar arrow icon to draw a popup menu icon
				return L"SCROLLBAR";
			}
		}

		// Fallback
		return L"BUTTON";
	}

};
