//
// Win32_Keyboard.cpp - Win32 keyboard routines
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_Keyboard.h"

#include "..\Errors.h"

namespace vgui
{
	//
	// Keyboard input handling and translation.
	//

	typedef struct
	{
		public:

#ifdef _DEBUG
			// The windows VK code - this is what we use to indirect into
			// the key lookup table.  This field is only present in debug
			// builds - in release builds it is not there and we don't
			// check for table mismatches.
			int m_VKCode;
#endif

			// The platform independent key code for m_Key
			TKeyCode m_KeyCode;
	}
	TKeyLookupEntry;

#ifdef _DEBUG
#define KEYENTRY(VKCode, VGUICode) { (VKCode), (VGUICode) }
#define KEY_UNUSED { 0, kc_Invalid }
#else
#define KEYENTRY(VKCode, VGUICode) { (VGUICode) }
#define KEY_UNUSED { kc_Invalid }
#endif

	//
	// The keycode lookup table
	//

	// This table is ordered by the Win32 VK codes.  Therefore
	// the entries must align with the VK values.

	static TKeyLookupEntry s_KeyLookup[] = 
	{
		// 00-0F
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEYENTRY(VK_BACK,		kc_Backspace),
		KEYENTRY(VK_TAB,		kc_Tab),
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEYENTRY(VK_RETURN,		kc_Return),		// Both Return and Enter keys
		KEY_UNUSED,
		KEY_UNUSED,

		// 10-1F
		KEYENTRY(VK_SHIFT,		kc_Shift),
		KEYENTRY(VK_CONTROL,	kc_Control),
		KEYENTRY(VK_MENU,		kc_Alt),
		KEYENTRY(VK_PAUSE,		kc_Invalid),

		KEYENTRY(VK_CAPITAL,	kc_Invalid),		// Caps lock
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEYENTRY(VK_ESCAPE,		kc_Escape),

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		// 20-2F
		KEYENTRY(VK_SPACE,		kc_Space),
		KEYENTRY(VK_PRIOR,		kc_Invalid),		// PageUp
		KEYENTRY(VK_NEXT,		kc_Invalid),		// PageDown
		KEYENTRY(VK_END,		kc_Invalid),

		KEYENTRY(VK_HOME,		kc_Invalid),
		KEYENTRY(VK_LEFT,		kc_Left),
		KEYENTRY(VK_UP,			kc_Up),
		KEYENTRY(VK_RIGHT,		kc_Right),

		KEYENTRY(VK_DOWN,		kc_Down),
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEYENTRY(VK_SNAPSHOT,	kc_Invalid),		// PrintScreen
		KEYENTRY(VK_INSERT,		kc_Invalid),
		KEYENTRY(VK_DELETE,		kc_Delete),
		KEYENTRY(VK_HELP,		kc_Invalid),

		// 30-3F
		KEYENTRY('0',			kc_0),
		KEYENTRY('1',			kc_1),
		KEYENTRY('2',			kc_2),
		KEYENTRY('3',			kc_3),

		KEYENTRY('4',			kc_4),
		KEYENTRY('5',			kc_5),
		KEYENTRY('6',			kc_6),
		KEYENTRY('7',			kc_7),

		KEYENTRY('8',			kc_8),
		KEYENTRY('9',			kc_9),
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		// 40-4F
		KEY_UNUSED,
		KEYENTRY('A',			kc_A),
		KEYENTRY('B',			kc_B),
		KEYENTRY('C',			kc_C),

		KEYENTRY('D',			kc_D),
		KEYENTRY('E',			kc_E),
		KEYENTRY('F',			kc_F),
		KEYENTRY('G',			kc_G),

		KEYENTRY('H',			kc_H),
		KEYENTRY('I',			kc_I),
		KEYENTRY('J',			kc_J),
		KEYENTRY('K',			kc_K),

		KEYENTRY('L',			kc_L),
		KEYENTRY('M',			kc_M),
		KEYENTRY('N',			kc_N),
		KEYENTRY('O',			kc_O),

		// 50-5F
		KEYENTRY('P',			kc_P),
		KEYENTRY('Q',			kc_Q),
		KEYENTRY('R',			kc_R),
		KEYENTRY('S',			kc_S),

		KEYENTRY('T',			kc_T),
		KEYENTRY('U',			kc_U),
		KEYENTRY('V',			kc_V),
		KEYENTRY('W',			kc_W),

		KEYENTRY('X',			kc_X),
		KEYENTRY('Y',			kc_Y),
		KEYENTRY('Z',			kc_Z),
		KEYENTRY(VK_LWIN,		kc_Invalid),		// Left Windows Logo key	(MS Natural Keyboard)

		KEYENTRY(VK_RWIN,		kc_Invalid),		// Right Windows Logo key	(MS Natural Keyboard)
		KEYENTRY(VK_APPS,		kc_Invalid),		// Applications key			(MS Natural Keyboard)
		KEY_UNUSED,
		KEY_UNUSED,

		// 60-6F
		KEYENTRY(VK_NUMPAD0,	kc_Invalid),
		KEYENTRY(VK_NUMPAD1,	kc_Invalid),
		KEYENTRY(VK_NUMPAD2,	kc_Invalid),
		KEYENTRY(VK_NUMPAD3,	kc_Invalid),

		KEYENTRY(VK_NUMPAD4,	kc_Invalid),
		KEYENTRY(VK_NUMPAD5,	kc_Invalid),
		KEYENTRY(VK_NUMPAD6,	kc_Invalid),
		KEYENTRY(VK_NUMPAD7,	kc_Invalid),

		KEYENTRY(VK_NUMPAD8,	kc_Invalid),
		KEYENTRY(VK_NUMPAD9,	kc_Invalid),
		KEYENTRY(VK_MULTIPLY,	kc_Invalid),		// NumPad: Multiply
		KEYENTRY(VK_ADD,		kc_Invalid),		// NumPad: Add

		KEY_UNUSED,
		KEYENTRY(VK_SUBTRACT,	kc_Invalid),		// NumPad: Subtract
		KEYENTRY(VK_DECIMAL,	kc_Invalid),		// NumPad: Decimal Point
		KEYENTRY(VK_DIVIDE,		kc_Invalid),		// NumPad: Divide

		// 70-7F
		KEYENTRY(VK_F1,			kc_F1),
		KEYENTRY(VK_F2,			kc_F2),
		KEYENTRY(VK_F3,			kc_F3),
		KEYENTRY(VK_F4,			kc_F4),

		KEYENTRY(VK_F5,			kc_F5),
		KEYENTRY(VK_F6,			kc_F6),
		KEYENTRY(VK_F7,			kc_F7),
		KEYENTRY(VK_F8,			kc_F8),

		KEYENTRY(VK_F9,			kc_F9),
		KEYENTRY(VK_F10,		kc_F10),
		KEYENTRY(VK_F11,		kc_F11),
		KEYENTRY(VK_F12,		kc_F12),

		KEYENTRY(VK_F13,		kc_F13),
		KEYENTRY(VK_F14,		kc_F14),
		KEYENTRY(VK_F15,		kc_F15),
		KEYENTRY(VK_F16,		kc_F16),

		// 80-8F
		KEYENTRY(VK_F17,		kc_F17),
		KEYENTRY(VK_F18,		kc_F18),
		KEYENTRY(VK_F19,		kc_F19),
		KEYENTRY(VK_F20,		kc_F20),

		KEYENTRY(VK_F21,		kc_F21),
		KEYENTRY(VK_F22,		kc_F22),
		KEYENTRY(VK_F23,		kc_F23),
		KEYENTRY(VK_F24,		kc_F24),

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		// 90-9F
		KEYENTRY(VK_NUMLOCK,	kc_Invalid),
		KEYENTRY(VK_SCROLL,		kc_Invalid),		// Scroll-Lock
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		// A0-AF
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		// B0-BF
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		// C0-CF
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		// D0-DF
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,

		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
		KEY_UNUSED,
	};

	// We need to know how big the lookup table is.
	static int s_nKeyEntries = sizeof(s_KeyLookup) / sizeof(TKeyLookupEntry);


#ifdef _DEBUG
	void CheckKeyCodeLookupTable()
	{
		static bool bKeyTableChecked = false;

		if (bKeyTableChecked)
			return;

		//
		// Check the table for errors
		//
		for (int iKey = 0; iKey < s_nKeyEntries; iKey++)
		{
			// If the m_VKCode member is not the same as the position in
			// the array, then the key translation will not work, as we
			// use the VK_ value to index into the s_KeyLookup array.
			if ((s_KeyLookup[iKey].m_VKCode != iKey) &&
				(s_KeyLookup[iKey].m_VKCode != 0))
			{
				COMPLAIN("Error in key translation table!");
			}
		}

		bKeyTableChecked = true;
	}
#endif



	TKeyCode Win32_TranslateVKToKeyCode(int VK)
	{
#ifdef _DEBUG
		// In debug modes, we check our key translation table for errors.
		CheckKeyCodeLookupTable();
#endif

		// Convert VK code to vGUI TKeyCode value...
		if (VK >= s_nKeyEntries)
		{
			if (VK != 255)
			{
				// Unknown key code - the table should be extended to cope with this!
				COMPLAIN("TranslateKey(): Unknown VK code passed in.");
			}
			else
			{
				// On some systems, you get VK code 255 send when using 
				// arrow keys...I have no idea why, but we just ignore it
				// for now, otherwise it's very annoying.
				// TODO: Investigate why this happens.
			}

			return kc_Invalid;
		}
		else
		{
			// Translate VK_ code into vGUI platform independent value.
			return s_KeyLookup[VK].m_KeyCode;
		}
	}

	int Win32_TranslateKeyCodeToVK(TKeyCode KeyCode)
	{
#ifdef _DEBUG
		// In debug modes, we check our key translation table for errors.
		CheckKeyCodeLookupTable();
#endif

		// Search for keycode value - might be worth making a reverse lookup 
		// table if this ever slows anything down, but I doubt it.
		for (int VK = 0; VK < s_nKeyEntries; VK++)
		{
			if (s_KeyLookup[VK].m_KeyCode == KeyCode)
				return VK;
		}

		COMPLAIN("Win32_TranslateKeyCodeToVK() - KeyCode not found!");
		return 0;
	}

	bool IsModifierPressed(TKeyModifiers Modifier)
	{
		int VK = 0;

		if (Modifier & km_Shift)
			VK = VK_SHIFT;
		else if (Modifier & km_Control)
			VK = VK_CONTROL;
		else if (Modifier & km_Alt)
			VK = VK_MENU;
		else 
			return false;
			
		return ((::GetAsyncKeyState(VK) & 0x8000) != 0);
	}

	bool IsKeyPressed(TKeyCode KeyCode)
	{
		int VK = Win32_TranslateKeyCodeToVK(KeyCode);

		return ((::GetAsyncKeyState(VK) & 0x8000) != 0);
	}

};

