//
// Win32_ListBoxImp.cpp - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

/********************************************************************

	IMPLEMENTATION ISSUES

	The ListView control offers some challenges, in particular:

		* The ListView control provides no way to get the length of
		  an item's text label.  Therefore when getting the text 
		  of an item, we have to repeatedly increase the buffer size
		  until the label fits.

		* The LVS_SORTASCENDING and LVS_SORTDESCENDING window styles
		  do not work correctly - the control does not sort items
		  properly, therefore we always use a sorting callback
		  function with the LVM_SORTITEMS message.
		  See Microsoft Knowledge Base Article Q191295 for details
		  of the sordid mess.

********************************************************************/

#include "StdHdr.h"

#include "Win32_ListBoxImp.h"

#include "Win32_Headers.h"
#include "Win32_GCImp.h"
#include "Win32_MessageInfo.h"

#include "..\Errors.h"

#include "..\OwnerDrawListControl.h"
#include "..\ListBox.h"
#include "..\MessageParams.h"

#include "vos\Trace.h"

//
// Hack for old versions of commctrl.h
//

#ifndef LVM_SORTITEMSEX

#define LVM_SORTITEMSEX          (LVM_FIRST + 81)
#define ListView_SortItemsEx(hwndLV, _pfnCompare, _lPrm) \
  (BOOL)SNDMSG((hwndLV), LVM_SORTITEMSEX, (WPARAM)(LPARAM)(_lPrm), (LPARAM)(PFNLVCOMPARE)(_pfnCompare))

#endif

//
// Macro to check for valid window handle/state
//
#define ASSERT_VALID_WINDOW()							\
		if (!IsValid())									\
		{												\
			COMPLAIN("Invalid window handle used!");	\
			return;										\
		}

#define ASSERT_VALID_WINDOW_RETURN(RetVal)				\
		if (!IsValid())									\
		{												\
			COMPLAIN("Invalid window handle used!");	\
			return (RetVal);							\
		}

namespace vgui
{
	Win32_ListBoxImp::Win32_ListBoxImp() : m_ListType(ListType_Unattached),
										   m_iSortColumn(0),
										   m_bUseCustomTextColours(false)
	{
		// TCILB
	}

	void Win32_ListBoxImp::Attach(HWND hWnd)
	{
		// We have to check that the class name is correct.
		if (IsWindowOfClass(hWnd, "LISTBOX"))
		{
			m_ListType = ListType_ListBox;
		}
		else
		{
			// Not a vanilla ListBox - try a ListView control...
			if (IsWindowOfClass(hWnd, "SysListView32"))
			{
				// ListView control is in use...
				m_ListType = ListType_ListView;

				// If the listview is set to the 'report' style, insert
				// a column, otherwise if the user doesn't set the number
				// of columns, they'll never see anything!
				// (Thanks MS! - again)
				LONG lStyle = ::GetWindowLong(hWnd, GWL_STYLE);
				if ((lStyle & LVS_TYPEMASK) == LVS_REPORT)
				{
					// Create a single column, giving it a name that tells the developer
					// how to set up the control properly!
					//
					// Also, make the column 10000 pixels wide, so it always
					// reaches all the way across.
					LVCOLUMN lvc;
					ZeroMemory(&lvc, sizeof(lvc));
					lvc.mask = LVCF_SUBITEM | LVCF_TEXT | LVCF_WIDTH;
					lvc.iSubItem = 1;
					lvc.pszText = "Call SetNumberOfColumns() and SetColumnName()! :-)";
					lvc.cx = 10000;
					ListView_InsertColumn(hWnd, 0, &lvc);
				}
			}
			else
			{
				// Trying to attach to the wrong kind of control - it must be a
				// list control of some kind
				vguiERROR();
				return;
			}
		}

		// Ok, we allow it, so pass down to base class.
		ListBoxImp::Attach(hWnd);
	}

	void Win32_ListBoxImp::CleanUpItemInfoData()
	{
		// Dispose of all ItemInfo structures
		int cItems = GetCount();

		for (int iItem = 0; iItem < cItems; iItem++)
		{
			delete GetItemInfo(iItem);
		}
	}

	LRESULT Win32_ListBoxImp::WindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		if (uMsg == WM_DESTROY)
		{
			CleanUpItemInfoData();
		}

		return Win32_WindowImp::WindowProc(uMsg, wParam, lParam);
	}

	void Win32_ListBoxImp::Clear()
	{
		// Dispose of all ItemInfo structures
		CleanUpItemInfoData();

		// Now clear out the control's contents
		if (IsListBox())
		{
			SendMessage(LB_RESETCONTENT, 0, 0);
		}
		else if (IsListView())
		{
			ListView_DeleteAllItems(GetHWND());
		}
	}

	// Add and delete items
	bool Win32_ListBoxImp::AddItem(const String& ItemName, 
								   int* piItem,
								   UserData TheUserData)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		int iNewItem;

		ItemInfo* pItemInfo = new ItemInfo(TheUserData);
		
		if (IsListBox())
		{
			// Add the string
			iNewItem = SendMessage(LB_ADDSTRING, 0, (LPARAM) ItemName.c_str());

			// Check for problems
			if (iNewItem < 0)
			{
				// Negative index => error

			#ifdef _DEBUG
				// Make sure people check for errors by setting index to 
				// an outlandish value so they're more likely to notice it
				// in a debugger.  Step through your code (tm) :-)
				if (piItem != NULL)
					*piItem = 999999;
			#endif

				delete pItemInfo;
				return false;
			}
			else
			{
	#if _DEBUG
				// Check the text is entered into the control item
				//
				// NB. A common reason for this to fail is that you are using
				//     an owner draw list box, but that you have not specified
				//     the LBS_HASSTRINGS style for the listbox in the dialog 
				//	   editor - if you don't do this, the Windows listbox control
				//	   does not maintain strings for each item for you, so you'll
				//     get rubbish back from GetText().
				//
				TCHAR *pCheckBuffer = new TCHAR[ItemName.GetLength() + 1];
				SendMessage(LB_GETTEXT, iNewItem, (LPARAM) pCheckBuffer);
				COMPLAIN_IF(ItemName != pCheckBuffer, 
							"Failed to add text to list box!");
				delete [] pCheckBuffer;
	#endif
			}

			// Ok, we managed to insert the item, so set the user data
			if (SendMessage(LB_SETITEMDATA, iNewItem, (LPARAM) pItemInfo) == LB_ERR)
			{
				// Something went wrong.
			#ifdef _DEBUG
				// See above.
				if (piItem != NULL)
					*piItem = 999999;
			#endif
				delete pItemInfo;
				return false;
			}
		}
		else if (IsListView())
		{
			// Set up item data
			LVITEM Item;
			ZeroMemory(&Item, sizeof(Item));
			Item.mask = LVIF_PARAM | LVIF_TEXT;

			// Add item to end of list.
			Item.iItem = INT_MAX;

			// Set user data for the item (actually our own object pointer)
			Item.lParam = (LPARAM) pItemInfo;

			// Set the text for the item
			Item.pszText = (char*) ItemName.c_str();
			
			// Ok, add the item
			iNewItem = ListView_InsertItem(GetHWND(), &Item);

			if (iNewItem < 0)
			{
				// Negative result => error.

				#ifdef _DEBUG
					// Make sure people check for errors by setting index to 
					// an outlandish value so they're more likely to notice it
					// in a debugger.  Step through your code (tm) :-)
					if (piItem != NULL)
						*piItem = 999999;
				#endif

				delete pItemInfo;
				return false;
			}
		}
			
		// If we get this far, it's all worked
		if (piItem != NULL)
			*piItem = iNewItem;
		return true;
	}

	bool Win32_ListBoxImp::InsertItemAfter(int iAfterItem,
										   const String& ItemName, 
										   int* piItem,
										   UserData TheUserData)
	{
		// Add the string - InsertItemAt takes the desired index of
		// the new string, so we add 1 to the value passed in to insert
		// it after the specified item.
		return InsertItemAt(iAfterItem + 1, ItemName, piItem, TheUserData);
	}

	bool Win32_ListBoxImp::InsertItemBefore(int iBeforeItem,
											const String& ItemName, 
											int* piItem,
											UserData TheUserData)
	{
		return InsertItemAt(iBeforeItem, ItemName, piItem, TheUserData);
	}

	/**************************************************************************

	>	bool Win32_ListBoxImp::InsertItemAt(int iInsertAt,
											const String& ItemName, 
											int* piItem,
											UserData TheUserData)


		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		iInsertAt - the desired position for the item.
								i.e. 0 results in the item being added at the
									 beginning of the list, and GetCount()+1
									 will result in it being added to the end.
					ItemName - the text to put in the list item
					TheUserData - user specific data to store with the item.
		Outputs:	piItem - the new item index.
		Returns:	true => item added ok;
					false otherwise.

		Purpose:	Add an item to a listbox at a specific position.

	**************************************************************************/

	bool Win32_ListBoxImp::InsertItemAt(int iInsertAt,
										const String& ItemName, 
										int* piItem,
										UserData TheUserData)
	{
		ItemInfo* pItemInfo = new ItemInfo(TheUserData);

		if (InsertItemAt(iInsertAt, ItemName, piItem, pItemInfo))
		{
			// It worked
			return true;
		}
		else
		{
			// Didn't work - clean up the item we allocated
			delete pItemInfo;
			return false;
		}
	}

	/**************************************************************************

	>	bool Win32_ListBoxImp::InsertItemAt(int iInsertAt,
											const String& ItemName, 
											int* piItem,
											ItemInfo* pItemInfo)


		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		iInsertAt - the desired position for the item.
								i.e. 0 results in the item being added at the
									 beginning of the list, and GetCount()+1
									 will result in it being added to the end.
					ItemName - the text to put in the list item
					pItemInfo - the ItemInfo object to attach to this item
		Outputs:	piItem - the new item index.
		Returns:	true => item added ok;
					false otherwise.

		Purpose:	Add an item to a listbox at a specific position.

	**************************************************************************/

	bool Win32_ListBoxImp::InsertItemAt(int iInsertAt,
										const String& ItemName, 
										int* piItem,
										ItemInfo* pItemInfo)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		int iNewItem = -1;

		if (IsListBox())
		{
			iNewItem = SendMessage(LB_INSERTSTRING, 
								   iInsertAt,
								   (LPARAM) ItemName.c_str());

			// Check for problems
			if (iNewItem < 0)
			{
				// Negative index => error

			#ifdef _DEBUG
				// Make sure people check for errors by setting index to 
				// an outlandish value so they're more likely to notice it
				// in a debugger.  Step through your code (tm) :-)
				if (piItem != NULL)
					*piItem = 999999;
			#endif

				return false;
			}

			// Ok, we managed to insert the item, so set the user data
			if (SendMessage(LB_SETITEMDATA, iNewItem, (LPARAM) pItemInfo) == LB_ERR)
			{
				// Something went wrong.
			#ifdef _DEBUG
				// See above.
				if (piItem != NULL)
					*piItem = 999999;
			#endif

				return false;
			}
		}
		else if (IsListView())
		{
			// Set up item data
			LVITEM Item;
			ZeroMemory(&Item, sizeof(Item));
			Item.mask = LVIF_PARAM | LVIF_TEXT;

			// Add item to end of list.
			Item.iItem = iInsertAt;

			// Set user data for the item
			Item.lParam = (LPARAM) pItemInfo;

			// Set the text for the item
			Item.pszText = (char*) ItemName.c_str();
			
			// Ok, add the item
			iNewItem = ListView_InsertItem(GetHWND(), &Item);

			if (iNewItem < 0)
			{
				// Negative result => error.

				#ifdef _DEBUG
					// Make sure people check for errors by setting index to 
					// an outlandish value so they're more likely to notice it
					// in a debugger.  Step through your code (tm) :-)
					if (piItem != NULL)
						*piItem = 999999;
				#endif

				return false;
			}
		}
			
		// If we get this far, it's all worked
		if (piItem != NULL)
			*piItem = iNewItem;

		return true;
	}

	bool Win32_ListBoxImp::DeleteItem(int iItem)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		if ((iItem < 0) || (iItem >= GetCount()))
		{
			COMPLAIN("DeleteItem() - illegal index!");
			return false;
		}

		// Dispose of our user data pointer
		delete GetItemInfo(iItem);

		bool bSuccess = false;

		if (IsListBox())
		{
			// Negative values (e.g. LB_ERR) mean there was a problem.
			bSuccess = (SendMessage(LB_DELETESTRING, iItem, 0) >= 0);
		}
		else
		{
			bSuccess = (ListView_DeleteItem(GetHWND(), iItem) == TRUE);
		}

		// Deleting non-existent items is something we want to know about.
		ASSERT(bSuccess);

		return bSuccess;
	}

	ItemInfo* Win32_ListBoxImp::GetItemInfo(int iItem) const
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		// Ask control for the data
		if (IsListBox())
		{
			return (ItemInfo*) SendMessage(LB_GETITEMDATA, iItem, 0);
		}
		else if (IsListView())
		{
			LV_ITEM Item;
			ZeroMemory(&Item, sizeof(Item));
			Item.mask = LVIF_PARAM;
			Item.iItem = iItem;
			if (ListView_GetItem(GetHWND(), &Item))
			{
				return (ItemInfo*) Item.lParam;
			}
		}
		else
		{
			// Unknown type of control
			vguiERROR();
		}

		// Error getting item data
		COMPLAIN("List item has no ItemInfo object!");
		return NULL;
	}

	/**************************************************************************

	>	bool Win32_ListBoxImp::SetItemText(int iItem, 
										   const String& Text,
										   int nColumn)


		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		iItem - the index of the item to change.
					Text - the new text to use.
					nColumn - the 0-based index of the column text to change
		Returns:	true => item text changed successfully;
					false otherwise.

		Purpose:	Change the text for one column field of one list item 
					(in a multi-column list control).

	**************************************************************************/

	bool Win32_ListBoxImp::SetItemText(int iItem, const String& Text, int iColumn)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		if (IsListBox())
		{
			COMPLAIN_IF(iColumn != 0, "Non-zero column index used on single-column list control!");

			// Listbox control does not support changing item text, so we add a new
			// item and delete the old one.

			// First, get the ItemInfo
			ItemInfo* pItemInfo = GetItemInfo(iItem);

			if (pItemInfo == NULL)
			{
				// Unable to access item
				vguiERROR();
				return false;
			}

			// Insert new item before old one
			int iNewItem;
			if (!InsertItemAt(iItem, Text, &iNewItem, pItemInfo))
			{
				// Unable to insert new item
				vguiERROR();
				return false;
			}

			// New item should be inserted where the old one was
			ASSERT(iItem == iNewItem);

			// Delete the old one
			return DeleteItem(iItem + 1);
		}
		else if (IsListView())
		{
			// Ok, change the item
			ListView_SetItemText(GetHWND(), iItem, iColumn, (char*) Text.c_str());
			return true;
		}
		else
		{
			// Unknown list control type
			vguiERROR();
			return false;
		}
	}

	bool Win32_ListBoxImp::SetItemData(int iItem, UserData TheUserData)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		ItemInfo* pItemInfo = GetItemInfo(iItem);

		if (pItemInfo != NULL)
		{
			pItemInfo->m_UserData = TheUserData;
			return true;
		}
		else
		{
			// Something went wrong.
			return false;
		}
	}


	int Win32_ListBoxImp::GetCount() const
	{
		ASSERT_VALID_WINDOW_RETURN(0);

		if (IsListBox())
		{
			// Nice simple case - but it can still error according to Win32 docs...
			int cItems = (int) SendMessage(LB_GETCOUNT, 0, 0);

			if (cItems < 0)
			{
				// Just tell them there aren't any, but shout about it.
				vguiERROR();
				return 0;
			}

			// Return actual count
			return cItems;
		}
		else if (IsListView())
		{
			return ListView_GetItemCount(GetHWND());
		}

		// Error - unknown control type!
		vguiERROR();
		return -1;
	}

	bool Win32_ListBoxImp::GetItemUserData(int iItem, UserData* pTheUserData) const
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		ItemInfo* pItemInfo = GetItemInfo(iItem);

		if (pItemInfo != NULL)
		{
			*pTheUserData = pItemInfo->m_UserData;
			return true;
		}
		else
		{
			// Error getting item data
			return false;
		}
	}

	// Change currently selected item
	bool Win32_ListBoxImp::SetSelectedItem(int iItem)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		if (IsListBox())
		{
			int iSelectedItem = (int) SendMessage(LB_SETCURSEL, iItem, 0);

			if (iSelectedItem >= 0)
			{
				// Tell caller it worked.
				return true;
			}
			else
			{
				// Some kind of failure - illegal index? - so complain.
				vguiERROR();
				return false;
			}
		}
		else if (IsListView())
		{
			ListView_SetItemState(GetHWND(), iItem, LVIS_SELECTED, LVIS_SELECTED);
			return true;
		}
		else
		{
			// Unknown list control type
			vguiERROR();
			return false;
		}
	}

	// Access currently selected item
	bool Win32_ListBoxImp::GetSelectedItem(int* piSelectedItem) const
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		if (IsListBox())
		{
			// Ask control for current selection index
			int iItem = (int) SendMessage(LB_GETCURSEL, 0, 0);

			if (iItem >= 0)
			{
				*piSelectedItem = iItem;
				return true;
			}
			else
			{
				// Some kind of failure, or no item is selected.
				return false;
			}
		}
		else if (IsListView())
		{
			// Find first selected item
			int iSelectedItem = ListView_GetNextItem(GetHWND(), -1, LVNI_SELECTED);

			if (iSelectedItem >= 0)
			{
				*piSelectedItem = iSelectedItem;
				return true;
			}
			else
			{
				// No selected item
				return false;
			}
		}
		else
		{
			// Unknown list control type
			vguiERROR();
			return false;
		}
	}

	// Find an item
	bool Win32_ListBoxImp::FindItem(const String& ItemName, int* piFoundItem)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		int iItem = -1;

		if (IsListBox())
		{
			iItem = (int) SendMessage(LB_FINDSTRINGEXACT, 
									  -1,		// Start from beginning
									  (LPARAM) ItemName.c_str());
		}
		else if (IsListView())
		{
			LV_FINDINFO fi;
			ZeroMemory(&fi, sizeof(fi));
			fi.flags = LVFI_STRING;
			fi.psz = ItemName.c_str();

			// Search for this text from the beginning (-1 => from beginning)
			iItem = ListView_FindItem(GetHWND(), -1, &fi);
		}
		else
		{
			// Unknown list control type
			vguiERROR();
		}

		if (iItem >= 0)
		{
			// Success
			*piFoundItem = iItem;
			return true;
		}
		else
		{
			// Failure
			return false;
		}
	}

	/**************************************************************************

	>	int CALLBACK Win32_ListBoxImp::CompareFunc(LPARAM lParam1, 
												   LPARAM lParam2, 
												   LPARAM lParamSort)


		Author:		Tim Browse
		Date:		6/4/2002
		Inputs:		lParam1 - the index of the first item being compared.
					lParam2 - the index of the second item being compared.
					lParamSort - the parameter that was passed to the
								 ListView_SortItem macro when this sorting
								 function was installed.

		Returns:	The comparison function must return a negative value if the 
					first item should precede the second, a positive value if 
					the first item should follow the second, or zero if the 
					two items are equivalent. 

		Purpose:	Implement sorting for Win32 ListView controls.  This
					function is called when a listview control is to be sorted.

					See documentation of the Win32 LVM_SORTITEMSEX message for
					more details.

	**************************************************************************/

	int CALLBACK Win32_ListBoxImp::CompareFunc(LPARAM lParam1, 
											   LPARAM lParam2, 
											   LPARAM lParamSort)
	{
		Win32_ListBoxImp* pListBoxImp = (Win32_ListBoxImp*) lParamSort;

		if (pListBoxImp != NULL)
		{
			// Upcast is a necessary evil.  Probably.
			const ListBox* pListBox = (const ListBox*) GetWindowFromHWND(pListBoxImp->GetHWND());

			// Use base class
			if (pListBox != NULL)
			{
				int iResult = pListBox->CompareItems((int) lParam1, 
													 (int) lParam2,
													 pListBoxImp->m_iSortColumn);

				// Reverse sort direction if required
				if (pListBox->GetSortOrder() == ListControl::SortOrder_Descending)
					return -iResult;
				else
					return iResult;
			}
		}

		// Unable to get at list box
		vguiERROR();
		return 0;
	}

	/**************************************************************************
	
	>	void Win32_ListBoxImp::EnableSorting()
	
		Author:		Tim Browse
		Date:		2/4/2002
	
		Purpose:	Enable sorting on this list box - items will be kept
					in sorted order in the list box.
	
	**************************************************************************/
	
	void Win32_ListBoxImp::EnableSorting()
	{
		if (IsListBox())
		{
			HWND hWnd = GetHWND();

			// Set the window style to specify sorting
			LONG lStyle = ::GetWindowLong(hWnd, GWL_STYLE);
			lStyle |= LBS_SORT;
			::SetWindowLong(hWnd, GWL_STYLE, lStyle);
		}
		else if (IsListView())
		{
			HWND hWnd = GetHWND();

			// Set the window style to disable auto-sorting
			LONG lStyle = ::GetWindowLong(hWnd, GWL_STYLE);
			lStyle &= ~(LVS_SORTDESCENDING | LVS_SORTASCENDING);
			::SetWindowLong(hWnd, GWL_STYLE, lStyle);

			// Now set up sorting callback function
			ListView_SortItemsEx(hWnd, CompareFunc, (LPARAM) this);
		}
		else
		{
			// Unknown list control type
			vguiERROR();
		}
	}

	/**************************************************************************
	
	>	void Win32_ListBoxImp::DisableSorting()
	
		Author:		Tim Browse
		Date:		2/4/2002
	
		Purpose:	Disable sorting on this list box - items will be kept in
					the order in which they are added.
	
	**************************************************************************/
	
	void Win32_ListBoxImp::DisableSorting()
	{
		if (IsListBox())
		{
			HWND hWnd = GetHWND();

			// Set the window style to specify no sorting
			LONG lStyle = ::GetWindowLong(hWnd, GWL_STYLE);
			lStyle &= ~LBS_SORT;
			::SetWindowLong(hWnd, GWL_STYLE, lStyle);
		}
		else if (IsListView())
		{
			HWND hWnd = GetHWND();

			// Set the window style to specify no sorting
			LONG lStyle = ::GetWindowLong(hWnd, GWL_STYLE);
			lStyle &= ~(LVS_SORTDESCENDING | LVS_SORTASCENDING);
			::SetWindowLong(hWnd, GWL_STYLE, lStyle);

			// Disable any sorting callback function
			ListView_SortItemsEx(hWnd, NULL, 0);
		}
		else
		{
			// Unknown list control type
			vguiERROR();
		}
	}

	uint Win32_ListBoxImp::GetNumberOfColumns()
	{
		ASSERT_VALID_WINDOW_RETURN(1);

		if (!IsListView())
		{
			// This operation is not allowed unless we have a ListView control.
			vguiERROR();
			return 1;
		}

		// Find out how many columns we have.
		uint cExistingColumns = 1;
		HWND hwndHeader = ListView_GetHeader(GetHWND());
		if (hwndHeader != NULL)
		{
			cExistingColumns = Header_GetItemCount(hwndHeader);
		}

		return cExistingColumns;
	}

	/**************************************************************************
	
	>	bool Win32_ListBoxImp::SetNumberOfColumns(uint nColumns)
	
		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		nColumns - the number of columns that the list control 
							   should have.
		Returns:	true => the columns were set up successfully;
					false otherwise.
	
		Purpose:	Change the number of columns in a multi-column list control.
	
	**************************************************************************/
	
	bool Win32_ListBoxImp::SetNumberOfColumns(uint cRequiredColumns)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		if (!IsListView())
		{
			// This operation is not allowed unless we have a ListView control.
			vguiERROR();
			return false;
		}

		// We must have at least 1 column!
		if (cRequiredColumns < 1)
		{
			vguiERROR();
			return false;
		}

		// Find out how many columns we have.
		uint cExistingColumns = GetNumberOfColumns();

		// So, do we need to add some, delete some, or leave the list control 
		// unchanged?
		if (cExistingColumns == cRequiredColumns)
		{
			// No change needed
			return true;
		}
		else if (cExistingColumns < cRequiredColumns)
		{
			// We need to add some columns
			int nColumnsToAdd = cRequiredColumns - cExistingColumns;

			// The 0-based index for the next new column to add
			int iNewColumn = cExistingColumns;

			// Set up a useful default column layout
			LV_COLUMN lvc;
			ZeroMemory(&lvc, sizeof(lvc));
			lvc.cx = 75;
			lvc.iSubItem = iNewColumn;
			lvc.mask = LVCF_WIDTH | LVCF_SUBITEM;

			// If we don't set the text field when we create the column,
			// then it seems to prevent us from ever displaying text in
			// the column name (God knows why!), so we pass in an empty
			// string here.
			lvc.pszText = "";
			lvc.mask |= LVCF_TEXT;

			while (nColumnsToAdd > 0)
			{
				ListView_InsertColumn(GetHWND(), iNewColumn, &lvc);
				iNewColumn++;
				lvc.iSubItem = iNewColumn;
				nColumnsToAdd--;
			}

			// New columns added
			return true;
		}
		else
		{
			// We need to delete some columns
			int nColumnsToDelete = cExistingColumns - cRequiredColumns;

			while (nColumnsToDelete > 0)
			{
				ListView_DeleteColumn(GetHWND(), cExistingColumns);
				nColumnsToDelete--;
			}

			// All done.
			return true;
		}
	}

	/**************************************************************************
	
	>	void Win32_ListBoxImp::ShowColumnHeaders(bool bShow)
	
		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		bShow - true => show the headers
							false => hide the headers
	
		Purpose:	Enable/disable display of column headers (this does not
					affect whether or not multiple columns are shown - just
					the headers).
	
	**************************************************************************/
	
	void Win32_ListBoxImp::ShowColumnHeaders(bool bShow)
	{
		ASSERT_VALID_WINDOW()

		if (!IsListView())
		{
			// This operation is not allowed unless we have a ListView control.
			vguiERROR();
			return;
		}

		// Cet current window style
		LONG lOldStyle = ::GetWindowLong(GetHWND(), GWL_STYLE);
		LONG lNewStyle = lOldStyle;

		// Twiddle the bit that controls whether the column headers are displayed...
		if (bShow)
		{
			lNewStyle &= ~LVS_NOCOLUMNHEADER;
		}
		else
		{
			lNewStyle |= LVS_NOCOLUMNHEADER;
		}

		// If we changed anything, update the style and cause a redraw.
		if (lNewStyle != lOldStyle)
		{
			::SetWindowLong(GetHWND(), GWL_STYLE, lNewStyle);
			::InvalidateRect(GetHWND(), NULL, TRUE);
		}
	}

	/**************************************************************************
	
	>	bool Win32_ListBoxImp::SetColumnName(uint iColumn, const String& ColumnName)
	
		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		iColumn - the 0-based index of the column to rename
					ColumnName - the new name to use for the column
		Returns:	true => column rename successful;
					false otherwise.
	
		Purpose:	Change the name of a particular column in the control.
	
	**************************************************************************/
	
	bool Win32_ListBoxImp::SetColumnName(uint iColumn, const String& ColumnName)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		if (!IsListView())
		{
			// This operation is not allowed unless we have a ListView control.
			vguiERROR();
			return false;
		}

#ifdef _DEBUG
		if (iColumn >= GetNumberOfColumns())
		{
			COMPLAIN("SetColumnName() - column does not exist!");
			return false;
		}
#endif

		// Change just the text for this column.
		LV_COLUMN lvc;
		ZeroMemory(&lvc, sizeof(lvc));
		lvc.mask = LVCF_TEXT;
		lvc.pszText = (char*) ColumnName.c_str();

		return (ListView_SetColumn(GetHWND(), iColumn, &lvc) == TRUE);
	}

	/**************************************************************************
	
	>	bool Win32_ListBoxImp::SetColumnWidth(uint iColumn, uint WidthInPixels)
	
		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		iColumn - the 0-based index of the column to resize
					WidthInPixels - new size of the column
		Returns:	true => column resized successfully;
					false otherwise.
	
		Purpose:	Set the width of the specified column.
	
	**************************************************************************/
	
	bool Win32_ListBoxImp::SetColumnWidth(uint iColumn, uint WidthInPixels)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		if (!IsListView())
		{
			// This operation is not allowed unless we have a ListView control.
			vguiERROR();
			return false;
		}

		return (ListView_SetColumnWidth(GetHWND(), iColumn, WidthInPixels) == TRUE);
	}

	/**************************************************************************

	>	uint Win32_ListBoxImp::GetColumnWidth(uint iColumn)


		Author:		Tim Browse
		Date:		5/5/2002
		Inputs:		iColumn - the column to read the width from
		Returns:	Width of the specified column, in pixels.

		Purpose:	Find out how wide a particular column is in a listbox

	**************************************************************************/

	uint Win32_ListBoxImp::GetColumnWidth(uint iColumn)
	{
		return (uint) ListView_GetColumnWidth(GetHWND(), iColumn);
	}

	/**************************************************************************
	
	>	bool Win32_ListBoxImp::AutoSizeColumn(uint iColumn)
	
		Author:		Tim Browse
		Date:		29/4/2002
		Inputs:		iColumn - the column to resize
		Returns:	true => column resized ok;
					false => otherwise
	
		Purpose:	Resize a column to fit the items in it.  This is similar
					behaviour to what happens when you double-click on 
					divider line in the column header of an MS-Windows
					list view control.
	
	**************************************************************************/
	
	bool Win32_ListBoxImp::AutoSizeColumn(uint iColumn)
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		if (!IsListView())
		{
			// This operation is not allowed unless we have a ListView control.
			vguiERROR();
			return false;
		}

		// There are two ways to auto-size columns - the first is to size
		// so that all items fit, the other is so that the column header
		// text fit.
		//
		// For aesthetic purposes, we try both of these, then use the
		// large of the two widths that resulted.  This means we always see
		// the whole column header, and we always see all the items.
		HWND hListView = GetHWND();

		::LockWindowUpdate(hListView);

			// NB. Currently, we don't use the LVSCW_AUTOSIZE_USEHEADER flag on the
			//	   last column, as it screws up, and makes the column fill the 
			//	   rest of the control width.
			bool bResult = (ListView_SetColumnWidth(hListView, iColumn, LVSCW_AUTOSIZE) == TRUE);

			if (bResult)
			{
				bool bIsLastColumn = iColumn == (GetNumberOfColumns() - 1);

				if (!bIsLastColumn)
				{
					int ColumnWidth1 = ListView_GetColumnWidth(hListView, iColumn);

					bResult = (ListView_SetColumnWidth(hListView, iColumn, LVSCW_AUTOSIZE_USEHEADER) == TRUE);

					if (bResult)
					{
						int ColumnWidth2 = ListView_GetColumnWidth(hListView, iColumn);

						// Ok, pick the widest setting, and use it.
						int ColumnWidth = max(ColumnWidth1, ColumnWidth2);

						bResult = (ListView_SetColumnWidth(hListView, iColumn, ColumnWidth) == TRUE);
					}
				}
			}
		::LockWindowUpdate(NULL);

		return bResult;
	}

	bool Win32_ListBoxImp::GetItemText(int iItem, String* pItemName, int iColumn) const
	{
		if (IsListView())
		{
			return GetItemText_ListView(iItem, pItemName, iColumn);
		}
		else if (IsListBox())
		{
			COMPLAIN_IF(iColumn != 0, "Non-zero column specified for single-column list control!");
			return GetItemText_ListBox(iItem, pItemName);
		}
		else
		{
			// Unknown list control type.
			vguiERROR();
			return false;
		}
	}

	// Find info about an item
	bool Win32_ListBoxImp::GetItemText_ListBox(int iItem, String* pItemName) const
	{
		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		if (IsListView())
		{
			// Use column-based function, because getting the text from
			// a listview item is unnaturally complex, and I don't want the
			// code duplicated.
			return GetItemText(iItem, pItemName, 0);
		}

		if (!IsListBox())
		{
			// Unknown control type!
			vguiERROR();
			return false;
		}

		// Use a buffer on the stack for most purposes
		const int kBufferLength = 80;

		// Find out length of string (add 1 because the control doesn't
		// include the null terminator in the length calculation).
		int cTextLength = SendMessage(LB_GETTEXTLEN, iItem, 0) + 1;

		// Check for errors
		if (cTextLength < 0)
		{
			// Error - complain about this as it's probably an 
			// algorithmic error from the client.
			vguiERROR();
			return false;
		}

		bool bSuccess = false;

		// Is our on-stack buffer sufficient?
		if (cTextLength <= kBufferLength)
		{
			// Yes, so just get the text.
			TCHAR NameBuffer[kBufferLength+1];

			if (SendMessage(LB_GETTEXT, iItem, (LPARAM) NameBuffer) >= 0)
			{
				// It worked - assign name to caller's string
				*pItemName = NameBuffer;
				bSuccess = true;
			}
		}
		else
		{
			// We need a bigger buffer so dynamically allocate one
			// big enough to cope with exceptional cases (we do it in situ
			// in the caller's string object).
			vos::tchar* pBuffer = pItemName->LockBuffer(cTextLength + 1);

			// Extract the item string into this buffer
			bSuccess = (SendMessage(LB_GETTEXT, iItem, (LPARAM) pBuffer) >= 0);

			// All done
			pItemName->UnlockBuffer();
		}

		// Tell caller if it worked
		return bSuccess;
	}

	/**************************************************************************
	
	>	bool Win32_ListBoxImp::GetItemText(int iItem, String* pItemName, int nColumn) const
	
		Author:		Tim Browse
		Date:		2001/10/28
		Inputs:		iItem - the index of the item to change.
					nColumn - the 0-based index of the column text to change
		Outputs:	pItemName - the text of the specified field of the specified
								item.
		Returns:	true => item text read successfully;
					false otherwise.
	
		Purpose:	Get the text from the specified column of the specified
					item (in a multi-column list control).
	
	**************************************************************************/
	
	bool Win32_ListBoxImp::GetItemText_ListView(int iItem, String* pItemName, int nColumn) const
	{
		if (!IsListView())
		{
			// This operation is not allowed unless we have a ListView control.
			vguiERROR();
			return false;
		}

		ASSERT_VALID_WINDOW_RETURN(false);

		// Validate item index
		ASSERT(iItem >= 0);
		ASSERT(iItem < GetCount());

		if (!IsListBox() && !IsListView())
		{
			// Unknown control type!
			vguiERROR();
			return false;
		}

		// Use a buffer on the stack for most purposes
		const int kBufferLength = 80;
		TCHAR NameBuffer[kBufferLength+1];

		// Find out length of string
		int cTextLength;
		
		LVITEM Item;
		ZeroMemory(&Item, sizeof(Item));
		Item.iItem = iItem;
		Item.iSubItem = nColumn;
		Item.mask = LVIF_TEXT;
		Item.pszText = NameBuffer;
		Item.cchTextMax = kBufferLength;

		cTextLength = SendMessage(LVM_GETITEMTEXT, (WPARAM) iItem, (LPARAM) &Item);

		// Was our on-stack buffer sufficient?
		if (cTextLength < (kBufferLength - 1))
		{
			// Yes, so just return the text.
			*pItemName = NameBuffer;
		}
		else
		{
			// We need a bigger buffer, so dynamically allocate one
			// big enough to cope with exceptional cases, until we make a
			// buffer big enough to hold the text.
			//
			// This code may look nasty, but it's the only way to do it as
			// far as I can tell - and I looked pretty hard.  This method is
			// taken from the way that MFC does it (so obviously MS couldn't
			// figure out how to get the size of the text first either).
			//

			int BufSize = kBufferLength;

			do
			{
				// Double buffer size 
				BufSize *= 2;
				Item.cchTextMax = BufSize;

				// And try to get the text
				Item.pszText = pItemName->LockBuffer(BufSize+1);

					cTextLength = SendMessage(LVM_GETITEMTEXT, (WPARAM) iItem, (LPARAM) &Item);

				pItemName->UnlockBuffer();
			}
			while (cTextLength == (BufSize - 1));
		}

		// Tell caller it worked
		return true;
	}

	/**************************************************************************

	>	ControlTranslation Win32_ListBoxImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																		  TMessageType* pMsgType,
																		  LRESULT* pResult,
																		  MessageParams* pParams)

		Author:		Tim Browse
		Date:		22/02/2001
		Inputs:		uMsg - the Win32 notify code
					nCtrlID - the control that this message is for.
					lParam, wParam - the Win32 message data.
		Outputs:	The vgui notification message.

		Returns:	ControlTranslation_OK      - Message translated ok, and
												 it should be delivered.

					ControlTranslation_Unknown - Message not recognised 
												 by the control.

					ControlTranslation_Handled - Message recognised and 
												 automatically handled by 
												 control - do not pass the 
												 message on.

		Purpose:	Translate Win32 listbox notification messages into
					standardised vgui messages.

	**************************************************************************/

	ControlTranslation Win32_ListBoxImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																	  TMessageType* pMsgType,
																	  LRESULT* pResult,
																	  MessageParams* pParams)
	{
		//
		// TODO: add translation for ListView control messages...
		//
		switch (MsgInfo.m_uNotifyCode)
		{
			// Translate to vgui message
			case LBN_SELCHANGE:
			{
				*pMsgType = mt_LBN_SELCHANGED;	
				return ControlTranslation_OK;
			}

			case LBN_DBLCLK:
			case NM_DBLCLK:
			{
				*pMsgType = mt_LBN_DOUBLECLICK;
				return ControlTranslation_OK;
			}

			// NB. 
			//
			// Lame ListView API again - it can't just tell us when the selection
			// is changing, because that would be too useful and simple - it
			// tells us when *any* aspect of the item changes, and then we
			// have to check the details to know if the selection state
			// of the item has changed.
			//
			// Notice how this crap low level API prevents us from getting
			// the high level picture - i.e. that the selection has changed.
			// With this method, we usually get two selection change messages;
			// one for the item being deselected, and one for the item being
			// selected.  We can't work out when to send just one message, as
			// only the ListView control has that info, and it won't tell us.
			// </rant>
			//
			case LVN_ITEMCHANGED:
			{
				NM_LISTVIEW* pNMListView = (NM_LISTVIEW*) MsgInfo.m_Msg.lParam;

				if (pNMListView->uChanged & LVIF_STATE)
				{
					// Item state has changed - how?

					// Is the item being selected or deselected?
					DWORD dwOldState = pNMListView->uOldState & LVIS_SELECTED;
					DWORD dwNewState = pNMListView->uNewState & LVIS_SELECTED;

					if (dwNewState != dwOldState)
					{
						// Selection state has changed
						*pMsgType = mt_LBN_SELCHANGED;	
						return ControlTranslation_OK;
					}

					// Does the list have checkboxes?
					if (ListView_GetExtendedListViewStyle(m_hWnd) && LVS_EX_CHECKBOXES)
					{
						// We have checkboxes enabled - has the state image changed?
						DWORD dwOldStateImage = pNMListView->uOldState & LVIS_STATEIMAGEMASK;
						DWORD dwNewStateImage = pNMListView->uNewState & LVIS_STATEIMAGEMASK;

						if (dwNewStateImage != dwOldStateImage)
						{
							// Selection state has changed
							*pMsgType = mt_LBN_CHECKBOXCHANGED;
							pParams->m_Int1 = pNMListView->iItem;
							return ControlTranslation_OK;
						}
					}
				}

				break;
			}
			
			case LVN_COLUMNCLICK:
			{
				// Even though LVN_COLUMNCLICK is only for ListView controls,
				// we just make sure here that this is what we have, as those
				// jokers at Microsoft seem to like overloading 32-bit message
				// values (after all, there's only 4 billion - we might run out!)
				if (IsListView())
				{
					// User has clicked a column heading - they want to change
					// the sort column
					NM_LISTVIEW *pNm = (NM_LISTVIEW *) MsgInfo.m_Msg.lParam;

					// Make a note of which column is used for sorting...if it's
					// the same as last time, just flip the ascending/descending
					// sort flag.
					ListControl* pListControl = (ListControl*) m_pWindow;

					if (m_iSortColumn == pNm->iSubItem)
					{
						if (pListControl->GetSortOrder() == ListControl::SortOrder_Ascending)
							pListControl->SetSortOrder(ListControl::SortOrder_Descending);
						else
							pListControl->SetSortOrder(ListControl::SortOrder_Ascending);
					}
					else
					{
						// If we're changing columns, we always sort
						// ascending to start with
						m_iSortColumn = pNm->iSubItem;
						pListControl->SetSortOrder(ListControl::SortOrder_Ascending);
					}

					// And set up listview to sort.
					ListView_SortItemsEx(pNm->hdr.hwndFrom,
									     CompareFunc,
									     (LPARAM) this);

					// Sanity check
					COMPLAIN_IF(GetHWND() != pNm->hdr.hwndFrom,
								"ListBox window handle does not match!");

					// This is all that is needed for this message
					return ControlTranslation_Handled;
				}
				else
				{
					// We don't know what this is.
					return ControlTranslation_Unknown;
				}
			}

			//
			// Owner draw messages
			//
			case WM_MEASUREITEM:
			case WM_DRAWITEM:
			{
				// Should we be doing owner draw messages?
				COMPLAIN_IF(m_pWindow == NULL, "TranslateControlNotification(): NULL window");

				ListBox* pListBox = (ListBox*) m_pWindow;
				OwnerDrawListControl* pHandler = pListBox->GetOwnerDrawHandler();

				if (pHandler == NULL)
				{
					if (IsListBox())
					{
						// Deal with item measurement
						if (MsgInfo.m_uNotifyCode == WM_MEASUREITEM)
						{
							// Do not respond - let control use default size
							return ControlTranslation_Unknown;
						}
						else
						{
							LPDRAWITEMSTRUCT pDIS = (LPDRAWITEMSTRUCT) MsgInfo.m_Msg.lParam;

							ItemInfo* pItemInfo = GetItemInfo(pDIS->itemID);

							if (pItemInfo != NULL)
							{
								DrawListBoxItem(pDIS, pItemInfo);
								return ControlTranslation_Handled;
							}
							else
							{
								// Get control to draw item
								return ControlTranslation_Unknown;
							}
						}
					}

					// We do not have owner draw handler, so we should not be getting
					// this message!
					COMPLAIN("TranslateControlNotification(): Owner draw message received in error.");
					return ControlTranslation_Unknown;
				}
				
				// Ok, we need to call the window handler function with the decoded
				// parameters.
				bool bHandled;

				if (MsgInfo.m_uNotifyCode == WM_DRAWITEM)
				{
					bHandled = CallDrawItem(pHandler, MsgInfo.m_Msg.lParam);
				}
				else
				{
					bHandled = CallMeasureItem(pHandler, MsgInfo.m_Msg.lParam);
				}

				if (bHandled)
				{
					// We have handled this message ourselves
					return ControlTranslation_Handled;
				}
				else
				{
					// Message not handled
					return ControlTranslation_Unknown;
				}
			}

			case NM_CUSTOMDRAW:
			{
				// We only respond to custom drawing if the client
				// has called SetRowColour() or something similar.
				if (m_bUseCustomTextColours)
				{
					return HandleCustomDraw((NMLVCUSTOMDRAW*) MsgInfo.m_Msg.lParam, pResult);
				}
				break;
			}
			
			// Currently unhandled listbox control notifications
			case LBN_SELCANCEL:
			case LBN_SETFOCUS:
			case LBN_KILLFOCUS:
			{
				return ControlTranslation_Unknown;
			}
		}

		// Message not recognised
		return ControlTranslation_Unknown;
	}

	/**************************************************************************
	
	>	void Win32_ListBoxImp::DrawListBoxItem(DRAWITEMSTRUCT* pDIS, 
											   ItemInfo* pItemInfo)
	
		Author:		Tim Browse
		Date:		5/8/2002
		Inputs:		pDIS - the DRAWITEMSTRUCT passed in with the WM_DRAWITEM message
					pItemInfo - the item info for the item to be drawn.
	
		Purpose:	Provide a standard rendering routine for drawing items
					in a standard Win32 listbox.  The only extra functionality
					of this function is to optionally render the item text in
					a special colour - as specified by SetRowColour().

					This function is called when a ListBox control receives
					a WM_DRAWITEM message, and no owner draw handler has been
					registered.
	
	**************************************************************************/
	
	void Win32_ListBoxImp::DrawListBoxItem(DRAWITEMSTRUCT* pDIS, ItemInfo* pItemInfo)
	{
		// Much of this code is based on MS sample code - for details, 
		// see "HOWTO: Make a List Box Item Unavailable for Selection" in MSDN.
		if ((int)(pDIS->itemID) < 0)
		{
			switch (pDIS->CtlType)
			{
				case ODT_LISTBOX:
					if ((pDIS->itemAction) & (ODA_FOCUS))
						DrawFocusRect(pDIS->hDC, &pDIS->rcItem);
					break;
					
				case ODT_COMBOBOX:
					if ((pDIS->itemState) & (ODS_FOCUS))
						DrawFocusRect(pDIS->hDC, &pDIS->rcItem);
					break;
			}

			return;
		}

		// First, draw background
		HBRUSH hBrush;
		if (pDIS->itemState & ODS_SELECTED)
			hBrush = ::GetSysColorBrush(COLOR_HIGHLIGHT);
		else
			hBrush= ::GetSysColorBrush(COLOR_WINDOW);

		::FillRect(pDIS->hDC, &pDIS->rcItem, hBrush);

		// Draw the text of the item
		vos::String ItemText;
		GetItemText(pDIS->itemID, &ItemText, 0);

		// Find out which text colour to use - either the standard
		// text/highlight colours, or a special colour the user has
		// requested for this particular item in the control.
		COLORREF TextColour;
		if (pDIS->itemState & ODS_SELECTED)
		{
			TextColour = ::GetSysColor(COLOR_HIGHLIGHTTEXT);
		}
		else
		{
			if (pItemInfo->m_bUseColour)
			{
				TextColour = RGB(pItemInfo->m_Colour.GetRed(),
								 pItemInfo->m_Colour.GetGreen(),
								 pItemInfo->m_Colour.GetBlue());
			}
			else
			{
				TextColour = ::GetSysColor(COLOR_WINDOWTEXT);
			}
		}

		//
		// Ok, we have the text colour, so let's render the text.
		//
		// NB. If we don't use SaveDC/RestoreDC here, then DrawFocusRect
		//	   does not render correctly.
		int iDCState = ::SaveDC(pDIS->hDC);

			::SetTextColor(pDIS->hDC, TextColour);
			::SetBkMode(pDIS->hDC, TRANSPARENT);
			
			::TextOut(pDIS->hDC,
					  pDIS->rcItem.left, pDIS->rcItem.top,
					  ItemText.c_str(),
					  ItemText.GetLength());

		::RestoreDC(pDIS->hDC, iDCState);

		// Draw focus rect if necessary (do this last as
		// it uses XOR drawing)
		if (pDIS->itemState & ODS_FOCUS)
			::DrawFocusRect(pDIS->hDC, &pDIS->rcItem);
	}

	/**************************************************************************
	
	>	ControlTranslation Win32_ListBoxImp::HandleCustomDraw(NMLVCUSTOMDRAW* pNMLVCD,
															  LRESULT* pResult)
	
		Author:		Tim Browse
		Date:		5/8/2002
		Inputs:		pNMLVCD - the custom draw data for the ListView
		Outputs:	pResult - result code for this message
		Returns:	Indicates how to deal with this message.
	
		Purpose:	Process a ListView control's NM_CUSTOMDRAW messages.
					
					We process these when the SetRowColour() function has been
					called on this control, as we need to render the text
					of some or all of the items in specific colours.
	
	**************************************************************************/
	
	ControlTranslation Win32_ListBoxImp::HandleCustomDraw(NMLVCUSTOMDRAW* pNMLVCD,
														  LRESULT* pResult)
	{
		switch (pNMLVCD->nmcd.dwDrawStage)
		{
			case CDDS_PREPAINT:
			{
				// We want to take part in redraw of this item...
				*pResult = CDRF_NOTIFYITEMDRAW;
				return ControlTranslation_UseReturnCode;
			}

			case CDDS_ITEMPREPAINT:
			{
				// Ok, change the text colour for this item (if necessary)
				// and tell the control about the changes we make...
				ItemInfo* pItemInfo = (ItemInfo*) pNMLVCD->nmcd.lItemlParam;

				if ((pItemInfo != NULL) && (pItemInfo->m_bUseColour))
				{
					pNMLVCD->clrText = RGB(pItemInfo->m_Colour.GetRed(),
										   pItemInfo->m_Colour.GetGreen(),
										   pItemInfo->m_Colour.GetBlue());

					*pResult = CDRF_NEWFONT;
					return ControlTranslation_UseReturnCode;
				}

				break;
			}
		}

		// Otherwise, we don't want to handle this message
		return ControlTranslation_Unknown;
	}

	void Win32_ListBoxImp::SetRowColour(uint Row, const Colour& RowColour)
	{
		ItemInfo* pItemInfo = GetItemInfo(Row);

		if (pItemInfo != NULL)
		{
			pItemInfo->m_bUseColour = true;
			pItemInfo->m_Colour = RowColour;
			m_bUseCustomTextColours = true;

#ifdef _DEBUG
			if (IsListBox())
			{
				// We need this control to be owner draw, so we can draw
				// the text in the right colour
				LONG lStyle = ::GetWindowLong(m_hWnd, GWL_STYLE);

				LONG lRequiredStyle =  LBS_OWNERDRAWFIXED | LBS_HASSTRINGS;
				LONG lStyleMask = lRequiredStyle | LBS_OWNERDRAWVARIABLE;

				if ((lStyle & lStyleMask) != lRequiredStyle)
				{
					// Tell the developer what they need to do to make this work.
					//
					// NB. Sadly, we can't change the style bits ourselves here automatically, 
					//	   because the listbox only uses them when it initialises 
					//	   itself (as far as I can tell).
					COMPLAIN("ListBox::SetRowColour() - to use this function, the list "
							 "box control must have LBS_HASSTRINGS and LBS_OWNERDRAWFIXED "
							 "styles, but it the control you are using does not.\r\n\r\n"
							 "The text colour change will not work until you correct the "
							 "list control's style bits.");
				}
			}
#endif
		}
	}

	void Win32_ListBoxImp::UseDefaultRowColour(uint Row)
	{
		ItemInfo* pItemInfo = GetItemInfo(Row);

		if (pItemInfo != NULL)
		{
			pItemInfo->m_bUseColour = false;
		}
	}
	
	/**************************************************************************

	>	bool Win32_ListBoxImp::CallDrawItem(OwnerDrawListControl* pHandler,
											LPARAM lParam)

		Author:		Tim Browse
		Date:		10/03/2001
		Inputs:		pHandler - the owner draw handler object
					lParam - the LPARAM from the WM_DRAWITEM message.
		Returns:	true => message handled;
					false otherwise.

		Purpose:	Dispatch the data from the WM_DRAWITEM to the ListBox
					OnDrawItem() handler function, converting it to a platform
					independent format on the way through.

					NB.

					There's a fairly unpleasant performance issue here that
					might become a problem with lists that have a large number
					of items visible at once.  Basically, we have to create
					a brand new vGUI GC object everytime we draw a list item.

					However, I'm not going to address it until it becomes a 
					problem, as the solution would involve complex state
					logic, or a lightweight GC object, neither of which are
					trivial to implement right now, and I'm not sure if I'll
					even notice a problem in these days when a 300 MHz
					Pentium is considered so slow that you can't even buy a
					CPU that runs twice that fast anymore.

					Basically, this code is inelegant, but it enables us to
					have an elegant solution, so I'm putting up with it
					(i.e. it allows us to have platform independent owner-draw
					list classes, which is worth the unpleasantness).

	**************************************************************************/

	bool Win32_ListBoxImp::CallDrawItem(OwnerDrawListControl* pHandler, 
										LPARAM lParam)
	{
		if (!IsListBox())
		{
			// This type of control not supported as owner draw yet...
			// TODO: implement owner-draw for ListView controls.
			vguiERROR();
			return false;
		}

		// Decode parameters...
		DRAWITEMSTRUCT* pDI = (DRAWITEMSTRUCT*) lParam;

		// Check we have correct object type
		COMPLAIN_IF(pDI->CtlType != ODT_LISTBOX,
					"CallDrawItem() - wrong type of control found (not listbox).");

		int OwnerDrawItemState = TranslateListBoxItemState(pDI->itemState);

		// Preserve HDC state (required by owner-draw API)
		int iSavedDCState = ::SaveDC(pDI->hDC);

		// Attach HDC to a GC object
		GC gc;
		Win32_GCImp* pWin32GC = (Win32_GCImp*) gc.GetGCImp();
		pWin32GC->Attach(pDI->hDC);

		bool bResult = pHandler->OnDrawItem(&gc, 
											pDI->itemID,
											(UserData) pDI->itemData,
											(Rect&) pDI->rcItem,
											OwnerDrawItemState);

		// Detach HDC from our GC object
		pWin32GC->Detach();

		// Restore HDC
		::RestoreDC(pDI->hDC, iSavedDCState);

		// Tell caller how it went
		return bResult;
	}

	
	/**************************************************************************

	>	bool Win32_ListBoxImp::CallMeasureItem(OwnerDrawListControl* pHandler,
											   LPARAM lParam)

		Author:		Tim Browse
		Date:		10/03/2001
		Inputs:		pHandler - the owner draw handler object
					lParam - the LPARAM from the WM_MEASUREITEM message.
		Returns:	true => message handled;
					false otherwise.

		Purpose:	Dispatch the data from the WM_MEASUREITEM to the ListBox
					OnMeasureItem() handler function, converting it to a platform
					independent format on the way through.

	**************************************************************************/

	bool Win32_ListBoxImp::CallMeasureItem(OwnerDrawListControl* pHandler,
										   LPARAM lParam)
	{
		if (!IsListBox())
		{
			// This type of control not supported as owner draw yet...
			// TODO: implement owner-draw for ListView controls.
			vguiERROR();
			return false;
		}

		// Decode parameters...
		MEASUREITEMSTRUCT* pMI = (MEASUREITEMSTRUCT*) lParam;

		// Check we have correct object type
		COMPLAIN_IF(pMI->CtlType != ODT_LISTBOX,
					"CallMeasureItem() - wrong type of control found (not listbox).");

		return pHandler->OnMeasureItem(pMI->itemID,
									   (UserData) pMI->itemData,
									   &pMI->itemHeight);
	}

	
	/**************************************************************************

	>	int Win32_ListBoxImp::TranslateListBoxItemState(UINT itemState)

		Author:		Tim Browse
		Date:		10/03/2001
		Inputs:		itemState - state of a list item in Win32 format 
								(see MEASUREITEMSTRUCT in Win32 help).
		Returns:	vGUI item state value - see TOwnerDrawItemState.

		Purpose:	Translates Win32 item state bitfield into platform
					independent format.

	**************************************************************************/

	int Win32_ListBoxImp::TranslateListBoxItemState(UINT itemState)
	{
		int State = 0;

		if (itemState & ODS_SELECTED)
		{
			State |= OwnerDrawItemState_Selected;
		}

		if (itemState & ODS_FOCUS)
		{
			State |= OwnerDrawItemState_HasFocus;
		}

		if (itemState & ODS_DISABLED)
		{
			State |= OwnerDrawItemState_Disabled;
		}

		return State;
	}

	
	/**************************************************************************

	>	void Win32_ListBoxImp::RemeasureItems()

		Author:		Tim Browse
		Date:		24/03/2001

		Purpose:	Force an update of all list items in an owner-draw list 
					box, so that they are sized correctly.

					(See below for faster version of this function)

	**************************************************************************/

	void Win32_ListBoxImp::RemeasureItems()
	{
		if (!IsListBox())
		{
			// This type of control not supported as owner draw yet...
			// TODO: implement owner-draw for ListView controls.
			vguiERROR();
			return;
		}

		// Get the owner draw handler
		ListBox* pListBox = (ListBox*) m_pWindow;
		OwnerDrawListControl* pHandler = pListBox->GetOwnerDrawHandler();

		if (pHandler == NULL)
		{
			// We do not have an owner draw handler, so we should not be 
			// asked to do this!
			COMPLAIN("RemeasureItems() - called on list control that is not owner drawn!");
			return;
		}

		// Resize each item
		int cItems = GetCount();
		UserData TheUserData;

		for (int iItem = 0; iItem < cItems; iItem++)
		{
			uint NewHeight;
			GetItemUserData(iItem, &TheUserData);
			if (pHandler->OnMeasureItem(iItem, TheUserData, &NewHeight))
			{
				// Client handler updated size, so tell listbox item.
				SendMessage(LB_SETITEMHEIGHT, (WPARAM) iItem, (LPARAM) NewHeight);
			}
		}
	}

	/**************************************************************************

	>	void Win32_ListBoxImp::RemeasureItems(int NewHeight)

		Author:		Tim Browse
		Date:		24/03/2001
		Inputs:		NewHeight - the new height to use for all items in the list.

		Purpose:	Force an update of all list items in an owner-draw list 
					box, so that they are sized correctly.

					This version is faster, because the height is specified
					directly, rather than being determined by calling
					OnMeasureItem() for each item (unlike the function above).

	**************************************************************************/

	void Win32_ListBoxImp::RemeasureItems(int NewHeight)
	{
		if (!IsListBox())
		{
			// This type of control not supported as owner draw yet...
			// TODO: implement owner-draw for ListView controls.
			vguiERROR();
			return;
		}

		// Resize each item
		int cItems = GetCount();

		for (int iItem = 0; iItem < cItems; iItem++)
		{
			SendMessage(LB_SETITEMHEIGHT, (WPARAM) iItem, (LPARAM) NewHeight);
		}
	}

	/**************************************************************************
	
	>	void Win32_ListBoxImp::EnableCheckboxes(bool bEnable)
	
		Author:		Tim Browse
		Date:		15/8/2002
		Inputs:		bEnable - true => show checkboxes
							 false => hide checkboxes
	
		Purpose:	Enable/disable the use/display of checkboxes next to
					items in a list box control.
	
	**************************************************************************/
	
	void Win32_ListBoxImp::EnableCheckboxes(bool bEnable)
	{
		if (!IsListView())
		{
			// This type of control not supported as having checkboxes yet...
			// TODO: implement checkboxes for all list controls
			vguiERROR();
			return;
		}

		// TODO: Cope with older versions of the listview control that don't
		//		 support the LVS_EX_CHECKBOXES style. (i.e. do owner draw checkboxes)
		DWORD dwNewStyle = 0;
		if (bEnable)
			dwNewStyle = LVS_EX_CHECKBOXES;

		ListView_SetExtendedListViewStyleEx(m_hWnd, LVS_EX_CHECKBOXES, dwNewStyle);

		// That really should have worked...
		COMPLAIN_IF((ListView_GetExtendedListViewStyle(m_hWnd) & LVS_EX_CHECKBOXES) == 0,
					"EnableCheckboxes() - an error occurred when enabling the checkbox "
					"functionality on the listview control.");

	}

	/**************************************************************************
	
	>	void Win32_ListBoxImp::SetCheckState(int iItem, ControlState ItemState)
	
		Author:		Tim Browse
		Date:		15/8/2002
		Inputs:		iItem - the item to change
					ItemState - the new state of the item 
								(cs_Checked or cs_Unchecked)
	
		Purpose:	Set the checkbox state on an item in a listbox.
	
	**************************************************************************/
	
	void Win32_ListBoxImp::SetCheckState(int iItem, ControlState ItemState)
	{
		if (!IsListView())
		{
			// This type of control not supported as having checkboxes yet...
			// TODO: implement checkboxes for all list controls
			vguiERROR();
			return;
		}

		// Check for checkbox style on the control
		COMPLAIN_IF((ListView_GetExtendedListViewStyle(m_hWnd) & LVS_EX_CHECKBOXES) == 0,
					"SetCheckState() error - "
					"The listview control does not have the correct style "
					"applied to support checkboxes.\r\n"
					"You should add the LVS_EX_CHECKBOXES style to the control "
					"in the resource editor.");

		// Check index is within the current range
		COMPLAIN_IF(iItem >= GetCount(), "SetCheckState() - illegal index");

		// Ok, let's set the check state
		BOOL fCheck = FALSE;
		if (ItemState == cs_Checked)
			fCheck = TRUE;

		ListView_SetCheckState(m_hWnd, iItem, fCheck);
	}

	/**************************************************************************
	
	>	ControlState Win32_ListBoxImp::GetCheckState(int iItem)
	
		Author:		Tim Browse
		Date:		15/8/2002
		Inputs:		iItem - the item to examine
		Returns:	cs_Checked - the item is checked
					cs_Unchecked - the item is not checked, or the control
								   does not support checkbox functionality
								   (in which case a COMPLAIN error will occur).
	
		Purpose:	Get the checkbox state on an item.
	
	**************************************************************************/
	
	ControlState Win32_ListBoxImp::GetCheckState(int iItem)
	{
		if (!IsListView())
		{
			// This type of control not supported as having checkboxes yet...
			// TODO: implement checkboxes for all list controls
			vguiERROR();
			return cs_Unchecked;
		}

		// Check for checkbox style on the control
		COMPLAIN_IF((ListView_GetExtendedListViewStyle(m_hWnd) & LVS_EX_CHECKBOXES) == 0,
					"SetCheckState() error - "
					"The listview control does not have the correct style "
					"applied to support checkboxes.\r\n"
					"You should add the LVS_EX_CHECKBOXES style to the control "
					"in the resource editor.");

		// Check index is within the current range
		COMPLAIN_IF(iItem >= GetCount(), "GetCheckState() - illegal index");

		// Ok, let's get the check state
		return ListView_GetCheckState(m_hWnd, iItem) ? cs_Checked : cs_Unchecked;
	}
	
}




