//
// Win32_ListBoxImp.h - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_Win32_ListBoxImp_h
#define vgui_Win32_ListBoxImp_h

#include "..\vgui.h"

#include "..\ListBoxImp.h"

#include "..\Colour.h"

#include <commctrl.h>

namespace vgui
{
	class OwnerDrawListControl;

	// 
	// Specifies what kind of list box control we are attached to.
	//
	typedef enum
	{
		ListType_Unattached,
		ListType_ListBox,			// 'LISTBOX'		- standard Windows listbox
		ListType_ListView,			// 'SysListView32'  - Win95 style list view control
	}
	TListType;

	struct ItemInfo
	{
		// Constructor
		ItemInfo() : m_UserData(0), m_bUseColour(false) {}
		ItemInfo(UserData TheUserData) : m_UserData(TheUserData), m_bUseColour(false) {}

		// UserData associated with this item
		UserData m_UserData;

		// Colour to render the item with (if m_bUseColour is true)
		Colour m_Colour;

		// true => item should be rendered using m_Colour
		bool m_bUseColour;
	};

	class Win32_ListBoxImp : public ListBoxImp
	{
		public:
			//
			// Constructor
			//
			Win32_ListBoxImp();

			// We veto attachments to make sure they're the right
			// type of control.
			virtual void Attach(HWND hWnd);

		public:
			virtual void Clear();

			// Add and delete items

			virtual bool AddItem(const String& ItemName,
								 int* piItem,
								 UserData TheUserData);

			virtual bool InsertItemAfter(int iAfterItem,
										 const String& ItemName, 
										 int* piItem,
										 UserData TheUserData);

			virtual bool InsertItemBefore(int iBeforeItem,
										  const String& ItemName, 
										  int* piItem,
										  UserData TheUserData);

			virtual bool DeleteItem(int iItem);

			virtual bool SetItemText(int iItem, const String& Text, int iColumn);
			virtual bool SetItemData(int iItem, UserData TheUserData);

			// Find out how many items there are in the list
			virtual int GetCount() const;

			// Find info about an item
			virtual bool GetItemText(int iItem, String* pItemName, int iColumn) const;
			virtual bool GetItemUserData(int iItem, UserData* pTheUserData) const;

			// Change currently selected item
			virtual bool SetSelectedItem(int iItem);

			// Access currently selected item
			virtual bool GetSelectedItem(int* piItem) const;

			// Find an item
			virtual bool FindItem(const String& ItemName, int* piFoundItem);

			// Checkbox control
			virtual void EnableCheckboxes(bool bEnable);
			virtual void SetCheckState(int iItem, ControlState ItemState);
			virtual ControlState GetCheckState(int iItem);

			// Sorting control
			virtual void EnableSorting();
			virtual void DisableSorting();

			//
			// Multi-column support
			//

			virtual bool SetNumberOfColumns(uint nColumns);
			virtual uint GetNumberOfColumns();
			virtual void ShowColumnHeaders(bool bShow);
			virtual bool SetColumnName(uint iColumn, const String& ColumnName);
			virtual bool SetColumnWidth(uint iColumn, uint WidthInPixels);
			virtual uint GetColumnWidth(uint iColumn);
			virtual bool AutoSizeColumn(uint iColumn);

			//
			// Redraw control
			//
			virtual void SetRowColour(uint Row, const Colour& RowColour);
			virtual void UseDefaultRowColour(uint Row);
			
			// Owner draw lists - refresh item metrics
			virtual void RemeasureItems();
			virtual void RemeasureItems(int NewHeight);

		protected:
			void CleanUpItemInfoData(); 
			ItemInfo* GetItemInfo(int iItem) const;
			
			// Helper functions for getting list item text
			// (because MS make it so bloody complicated)
			bool GetItemText_ListBox(int iItem, String* pItemName) const;
			bool GetItemText_ListView(int iItem, String* pItemName, int nColumn) const;

			// Helper functions for adding items
			bool InsertItemAt(int iInsertAt,
							  const String& ItemName, 
							  int* piItem,
							  UserData TheUserData);

			bool InsertItemAt(int iInsertAt,
							  const String& ItemName, 
							  int* piItem,
							  ItemInfo* pItemInfo);

			// Translating Win32 listbox notification messages
			virtual ControlTranslation TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																    TMessageType* pMsgType,
																    LRESULT* pResult,
																    MessageParams* pParams);

			// Owner draw handling
			bool CallDrawItem(OwnerDrawListControl* pHandler, LPARAM lParam);
			bool CallMeasureItem(OwnerDrawListControl* pHandler, LPARAM lParam);
			int TranslateListBoxItemState(UINT itemState);

			void DrawListBoxItem(DRAWITEMSTRUCT* pDIS, ItemInfo* pItemInfo);
			ControlTranslation HandleCustomDraw(NMLVCUSTOMDRAW* pNMLVCD,
												LRESULT* pResult);
			
			// Type checking
			inline bool IsListBox()  const { return m_ListType == ListType_ListBox;  }
			inline bool IsListView() const { return m_ListType == ListType_ListView; }

			// Sorting support function
			static int CALLBACK CompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);

			// Hook window destruction, so we can clean up
			LRESULT WindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam);

		private:
			// The type of control we are attached to.
			TListType m_ListType;

			// The column we should sort on (for multi-column list views)
			int m_iSortColumn;

			// Indicates whether custom text colours have been requested
			bool m_bUseCustomTextColours;
	};
};

#endif // !vgui_Win32_ListBoxImp_h


