//
// Win32_MDIApplicationImp.cpp - Application Implementation Interface 
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_MDIApplicationImp.h"

#include "..\DocumentApplication.h"
#include "..\GC.h"
#include "..\Dialog.h"
#include "..\View.h"
#include "Win32_GCImp.h"
#include "Win32_WindowImp.h"
#include "Win32_WindowInfo.h"


#include "Win32_Headers.h"
#include <commctrl.h>
#include <shellapi.h>
#include <objbase.h>

#include "..\Errors.h"


static char *szCommandLine = NULL;

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	HWND Win32_MDIApplicationImp::m_hwndMDIClient = NULL;

	//
	// The first ID to use for the MDI child windows.
	// (Value taken from MFC source code; I assume it's reasonable).
	//
	const int kIDFirstChild = 0xFF00;

	/**************************************************************************

		Win32_MDIApplicationImp::Win32_MDIApplicationImp()

		Author:		Tim Browse
		Date:		31/12/99

		Purpose:	Constructor

	**************************************************************************/

	Win32_MDIApplicationImp::Win32_MDIApplicationImp()
	{
		m_szMDIChildClass[0] = 0;
	}

	/**************************************************************************

		void Fn()

		Author:		Tim Browse
		Date:		31/12/99

		Purpose:	Destructor

	**************************************************************************/

	Win32_MDIApplicationImp::~Win32_MDIApplicationImp()
	{
	}

	/**************************************************************************

		LRESULT CALLBACK Win32_MDIApplicationImp::vGuiFrameWndProc(HWND hWnd, 
																   UINT message, 
																   WPARAM wParam, 
																   LPARAM lParam)

		Author:		Tim Browse
		Date:		31/12/99
		Inputs:		Standard WindowProc parameters
		Returns:	Standard WindowProc return.

		Purpose:	WindowProc for the MDI Frame window.

					Handles creation of the MDIClient window - this is done
					when the WM_CREATE message is sent to the main frame window.

					It provides all the required frame window handling, 
					including passing on messages that should be passed on even
					if they are handled.

	**************************************************************************/

	LRESULT CALLBACK Win32_MDIApplicationImp::vGuiFrameWndProc(HWND hWnd, 
															   UINT message, 
															   WPARAM wParam, 
															   LPARAM lParam)
	{
		// Handle WM_CREATE so that we create the MDI Client window
		if (message == WM_CREATE)
		{
			Win32_MDIApplicationImp *pWin32MDIApp = (Win32_MDIApplicationImp *) GetWin32App();
			pWin32MDIApp->CreateMDIClient(hWnd);

			// All done
			return 0;
		}

		// Is this message handled?
		LRESULT lResult = vGuiRawWndProc(hWnd, message, wParam, lParam);

		if (lResult != kWndProc_NotHandled)
		{
			// Yes - does the MDI FrameProc need this message anyway?
			switch (message)
			{
				case WM_COMMAND:
				case WM_MENUCHAR:
				case WM_SETFOCUS:
				case WM_SIZE:
					// Yes - we must pass this through to the DefFrameproc
					// (See Win32 docs for MDI apps)
					break;

				default:
					// Otherwise, we have handled the message and are done.
					return lResult;
			}
		}

		// Fall back to default behaviour for MDI frame window
		return ::DefFrameProc(hWnd, m_hwndMDIClient, message, wParam, lParam);
	}

	/**************************************************************************

		LRESULT CALLBACK Win32_MDIApplicationImp::vGuiMDIChildWndProc(HWND hWnd, 
																	  UINT message, 
																	  WPARAM wParam, 
																	  LPARAM lParam)
		Author:		Tim Browse
		Date:		1/1/00
		Inputs:		Standard WindowProc parameters
		Returns:	Standard WindowProc return.

		Purpose:	WindowProc for the MDI child window.

					It provides all the required child window handling, 
					including passing on messages that should be passed on even
					if they are handled.

	**************************************************************************/

	LRESULT CALLBACK Win32_MDIApplicationImp::vGuiMDIChildWndProc(HWND hWnd, 
																  UINT message, 
																  WPARAM wParam, 
																  LPARAM lParam)
	{
		// Check for messages that we over-ride the default behaviour for.
		switch (message)
		{
			case WM_CREATE:
			{
				// Get our aplication-defined data...
				CREATESTRUCT* pCS = (CREATESTRUCT*) lParam;
				MDICREATESTRUCT* pMDICS = (MDICREATESTRUCT*) pCS->lpCreateParams;

				// Put the View pointer into our window structure...
				//
				// NB. We passed it in as an LPARAM to CreateMDIWindow() - see
				//     CreateWin32Window().
				Win32_WindowInfo::Attach(hWnd, (View*) pMDICS->lParam);
				break;
			}

			case WM_CLOSE:
			{
				// Get the view object.
				View* pView;
				
				if (!Win32_WindowInfo::GetAttachedView(hWnd, &pView))
				{
					COMPLAIN("Win32_MDIApplicationImp: no View found attached when processing WM_CLOSE");
					break;
				}

				// View is to be closed - first, ask view if this is ok.
				if (pView->OnClose())
				{
					// View has handled closing itself - prevent further 
					// processing of this message.
					return 0;
				}

				// Tell the app that this child (View) should be closed...
				DocumentApplication* pDocApp = (DocumentApplication*) GetApp();
				
				if (pDocApp->OnCloseView(pView))
				{
					// Use standard MDI processing for closing a child window.
					return ::DefMDIChildProc(hWnd, message, wParam, lParam);
				}
				else
				{
					// View was not closed - no further processing needed
					return 0;
				}
			}

			case WM_CHILDACTIVATE:
			{
				// Tell the app that this child (View) is active.
				View* pView;
				
				if (!Win32_WindowInfo::GetAttachedView(hWnd, &pView))
				{
					COMPLAIN("Win32_MDIApplicationImp: no View found attached when processing WM_CHILDACTIVATE");
					break;
				}

				DocumentApplication* pDocApp = (DocumentApplication*) GetApp();
				pDocApp->NotifyViewActivated(pView);

				// Pass through for normal handling - we just want to be
				// kept informed, rather than taking over this message.
				break;
			}
		}

		// Try standard message processing
		LRESULT lResult = vGuiRawWndProc(hWnd, message, wParam, lParam);

		if (message == WM_DESTROY)
		{
			// Special case - remove our view pointer from the window
			Win32_WindowInfo::Detach(hWnd);
		}

		// Was this message handled?
		if (lResult != kWndProc_NotHandled)
		{
			// Yes - does the MDIChildProc need this message anyway?
			switch (message)
			{
				case WM_CHILDACTIVATE:
				case WM_GETMINMAXINFO:
				case WM_MENUCHAR:
				case WM_MOVE:
				case WM_SETFOCUS:
				case WM_SIZE:
				case WM_SYSCOMMAND:
					// Yes - we must pass this through to the DefMDIChildWndProc
					// (See Win32 docs for MDI apps)
					break;

				default:
					// Otherwise, we have handled the message and are done.
					return lResult;
			}
		}

		// Fall back to default behaviour for MDI child window
		return ::DefMDIChildProc(hWnd, message, wParam, lParam);
	}


	/**************************************************************************

		void Fn()

		Author:		Tim Browse
		Date:		31/12/99
		Inputs:		
		Outputs:	
		Returns:	
		Throws:		

		Purpose:	

	**************************************************************************/

	void Win32_MDIApplicationImp::CreateMDIClient(HWND hFrameWnd)
	{
		CLIENTCREATESTRUCT ccs;
		ZeroMemory(&ccs, sizeof(ccs));

		// Retrieve the handle to the Window menu and assign the 
		// first child window identifier.
		ccs.hWindowMenu = FindWindowMenu(hFrameWnd);
		ccs.idFirstChild = kIDFirstChild;

		// Create the MDI client window.  
		m_hwndMDIClient = ::CreateWindow("MDICLIENT",
										 (LPCTSTR) NULL,
										 WS_CHILD | WS_CLIPCHILDREN | WS_VSCROLL | WS_HSCROLL,// | MDIS_ALLCHILDSTYLES,
										 0, 0, 0, 0, 
										 hFrameWnd, 
										 0,
										 GetInstance(), 
										 (LPSTR) &ccs);

		::ShowWindow(m_hwndMDIClient, SW_SHOW);     
	}

	/**************************************************************************

		void Fn()

		Author:		Tim Browse
		Date:		31/12/99
		Inputs:		
		Outputs:	
		Returns:	
		Throws:		

		Purpose:	

	**************************************************************************/

	HMENU Win32_MDIApplicationImp::FindWindowMenu(HWND hFrameWnd)
	{
		// Find the main menu.
		HMENU hMainMenu = ::GetMenu(hFrameWnd);
		
		// If it didn't work, we can't help
		if (hMainMenu == NULL)
			return NULL;

		// This is attempting to guess which sub-menu is the Window menu.
		// The Windows user interface guidelines say that the right-most
		// menu on the menu bar should be Help and Window should be one
		// to the left of that.
		int iMenu = ::GetMenuItemCount(hMainMenu) - 2;

		// If this assertion fails, your menu bar does not follow the guidelines
		// so you will have to override this function.
		COMPLAIN_IF(iMenu < 0, "Unable to find the Window menu for MDI frame.");
		HMENU hWindowMenu = ::GetSubMenu(hMainMenu, iMenu);
		COMPLAIN_IF(hWindowMenu == NULL, "Unable to find the Window menu for MDI frame.");

		// All done
		return hWindowMenu;
	}

	/**************************************************************************

		void Fn()

		Author:		Tim Browse
		Date:		??/??/??
		Inputs:		
		Outputs:	
		Returns:	
		Throws:		

		Purpose:	

	**************************************************************************/

	//
	//  PURPOSE: Registers the window class.
	//
	//  COMMENTS:
	//
	//    This function and its usage is only necessary if you want this code
	//    to be compatible with Win32 systems prior to the 'RegisterClassEx'
	//    function that was added to Windows 95. It is important to call this function
	//    so that the application will get 'well formed' small icons associated
	//    with it.
	//
	ATOM Win32_MDIApplicationImp::RegisterMainWindowClass()
	{
		// Get application object
		Application *pApp = GetApp();

		//
		// First, we register a window class for MDI child windows that we
		// will be creating.
		//

		// Make a unique name for the class
		wsprintf(m_szMDIChildClass, "vguiMDIChild_%x", (int) ::GetModuleHandle(NULL));

		WNDCLASSEX wcex;

		ZeroMemory(&wcex, sizeof(wcex));
		wcex.cbSize = sizeof(WNDCLASSEX); 

		wcex.style			= 0;
		wcex.lpfnWndProc	= (WNDPROC) vGuiMDIChildWndProc;
		wcex.cbClsExtra		= 0;
		wcex.cbWndExtra		= 0;
		wcex.hInstance		= m_hInstance;
		wcex.hIcon			= LoadIcon(wcex.hInstance, pApp->GetIconID().GetID());
		wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
		wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
		wcex.lpszMenuName	= NULL;
		wcex.lpszClassName	= m_szMDIChildClass;

		::RegisterClassEx(&wcex);

		//
		// Now, we register the main window class...
		//

		// Make a unique name for the class
		wsprintf(m_szWindowClass, "vguiWindow_%x", (int) ::GetModuleHandle(NULL));

		ZeroMemory(&wcex, sizeof(wcex));
		wcex.cbSize = sizeof(WNDCLASSEX); 

		wcex.style			= 0;
		wcex.lpfnWndProc	= (WNDPROC) vGuiFrameWndProc;
		wcex.cbClsExtra		= 0;
		wcex.cbWndExtra		= 0;
		wcex.hInstance		= m_hInstance;
		wcex.hIcon			= LoadIcon(wcex.hInstance, pApp->GetIconID().GetID());
		wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
		wcex.hbrBackground	= (HBRUSH)(COLOR_APPWORKSPACE+1);
		wcex.lpszMenuName	= pApp->GetMenuID().GetID();
		wcex.lpszClassName	= m_szWindowClass;
		wcex.hIconSm		= LoadIcon(wcex.hInstance, pApp->GetSmallIconID().GetID());

		return RegisterClassEx(&wcex);
	}

	HWND Win32_MDIApplicationImp::CreateWin32Window(Window *pWindow,
												    const Window* pParent,
												    TWindowStyle Style, 
												    const String* psWindowTitle)
	{
		// We only over-ride document view windows
		if (Style != ws_DocumentView)
		{
			// Defer to base class
			return Win32_ApplicationImp::CreateWin32Window(pWindow, 
														   pParent, 
														   Style, 
														   psWindowTitle);
		}

		LPCTSTR szWindowTitle = NULL;

		if (psWindowTitle != NULL)
		{
			szWindowTitle = psWindowTitle->c_str();
		}

		// This type of window should always be a View-derived class...
		const View* pView = dynamic_cast<const View*>(pWindow);

		if (pView == NULL)
		{
			// Cast didn't work, so the window is not a view, which is not
			// allowed - MDI child windows *must* be View objects.
			COMPLAIN("CreateWin32Window(): attempt made to make a View class window "
					 "without a View object.");
			return NULL;
		}

		// TODO: Allow customisation of initial state (e.g. scrollbars, maximised, etc)
		HWND hWnd = ::CreateMDIWindow(m_szMDIChildClass,
									  (LPTSTR) szWindowTitle,
									  0,
									  CW_USEDEFAULT, CW_USEDEFAULT,
									  CW_USEDEFAULT, CW_USEDEFAULT,
									  m_hwndMDIClient,
									  GetInstance(),
									  (LPARAM) pView);

		// vGuiMDIChildWndProc() will fill in the Window pointer in the 
		// GWL_USERDATA slot in response to WM_CREATE.

		// Pass back window handle to caller
		return hWnd;
	}

	WPARAM Win32_MDIApplicationImp::DoMessageLoop()
	{
		// Message loop for MDI apps
		MSG msg;

		while (!m_bQuitReceived && ::GetMessage(&msg, NULL, 0, 0)) 
		{
			if (!::TranslateMDISysAccel(m_hwndMDIClient, &msg) && 
				!::TranslateAccelerator(m_hMainWindow, m_hAccelTable, &msg)) 
			{
				::TranslateMessage(&msg);
				::DispatchMessage(&msg);
			}
		}

		return msg.wParam;
	}

	Win32_MDIApplicationImp* GetWin32MDIApp()
	{
		return (Win32_MDIApplicationImp *) GetWin32App();
	}

};
