//
// Win32_SliderImp.cpp - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_SliderImp.h"

#include "Win32_MessageInfo.h"

#include "vgui\Errors.h"

#include <commctrl.h>

namespace vgui
{
	void Win32_SliderImp::Attach(HWND hWnd)
	{
		// We have to check that the class name is correct.
		if (!IsWindowOfClass(hWnd, "msctls_trackbar32"))
		{
			// Trying to attach to the wrong kind of control - it must be an edit field.
			vguiERROR();
			return;
		}

		// Ok, we allow it, so pass down to base class.
		SliderImp::Attach(hWnd);
	}

	/**************************************************************************
	
	>	void Win32_SliderImp::SetRange(int Min, int Max)
	
		Author:		Tim Browse
		Date:		30/3/2002
		Inputs:		Min - minimum value of slider
					Max - maximum value of slider
	
		Purpose:	Set slider range.
	
	**************************************************************************/
	
	void Win32_SliderImp::SetRange(int Min, int Max)
	{
		// Slider will be redrawn
		BOOL fRedraw = TRUE;

		// Tell slider about range update
		SendMessage(TBM_SETRANGE, (WPARAM) fRedraw, (LPARAM) MAKELONG(Min, Max)); 
	}

	/**************************************************************************
	
	>	void Win32_SliderImp::GetRange(int* pMin, int* pMax) const
	
		Author:		Tim Browse
		Date:		30/3/2002
		Outputs:	pMin - the minimum value of slider
					pMax - the maximum value of slider
	
		Purpose:	Get the max/min range of the slider.
	
	**************************************************************************/
	
	void Win32_SliderImp::GetRange(int* pMin, int* pMax) const
	{
		// Ask control for range values
		*pMin = SendMessage(TBM_GETRANGEMIN, 0, 0);
		*pMax = SendMessage(TBM_GETRANGEMAX, 0, 0);
	}

	/**************************************************************************
	
	>	void Win32_SliderImp::SetPos(int Pos)
	
		Author:		Tim Browse
		Date:		30/3/2002
		Inputs:		Pos - new slider position
	
		Purpose:	Set the current position for the slider - should be within
					the range set by SetRange().
	
	**************************************************************************/
	
	void Win32_SliderImp::SetPos(int Pos)
	{
		SendMessage(TBM_SETPOS, (WPARAM) TRUE, (LPARAM) Pos);
	}

	/**************************************************************************
	
	>	int Win32_SliderImp::GetPos() const
	
		Author:		Tim Browse
		Date:		30/3/2002
		Returns:	The current slider position
	
		Purpose:	Get the current slider position.
	
	**************************************************************************/
	
	int Win32_SliderImp::GetPos() const
	{
		return (int) SendMessage(TBM_GETPOS, 0, 0);
	}
	
	/**************************************************************************

	>	ControlTranslation Win32_SliderImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																	     TMessageType* pMsgType,
																	     LRESULT* pResult,
																	     MessageParams* pParams)

		Author:		Tim Browse
		Date:		22/02/2001
		Inputs:		See Win32_WindowImp::TranslateControlNotification().
		Outputs:	See Win32_WindowImp::TranslateControlNotification().
		Returns:	See Win32_WindowImp::TranslateControlNotification().

		Purpose:	Translate Win32 slider notification messages into
					standardised vgui messages.

	**************************************************************************/

	ControlTranslation Win32_SliderImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																	 TMessageType* pMsgType,
																	 LRESULT* pResult,
																     MessageParams* pParams)
	{
		switch (MsgInfo.m_uNotifyCode)
		{
			// Sliders sent notifications via WM_[HV]SCROLL messages...
			case WM_HSCROLL:
			case WM_VSCROLL:
			{
				switch (LOWORD(MsgInfo.m_Msg.wParam))
				{
					// Messages we understand...
					case TB_BOTTOM:			*pMsgType = mt_SLN_MINIMUM;		break;
					case TB_TOP:			*pMsgType = mt_SLN_MAXIMUM;		break;

					case TB_THUMBTRACK:		*pMsgType = mt_SLN_TRACKPOS;	break;
					case TB_ENDTRACK:		*pMsgType = mt_SLN_ENDTRACK;	break;

					case TB_LINEDOWN:		*pMsgType = mt_SLN_DOWN;		break;
					case TB_LINEUP:			*pMsgType = mt_SLN_UP;			break;

					case TB_PAGEDOWN:		*pMsgType = mt_SLN_PAGEDOWN;	break;
					case TB_PAGEUP:			*pMsgType = mt_SLN_PAGEUP;		break;

					case TB_THUMBPOSITION:	*pMsgType = mt_SLN_POSCHANGE;	break;

					default:
						return ControlTranslation_Unknown;
				}

				break;
			}

			default:
				return ControlTranslation_Unknown;
		}

		// Message converted
		return ControlTranslation_OK;
	}

	bool Win32_SliderImp::Create(Window* pParent, 
								  const Rect& rcPosition,
								  ControlID ID)
	{
		// Access the Win32 object for the parent
		Win32_WindowImp* pParentImp = GetImp(pParent);

		if (pParentImp != NULL)
		{
			// Create an edit box in the specified window
			HWND hWnd = pParentImp->CreateChildControl(m_pWindow,
			 									    "msctls_trackbar32", "",
												    WS_BORDER, 0,
													rcPosition,
												    NOCTL);
			if (hWnd != NULL)
			{
				Attach(hWnd);
				return true;
			}
		}

		// Failed to create window
		return false;
	}
};