//
// Win32_StaticControl.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_StaticControl.h"

#include "Win32_WindowClasses.h"
#include "Win32_ApplicationImp.h"
#include "Win32_Theme.h"

#include "..\Errors.h"

#include <tmschema.h>

namespace vgui
{
	/**************************************************************************
	
	>	Win32_StaticControl::Win32_StaticControl()
	
		Author:		Tim Browse
		Date:		20/8/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	Win32_StaticControl::Win32_StaticControl() : m_Image(StaticControlImage_None)
	{
		// TCILB
	}

	/**************************************************************************

	>	void Win32_StaticControl::Create(HWND hParent,
 										 const Rect& Position,
										 ControlID ID)


		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hParent - the window to create the control in.
					Position - where to create the control
					ID - the control ID that the new control should use.

		Purpose:	Create a static control in the specified window, at the
					specified position, with the supplied control ID.

	**************************************************************************/

	void Win32_StaticControl::Create(HWND hParent,
 									 const Rect& Position,
									 ControlID ID)
	{
		DWORD dwStyle = WS_CHILDWINDOW|WS_VISIBLE;

		m_hWnd = ::CreateWindow(wcszStaticControlClass, 
    							"",
    							dwStyle,
    							Position.left, Position.top,
    							Position.GetWidth(), Position.GetHeight(),
    							hParent,
    							(HMENU) ID,
    							GetWin32App()->GetInstance(), 
    							NULL);

		if (m_hWnd != NULL)
		{
			Attach(m_hWnd);
		}
		else
		{
			COMPLAIN("Win32_StaticControl::Create() - failed to create control.");
		}
	}

	/**************************************************************************
	
	>	void Win32_StaticControl::OnRedraw(HWND hWnd, PAINTSTRUCT* pPS)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hWnd - the control window being redrawn
					pPS - the PAINTSTRUCT to use to redraw the control.
	
		Purpose:	Provide simple redrawing for various types of control
					images.

					Currently only resize gripper appearance is supported.
	
	**************************************************************************/
	
	void Win32_StaticControl::OnRedraw(HWND hWnd, PAINTSTRUCT* pPS)
	{
		switch (m_Image)
		{
			case StaticControlImage_ResizeGripper:
			{
				RECT rcClient;
				::GetClientRect(hWnd, &rcClient);

				HTHEME hTheme;
				if (GetThemeHandle(&hTheme))
				{
					// Use XP theme to draw the gripper
					Win32_Theme::DrawThemeBackground(hTheme, pPS->hdc,
													 SBP_SIZEBOX, SZB_RIGHTALIGN,
													 &rcClient, &pPS->rcPaint);
				}
				else
				{
					// Otherwise, use standard Windows 95 API to draw
					// the gripper.
					::DrawFrameControl(pPS->hdc, &rcClient, 
									   DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
				}

				break;
			}
		}
	}

	/**************************************************************************
	
	>	void Win32_StaticControl::Display(TStaticControlImage Image)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		Image - the appearance to use for this control.
	
		Purpose:	Change the display appearance of this control object.
	
	**************************************************************************/
	
	void Win32_StaticControl::Display(TStaticControlImage Image)
	{
		m_Image = Image;

		// Invalidate the theme handle, as we may have just changed the
		// class of the control
		InvalidateTheme();

		// Ensure control is redrawn because of this change.
		if (m_hWnd != NULL)
		{
			::InvalidateRect(m_hWnd, NULL, TRUE);
		}
	}

	/**************************************************************************
	
	>	LPCWSTR Win32_StaticControl::GetThemeClass() const
	
		Author:		Tim Browse
		Date:		20/8/2002
		Returns:	Name of the theme class (wide character string)
	
		Purpose:	Get the theme 'class' for this control.  This lets the XP
					theme API know what type of control this is, so it knows
					how to behave when DrawThemeBackground() etc are called.
	
	**************************************************************************/
	
	LPCWSTR Win32_StaticControl::GetThemeClass() const
	{
		switch (m_Image)
		{
			case StaticControlImage_ResizeGripper:
			{
				return L"SCROLLBAR";
			}

			default:
			{
				return L"WINDOW";
			}
		}
	}
};



