//
// Win32_TabControlImp.cpp - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_TabControlImp.h"

#include "Win32_ApplicationImp.h"
#include "Win32_DialogImp.h"
#include "Win32_Headers.h"
#include "Win32_WindowClasses.h"
#include "Win32_MessageInfo.h"

#include "..\Rect.h"

#include "vos\Errors.h"
#include "vos\Trace.h"

#include <commctrl.h>

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	/**************************************************************************
	
	>	Win32_TabControlImp::Win32_TabControlImp()
	
		Author:		Tim Browse
		Date:		23/6/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	Win32_TabControlImp::Win32_TabControlImp() : m_cTabs(0)
	{
		// TCILB
	}

	void Win32_TabControlImp::Attach(HWND hWnd)
	{
		// We have to check that the class name is correct.
		if (!IsWindowOfClass(hWnd, wcszTabControlClass))
		{
			// Trying to attach to the wrong kind of control - it must be a button.
			vguiERROR();
			return;
		}

		// Ok, we allow it, so pass down to base class.
		TabControlImp::Attach(hWnd);
	}

	bool Win32_TabControlImp::Create(Window* pParent, 
									 const Rect& Position,
									 ControlID ID)
	{
		// Ask parent window to make child control
		Win32_WindowImp* pParentImp = GetImp(pParent);

		if (pParentImp != NULL)
		{
			m_hWnd = pParentImp->CreateChildControl(m_pWindow,
													wcszTabControlClass, "",
													0, 0,
													Position, 
													ID);
		}

		return (m_hWnd != NULL);
	}

	/**************************************************************************

	>	ControlTranslation 
			Win32_TabControlImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
															  TMessageType* pMsgType,
															  LRESULT* pResult,
															  MessageParams* pParams)

		Author:		Tim Browse
		Date:		21/06/2002
		Inputs:		See Win32_WindowImp::TranslateControlNotification().
		Outputs:	See Win32_WindowImp::TranslateControlNotification().
		Returns:	See Win32_WindowImp::TranslateControlNotification().

		Purpose:	Translate Win32 button notification messages into
					standardised vgui messages.

	**************************************************************************/

	ControlTranslation 
		Win32_TabControlImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
														  TMessageType* pMsgType,
														  LRESULT* pResult,
														  MessageParams* pParams)
	{
		if (MsgInfo.m_uNotifyCode == TCN_SELCHANGE)
		{
			// Find out which tab is the new selected tab
			int iTab = TabCtrl_GetCurSel(m_hWnd);
			COMPLAIN_IF(iTab < 0, "Current tab index is negative!");

			// Show/hide the child dialogs as appropriate
			ShowTab(iTab);

			// Send a message to the dialog about this, and then exit
			*pMsgType = mt_TAB_SELECTED;
			return ControlTranslation_OK;
		}

		// Message unknown to TabControl
		return ControlTranslation_Unknown;
	}

	void Win32_TabControlImp::ShowTab(uint iTab)
	{
		// Show/hide the child dialogs as appropriate
		m_Tabs.ShowTab(iTab);

		const Win32_TabInfo* pTabInfo;
		if (m_Tabs.FindByIndex(iTab, &pTabInfo))
		{
			AdjustTabPosition(pTabInfo);
		}
	}


	/**************************************************************************

	>	bool Win32_TabControlImp::AddTab(const DialogResourceID& TabID, 
						 				 const vos::String& TabName)


		Author:		Tim Browse
		Date:		21/6/2002
		Inputs:		TabID - the dialog template to add as a new tab
					TabName - name of the tab to display on the tab itself.
		Returns:	true => tab added ok
					false => otherwise (e.g. template not found)

		Purpose:	Add a new tab to the control, using the specified dialog
					template to fill the tab, and the specified name to use
					as the tab's name.

	**************************************************************************/

	bool Win32_TabControlImp::AddTab(const DialogResourceID& TabID, 
					 				 const vos::String& TabName)
	{
		// Create the new child dialog box. 
		//
		// NB. We make it a child of the tab's parent - if we make it a child
		//	   of the tab, then sometimes/often the Windows dialog code gets
		//	   confused and sends us an infinite number of WM_GETDLGCODE 
		//	   messages, and the app freezes.  See MSKB article Q149501 for
		//	   an example of this happening (NB. the suggested workaround in
		//	   that article does NOT work in this general case).
		//
		Dialog* pChildDialog = new Dialog(TabID, m_pWindow->GetParent());

		if (!pChildDialog->Create())
		{
			COMPLAIN("Unable to create dialog for tab controls!");
			return false;
		}

		// Hide window initially
		pChildDialog->ShowWindow(sw_HIDE);

		// Attach to parent, and route control messages in the child 
		// dialog to our parent.
		Win32_DialogImp* pWin32Dialog = (Win32_DialogImp*) pChildDialog->GetImp();
		pWin32Dialog->SetControlHandler(m_pParent);

		// Move this window to be above the tab control in window Z order,
		// so that redraw/tab ordering in dialogs works as expected.
		HWND hwndInsertAfter = ::GetWindow(m_hWnd, GW_HWNDPREV);
		if (hwndInsertAfter == NULL)
			hwndInsertAfter = HWND_TOP;

		HWND hChildDialog = pWin32Dialog->GetWindowHandle();

		::SetWindowPos(hChildDialog, hwndInsertAfter, 0, 0, 0, 0, 
					   SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOREPOSITION|SWP_NOSIZE);
		
		// Check to see if the client has set the 'Control' flag, which means
		// the child window will play nice with the others and take part in
		// tab ordering operations etc.
		COMPLAIN_IF(((::GetWindowLong(hChildDialog, GWL_STYLE) & DS_CONTROL) == 0),
					"TabControl::AddTab() - the dialog supplied does not have the DS_CONTROL "
					"style bit set.  This means the controls in the tab will "
					"not work correctly with tab ordering (i.e. using the TAB key "
					"to move between controls will not result in the correct order).");

		// Ok, add this as a new tab to the Win32 control
		TCITEM TabItem;
		ZeroMemory(&TabItem, sizeof(TabItem));
		TabItem.mask = TCIF_TEXT|TCIF_IMAGE;
		TabItem.iImage = -1;

		// Strictly speaking, we shouldn't use this cast, but MS couldn't
		// be bothered to declare their API properly...
		TabItem.pszText = (char*) TabName.c_str();

		int iNewTab = m_cTabs;
		iNewTab = SendMessage(TCM_INSERTITEM, (WPARAM) iNewTab, (LPARAM) &TabItem);

		// Did it work?
		if (iNewTab == -1)
		{
			// No...
			delete pChildDialog;
			return false;
		}

		// Tab added ok, so add tab details to our own list
		m_Tabs.AddTab(pChildDialog, TabID, iNewTab);
		::InvalidateRect(m_hWnd, NULL, TRUE);
		m_cTabs++;

		// If it's the first tab to be added, then show it
		if (m_cTabs == 1)
			ShowTab(iNewTab);
		
		return true;
	}

	/**************************************************************************

	>	bool Win32_TabControlImp::SelectTab(const DialogResourceID& TabID)

		Author:		Tim Browse
		Date:		21/6/2002
		Inputs:		TabID - the ID of the tab to select
		Returns:	true => tab found and selected ok;
					false => otherwise

		Purpose:	Select the tab that uses the specified dialog template
					ID.

	**************************************************************************/

	bool Win32_TabControlImp::SelectTab(const DialogResourceID& TabID)
	{
		// Find the tab index for this resource ID
		const Win32_TabInfo* pTabInfo;
		if (m_Tabs.FindByID(TabID, &pTabInfo))
		{
			SendMessage(TCM_SETCURSEL, (WPARAM) pTabInfo->m_Index, 0);

			// TCM_SETCURSEL doesn't cause a TCN_SELCHANGE notification,
			// so we show the correct tab manually.
			ShowTab(pTabInfo->m_Index);
			return true;
		}
		else
		{
			COMPLAIN("Internal error: Unable to find the tab for this dialog ID!");
			return false;
		}
	}

	/**************************************************************************

	>	DialogResourceID Win32_TabControlImp::GetSelectedTab() const

		Author:		Tim Browse
		Date:		21/6/2002
		Returns:	ID of the currently selected tab in the control (if there
					is no currently selected tab, the null resource ID is
					returned - use IsNull() to check for this).

		Purpose:	Find out which tab is currently selected (i.e. the tab
					which is visible) in the control.

	**************************************************************************/

	DialogResourceID Win32_TabControlImp::GetSelectedTab() const
	{
		// Get the new selection index
		int iSel = TabCtrl_GetCurSel(m_hWnd); 

		// Get the dialog ID for this tab
		const Win32_TabInfo* pTabInfo;
		if (m_Tabs.FindByIndex(iSel, &pTabInfo))
		{
			return pTabInfo->m_ResourceID;
		}
		else
		{
			COMPLAIN("Internal error: Unable to find the dialog for this tab!");
			return 0;
		}
	}

	bool Win32_TabControlImp::OnMove(int, int)
	{
		AdjustCurrentTabPosition();
		return false;
	}

	bool Win32_TabControlImp::OnResize(int NewWidth, int NewHeight)
	{
		AdjustCurrentTabPosition();
		return false;
	}

	void Win32_TabControlImp::AdjustCurrentTabPosition()
	{
		// Find currently selected tab...
		int iTab = TabCtrl_GetCurSel(m_hWnd); 

		const Win32_TabInfo* pTabInfo;
		if (m_Tabs.FindByIndex(iTab, &pTabInfo))
		{
			AdjustTabPosition(pTabInfo);
		}
	}

	void Win32_TabControlImp::AdjustTabPosition(const Win32_TabInfo* pTabInfo)
	{
		// And now resize it to fit the current tab control size
		vgui::Rect rcTabControl;
		GetPosition(&rcTabControl);
		vgui::Rect rcTabDisplay(rcTabControl);
		TabCtrl_AdjustRect(m_hWnd, FALSE, (RECT*) &rcTabDisplay);

		// Move dialog to here
		pTabInfo->m_pDialog->Move(rcTabDisplay);
	}

};
