//
// Win32_TabList.cpp - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_TabList.h"

#include "Win32_WindowImp.h"

#include "Win32_WindowClasses.h"

#include "vos\Errors.h"
#include "vos\Trace.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	/**************************************************************************
	
	>	Win32_TabList::~Win32_TabList()
	
		Author:		Tim Browse
		Date:		13/7/2002
	
		Purpose:	Destructor.  This dialogs registered with the tab list
					(via AddTab()) are destroyed and deleted.
	
	**************************************************************************/
	
	Win32_TabList::~Win32_TabList()
	{
		// Clean up the dialogs used for this tab list.
		for (TabInfoVector::const_iterator it = m_Tabs.begin();
			 it != m_Tabs.end();
			 ++it)
		{
			it->m_pDialog->Destroy();
			delete it->m_pDialog;
		}
	}

	/**************************************************************************

	>	void Win32_TabList::AddTab(Dialog* m_pDialog,
								   const DialogResourceID& ID,
								   uint Index)


		Author:		Tim Browse
		Date:		21/6/2002
		Inputs:		pDialog - the actual dialog object for this tab
					ID - the resource ID where the template dat acame from.
					Index - the integer index of the tab, used to identify the
							tab to the Win32 API.

		Purpose:	Add info about a tab to the list.

	**************************************************************************/

	void Win32_TabList::AddTab(Dialog* pDialog,
							   const DialogResourceID& ID,
							   uint Index)
	{
		Win32_TabInfo Info;
		Info.m_pDialog = pDialog;
		Info.m_ResourceID = ID;
		Info.m_Index = Index;
		m_Tabs.push_back(Info);
	}

	/**************************************************************************
	
	>	void Win32_TabList::ShowTab(uint iTab)
	
		Author:		Tim Browse
		Date:		23/6/2002
		Inputs:		iTab - index of the tab to show
	
		Purpose:	Show the specified tab, and ensure all other tabs are
					hidden.
	
	**************************************************************************/
	
	void Win32_TabList::ShowTab(uint iTab)
	{
		for (TabInfoVector::const_iterator it = m_Tabs.begin();
			 it != m_Tabs.end();
			 ++it)
		{
			 if (it->m_Index == iTab)
			 {
				 it->m_pDialog->ShowWindow(sw_SHOW);
				 ::EnableWindow(it->m_pDialog->GetWindowHandle(), TRUE);

				 // Give first control of the tab dialog the focus
				 // (according to tab order of dialog)
				 GiveFocusToFirstControl(it->m_pDialog->GetWindowHandle());
			 }
			 else
			 {
				 it->m_pDialog->ShowWindow(sw_HIDE);
				 ::EnableWindow(it->m_pDialog->GetWindowHandle(), FALSE);
			 }
		}
	}

	/**************************************************************************
	
	>	void Win32_TabList::GiveFocusToFirstControl(HWND hDialog)
	
		Author:		Tim Browse
		Date:		13/7/2002
		Inputs:		hDialog - the dialog to give focus to
	
		Purpose:	Find the first appropriate control in the dialog, and
					give it the focus.

					We have to be careful when giving controls focus, because
					for some controls (e.g. unselected radio buttons) it can
					change the control's state.

	**************************************************************************/
	
	void Win32_TabList::GiveFocusToFirstControl(HWND hDialog)
	{
		HWND hChild = ::GetWindow(hDialog, GW_CHILD);
		HWND hControl = ::GetWindow(hChild, GW_HWNDFIRST);

		// Check each control in turn until we find a 'safe' one
		for (hControl = ::GetWindow(hChild, GW_HWNDFIRST);
			 hControl != NULL;
			 hControl = ::GetWindow(hControl, GW_HWNDNEXT))
		{
			// Check to see if we can give focus to this control
			// safely...

			if (Win32_WindowImp::IsWindowOfClass(hControl, wcszStaticControlClass))
			{
				// It's a static label - we can give this focus, but it
				// might result in the control after it being given focus,
				// which might be innapropriate, so we skip it.
				continue;
			}

			if (Win32_WindowImp::IsWindowOfClass(hControl, wcszButtonControlClass))
			{
				// It's a button - is it a radio button?
				if (Win32_WindowImp::WindowHasStyle(hControl, BS_RADIOBUTTON) ||
					Win32_WindowImp::WindowHasStyle(hControl, BS_AUTORADIOBUTTON))
				{
					// Yes, so if we give it focus, it will select the
					// radio button, which may change the radio group's
					// setting.
					//
					// We don't want this, so if this radio button is not
					// selected, we skip this control.
					if (::SendMessage(hControl, BM_GETCHECK, 0, 0) != BST_CHECKED)
					{
						continue;
					}
				}
			}

			// If we get here, we've found a control that it is safe
			// to give focus to
			::SetFocus(hControl);
			return;
		}
	}

	/**************************************************************************

	>	bool Win32_TabList::FindByID(const DialogResourceID& ID, 
									 const Win32_TabInfo** ppTabInfo) const


		Author:		Tim Browse
		Date:		21/6/2002
		Inputs:		ID - the ID to search for
		Outputs:	ppTabInfo - the tab info, if found
		Returns:	true => the tab information was found ok
					false => tab info not found.

		Purpose:	Find a tab in the control, based on the resource ID that
					was used to load the dialog template for the tab.

	**************************************************************************/

	bool Win32_TabList::FindByID(const DialogResourceID& ID, 
								 const Win32_TabInfo** ppTabInfo) const
	{
		TabInfoVector::const_iterator it = m_Tabs.begin();

		while (it != m_Tabs.end())
		{
			if (it->m_ResourceID == ID)
			{
				// Found the tab
				*ppTabInfo = &(*it);
				return true;
			}

			++it;
		}

		// Not found
		return false;
	}

	/**************************************************************************

	>	bool Win32_TabList::FindByIndex(uint Index, 
										const Win32_TabInfo** ppTabInfo)


		Author:		Tim Browse
		Date:		21/6/2002
		Inputs:		Index - the tab index to search for
		Outputs:	ppTabInfo - the tab info, if found
		Returns:	true => the tab information was found ok
					false => tab info not found.

		Purpose:	Find a tab in the control, based on the index of the tab
					in the tab control.

	**************************************************************************/

	bool Win32_TabList::FindByIndex(uint Index, 
									const Win32_TabInfo** ppTabInfo) const
	{
		TabInfoVector::const_iterator it = m_Tabs.begin();

		while (it != m_Tabs.end())
		{
			if (it->m_Index == Index)
			{
				// Found the tab
				*ppTabInfo = &(*it);
				return true;
			}

			++it;
		}

		// Not found
		return false;
	}
};

