//
// Win32_Theme.cpp - Support for Windows XP themes
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_Theme.h"

#include "..\Errors.h"

namespace vgui
{
	//
	// Function pointer types for the theme API
	//
	typedef BOOL (STDAPICALLTYPE *tfnIsThemeActive)();
	typedef HTHEME (STDAPICALLTYPE *tfnOpenThemeData)(HWND, LPCWSTR);
	typedef HRESULT (STDAPICALLTYPE *tfnCloseThemeData)(HTHEME);
	typedef HRESULT (STDAPICALLTYPE *tfnDrawThemeBackground)(HTHEME, HDC,
															 int, int,
															 const RECT*,
															 const RECT*);

	//
	// Function pointers for the Theme API
	//
	// These are left NULL if we cannot load the theme library.
	static tfnIsThemeActive s_pfnIsThemeActive = NULL;
	static tfnOpenThemeData s_pfnOpenThemeData = NULL;
	static tfnCloseThemeData s_pfnCloseThemeData = NULL;
	static tfnDrawThemeBackground s_pfnDrawThemeBackground = NULL;


	void Win32_Theme::Initialise()
	{
		// Try to load the theme library
		HMODULE hUxTheme = ::LoadLibrary("uxtheme.dll");

		// if it worked, get pointers to the functions we want to use
		if (hUxTheme != NULL)
		{
			s_pfnIsThemeActive = (tfnIsThemeActive) ::GetProcAddress(hUxTheme, "IsThemeActive");
			s_pfnOpenThemeData = (tfnOpenThemeData) ::GetProcAddress(hUxTheme, "OpenThemeData");
			s_pfnCloseThemeData	= (tfnCloseThemeData) ::GetProcAddress(hUxTheme, "CloseThemeData");
			s_pfnDrawThemeBackground = (tfnDrawThemeBackground) ::GetProcAddress(hUxTheme, "DrawThemeBackground");
		}
		else
		{
			// This platform does not support XP themes
		}
	}

	bool Win32_Theme::IsThemeActive()
	{
		if (s_pfnIsThemeActive != NULL)
			return ((*s_pfnIsThemeActive)() != 0);
		else
			return false;
	}

	HTHEME Win32_Theme::OpenThemeData(HWND hWnd, LPCWSTR pszClassList)
	{
		if (s_pfnOpenThemeData != NULL)
			return (*s_pfnOpenThemeData)(hWnd, pszClassList);
		else
		{
			COMPLAIN("OpenThemeData() - no theme API available! "
					 "Did you call IsThemeActive()?");
			return NULL;
		}
	}

	HRESULT Win32_Theme::CloseThemeData(HTHEME hTheme)
	{
		if (s_pfnCloseThemeData != NULL)
			return (*s_pfnCloseThemeData)(hTheme);
		else
		{
			COMPLAIN("CloseThemeData() - no theme API available! "
					 "Did you call IsThemeActive()?");
			return NULL;
		}
	}

	HRESULT Win32_Theme::DrawThemeBackground(HTHEME hTheme,
										     HDC hdc,
										     int iPartId,
										     int iStateId,
										     const RECT* pRect,
										     const RECT* pClipRect)
	{
		if (s_pfnCloseThemeData != NULL)
			return (*s_pfnDrawThemeBackground)(hTheme, hdc, iPartId, iStateId,
											  pRect, pClipRect);
		else
		{
			COMPLAIN("DrawThemeBackground() - no theme API available! "
					 "Did you call IsThemeActive()?");
			return NULL;
		}
	}
};

