//
// Win32_ThemedObject.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_ThemedObject.h"

#include "Win32_Theme.h"

#include "..\Controls.h"

#include <tmschema.h>

namespace vgui
{
	/**************************************************************************
	
	>	Win32_ThemedObject::Win32_ThemedObject()
	
		Author:		Tim Browse
		Date:		21/8/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	Win32_ThemedObject::Win32_ThemedObject() : m_hTheme(NULL),
											   m_hWnd(NULL)
	{
		// TCILB
	}

	/**************************************************************************
	
	>	Win32_ThemedObject::~Win32_ThemedObject()
	
		Author:		Tim Browse
		Date:		20/8/2002
	
		Purpose:	Destructor - cleans up an XP Theme handle used.
	
	**************************************************************************/
	
	Win32_ThemedObject::~Win32_ThemedObject()
	{
		FreeThemeHandle();
	}

	/**************************************************************************
	
	>	LPCWSTR Win32_ThemedObject::GetThemeClass() const
	
		Author:		Tim Browse
		Date:		21/8/2002
		Returns:	Name of this object's theme class.
	
		Purpose:	Get the theme class for this object (see Win32 function
					OpenThemeData() for an explanation of theme classes).
	
	**************************************************************************/
	
	LPCWSTR Win32_ThemedObject::GetThemeClass() const
	{
		return L"WINDOW";
	}

	/**************************************************************************
	
	>	void Win32_ThemedObject::InvalidateTheme()
	
		Author:		Tim Browse
		Date:		20/8/2002
	
		Purpose:	This should be called when the theme class of the object
					has changed.  The next time UsingTheme() is called, a
					new theme handle will be obtained.

					(This assumes you handle multiple theme classes in your
					GetThemeClass() function - see Win32_StaticControl for
					an example of how to do this.)
	
	**************************************************************************/
	
	void Win32_ThemedObject::InvalidateTheme()
	{
		FreeThemeHandle();
	}

	/**************************************************************************
	
	>	bool Win32_ThemedObject::UsingTheme()
	
		Author:		Tim Browse
		Date:		20/8/2002
		Returns:	true => XP theme is active, and m_hTheme is a handle to it.
					false => XP theme not in use - use standard redraw methods.
	
		Purpose:	Find out if an XP UI theme is in use.

					If a theme is in use, then GetThemeHandle() can be used to
					redraw themed UI components, using DrawThemeBackground()
					etc.
	
	**************************************************************************/
	
	bool Win32_ThemedObject::UsingTheme()
	{
		HTHEME hTheme;
		return GetThemeHandle(&hTheme);
	}

	/**************************************************************************
	
	>	bool Win32_ThemedObject::GetThemeHandle(HTHEME* phTheme)
	
		Author:		Tim Browse
		Date:		21/8/2002
		Outputs:	phTheme - the HTHEME to use for this object, if themes
							  are active for this app.
		Returns:	true => themes active, use the handle in phTheme
					false => themes not active for this app
	
		Purpose:	Try to get the theme handle to use for this object
					(to pass to DrawThemeBackground(), etc)
	
	**************************************************************************/
	
	bool Win32_ThemedObject::GetThemeHandle(HTHEME* phTheme)
	{
		if (m_hTheme == NULL)
		{
			// No theme handle - try to get one
			if (Win32_Theme::IsThemeActive())
			{
				m_hTheme = Win32_Theme::OpenThemeData(m_hWnd, GetThemeClass());
			}
		}

		if (m_hTheme != NULL)
		{
			*phTheme = m_hTheme;
			return true;
		}
		else
		{
			return false;
		}
	}

	/**************************************************************************
	
	>	void Win32_ThemedObject::FreeThemeHandle()
	
		Author:		Tim Browse
		Date:		20/8/2002
	
		Purpose:	Close down the current theme handle, if we have one.
	
	**************************************************************************/
	
	void Win32_ThemedObject::FreeThemeHandle()
	{
		if (m_hTheme != NULL)
		{
			Win32_Theme::CloseThemeData(m_hTheme);
			m_hTheme = NULL;
		}
	}

	/**************************************************************************
	
	>	void Win32_ThemedObject::DrawCheckBox(HDC hDC, 
											  const Rect& rcCheckBox, 
											  ControlState State,
											  bool bHighlight = false)
	
		Author:		Tim Browse
		Date:		21/8/2002
		Inputs:		hDC - device context to render into
					rcCheckBox - area to draw the checkbox into (it will be rendered
								 in the centre of this area)
					State - state of the checkbox to render.
					bHighlight - if true then highlight the button (assuming
								 such functionality is available).
		Purpose:	Draw a checkbox using the current theme, or using standard
					rendering routines if Themes are not available.
	
	**************************************************************************/
	
	void Win32_ThemedObject::DrawCheckBox(HDC hDC, 
										  const Rect& rcCheckBox, 
										  ControlState State,
										  bool bHighlight)
	{
		if (UsingTheme())
		{
			//
			// Use XP theme to draw the checkbox button...
			//
			int iCheckState = bHighlight ? CBS_CHECKEDHOT : CBS_CHECKEDNORMAL;

			switch (State)
			{
				case cs_Unchecked:
				{
					iCheckState = bHighlight ? CBS_UNCHECKEDHOT : CBS_UNCHECKEDNORMAL;	
					break;
				}

				case cs_Undefined:	
				{
					iCheckState = bHighlight ? CBS_MIXEDHOT : CBS_MIXEDNORMAL;		
					break;
				}
			}

			Win32_Theme::DrawThemeBackground(m_hTheme, hDC, BP_CHECKBOX, 
											 iCheckState, 
											 (RECT*) &rcCheckBox, 
											 (RECT*) &rcCheckBox);
		}
		else
		{
			// Otherwise, use standard Windows 95 API to draw
			// the checkbox
			UINT uCheckState = DFCS_BUTTONCHECK;
			if (State == cs_Checked)
				uCheckState |= DFCS_CHECKED;

			// Calculate where to draw checkboxes - unless we
			// use XP themes, we assume they are always 13x13, as
			// this seems to be the received wisdom from usenet.
			// (Unlike the UxTheme API, DrawFrameControl() will scale
			// up the checkbox if you provide a bigger area to draw in,
			// which looks wrong).
			const int kCheckBoxSize = 13;

			Rect rcCorrectArea(rcCheckBox);
			rcCorrectArea.left += (rcCheckBox.GetWidth() - kCheckBoxSize) / 2;
			rcCorrectArea.top  += (rcCheckBox.GetHeight() - kCheckBoxSize) / 2;
			rcCorrectArea.SetWidth(kCheckBoxSize);
			rcCorrectArea.SetHeight(kCheckBoxSize);

			::DrawFrameControl(hDC, (RECT*) &rcCorrectArea, DFC_BUTTON, uCheckState);
		}
	}

	/**************************************************************************

	>	void Win32_ThemedObject::DrawButtonEdge(HDC hDC, 
												const Rect& rcButton, 
												ButtonState State,
												bool bHighlight = false)


		Author:		Tim Browse
		Date:		21/8/2002
		Inputs:		hDC - the DC to use when drawing the button
					rcButton - position of the button to draw
					State - the state of the button (normal or pressed)
					bHighlight - if true then highlight the button (assuming
								 such functionality is available).

		Purpose:	Draw the edge of a button control, using XP themes
					where possible.

	**************************************************************************/

	void Win32_ThemedObject::DrawButtonEdge(HDC hDC, 
											const Rect& rcButton, 
											ButtonState State,
											bool bHighlight)
	{
		if (UsingTheme())
		{
			// Use XP theme to draw the button edge...
			int iButtonState =  bHighlight ? PBS_HOT : PBS_NORMAL;
			if (State == bs_Pressed)
				iButtonState = PBS_PRESSED;

			Win32_Theme::DrawThemeBackground(m_hTheme, hDC,
											 BP_PUSHBUTTON, 
											 iButtonState,
											 (RECT*) &rcButton, 
											 (RECT*) &rcButton);
		}
		else
		{
			// Otherwise, use standard Windows 95 API to draw
			// the button
			UINT uButtonState = DFCS_BUTTONPUSH;
			if (State == bs_Pressed)
				uButtonState |= DFCS_PUSHED;

			if (bHighlight)
				uButtonState |= DFCS_HOT;

			::DrawFrameControl(hDC, (RECT*) &rcButton, 
							   DFC_BUTTON, uButtonState);
		}
	}

	void Win32_ThemedObject::DrawPopupMenuIcon(HDC hDC, const Rect& rcButton, 
											   ButtonState State, bool bHighlight)
	{
		if (UsingTheme())
		{
			// Use XP theme to draw the button...
			int iButtonState =  bHighlight ? ABS_RIGHTHOT : ABS_RIGHTNORMAL;
			if (State == bs_Pressed)
				iButtonState = ABS_RIGHTPRESSED;

			Win32_Theme::DrawThemeBackground(m_hTheme, hDC,
											 SBP_ARROWBTN, 
											 iButtonState,
											 (RECT*) &rcButton, 
											 (RECT*) &rcButton);
		}
		else
		{
			// Otherwise, use standard Windows 95 API to draw
			// the button
			UINT uButtonState = DFCS_SCROLLRIGHT;
			if (State == bs_Pressed)
				uButtonState |= DFCS_PUSHED;

			if (bHighlight)
				uButtonState |= DFCS_HOT;

			::DrawFrameControl(hDC, (RECT*) &rcButton, 
							   DFC_SCROLL, uButtonState);
		}
	}
};


