//
// Win32_WindowImp.cpp - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_WindowImp.h"

#include "..\Application.h"
#include "Win32_ApplicationImp.h"
#include "..\Rect.h"
#include "..\Timer.h"
#include "..\Font.h"
#include "..\Mouse.h"
#include "..\Size.h"
#include "..\WaitCursor.h"
#include "..\MessageParams.h"

#include "Win32_GCImp.h"
#include "Win32_FontImp.h"
#include "Win32_Headers.h"
#include "Win32_Debug.h"
#include "Win32_Theme.h"
#include "Win32_WindowInfo.h"
#include "Win32_MessageInfo.h"
#include "Win32_Keyboard.h"

#include <zmouse.h>

#include "vOS\Errors.h"
#include "vOS\Trace.h"

#include "3rdparty\MinimizeToTray\MinimizeToTray.h"

//
// Macros to check for valid window handle/state
//
#define ASSERT_VALID_WINDOW()							\
		if (!IsValid())									\
		{												\
			COMPLAIN("Invalid window handle used!");	\
			return;										\
		}

#define ASSERT_VALID_WINDOW_RETURN(RetVal)				\
		if (!IsValid())									\
		{												\
			COMPLAIN("Invalid window handle used!");	\
			return (RetVal);							\
		}


namespace vgui
{
    Win32_WindowImp::Win32_WindowImp() : m_pParent(NULL),
    									 m_pSuperWndProc(NULL),
    									 m_bSubClassed(false),
										 m_dwWindowStyle(0),
										 m_WindowMode(WindowMode_Normal),
    									 m_ScrollBarStatus(0),
    									 m_pFont(NULL),
										 m_bAlive(false),
										 m_pControlHandlerWindow(NULL)
    {
		// TCILB
    }


    Win32_WindowImp::~Win32_WindowImp()
    {
		// Destroy the window font if we created one
		delete m_pFont;
		
    	// Ensure we are detached before destroying ourselves
    	Detach();
    	Destroy();
    }

    bool Win32_WindowImp::Create(Window *pWindow,
    							 Window* pParent,
    							 TWindowStyle Style, 
    							 const StringResourceID* prWindowTitle)
    {
    	// Load window title text
    	String strWindowName;

    	if (prWindowTitle != NULL)
    	{
    		prWindowTitle->Load(&strWindowName);
    	}

    	// pass on to main function.
    	return Create(pWindow, pParent, Style, &strWindowName);
    }

    bool Win32_WindowImp::Create(Window *pWindow,
    							 Window* pParent,
    							 TWindowStyle Style, 
    							 const String* psWindowTitle)
    {
    	// Remember parent
    	m_pParent = pParent;

    	// Get the application object to create the window.
    	//
    	// NB. This is *one* way of doing it, that happens to make sense for
    	//     Win32 apps - for other OSs, it may be more appropriate to create
    	//     windows directly in this function.
    	//
    	m_hWnd = GetWin32App()->CreateWin32Window(pWindow, pParent, Style, psWindowTitle);

    	// Did it work?
    	if (m_hWnd != NULL)
    	{
			// Yes, this window is alive
			m_bAlive = true;

			// Remember our window style flags
			m_dwWindowStyle = ::GetWindowLong(m_hWnd, GWL_STYLE);

    		// Tell window it has been created
    		pWindow->OnCreate();

    		// Set up scroll bars (assumes none are currently visible)
    		pWindow->ShowScrollBar((TScrollBar) m_ScrollBarStatus);
    		SetScrollBars();

    		// Success
    		return true;
    	}

    	// Failed - tell caller
    	return false;
    }

    HWND Win32_WindowImp::CreateChildControl(Window* pControl,
    										 LPCTSTR szControlClass,
    										 LPCTSTR szWindowName,
    										 DWORD dwButtonStyle,
											 DWORD dwStyleEx,
    										 const Rect& Position,
    										 ControlID ID,
    										 bool bTabStop)
    {
    	if (ID == NOCTL)
    	{
    		// Caller wants us to specify a unique ID, so find one
    		ID = FindUniqueControlID();
    	}

    	// Get Win32 object for the control
    	Win32_WindowImp* pControlImp = Win32_WindowImp::GetImp(pControl);

    	// Create a button window at the requested position with the specified 
    	// control ID.
    	DWORD dwStyle = WS_CHILD | WS_VISIBLE | dwButtonStyle;

    	if (bTabStop)
    		dwStyle |= WS_TABSTOP;

    	// Set up scrollbar styles
    	if (pControlImp->m_ScrollBarStatus & sb_Horizontal)
    		dwStyle |= WS_HSCROLL;
    	if (pControlImp->m_ScrollBarStatus & sb_Vertical)
    		dwStyle |= WS_VSCROLL;

    	HWND hWnd = ::CreateWindowEx(dwStyleEx, 
									 szControlClass, 
    							     szWindowName,
    							     dwStyle,
    							     Position.left, Position.top,
    							     Position.GetWidth(), Position.GetHeight(),
    							     m_hWnd,
    							     (HMENU) ID,
    							     GetWin32App()->GetInstance(), 
    							     NULL);

    	if (hWnd != NULL)
    	{
    		// Make the control use the window's normal font for consistency.
			HFONT hParentFont = (HFONT) ::SendMessage(m_hWnd, WM_GETFONT, 0, 0);
			::SendMessage(hWnd, WM_SETFONT, (WPARAM) hParentFont, FALSE);

			// BODGE: do we need this line?  Attach() will set the window up
			//		  for us, and until then, we shouldn't get any messages
			//		  for the control, as we haven't subclassed it until then...
    		// Add pointer to vGui object to the window.
//			Win32_WindowInfo::Attach(hWnd, pControl);

			// The control window is alive!
			pControlImp->m_bAlive = true;

    		// Set up scroll bars (assumes none are currently visible)
    		ShowScrollBar((TScrollBar) m_ScrollBarStatus);
    	}

    	return hWnd;
    }

    void Win32_WindowImp::ShowWindow(TShowWindow Command)
    {
    	// Vector to Win32
    	if (IsAlive())
    	{
			// Find out the current visibility state of this window
			bool bWindowIsVisible = ::IsWindowVisible(m_hWnd) ? true: false;

			Application* pApp = GetApp();

			if (pApp->IsAccessory() && (pApp->GetMainWindow() == m_pWindow))
			{
				// Special case handling for an accessory app's main 
				// window - we restore from/minimise to the system tray...
				if (Command == sw_SHOW)
				{
					if (bWindowIsVisible)
					{
						// Window already visible - just make it the active window
						::ShowWindow(m_hWnd, SW_SHOW);
						::SetActiveWindow(m_hWnd);
						::SetForegroundWindow(m_hWnd);
					}
					else
					{
						// Window is invisible - do a 'full' restore action
						// from the system tray.
						RestoreWndFromTray(m_hWnd);
					}
				}
				else
				{
					// If window is visible, minimise it to the system tray.
					if (bWindowIsVisible)
						MinimizeWndToTray(m_hWnd);
				}
			}
			else
			{
    			// Normal window - translate to Windows codes
    			int nCmdShow = SW_SHOW;

    			if (Command == sw_HIDE)
    				nCmdShow = SW_HIDE;

    			::ShowWindow(m_hWnd, nCmdShow);
			}
    	}
    }


    void Win32_WindowImp::UpdateWindow()
    {
    	// Vector to Win32
    	if (IsAlive())
    	{
    		::UpdateWindow(m_hWnd);
    	}
    }

    void Win32_WindowImp::CentreWindow(HWND hwnd)
    {
    	if (hwnd != NULL)
    	{
    		WINDOWPLACEMENT wndpl;
    		ZeroMemory(&wndpl, sizeof(wndpl));
    		wndpl.length = sizeof(wndpl);

    		::GetWindowPlacement(hwnd, &wndpl);

    		// Work out size of window
    		int Width = wndpl.rcNormalPosition.right - wndpl.rcNormalPosition.left;
    		int Height = wndpl.rcNormalPosition.bottom - wndpl.rcNormalPosition.top;

    		// Get size of screen
    		// TODO: Multi-monitor support
    		HWND hDesktop = ::GetDesktopWindow();

    		Rect rcDesktop;
    		::GetClientRect(hDesktop, rcDesktop.GetPointer());

    		// Position window in centre of screen
    		wndpl.rcNormalPosition.left = rcDesktop.GetHalfWidth() - (Width / 2);
    		wndpl.rcNormalPosition.right = wndpl.rcNormalPosition.left + Width;

    		wndpl.rcNormalPosition.top = rcDesktop.GetHalfHeight() - (Height / 2);
    		wndpl.rcNormalPosition.bottom = wndpl.rcNormalPosition.top + Height;

    		// Because GetWindowPlacement is stupid sometimes
    		wndpl.length = sizeof(wndpl);

			// Now, a horrible hack - GetWindowPlacement() returns
			// wndpl.showCmd as SW_SHOWNORMAL if you call it while creating 
			// an invisible modeless dialog box (hnnnggg!).
			//
			// Therefore, we do a quick check to see if the window is actually
			// invisible, and if it is, then we set showCmd to SW_HIDE to make
			// sure it stays hidden.  Otherwise, the very act of centering a
			// window will make it visible.  Handy eh?
			if ((wndpl.showCmd == SW_SHOWNORMAL) && !::IsWindowVisible(hwnd))
			{
				wndpl.showCmd = SW_HIDE;
			}

    		// Reposition the window
    		::SetWindowPlacement(hwnd, &wndpl);
    	}
    }

    void Win32_WindowImp::CentreWindow()
    {
    	if (IsAlive())
    	{
    		CentreWindow(m_hWnd);
    	}
    }




    void Win32_WindowImp::Destroy()
    {
    	if (IsAlive())
    	{
    		::DestroyWindow(m_hWnd);
    	}
    }

    HWND Win32_WindowImp::GetHWND(const Window* pWindow)
    {
    	if (pWindow == NULL)
    		return NULL;

    	Win32_WindowImp *pImp = static_cast<Win32_WindowImp*>(pWindow->GetImp());

		// Only return a 'safe' window handle, i.e. one that is alive and valid.
		if (pImp->IsValid())
    		return pImp->GetHWND();
		else
			return NULL;
    }

    void Win32_WindowImp::Attach(HWND hWnd)
    {
    	// We can only attach if not already attached
    	ASSERT(m_hWnd == NULL);

    	if (m_hWnd != NULL)
    	{
    		return;
    	}

    	// Install window handle
    	m_hWnd = hWnd;

		// This window is now alive
		m_bAlive = true;

    	// Subclass window so we get to know all its comings and goings...
    	SubclassWindow(m_hWnd);
    }

    void Win32_WindowImp::Detach()
    {
    	if (m_hWnd != NULL) 
    	{
    		// If it is not a valid window, then assume it has been 
    		// destroyed by the OS, and do nothing.
    		if (::IsWindow(m_hWnd))
    		{
    			// Valid window - unsubclass it if necessary
    			UnSubclassWindow(m_hWnd);

				// Mark it as not active
				m_bAlive = false;

    			// And remove our window pointer from the window record
				Win32_WindowInfo::Detach(m_hWnd);
    		}

    		m_hWnd = NULL;
    	}
    }

    void Win32_WindowImp::SubclassWindow(HWND hWnd)
    {
    	// Don't bother to subclass a window if it's already under our control
    	// (e.g. a DialogControl derived window)
    	WNDPROC CurrentWndProc = (WNDPROC) ::GetWindowLong(hWnd, GWL_WNDPROC);

    	if (CurrentWndProc == Win32_ApplicationImp::vGuiWndProc)
		{
    		// It's one of ours, so don't touch it.
    		return;
		}

    	// Install our app-data in the window
		Win32_WindowInfo::Attach(hWnd, m_pWindow);

    	// Install our own WNDPROC
    	m_pSuperWndProc = (WNDPROC) ::SetWindowLong(hWnd, GWL_WNDPROC, (LONG) SubclassWndProc);

    	// Check that we haven't subclassed more than once - at the moment
    	// you cannot attach more than one Window object to a Win32 window.
    	ASSERT(m_pSuperWndProc != SubclassWndProc);

    	// Remember that we have subclassed this window
    	m_bSubClassed = true;
    }

    void Win32_WindowImp::UnSubclassWindow(HWND hWnd)
    {
    	// If we haven't subclassed this window, ignore the request.
    	if (!m_bSubClassed)
    		return;

    	// Check it's one of ours...
    	if (::GetWindowLong(hWnd, GWL_WNDPROC) != (LONG) SubclassWndProc)
    	{
    		// It's not subclasses by us, so don't touch it.
    		COMPLAIN("UnSubclassWindow: this window is not subclassed by vGui.");
    		return;
    	}

		// Remove window data
		Win32_WindowInfo::Detach(hWnd);

    	// Install the old WNDPROC
    	::SetWindowLong(hWnd, GWL_WNDPROC, (LONG) m_pSuperWndProc);

    	// We longer have a subclass on this window
    	m_bSubClassed = false;
    }

    Win32_WindowImp* Win32_WindowImp::GetImpFromHWND(HWND hWnd)
    {
    	Window *pWindow = GetWindowFromHWND(hWnd);

    	if (pWindow != NULL)
    	{
    		return static_cast<Win32_WindowImp*>(pWindow->GetImp());
    	}
    	else
    	{
    		// Window not found - this is valid, as we might just
			// be trying to see if a particular window is linked
			// to a vGUI object.
    		return NULL;
    	}
    }

    Window* Win32_WindowImp::GetWindowFromHWND(HWND hwnd)
    {
    	// Extract window pointer from window record user-data field.
    	Window *pWindow = NULL;
		Win32_WindowInfo::GetAttachedWindow(hwnd, &pWindow);
   		return pWindow;
    }

    LRESULT CALLBACK Win32_WindowImp::SubclassWndProc(HWND hWnd, 
    												  UINT nMsg, 
    												  WPARAM wParam, 
    												  LPARAM lParam)
    {
		WIN32_LOGMESSAGE(hWnd, nMsg, wParam, lParam);

    	// Find window implementation object
    	Win32_WindowImp *pImp = GetImpFromHWND(hWnd);

    	// If it worked, route message accordingly
    	if (pImp != NULL)
    	{
    		// The WindowProc will return kWndProc_NotHandled rather than call the default
    		// window proc, so only return the result if it processed the message.
    		LRESULT lResult = pImp->WindowProc(nMsg, wParam, lParam);

    		if (lResult != kWndProc_NotHandled)
			{
				WIN32_LOGMESSAGEHANDLED(hWnd, nMsg, lResult);
			}

			// If window is being destroyed, then we unsubclass it, to
			// make sure everything's cleared up.
			if (nMsg == WM_DESTROY)
			{
				pImp->UnSubclassWindow(hWnd);
			}

    		if (lResult != kWndProc_NotHandled)
			{
    			return lResult;
			}

    		// Otherwise, fall through to standard default Windows processing...
    	}
    	else
    	{
    		// It didn't work - something is seriously wrong, so complain about it.
    		// We are responsible for a window without our WindowImp object installed.
    		ASSERT(FALSE);
    	}

    	// Message not handled - revert to default handling
    	return ::DefWindowProc(hWnd, nMsg, wParam, lParam);
    }

#ifndef GET_WHEEL_DELTA_WPARAM
#define GET_WHEEL_DELTA_WPARAM(wparam) ((short)HIWORD (wparam))
#endif

    LRESULT Win32_WindowImp::WindowProc(UINT uMsg,
    									WPARAM wParam,
    									LPARAM lParam)
    {
    	// We don't handle messages unless we're alive
    	if (IsAlive())
    	{
			// Window should be valid if this object is alive
			ASSERT(::IsWindow(m_hWnd));

			// Check for obvious input messages
			if (((uMsg >= WM_MOUSEFIRST) && (uMsg <= WM_MOUSELAST)) ||
				((uMsg >= WM_KEYFIRST) && (uMsg <= WM_KEYLAST)))
			{
				// We don't count mouse movement as worthy of holding up
				// idle processing
				if (uMsg != WM_MOUSEMOVE)
				{
					NotifyUserInputMessageReceived();
				}
			}

    		switch (uMsg)
    		{
    			//
    			// General Window functions
    			//

    			case WM_DESTROY:
    			{
					Win32_ApplicationImp* pWin32App = GetWin32App();

    				if (m_hWnd == pWin32App->GetMainHWND())
    				{
    					// Main window is being destroyed, so we're done.
						pWin32App->OnDestroyMainWindow();
					}

					// Window is dying
					m_bAlive = false;

    				// Window being destroyed - let the OS-independent window object know.
    				m_pWindow->OnDestroy();

					// We don't need the HWND anymore
					m_hWnd = NULL;

					return 0;
    			}

    			case WM_SETFOCUS:
    			{
					NotifyUserInputMessageReceived();

    				if (m_pWindow->OnGainFocus())
    				{
    					// Message processed
    					return 0;
    				}
    				break;
    			}

    			case WM_KILLFOCUS:
    			{
    				if (m_pWindow->OnLoseFocus())
    				{
    					// Message processed
    					return 0;
    				}
    				break;
    			}

    			case WM_CLOSE:
    			{
    				// If window class doesn't handle this, then leave it to the
    				// default WndProc to destroy the window in a standard way.
    				if (m_pWindow->OnClose())
    				{
    					// Message processed
    					return 0;
    				}
    				break;
    			}

    			case WM_PAINT:
    			{
    				if (m_pWindow->OnRedraw())
    				{
    					// Message processed
    					return 0;
    				}

    				break;
    			}

				case WM_GETMINMAXINFO:
				{
					break;
				}

				case WM_WINDOWPOSCHANGING:
				{
					if (HandleWindowPosChanging((WINDOWPOS*) lParam))
						return 0;

					break;
				}

				case WM_MOVE:
				{
    				if (m_pWindow->OnMove(LOWORD(lParam), HIWORD(lParam)))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
				}

    			case WM_SIZE:
    			{
    				// TODO: use wParam to rationalise sending of different size types.
    				if (m_pWindow->OnResize(LOWORD(lParam), HIWORD(lParam)))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
    			}

				case WM_SYSCOMMAND:
				{
					if (wParam == SC_MINIMIZE)
					{
						// If this is the main app window, and the app is
						// curently configured as an accessory, then we
						// will minimise the app to the system tray.
						Win32_ApplicationImp* pWin32App = GetWin32App();

    					if ((m_hWnd == pWin32App->GetMainHWND()) &&
							GetApp()->IsAccessory())
    					{
							MinimizeWndToTray(m_hWnd);

							// Message processed
							return 0;
						}
					}
					break;
				}

    			case WM_COMMAND:
    			{
    				// Decode parameters
    				UINT uNotifyCode = HIWORD(wParam);

					if (uNotifyCode == 0)
					{
						// It's a menu command
						int nMenuCommand = LOWORD(wParam);

						// See if we should process this menu item
						if (m_pWindow->OnCommand(nMenuCommand))
						{
							// processed
							return 0;
						}
					}
					else
					{
    					// Extract item,control or accelerator ID
    					ControlID nCtrlID = LOWORD(wParam);

    					// handle of control 
    					HWND hwndCtl = (HWND) lParam;

    					// Try control notification messages
    					//
    					// (We don't bother passing in lParam or wParam, as there's 
    					// nothing useful in them that we haven't already used).
    					LRESULT lResult = DispatchControlNotification(uNotifyCode, nCtrlID, 0, 0);

    					if (lResult != kWndProc_NotHandled)
    					{
    						// Processed message
    						return lResult;
    					}
					}

    				break;
     			}

    			case WM_TIMER:
    			{
    				// Decode parameters
    				TimerID ID = (TimerID) wParam;

    				if (Timer::ServiceTimer(ID))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
    			}

				case WM_GETDLGCODE:
				{
					if (lParam == 0)
					{
						// Let the superclass have a go first...
						if (m_pSuperWndProc != NULL)
						{
							LRESULT lResult = ::CallWindowProc(m_pSuperWndProc, m_hWnd, uMsg, wParam, lParam);

							if (lResult != 0)
							{
								// Super-class has claimed some input
								return lResult;
							}
						}

						// Query - tell it we want everything - we'll choose when the
						// actual message arrives.
						return DLGC_WANTALLKEYS;
					}
					else
					{
						// Dialog manager is asking us what input we want...ask
						// Window object what it wants to do.
						NotifyUserInputMessageReceived();
						return ProcessDialogInputRequest(m_pWindow, (MSG*) lParam);
					}
				}

				case WM_KEYDOWN:
				case WM_KEYUP:
				{
					if (HandleRawKeyEvent(uMsg, wParam))
					{
    					// Message processed
    					return 0;
					}

					break;
				}

    			case WM_CHAR:
    			{
    				// Check to see if window wants this character message
    				if (m_pWindow->OnChar((uint16) wParam))
					{
    					// Message processed
    					return 0;
					}
					break;
				}

    			case WM_VSCROLL:
    			{
					NotifyUserInputMessageReceived();

					// Process this message, giving superclass a chance first...
					if (m_pSuperWndProc != NULL)
					{
						LRESULT lResult = ::CallWindowProc(m_pSuperWndProc, m_hWnd, uMsg, wParam, lParam);

						if (lResult == 0)
							return lResult;
					}

    				ProcessScrollBarUpdate(SB_VERT, LOWORD(wParam), HIWORD(wParam));
          
    				// Message processed
    				return 0;
    			}

    			case WM_HSCROLL:
    			{
					NotifyUserInputMessageReceived();

					// Process this message, giving superclass a chance first...
					if (m_pSuperWndProc != NULL)
					{
						LRESULT lResult = ::CallWindowProc(m_pSuperWndProc, m_hWnd, uMsg, wParam, lParam);

						if (lResult == 0)
							return lResult;
					}

    				ProcessScrollBarUpdate(SB_HORZ, LOWORD(wParam), HIWORD(wParam));
          
    				// Message processed
    				return 0;
    			}

    			case WM_LBUTTONDOWN:
    			{
    				vgui::Point Pos(LOWORD(lParam), HIWORD(lParam));

    				if (m_pWindow->OnMouseClick(MouseButton_Left | TranslateMouseModifiers(wParam), Pos))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
    			}

    			case WM_MBUTTONDOWN:
    			{
    				vgui::Point Pos(LOWORD(lParam), HIWORD(lParam));

    				if (m_pWindow->OnMouseClick(MouseButton_Middle | TranslateMouseModifiers(wParam), Pos))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
    			}

    			case WM_RBUTTONDOWN:
    			{
    				vgui::Point Pos(LOWORD(lParam), HIWORD(lParam));

    				if (m_pWindow->OnMouseClick(MouseButton_Right | TranslateMouseModifiers(wParam), Pos))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
    			}

    			case WM_LBUTTONUP:
    			{
    				vgui::Point Pos(LOWORD(lParam), HIWORD(lParam));

    				if (m_pWindow->OnMouseRelease(MouseButton_Left | TranslateMouseModifiers(wParam), Pos))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
    			}

    			case WM_MBUTTONUP:
    			{
    				vgui::Point Pos(LOWORD(lParam), HIWORD(lParam));

    				if (m_pWindow->OnMouseRelease(MouseButton_Middle | TranslateMouseModifiers(wParam), Pos))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
    			}

    			case WM_RBUTTONUP:
    			{
    				vgui::Point Pos(LOWORD(lParam), HIWORD(lParam));

    				if (m_pWindow->OnMouseRelease(MouseButton_Right | TranslateMouseModifiers(wParam), Pos))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
    			}

    			case WM_MOUSEMOVE:
    			{
    				vgui::Point Pos(LOWORD(lParam), HIWORD(lParam));

    				if (m_pWindow->OnMouseMove(Pos))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
    			}

				case WM_MOUSEWHEEL:
				{
					int WheelDelta = GET_WHEEL_DELTA_WPARAM(wParam);

    				if (m_pWindow->OnMouseWheel(WheelDelta))
    				{
    					// Message processed
    					return 0;
    				}

    				break;
				}
				
				case WM_CANCELMODE:
				{
					if (m_pWindow->OnLoseMouseCapture())
    				{
    					// Message processed
    					return 0;
    				}

					break;
				}

				// Cursor appearance
				case WM_SETCURSOR:
				{
					// If in full screen mode, then hide the cursor.
					if (m_WindowMode == WindowMode_FullScreen)
					{
						// Remove cursor
						::SetCursor(NULL);

    					// Message processed
						return 0;
					}

					// Ok, should we display an hourglass?
					if (WaitCursor::IsActive())
					{
						// Yes...
						::SetCursor(::LoadCursor(NULL, IDC_WAIT));
					}
					break;
				}
    			//
    			// Owner-draw control messages
    			//
    			case WM_DRAWITEM:
    			case WM_MEASUREITEM:
    			{
    				// Dispatch message to control using central method.
    				if (DispatchControlNotification(uMsg, (int) wParam, lParam, wParam) != kWndProc_NotHandled)
    				{
    					// Processed message
    					return 0;
    				}
    				break;
    			}

				case WM_THEMECHANGED:
				{
					// Invalidate our theme handle
					InvalidateTheme();
				}
    		}
    	}

    	// If not handled, defer to our default window proc
    	return DefaultWindowProc(uMsg, wParam, lParam);
    }

	bool Win32_WindowImp::HandleWindowPosChanging(WINDOWPOS* pWP)
	{
		//
		// Convert window size into client area - luckily, because Win32 is
		// a bit crap, we have to do this backwards...
		//
		
		// First, ask Windows how big this window would be with a client area
		// of 200x200...
		const int kClientSize = 200;

    	// Does the window have a menu?
    	BOOL bMenu = FALSE;
    	if (::GetMenu(m_hWnd) != NULL)
    	{
    		// Yes, it has a menu
    		bMenu = TRUE;
    	}

    	// Ask Windows where the window should be if we want the client area here.
		vgui::Rect rcDummy(0,0, kClientSize,kClientSize);

    	::AdjustWindowRectEx((RECT*) &rcDummy, 
    						 ::GetWindowLong(m_hWnd, GWL_STYLE), 
    						 bMenu,
    						 ::GetWindowLong(m_hWnd, GWL_EXSTYLE));

		// Use the resulting border sizes to adjust the new size rectangle.
		vgui::Size NewSize;
		NewSize.m_Width  = pWP->cx - (rcDummy.GetWidth() - kClientSize);
		NewSize.m_Height = pWP->cy - (rcDummy.GetHeight() - kClientSize);

		// Give the window object the chance to veto the resize operation.
		vgui::Size AdjustedSize = NewSize;

    	if (m_pWindow->OnResizing(&AdjustedSize))
		{
			// Message handled - was the resize adjusted?
			if (AdjustedSize != NewSize)
			{
				// Yes, so update the size.
				pWP->cx += (AdjustedSize.m_Width  - NewSize.m_Width);
				pWP->cy += (AdjustedSize.m_Height - NewSize.m_Height);
			}

			return true;
		}

		return false;
	}

	void Win32_WindowImp::OnControlResizing()
	{
		// Base class doesn't need to do anything
	}

	void Win32_WindowImp::OnControlResize()
	{
		// Base class doesn't need to do anything
	}
	
    int Win32_WindowImp::TranslateMouseButtons(WPARAM wParam)
    {
    	int MouseButtons = 0;

    	// Check for mouse buttons
    	if (wParam & MK_LBUTTON)
    		MouseButtons |= MouseButton_Left;

    	if (wParam & MK_MBUTTON)
    		MouseButtons |= MouseButton_Middle;

    	if (wParam & MK_RBUTTON)
    		MouseButtons |= MouseButton_Right;

    	return MouseButtons;
    }

    int Win32_WindowImp::TranslateMouseModifiers(WPARAM wParam)
    {
    	int Modifiers = 0;

    	// Check for modifier keys
    	if (wParam & MK_SHIFT)
    		Modifiers |= MouseButton_Shift;

    	if (wParam & MK_CONTROL)
    		Modifiers |= MouseButton_Control;

    	// TODO: Handle ALT key.
    	return Modifiers;
    }

    void Win32_WindowImp::ProcessScrollBarUpdate(int ScrollBar, 
    											 int ScrollUpdateCode,
    											 int NewScrollPos)
    {
    	SCROLLINFO si;
    	si.cbSize = sizeof(si);
    	si.fMask = SIF_PAGE | SIF_RANGE | SIF_POS;

    	::GetScrollInfo(m_hWnd, ScrollBar, &si);

    	int ScrollPos = si.nPos;

    	switch (ScrollUpdateCode)
    	{
    		case SB_LINEUP:
    		//case SB_LINELEFT:
    			if (ScrollBar == SB_HORZ)
    				ScrollPos -= m_pWindow->GetScrollDelta(sb_Horizontal);
    			else
    				ScrollPos -= m_pWindow->GetScrollDelta(sb_Vertical);;
    			break ;

    		case SB_LINEDOWN:
    		//case SB_LINERIGHT:
    			if (ScrollBar == SB_HORZ)
    				ScrollPos += m_pWindow->GetScrollDelta(sb_Horizontal);
    			else
    				ScrollPos += m_pWindow->GetScrollDelta(sb_Vertical);
    			break ;

    		case SB_PAGEUP:
    		//case SB_PAGELEFT:
    			ScrollPos -= si.nPage;
    			break ;

    		case SB_PAGEDOWN:
    		//case SB_PAGERIGHT:
    			ScrollPos += si.nPage;
    			break ;

    		case SB_TOP:
    		//case SB_LEFT:
    			ScrollPos = si.nMin;
    			break;

    		case SB_BOTTOM:
    		//case SB_RIGHT:
    			ScrollPos = si.nMax;
    			break;

    		case SB_THUMBTRACK:
    			// Live dragging
    			ScrollPos = NewScrollPos;
    			break;

    		default :
    			// We don't handle anything else
    			return;
    	}

		// Make sure the scroll position stays within scroll bar limits.
    	ScrollPos = max(si.nMin, min(ScrollPos, si.nMax));

		// If the scroll position has changed as a result of the above
		// calculations, then tell Windows about it.
    	if (ScrollPos != si.nPos)
    	{
    		si.nPos = ScrollPos;
    		si.fMask = SIF_POS;
    		::SetScrollInfo(m_hWnd, ScrollBar, &si, TRUE);

			// Force a redraw to reflect status.
			//
			// TODO: Use anything resembling intelligence to make
			//       this work efficiently, i.e. minimise redraw with
			//		 calls to ScrollWindowEx() etc.
    		::InvalidateRect(m_hWnd, NULL, TRUE);
    	}
    }

	/**************************************************************************

	>	ControlTranslation Win32_WindowImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																		 TMessageType* pMsgType,
																		 LRESULT* pResult,
																		 MessageParams* pParams)

		Author:		Tim Browse
		Date:		22/02/2001
		Inputs:		MsgInfo - details of the Win32 message
		Outputs:	pMsgType - The vgui notification message.
					pResult - the lRESULT code to return for this message
							  if (ControlTranslation_UseReturnCode is returned)
					pParams - any extra parameters the message handler for this
							  message might need.

		Returns:	ControlTranslation_OK      - Message translated ok, and
												 it should be delivered.

					ControlTranslation_Unknown - Message not recognised 
												 by the control.

					ControlTranslation_Handled - Message recognised and 
												 automatically handled by 
												 control - do not pass the 
												 message on.

					ControlTranslation_UseReturnCode - message was processed, and
													   the supplied return code 
													   (in pResult) should be
													   returned to Windows.

		Purpose:	Translate Win32 control notification messages into
					standardised vgui messages.

	**************************************************************************/

    ControlTranslation Win32_WindowImp::TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																	 TMessageType* pMsgType,
																	 LRESULT* pResult,
																	 MessageParams* pParams)
    {
    	// By default, we understand only button click notifications.
		//
    	// NB. If we do not translate them, then the 'Close' [X] buttons on
    	//     dialogs do not work, as they map to a BN_CLICK on ID 2
    	//     (2 being the standard IDCANCEL code).
    	//
    	//     Hence if we map button clicks here, a [X] button maps to a
    	//     Cancel click, which by default ends the dialog (because it
    	//     is by default mapped to the vgui::Dialog::OnCancel() handler
    	//	   function).

    	//
    	// Specific derived control classes will over-ride this function to
    	// provide their own translations.
    	//
    	switch (MsgInfo.m_uNotifyCode)
    	{
    		// Translate to vgui message
    		case BN_CLICKED:
			{
				*pMsgType = mt_BN_CLICKED;	
				return ControlTranslation_OK;
			}
    	}

    	// Otherwise, we leave it alone.
    	return ControlTranslation_Unknown;
    }

    LRESULT Win32_WindowImp::DispatchControlNotification(UINT uMsg, 
														 ControlID nCtrlID,
														 LPARAM lParam,
														 WPARAM wParam)
    {
		Win32_MessageInfo MsgInfo;
		// TODO: fill in message field correctly
		MsgInfo.m_Msg.message = 0;
		MsgInfo.m_ControlID = nCtrlID;
		MsgInfo.m_Msg.lParam = lParam;
		MsgInfo.m_Msg.wParam = wParam;
		MsgInfo.m_uNotifyCode = uMsg;

		// Set up message info structure
    	// Get control to perform specialised translation.
    	TMessageType MsgType;
		LRESULT lResult;
		MessageParams Params;
    	switch (TranslateControlNotification(MsgInfo, 
											 &MsgType, 
											 &lResult,
											 &Params))
    	{
    		case ControlTranslation_OK:
    		{
    			// We understand it, so try message map to send this notification 
				// message.
				//
				// NB. We vector through m_pControlHandlerWindow if it is available.
				//
				Window *pDestinationWindow = m_pControlHandlerWindow;

				if (pDestinationWindow == NULL)
					pDestinationWindow = m_pWindow;

    			if (pDestinationWindow->OnMessage(MsgType, 0, nCtrlID, &Params))
    			{
    				// Message processed
    				return 0;
    			}
    			break;
    		}

    		case ControlTranslation_Handled:
    		{
    			// Control has already handled this message
    			return 0;
    		}

			case ControlTranslation_UseReturnCode:
			{
    			// Control has already handled this message, and wants us to
				// use the specified return code
    			return lResult;
			}

    		case ControlTranslation_Unknown:
    			// This message is not recognised for the control
    			break;

    		default:
    			COMPLAIN("DispatchControlNotification(): Logic error - illegal enum found in switch statement.");
    			break;
    	}

    	// Not handled 
    	return kWndProc_NotHandled;
    }

    LRESULT Win32_WindowImp::DefaultWindowProc(UINT nMsg, 
    										   WPARAM wParam, 
    										   LPARAM lParam)
    {
    	// If we have subclassed an existing window, then defer to the super
    	// class window proc.
    	if ((m_pSuperWndProc != NULL) && (m_hWnd != NULL))
    		return ::CallWindowProc(m_pSuperWndProc, m_hWnd, nMsg, wParam, lParam);

    	// Otherwise, we leave it to the main WndProc that called us
    	return kWndProc_NotHandled;
    }

    Win32_WindowImp *Win32_WindowImp::GetImp(const Window *pWindow)
    {
    	// For NULL windows...
    	if (pWindow == NULL)
    		return NULL;

    	// Cast to Win32 implementation pointer.
    	Win32_WindowImp *pImp = dynamic_cast<Win32_WindowImp*>(pWindow->GetImp());

    	// If that didn't work, complain loudly in debug builds
    	ASSERT(pImp != NULL);

    	return pImp;
    }

    bool Win32_WindowImp::SetText(const String& Text)
    {
    	// For Win32, same as window text
    	if (IsAlive())
    	{
			// NB. do not remove comparison to TRUE (other non-zero values
			//	   are used to indicate failure).
    		return (::SetWindowText(m_hWnd, Text.c_str()) == TRUE);
    	}
		else
		{
			return false;
		}
    }

    bool Win32_WindowImp::GetText(String& Text)
    {
    	if (IsAlive())
    	{
    		// We use this small buffer on the stack for most cases
    		const int kBufLength = 128;
    		TCHAR TextBuf[kBufLength];

    		int nTextLength = ::GetWindowTextLength(m_hWnd);
    		
    		if (nTextLength < kBufLength)
    		{
    			// Use stack buffer
    			::GetWindowText(m_hWnd, TextBuf, kBufLength);
    			Text.assign(TextBuf);
    		}
    		else
    		{
    			// Allocate temporary buffer to get this data
    			TCHAR *pText = new TCHAR[nTextLength];
    			::GetWindowText(m_hWnd, pText, nTextLength);
    			Text.assign(pText);
    			delete [] pText;
    		}

			return true;
    	}
		else
		{
			return false;
		}
    }

	/**************************************************************************
	
	>	class ControlSearch
	
		Author:		Tim Browse
		Date:		14/7/2002
	
		Purpose:	Simple class to hold relevant info for searching for a
					control - see FindControl()
	
	**************************************************************************/
	
	class ControlSearch
	{
		public:
			// The ID of the control we're looking for
			ControlID m_ID;

			// The handle of the control, if it has been found
			HWND	  m_hControl;
	};

	/**************************************************************************
	
	>	HWND Win32_WindowImp::FindControl(HWND hParent, ControlID ID)
	
		Author:		Tim Browse
		Date:		14/7/2002
		Inputs:		hWindow - the window in which to search for this control
					ID - the ID of the control to search for
		Returns:	The HWND of the control, or NULL if not found.
	
		Purpose:	Find the specified control in this window and any of its
					children.
	
	**************************************************************************/
	
	HWND Win32_WindowImp::FindControl(HWND hWindow, ControlID ID)
	{
		// Check this window for the control
		HWND hControl = ::GetDlgItem(hWindow, ID);

		if (hControl == NULL)
		{
			// Ok, not found in this window, so check the child windows
			ControlSearch Search;
			Search.m_ID = ID;
			Search.m_hControl = NULL;
			::EnumChildWindows(hWindow, FindControlInWindow, (LPARAM) &Search);
			hControl = Search.m_hControl;
		}

		// Let the caller know what we found
		return hControl;
	}

	/**************************************************************************
	
	>	BOOL CALLBACK Win32_WindowImp::FindControlInWindow(HWND hwnd, LPARAM lParam)
	
		Author:		Tim Browse
		Date:		14/7/2002
		Inputs:		HWND - the handle of the child window
					lParam - actually a pointer to a ControlSearch object.
		Returns:	TRUE - keep searching;
					FALSE - control found, so stop enumerating child windows
	
		Purpose:	Callback used by FindControl() in conjunction with the Win32
					EnumChildWindows() API call, to check for a particular
					control in a window.
	
	**************************************************************************/
	
	BOOL CALLBACK Win32_WindowImp::FindControlInWindow(HWND hwnd, LPARAM lParam)
	{
		ControlSearch* pSearch = (ControlSearch*) lParam;

		// Check this child window for the control
		pSearch->m_hControl = ::GetDlgItem(hwnd, pSearch->m_ID);

		// If we haven't found it, keep on looking
		return (pSearch->m_hControl == NULL);
	}

    void Win32_WindowImp::Attach(Window *pParent,
    							 ControlID ID)
    {
		// Remember this parent
		m_pParent = pParent;

    	// Get parent's window handle, and use it to find control handle
    	HWND hControl = FindControl(GetHWND(pParent), ID);

    	// Complain if we didn't find the control
    	ASSERT(hControl != NULL);

    	// Attach to the control
    	Attach(hControl);
    }

    void Win32_WindowImp::AttachByHandle(Window *pParent,
    									 WindowHandle hWindow)
    {
		// Remember this parent
		m_pParent = pParent;

    	// Attach to the window
    	Attach(hWindow);
    }

    LRESULT Win32_WindowImp::SendMessage(UINT nMsg, 
    									 WPARAM wParam, 
    									 LPARAM lParam) const
    {
    	// If this ASSERT goes off, then either the window has
    	// not been created yet, or you have not called Attach() to 
    	// attach a vGUI dialog control object to its GUI object counterpart.
		ASSERT(IsValid());

    	// Send the message
    	return ::SendMessage(m_hWnd, nMsg, wParam, lParam);
    }

    bool Win32_WindowImp::IsWindowOfClass(HWND hWnd, LPCTSTR szClassName)
    {
    	// Validate hwnd
    	ASSERT(hWnd != NULL);

    	// Get the class name
    	const int kBufferSize = 128;
    	TCHAR szWindowClass[kBufferSize];

    	if (::GetClassName(hWnd, szWindowClass, kBufferSize-1) == 0)
    	{
    		// Error getting class name - shouldn't happen in normal practice.
    		vguiERROR();
    		return false;
    	}

    	// Compare class names (ignoring case)
    	return (_stricmp(szClassName, szWindowClass) == 0);
    }

    void Win32_WindowImp::GetArea(Rect* pArea)
    {
		ASSERT_VALID_WINDOW();

    	::GetWindowRect(m_hWnd, (RECT*) pArea);
    }

    void Win32_WindowImp::GetClientArea(Rect* pClientArea)
    {
		ASSERT_VALID_WINDOW();

    	::GetClientRect(m_hWnd, (RECT*) pClientArea);
    }

    void Win32_WindowImp::ClientToWindow(Window *pWindow, 
    									 Point* pPoints, 
    									 uint cPoints)
    {
		ASSERT_VALID_WINDOW();

    	::MapWindowPoints(m_hWnd, GetHWND(pWindow),
    					  (POINT*) pPoints, cPoints);
    }

    void Win32_WindowImp::ForceRedraw(const Rect* pRect)
    {
		ASSERT_VALID_WINDOW();

    	if (pRect == NULL)
    	{
    		// Force redraw of entire window
    		::InvalidateRect(m_hWnd, NULL, TRUE);
    	}
    	else
    	{
    		// Convert to GDI coords, taking account of scroll offsets
    		Rect RedrawArea = *pRect;
    		RedrawArea.Translate(-::GetScrollPos(m_hWnd, SB_HORZ),
    							 -::GetScrollPos(m_hWnd, SB_VERT));

    		::InvalidateRect(m_hWnd, (const RECT*) pRect, TRUE);
    	}
    }


	/**************************************************************************
	
	>	void Win32_WindowImp::LockRedraw()
	
		Author:		Tim Browse
		Date:		2001/10/28
	
		Purpose:	Prevent window redraw occurring in the window until
					UnlockRedraw() is called.
	
	**************************************************************************/
	
	void Win32_WindowImp::LockRedraw()
	{
		::LockWindowUpdate(GetHWND());
	}

	/**************************************************************************
	
	>	void Win32_WindowImp::UnlockRedraw()
	
		Author:		Tim Browse
		Date:		2001/10/28
	
		Purpose:	Unlock the window so it can be redrawn again, after
					LockRedraw() has been called().
	
	**************************************************************************/
	
	void Win32_WindowImp::UnlockRedraw()
	{
		::LockWindowUpdate(NULL);
	}
	
	/**************************************************************************

	>	void Win32_WindowImp::SetWindowMode(TWindowMode Mode)

		Author:		Tim Browse
		Date:		01/08/2001
		Inputs:		Mode - WindowMode_Normal or WindowMode_FullScreen.

		Purpose:	Change the window style according to whether the window
					is normal or in fullscreen mode.

					This is really only needed for DirectX apps.

	**************************************************************************/

	void Win32_WindowImp::SetWindowMode(TWindowMode Mode)
	{
		m_WindowMode = Mode;

		if (m_WindowMode == WindowMode_Normal)
		{
			// Set windowed-mode style
			::SetWindowLong(m_hWnd, GWL_STYLE, m_dwWindowStyle);
		}
		else
		{
			// Set fullscreen-mode style
			::SetWindowLong(m_hWnd, GWL_STYLE, WS_POPUP|WS_SYSMENU|WS_VISIBLE);
		}
	}
	
    void Win32_WindowImp::GetPosition(Rect* pWindowPosition)
    {
		ASSERT_VALID_WINDOW();

		GetPosition(m_hWnd, pWindowPosition);
    }

    void Win32_WindowImp::GetPosition(HWND hWnd, Rect* pWindowPosition)
    {
    	// Get position in screen co-ordinates
    	::GetWindowRect(hWnd, (RECT*) pWindowPosition);

    	// Translate to parent window co-ordinates
    	HWND hParent = ::GetParent(hWnd);

    	if (hParent != NULL)
    	{
    		::ScreenToClient(hParent, (POINT*) pWindowPosition);
    		::ScreenToClient(hParent, (POINT*) &(pWindowPosition->right));
    	}
    }
	
	void Win32_WindowImp::SetPosition(HWND hWnd, const Rect& WindowPosition)
    {
		::MoveWindow(hWnd, WindowPosition.left, WindowPosition.top,
					 WindowPosition.GetWidth(), WindowPosition.GetHeight(), 
					 TRUE);
    }
	
	void Win32_WindowImp::SetSafePosition(const Rect& NewPosition, bool bForceRedraw)
	{
		// TODO: make sure window is on screen!
		Move(NewPosition, bForceRedraw);
	}

    void Win32_WindowImp::Move(const Rect& NewPosition, bool bForceRedraw)
    {
		ASSERT_VALID_WINDOW();

    	::MoveWindow(m_hWnd, 
    				 NewPosition.left, NewPosition.top,
    				 NewPosition.GetWidth(), NewPosition.GetHeight(),
    				 bForceRedraw ? TRUE : FALSE);
    }

    void Win32_WindowImp::Resize(uint NewWidth, uint NewHeight)
    {
		ASSERT_VALID_WINDOW();

    	// Get current window position
    	Rect WindowPosition;
    	GetPosition(&WindowPosition);

    	// Change width/height
    	::MoveWindow(m_hWnd, 
    				 WindowPosition.left, WindowPosition.top,
    				 NewWidth, NewHeight,
    				 TRUE);
    }

    void Win32_WindowImp::SetClientArea(const Rect& NewClientArea)
    {
		ASSERT_VALID_WINDOW();

    	// Move client rect to correct place on screen
    	Rect OldClientRect;
    	::GetClientRect(m_hWnd, (RECT*) &OldClientRect);

    	// Get current client origin in screen co-ordinates.
    	::ClientToScreen(m_hWnd, (POINT*) &OldClientRect);

    	// Move new client area to align origin with old area
    	Rect NewClientRect = NewClientArea;
    	NewClientRect.Translate(OldClientRect.left, OldClientRect.top);

    	// Does the window have a menu?
    	BOOL bMenu = FALSE;
    	if (::GetMenu(m_hWnd) != NULL)
    	{
    		// Yes, it has a menu
    		bMenu = TRUE;
    	}

    	// Ask Windows where the window should be if we want the client area here.
    	::AdjustWindowRectEx((RECT*) &NewClientRect, 
    						 ::GetWindowLong(m_hWnd, GWL_STYLE), 
    						 bMenu,
    						 ::GetWindowLong(m_hWnd, GWL_EXSTYLE));

    	// Ok, so move the window to here
    	::MoveWindow(m_hWnd, 
    				 NewClientRect.left, NewClientRect.top,
    				 NewClientRect.GetWidth(), NewClientRect.GetHeight(),
    				 TRUE);
    }

    ControlID Win32_WindowImp::GetControlID()
    {
		ASSERT_VALID_WINDOW_RETURN(NOCTL);

		// Get the window ID
    	return (ControlID) ::GetWindowLong(m_hWnd, GWL_ID);
    }

    ControlID Win32_WindowImp::FindUniqueControlID()
    {
		ASSERT_VALID_WINDOW_RETURN(NOCTL);

    	// Start off at 1001, by convention
    	ControlID ID = 1001;

    	for(;;)
    	{
    		// Is this ID unique?  Check all the child windows...
    		HWND hChild = ::GetWindow(m_hWnd, GW_CHILD);

    		while (hChild != NULL)
    		{
    			if (::GetWindowLong(hChild, GWL_ID) == ID)
    			{
    				// ID is not unique
    				break;
    			}

    			// Try next window
    			hChild = ::GetWindow(hChild, GW_HWNDNEXT);
    		}

    		if (hChild == NULL)
    		{
    			// ID is unique
    			return ID;
    		}

    		// ID is not unique - try next one;
    		ID++;
    	}


    }

    WPARAM Win32_WindowImp::GetWin32State(ControlState State)
    {
    	WPARAM fCheck = BST_UNCHECKED;

    	switch (State)
    	{
    		case cs_Checked:		fCheck = BST_CHECKED;		break;
    		case cs_Undefined:		fCheck = BST_INDETERMINATE;	break;
    	}

    	return fCheck;
    }

    WindowHandle Win32_WindowImp::GetWindowHandle() const
    {
    	return GetHWND();
    }

    void Win32_WindowImp::ShowScrollBar(TScrollBar Bar)
    {
    	if (IsAlive())
    	{
    		// Window exists; caller wants to update it
    		if (Bar & sb_Horizontal)
    		{
    			if (::ShowScrollBar(m_hWnd, SB_HORZ, TRUE))
    				m_ScrollBarStatus |= sb_Horizontal;
    		}

    		if (Bar & sb_Vertical)
    		{
    			if (::ShowScrollBar(m_hWnd, SB_VERT, TRUE))
    				m_ScrollBarStatus |= sb_Vertical;
    		}
    	}
    	else
    	{
			// If the window handle is not NULL, then the window was alive
			// but has been destroyed, so this function should not be called.
			if (m_hWnd != NULL)
			{
				COMPLAIN("ShowScrollBar() called on destroyed window!");
				return;
			}

    		// Window not yet created; caller is specifying initial state
    		if (Bar & sb_Horizontal)
    			m_ScrollBarStatus |= sb_Horizontal;

    		if (Bar & sb_Vertical)
    			m_ScrollBarStatus |= sb_Vertical;
    	}
    }

    void Win32_WindowImp::HideScrollBar(TScrollBar Bar)
    {
    	if (IsAlive())
    	{
    		// Window exists; caller wants to update it
    		if (Bar & sb_Horizontal)
    		{
    			if (::ShowScrollBar(m_hWnd, SB_HORZ, FALSE))
    				m_ScrollBarStatus &= ~sb_Horizontal;
    		}

    		if (Bar & sb_Vertical)
    		{
    			if (::ShowScrollBar(m_hWnd, SB_VERT, FALSE))
    				m_ScrollBarStatus &= ~sb_Vertical;
    		}
    	}
    	else
    	{
			// If the window handle is not NULL, then the window was alive
			// but has been destroyed, so this function should not be called.
			if (m_hWnd != NULL)
			{
				COMPLAIN("HideScrollBar() called on destroyed window!");
				return;
			}

    		// Window not yet created; caller is specifying initial state
    		if (Bar & sb_Horizontal)
    			m_ScrollBarStatus &= ~sb_Horizontal;

    		if (Bar & sb_Vertical)
    			m_ScrollBarStatus &= ~sb_Vertical;
    	}
    }

    void Win32_WindowImp::EnableScrollBar(TScrollBar Bar, bool bEnable)
    {
		if (IsAlive())
		{
    		if (Bar & sb_Horizontal)
    		{
    			::EnableScrollBar(m_hWnd, SB_HORZ, 
    							  bEnable ? ESB_ENABLE_BOTH : ESB_DISABLE_BOTH);
    		}

    		if (Bar & sb_Vertical)
    		{
    			::EnableScrollBar(m_hWnd, SB_VERT, 
    							  bEnable ? ESB_ENABLE_BOTH : ESB_DISABLE_BOTH);
    		}
		}
		else
		{
			COMPLAIN("EnableScrollBar() called on dead window");
			return;
		}
    }

	bool Win32_WindowImp::IsScrollBarVisible(TScrollBar Bar)
	{
		if (IsAlive())
		{
			// CHeck window styles to see if scroll bar is on/off
    		if (Bar == sb_Horizontal)
    		{
				return (::GetWindowLong(m_hWnd, GWL_STYLE) & WS_HSCROLL) != 0;
    		}
    		else if (Bar == sb_Vertical)
    		{
				return (::GetWindowLong(m_hWnd, GWL_STYLE) & WS_VSCROLL) != 0;
    		}
			else
			{
				COMPLAIN("Unknown scroll bar type in IsScrollBarVisible()!");
			}
		}
		else
		{
			COMPLAIN("IsScrollBarVisible() called on dead window");
		}
		
		return false;
	}

    int Win32_WindowImp::GetScrollPos(TScrollBar Bar)
    {
		ASSERT_VALID_WINDOW_RETURN(0);

    	return ::GetScrollPos(m_hWnd, (Bar == sb_Vertical) ? SB_VERT : SB_HORZ);
    }

    void Win32_WindowImp::SetScrollPos(TScrollBar Bar, int Pos)
    {
		ASSERT_VALID_WINDOW();

    	::SetScrollPos(m_hWnd, (Bar == sb_Vertical) ? SB_VERT : SB_HORZ, Pos, TRUE);
    }

	TWindowPosture Win32_WindowImp::GetPosture()
	{
		ASSERT_VALID_WINDOW_RETURN(WindowPosture_Normal);

		WINDOWPLACEMENT wndpl;
		ZeroMemory(&wndpl, sizeof(wndpl));
		wndpl.length = sizeof(wndpl);
		::GetWindowPlacement(m_hWnd, &wndpl);

		switch (wndpl.showCmd)
		{
			case SW_SHOWMAXIMIZED:
				return WindowPosture_Maximised;

			case SW_SHOWMINIMIZED:
				return WindowPosture_Minimised;

			case SW_SHOWNORMAL:
				return WindowPosture_Normal;

			default:
				COMPLAIN("GetPosture() - Unknown type of window placement!");
				return WindowPosture_Normal;
		}
	}

	void Win32_WindowImp::SetPosture(TWindowPosture Posture)
	{
		ASSERT_VALID_WINDOW();

		// We just set the window state to be normal, unless we
		// find a non-normal posutre has been specified.
		int ShowCmd = SW_SHOWNORMAL;

		switch (Posture)
		{
			case WindowPosture_Maximised:
				ShowCmd = SW_MAXIMIZE;
				break;

			case WindowPosture_Minimised:
				ShowCmd = SW_MINIMIZE;
				break;
		}

		// Change window posture.
		::ShowWindow(m_hWnd, ShowCmd);
	}

    Font* Win32_WindowImp::GetFont()
    {
		ASSERT_VALID_WINDOW_RETURN(NULL);

    	// If we have no font object for this dialog, then make it.
    	if (m_pFont == NULL)
    	{
    		// Get the font handle used in the dialog
    		// (It will be NULL if it's the system font).
    		HFONT hFont = (HFONT) ::SendMessage(GetHWND(), WM_GETFONT, 0, 0);

    		// Make the font object.
    		m_pFont = new Font();

    		// Attach the Win32 font object to the vGUI object.
    		Win32_FontImp* pFontImp = Win32_FontImp::GetWin32_FontImp(m_pFont);
    		pFontImp->Attach(hFont);
    	}

    	// Return font object to caller
    	return m_pFont;
    }

    void Win32_WindowImp::SetWorkArea(const Rect& rcWorkArea)
    {
    	// NB. We allow the work area to be smaller than the client area - it's
    	//	   up to the app to sort that out.
    	m_rcWorkArea = rcWorkArea;

    	if (IsValid())
    	{
    		// Set up scroll bars based on work area
    		SetScrollBars();
    	}
    }

    void Win32_WindowImp::SetScrollBars()
    {
		ASSERT_VALID_WINDOW();

    	if (!m_rcWorkArea.IsEmpty())
    	{
    		// Set up scroll bars based on work area
    		SCROLLINFO si;
    		si.cbSize = sizeof(si);

    		Rect rcClient;
    		GetClientArea(&rcClient);

    		if (m_ScrollBarStatus & sb_Horizontal)
    		{
    			si.nMin = m_rcWorkArea.left;
    			si.nMax = m_rcWorkArea.right;
    			si.nPage = rcClient.GetWidth();
    			si.fMask = SIF_RANGE | SIF_PAGE;
    			::SetScrollInfo(m_hWnd, SB_HORZ, &si, TRUE);
    		}

    		if (m_ScrollBarStatus & sb_Vertical)
    		{
    			si.nMin = m_rcWorkArea.top;
    			si.nMax = m_rcWorkArea.bottom;
    			si.nPage = rcClient.GetHeight();
    			si.fMask = SIF_RANGE | SIF_PAGE;
    			::SetScrollInfo(m_hWnd, SB_VERT, &si, TRUE);
    		}
    	}
    }

    void Win32_WindowImp::GetWorkArea(Rect* pWorkArea)
    {
    	// If the work area has not been set, then we always use the client area.
    	if (m_rcWorkArea.IsEmpty())
    	{
    		GetClientArea(pWorkArea);
    		return;
    	}
    	else
    	{
    		*pWorkArea = m_rcWorkArea;
    	}
    }

    void Win32_WindowImp::SetFocus()
    {
    	if (IsValid())
    	{
    		::SetFocus(m_hWnd);
    	}
    }

	
	/**************************************************************************

	>	bool Win32_WindowImp::HasFocus()

		Author:		Tim Browse
		Date:		24/03/2001
		Returns:	true => window has input focus;
					false otherwise.

		Purpose:	Find out if this window object currently has the input focus.

	**************************************************************************/

	bool Win32_WindowImp::HasFocus()
	{
    	if (IsValid())
    	{
			return (::GetFocus() == m_hWnd);
		}

		// No window handle => we don't have focus.
		return false;
	}

	/**************************************************************************
	
	>	bool Win32_WindowImp::CaptureMouse()
	
		Author:		Tim Browse
		Date:		22/4/2002
		Returns:	true => mouse captured ok;
					false => otherwise
	
		Purpose:	Capture the mouse, so that all mouse input is directed
					to this window (regardless of which window the mouse is
					over).
	
					Call ReleaseMouse() when you no longer want all mouse
					input.

	**************************************************************************/
	
	bool Win32_WindowImp::CaptureMouse()
	{
		if (IsValid())
		{
			::SetCapture(m_hWnd);
			return true;
		}
		else
		{
			// Invalid window
			return false;
		}
	}

	/**************************************************************************
	
	>	void Win32_WindowImp::ReleaseMouse()
	
		Author:		Tim Browse
		Date:		22/4/2002
	
		Purpose:	Release captured mouse - see CaptureMouse().
	
	**************************************************************************/
	
	void Win32_WindowImp::ReleaseMouse()
	{
		if (IsValid())
		{
			::ReleaseCapture();
		}
	}

	/**************************************************************************
	
	>	void Win32_WindowImp::NotifyUserInputMessageReceived(DWORD dwTimeOfMessage = 0)
	
		Author:		Tim Browse
		Date:		25/7/2002
	
		Purpose:	Let the application know when an input message was last
					received.  
					
					See Win32_ApplicationImp::NotifyUserInputMessageReceived().
	
	**************************************************************************/
	
	void Win32_WindowImp::NotifyUserInputMessageReceived()
	{
		GetWin32App()->NotifyUserInputMessageReceived();
	}
	
	LRESULT Win32_WindowImp::ProcessDialogInputRequest(Window* pWindow, 
													   MSG* pMsg)
	{
		LRESULT lResult = 0;

		switch (pMsg->message)
		{
			case WM_CHAR:
			case WM_KEYDOWN:
			case WM_KEYUP:
			{
				// If window is subclassed, ask the super-class to process 
				// the message.
				if (m_pSuperWndProc != NULL)
				{
					lResult = ::CallWindowProc(m_pSuperWndProc, m_hWnd, 
											   WM_GETDLGCODE, 
											   0, (LPARAM) pMsg);
				}

				// Also, ask this vgui window what it wants to do
				if (pMsg->message == WM_CHAR)
				{
					if (pWindow->IsCharWanted((uint16) pMsg->wParam))
					{
						// message wanted
						lResult |= DLGC_WANTMESSAGE;
					}
				}
				else
				{
					TKeyCode KeyCode = Win32_TranslateVKToKeyCode(pMsg->wParam);

					if (pWindow->IsKeyWanted(KeyCode))
					{
						// message wanted
						lResult |= DLGC_WANTMESSAGE;
					}
				}
			}
		}

		// If we don't process this, lResult is 0, so we return lResult.
		//
		// This is one of those MS Windows messages where you return 0 if
		// you *don't* process it... :-)
		//
		return lResult;
	}

	bool Win32_WindowImp::HandleRawKeyEvent(UINT Msg,
											WPARAM wParam)
	{
		// Translate VK_ code into vGUI platform independent value.
		TKeyCode KeyCode = Win32_TranslateVKToKeyCode(wParam);

		if (KeyCode != kc_Invalid)
		{
			// Pass message to the Window to see if it claims it.
			if (Msg == WM_KEYDOWN)
			{
				if (m_pWindow->OnKeyDown(KeyCode, 0))
				{
					// message handled
					return true;
				}
			}
			else
			{
				if (m_pWindow->OnKeyUp(KeyCode, 0))
				{
					// message handled
					return true;
				}
			}
		}

		// Key not processed
		return false;
	}


	bool Win32_WindowImp::IsValid() const
	{
		return (IsAlive() && ::IsWindow(m_hWnd));
	}

	void Win32_WindowImp::SetControlHandler(Window* pHandlerWindow)
	{
		m_pControlHandlerWindow = pHandlerWindow;
	}

	Window* Win32_WindowImp::GetControlHandler()
	{
		return m_pControlHandlerWindow;
	}
	
	bool Win32_WindowImp::WindowHasStyle(HWND hWnd, DWORD dwStyleMask)
	{
		if (hWnd == NULL)
		{
			COMPLAIN("WindowHasStyle() - NULL HWND passed in!");
			return false;
		}

		// Check that the window has the style specified
		DWORD dwStyle = ::GetWindowLong(hWnd, GWL_STYLE);
		return ((dwStyle & dwStyleMask) == dwStyleMask);
	}

	typedef BOOL (STDAPICALLTYPE fnTrackMouseEvent)(TRACKMOUSEEVENT*);

	/**************************************************************************
	
	>	BOOL Win32_WindowImp::TrackMouseEvent(TRACKMOUSEEVENT* pTME)
	
		Author:		Tim Browse
		Date:		26/8/2002
		Inputs:		pTME - the details of the mouse event to track.
		Returns:	TRUE => mouse event is being tracked
					FALSE => event could not be tracked, or the API is not
							 available on the host.
	
		Purpose:	Provide a wrapper for the Win32 function TrackMouseEvent()
					(or rather, actually for _TrackMouseEvent) so that it can
					be called without having to statically link to comctl32.lib.

					Therefore if the function is not supported on a particular
					platform (e.g. Windows 95 original release) then the
					function fails robustly and just returns FALSE.

					Otherwise, the _TrackMouseEvent() function is called
					on behalf of the client.
	
	**************************************************************************/
	
	BOOL Win32_WindowImp::TrackMouseEvent(TRACKMOUSEEVENT* pTME)
	{
		static fnTrackMouseEvent* s_pfnTrackMouseEvent = NULL;
		static bool s_bTrackMouseEventNotAvailable = false;

		// If we don't have a pointer to the _TrackMouseEvent() function,
		// then we try to load COMCTL32.DLL and get one.  If we try this once
		// and it fails, then we don't try again - subsequent calls to this 
		// function  will fail immediately and return FALSE.
		if (s_pfnTrackMouseEvent == NULL)
		{
			if (!s_bTrackMouseEventNotAvailable)
			{
				// Try to load the function from comctrl32.dll
				HMODULE hModule = ::LoadLibrary("COMCTL32.DLL");

				if (hModule != NULL)
				{
					s_pfnTrackMouseEvent = (fnTrackMouseEvent*) ::GetProcAddress(hModule, "_TrackMouseEvent");
				}

				if (s_pfnTrackMouseEvent == NULL)
				{
					// Function not available - don't try to load it again
					s_bTrackMouseEventNotAvailable = true;
				}
			}
		}

		// Ok, use the function pointer if we have it; otherwise we fail
		if (s_pfnTrackMouseEvent != NULL)
			return (*s_pfnTrackMouseEvent)(pTME);
		else
			return FALSE;
	}

	/**************************************************************************

	>	BOOL Win32_WindowImp::TrackPopupMenuEx(HMENU hMenu,
											   UINT fuFlags,
											   int x, int y,
											   HWND hwnd,
											   LPTPMPARAMS lptpm)


		Author:		Tim Browse
		Date:		26/8/2002
		Inputs:		See Win32 TrackMenuPopupEx() documentation.
		Returns:	See Win32 TrackMenuPopupEx() documentation.

		Purpose:	Provide a wrapper for TrackPopupMenuEx() API, to avoid
					the problem mentioned in Microsoft Knowledge Base article
					Q135788.

	**************************************************************************/

	BOOL Win32_WindowImp::TrackPopupMenuEx(HMENU hMenu,
										   UINT fuFlags,
										   int x, int y,
										   HWND hwnd,
										   LPTPMPARAMS lptpm)
	{
		BOOL bResult = ::TrackPopupMenuEx(hMenu, fuFlags, x, y, hwnd, lptpm);

		// Fixes problem with TrackMenuPopEx - see MS KB article Q135788
		::PostMessage(hwnd, WM_NULL, 0, 0);

		return bResult;
	}

	/**************************************************************************
	
	>	void Win32_WindowImp::EnableFloatingBehaviour(bool bFloat)
	
		Author:		Tim Browse
		Date:		11/9/2002
		Inputs:		bFloat - true => make this window appear above all other
									 windows on screen, even if it does not
									 have focus.
							false => make window behave normally, i.e. other
									 windows can appear on top of it
	
		Purpose:	Control the behaviour of a window with respect to whether
					it appears above all 'normal' windows.
	
	**************************************************************************/
	
	void Win32_WindowImp::EnableFloatingBehaviour(bool bFloat)
	{
		::SetWindowPos(m_hWnd, 
					   bFloat ? HWND_TOPMOST : HWND_NOTOPMOST, 
					   0, 0, 0, 0,
					   SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOSIZE);
	}

};



