//
// Win32_WindowImp.h - Window Implementation Interface
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vgui_Win32_WindowImp_h
#define vgui_Win32_WindowImp_h

#include "..\vgui.h"

#include "..\WindowImp.h"
#include "..\Controls.h"
#include "..\Rect.h"

#include "Win32_ThemedObject.h"

namespace vgui
{
	class Win32_MessageInfo;
	
	//
	// Result of translating a Win32 control notification message
	//
	typedef enum
	{
		// Message translated ok, and it should be delivered.
		ControlTranslation_OK,

		// Message not recognised by the control
		ControlTranslation_Unknown,

		// Message recognised and automatically handled by control - do
		// not pass the message on.
		ControlTranslation_Handled,

		// Message recognised and handled by control, and the specified
		// return code should be returned for this message
		ControlTranslation_UseReturnCode
	}
	ControlTranslation;

	class Win32_WindowImp : public WindowImp,
							public Win32_ThemedObject
	{
		public:

			//
			// Constructor/Destructor
			//

			Win32_WindowImp();
			virtual ~Win32_WindowImp();

			//
			// Window creation/destruction
			//
			virtual bool Create(Window *pWindow,
								Window* pParent,
								TWindowStyle Style, 
								const StringResourceID* prWindowTitle);

			virtual bool Create(Window *pWindow,
								Window* pParent,
								TWindowStyle Style, 
								const String* psWindowTitle);

			virtual void Destroy();

			virtual void CentreWindow();

			//
			// Window manipulation
			//
			virtual void ShowWindow(TShowWindow Command);
			virtual void UpdateWindow();

			virtual void GetPosition(Rect* pWindowPosition);
			virtual void Move(const Rect& NewPosition, bool bForceRedraw);
			virtual void SetSafePosition(const Rect& NewPosition, bool bForceRedraw);
			virtual void SetClientArea(const Rect& NewClientArea);
			virtual void Resize(uint NewWidth, uint NewHeight);

			virtual void SetWorkArea(const Rect& WorkArea);
			virtual void GetWorkArea(Rect* pWorkArea);

			virtual void EnableFloatingBehaviour(bool bFloat);
			
			virtual bool SetText(const String& Text);
			virtual bool GetText(String& Text);

			virtual void SetFocus();
			virtual bool HasFocus();

			virtual bool CaptureMouse();
			virtual void ReleaseMouse();

			virtual WindowHandle GetWindowHandle() const;

			virtual void ShowScrollBar(TScrollBar Bar);
			virtual void HideScrollBar(TScrollBar Bar);

			virtual void EnableScrollBar(TScrollBar Bar, bool bEnable);

			virtual bool IsScrollBarVisible(TScrollBar Bar);

			virtual int GetScrollPos(TScrollBar Bar);
			virtual void SetScrollPos(TScrollBar Bar, int Pos);

			virtual TWindowPosture GetPosture();
			virtual void SetPosture(TWindowPosture Posture);

			virtual void Attach(Window *pParent, ControlID ID);
			virtual void AttachByHandle(Window *pParent, WindowHandle hWindow);

			virtual void Detach();

			//
			// Co-ordinate functions
			//
			virtual void ClientToWindow(Window *pWindow, Point* pPoints, uint cPoints);

			//
			// Redraw functions
			//
			virtual void ForceRedraw(const Rect* pRect = NULL);

			// Redraw lock/unlock
			virtual void LockRedraw();
			virtual void UnlockRedraw();

			// Full-screen support
			virtual void SetWindowMode(TWindowMode Mode);
			
			//
			// Get information
			//

			virtual void GetArea(Rect* pArea);
			virtual void GetClientArea(Rect* pClientArea);

			virtual ControlID GetControlID();

			virtual ControlID FindUniqueControlID();

			virtual Font* GetFont();

			// Access for other Win32 classes
			HWND GetHWND() const { return m_hWnd; }
			static HWND GetHWND(const Window*);

			// Create a child control in this window
			HWND CreateChildControl(Window* pControl,
									LPCTSTR szControlClass,
									LPCTSTR szWindowName,
									DWORD dwButtonStyle,
									DWORD dwStyleEx,
									const Rect& Position,
									ControlID ID,
									bool bTabStop = true);

			// Function to centre a window
			static void CentreWindow(HWND hwnd);

			//
			// Notify controls when they are auto-resized
			//
			virtual void OnControlResizing();
			virtual void OnControlResize();

			// Control notify message translation
			virtual ControlTranslation TranslateControlNotification(const Win32_MessageInfo& MsgInfo,
																	TMessageType* pMsgType,
																	LRESULT* pResult,
																	MessageParams* pParams);

			int TranslateMouseButtons(WPARAM wParam);
			int TranslateMouseModifiers(WPARAM wParam);

			// Set the window which should receive control messages
			// that are generated in this window.
			// (usually, the window receives its own messages, but this
			// allows you to route messages elsewhere, e.g. to a parent
			// window)
			void SetControlHandler(Window* pHandlerWindow);
			Window* GetControlHandler();

			// Check for correct window class/style
			static bool IsWindowOfClass(HWND hwnd, LPCTSTR szClassName);
			static bool WindowHasStyle(HWND hWnd, DWORD dwStyle);

			// Convenient functions for window positioning
			static void GetPosition(HWND hWnd, Rect* pWindowPosition);
			static void SetPosition(HWND hWnd, const Rect& WindowPosition);

			// Find a control in a window or any of its children
			static HWND FindControl(HWND hParent, ControlID ID);

			// Call TrackMouseEvent in a robust way that doesn't require
			// static linking to COMCTRL32.DLL
			static BOOL TrackMouseEvent(TRACKMOUSEEVENT* pTME);

			// Wrapper to fix problem with TrackPopupMenuEx() function
			static BOOL Win32_WindowImp::TrackPopupMenuEx(HMENU hMenu,
														  UINT fuFlags,
														  int x, int y,
														  HWND hwnd,
														  LPTPMPARAMS lptpm);

		protected:
			// Access to the implementation of other Window objects
			static Win32_WindowImp *GetImp(const Window *pWindow);

			// The GC implementation class is coupled to the Window implementation class,
			// as is the RadioGroup class
			friend class Win32_GCImp;
			friend class RadioGroup;
			friend class Win32_BrowseDialogImp;

		protected:

			static Window* GetWindowFromHWND(HWND);
			static Win32_WindowImp* GetImpFromHWND(HWND hWnd);

			virtual void Attach(HWND hWnd);
			void SubclassWindow(HWND hWnd);
			void UnSubclassWindow(HWND hWnd);

			// SendMessage analogue
			LRESULT SendMessage(UINT nMsg, WPARAM wParam, LPARAM lParam) const;

			// Convert a vgui button state to a Win32 state
			WPARAM GetWin32State(ControlState State);

			// Set scroll bars according to work area
			void SetScrollBars();

			// Handle WM_VSCROLL and WM_HSCROLL messages.
			void ProcessScrollBarUpdate(int ScrollBar, 
										int ScrollUpdateCode,
										int NewScrollPos);

			static void NotifyUserInputMessageReceived();

			// The WndProc that is used to subclass other windows
			static LRESULT CALLBACK SubclassWndProc(HWND hWnd, 
												    UINT nMsg, 
												    WPARAM wParam, 
												    LPARAM lParam);

			// Which defers to analogue member functions...
			virtual LRESULT WindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam);
			LRESULT DefaultWindowProc(UINT nMsg, WPARAM wParam, LPARAM lParam);

			// Grant application object access to wndprocs
			friend class Win32_ApplicationImp;

			// ALlow debug class to access our data
			friend class Win32_Debug;

			// Helper functions for dispatching different types of messages
			virtual LRESULT DispatchControlNotification(UINT uMsg, 
														ControlID nCtrlID,
														LPARAM lParam,
														WPARAM wParam);

			bool HandleWindowPosChanging(WINDOWPOS* pWP);

			bool HandleRawKeyEvent(UINT Msg, WPARAM wParam);
			TKeyCode TranslateKey(int VK);

			LRESULT ProcessDialogInputRequest(Window* pWindow, MSG* pMsg);

#ifdef _DEBUG
			void CheckKeyCodeLookupTable();
#endif

			// Find out if the window is still alive (i.e. it has been created
			// but not yet destroyed).
			inline bool IsAlive() const { return m_bAlive; }

			bool IsValid() const;

			static BOOL CALLBACK FindControlInWindow(HWND hwnd, LPARAM lParam);

		protected:
			//
			// Win32 implementation data
			//

			// Handle of the window is inherited from Win32_ThemedObject - even 
			// if this is non-NULL, the window may not be active - use IsAlive() 
			// to find out if the window has been destroyed yet or not.
			// (member variable is m_hWnd)

			// Parent of this window object
			Window *m_pParent;

			// The window that we send control notifications to - this
			// allows us to easily vector control messages from a child
			// dialog to a host parent dialog. (see ChildDialog class)
			// (If NULL, we just send the messages to m_pWindow)
			Window* m_pControlHandlerWindow;

			WNDPROC m_pSuperWndProc;
			bool m_bSubClassed;

			// Style used to create window (used to reinstate the
			// window style when switching from fullscreen back to
			// normal windowed mode).
			DWORD m_dwWindowStyle;
			
			// Current mode this window is in.
			TWindowMode m_WindowMode;

			// Flags that record which scroll bars are enabled
			int m_ScrollBarStatus;

			// Default font in the window
			Font* m_pFont;

			Rect m_rcWorkArea;

			// true if window has been created, and not yet destroyed
			bool m_bAlive;
	};
};

#endif // !vgui_Win32_WindowImp_h


