//
// Win32_WindowInfo.cpp - for attaching data to Win32 windows
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_WindowInfo.h"

#include "..\Errors.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vgui
{
	/**************************************************************************
	
	>	void Win32_WindowInfo::Attach(HWND hWnd, Window* pWindow)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hWnd - the window to attach the pointer to
					pWindow - the Window pointer to attach
	
		Purpose:	Attach a Window pointer to the specified window in a
					safe and reliable way.
	
	**************************************************************************/
	
	void Win32_WindowInfo::Attach(HWND hWnd, Window* pWindow)
	{
		if (::GetWindowLongPtr(hWnd, GWL_USERDATA) != NULL)
		{
			COMPLAIN("Win32_WindowInfo::Attach() - window is already using GWL_USERDATA field!");
		}
		else
		{
			Win32_WindowInfo* pInfo = new Win32_WindowInfo;
			pInfo->m_Type = WindowDataType_Window;
			pInfo->m_pWindow = pWindow;
			::SetWindowLongPtr(hWnd, GWL_USERDATA, (LONG_PTR) pInfo);
		}
	}

	/**************************************************************************
	
	>	void Win32_WindowInfo::Attach(HWND hWnd, View* pView)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hWnd - the window to attach the pointer to
					pView - the View pointer to attach
	
		Purpose:	Attach a View pointer to the specified window in a
					safe and reliable way.
	
	**************************************************************************/
	
	void Win32_WindowInfo::Attach(HWND hWnd, View* pView)
	{
		if (::GetWindowLongPtr(hWnd, GWL_USERDATA) != NULL)
		{
			COMPLAIN("Win32_WindowInfo::Attach() - window is already using GWL_USERDATA field!");
		}
		else
		{
			Win32_WindowInfo* pInfo = new Win32_WindowInfo;
			pInfo->m_Type = WindowDataType_View;
			pInfo->m_pView = pView;
			::SetWindowLongPtr(hWnd, GWL_USERDATA, (LONG_PTR) pInfo);
		}
	}

	/**************************************************************************
	
	>	void Win32_WindowInfo::Attach(HWND hControl, Win32_CustomControl* pControl)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hWnd - the window to attach the pointer to
					pControl - the pointer to attach
	
		Purpose:	Attach a Win32_CustomControl pointer to the specified
					window in a safe and reliable way.
	
	**************************************************************************/
	
	void Win32_WindowInfo::Attach(HWND hControl, Win32_CustomControl* pControl)
	{
		if (::GetWindowLongPtr(hControl, GWL_USERDATA) != NULL)
		{
			COMPLAIN("Win32_WindowInfo::Attach() - window is already using GWL_USERDATA field!");
		}
		else
		{
			Win32_WindowInfo* pInfo = new Win32_WindowInfo;
			pInfo->m_Type = WindowDataType_Control;
			pInfo->m_pControl = pControl;
			::SetWindowLongPtr(hControl, GWL_USERDATA, (LONG_PTR) pInfo);
		}
	}

	/**************************************************************************
	
	>	void Win32_WindowInfo::Detach(HWND hWnd)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hWnd - the window to detach the pointer from
	
		Purpose:	Detach a pointer that has been attached to a window by
					a previous call to one of the forms of Attach().

					This should always be called at some point, to avoid
					memory leaks.

					The function COMPLAINs if no pointer has been attached
					to this window.

	**************************************************************************/
	
	void Win32_WindowInfo::Detach(HWND hWnd)
	{
		Win32_WindowInfo* pInfo = (Win32_WindowInfo*) ::GetWindowLongPtr(hWnd, GWL_USERDATA);

		if (pInfo != NULL)
		{
			switch (pInfo->m_Type)
			{
				case WindowDataType_Control:
				case WindowDataType_Window:
				case WindowDataType_View:
				{
					// It's really our data, so remove it
					::SetWindowLongPtr(hWnd, GWL_USERDATA, 0);
					delete pInfo;
					break;
				}

				default:
				{
					COMPLAIN("Win32_WindowInfo::Detach() - GWL_USERDATA does not "
							 "contain valid vGUI data - is other code using this field?");
				}
			}
		}
		else
		{
			COMPLAIN("Win32_WindowInfo::Detach() - no window info data found!");
		}
	}

	/**************************************************************************
	
	>	bool Win32_WindowInfo::GetAttachedWindow(HWND hWnd, Window** ppWindow)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hWnd - the window to check
		Outputs:	ppWindow - Window pointer from the window (if found)
		Returns:	true => a Window pointer was found on the window
					false => the window has no attached Window pointer.
	
		Purpose:	Access the Window pointer that is attached to the specified
					window.
	
	**************************************************************************/
	
	bool Win32_WindowInfo::GetAttachedWindow(HWND hWnd, Window** ppWindow)
	{
		Win32_WindowInfo* pInfo = (Win32_WindowInfo*) ::GetWindowLongPtr(hWnd, GWL_USERDATA);

		if (pInfo != NULL)
		{
			if (pInfo->m_Type == WindowDataType_Window)
			{
				// It's really our data, so return it
				*ppWindow = pInfo->m_pWindow;
				return true;
			}
		}

		// If we get here, there is no window (but do not COMPLAIN, so 
		// that code can safely check for existence of the window.
		return false;
	}

	/**************************************************************************
	
	>	bool Win32_WindowInfo::GetAttachedWindow(HWND hWnd, Window** ppWindow)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hWnd - the window to check
		Outputs:	ppWindow - Window pointer from the window (if found)
		Returns:	true => a Window pointer was found on the window
					false => the window has no attached Window pointer.
	
		Purpose:	Access the Window pointer that is attached to the specified
					window.
	
	**************************************************************************/
	
	bool Win32_WindowInfo::GetAttachedView(HWND hWnd, View** ppView)
	{
		Win32_WindowInfo* pInfo = (Win32_WindowInfo*) ::GetWindowLongPtr(hWnd, GWL_USERDATA);

		if (pInfo != NULL)
		{
			if (pInfo->m_Type == WindowDataType_View)
			{
				// It's really our data, so return it
				*ppView = pInfo->m_pView;
				return true;
			}
		}

		// If we get here, there is no view (but do not COMPLAIN, so 
		// that code can safely check for existence of the view.
		return false;
	}

	/**************************************************************************
	
	>	bool Win32_WindowInfo::GetAttachedControl(HWND hControl, Win32_CustomControl** ppControl)
	
		Author:		Tim Browse
		Date:		20/8/2002
		Inputs:		hWnd - the window to check
		Outputs:	ppControl - Win32_CustomControl pointer from the window (if found)
		Returns:	true => a Win32_CustomControl pointer was found on the window
					false => the window has no attached Win32_CustomControl pointer.
	
		Purpose:	Access the Win32_CustomControl pointer that is attached to 
					the specified window.
	
	**************************************************************************/
	
	bool Win32_WindowInfo::GetAttachedControl(HWND hControl, Win32_CustomControl** ppControl)
	{
		Win32_WindowInfo* pInfo = (Win32_WindowInfo*) ::GetWindowLongPtr(hControl, GWL_USERDATA);

		if (pInfo != NULL)
		{
			if (pInfo->m_Type == WindowDataType_Control)
			{
				// It's really our data, so return it
				*ppControl = pInfo->m_pControl;
				return true;
			}
		}

		// If we get here, there is no control (but do not COMPLAIN, so 
		// that code can safely check for existence of the control.
		return false;
	}

};
